/**
 * @file cmapl_collection.h
 *
 * The CMAPL collection-level function declarations. These functions provide all the
 * functionality to access and parse any type of collection containing CM Archiver
 * files, such as directories containing archiver files or individual archiver files.
 */

#ifndef _CMAPL_COLLECTION_H_
#define _CMAPL_COLLECTION_H_

/**
 * @name cmapl_open_collection
 *
 * @param[in]           collection_def	Collection definition to pass in with path. Can be
 *                                      either an archive file, or a directory containing
 *					archive files. Also includes time range of data to
 *					retrieve.
 * @param[out]          collection       Handle of collection that's opened
 * @return              cmapl_error_t   Error flag
 *
 * This method takes a path to a specific archive or a path to a directory
 * containing archive files and parses it. It then returns a collection
 * handle that contains the data of the structure opened and passes it back to the
 * user through the collection parameter.
 *
 * If a directory path is passed in, then this method will only search 4
 * directories deep for archive files before it stops.
 *
 * This method is used to open only a single cluster of data. Data from multiple
 * clusters will need to be opened separately.
 *
 * The collection_def parameter should be of some type of cmapl_collection_def_t struct,
 * depending on the version of the file.
 *
 * The return value will be of type cmapl_error indicating either success or failure.
 */
extern cmapl_error_t
cmapl_open_collection(
                        void * collection_def,
                        cmapl_collection_handle_t ** collection);

/**
 * @name cmapl_close_collection
 *
 * @param[in]           collection      Handle of collection to close
 * @return              cmapl_error_t   Error flag
 *
 * This method takes a collection handle and closes it. All internal data the handle
 * contains is freed. All file pointers that the handle has opened will be closed
 * as well.
 *
 * The return value will be of type cmapl_error indicating either success or failure.
 */
extern cmapl_error_t
cmapl_close_collection(
                        cmapl_collection_handle_t * collection);

/**
 * @name cmapl_get_collection_info
 *
 * @param[in]           collection      Handle of collection to parse
 * @param[out]          collection_info	Metadata of the collection
 * @return              cmapl_error_t   Error flag
 *
 * This method takes in a handle of a collection, and retrieves the metadata of
 * the collection.
 *
 * The collection_info parameter should be of some type of cmapl_collection_def_t struct,
 * depending on the version of the file. This struct will be freed if the collection
 * is closed.
 *
 * The collection_info struct cannot be modified by the user in any way. Doing so
 * will corrupt internal data.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_get_collection_info(
			cmapl_collection_handle_t * collection,
			void * collection_info);

/**
 * @name cmapl_next_collection_record
 *
 * @param[in]           collection      Handle of collection to parse
 * @param[in]		order_type	Specifies the order to return the records in
 * @param[out]          next_record     Handle of next record data
 * @pre                                 Pass in NULL to next_record to get the
 *					first record in the collection
 * @return              cmapl_error_t   Error flag
 *
 * This method takes in a handle of a collection, and retrieves a handle of the next
 * chronological record from that collection, regardless of node, process, or store. The store
 * information is passed back to the user through store_info to help the user determine which
 * node, process, and store this record belongs to. If no record exists, then NULL is returned.
 *
 * The order_type parameter specifies whether the records are returned in linear order (chronological
 * timestamp order across the whole collection) or in raw order (the order the records were written
 * to each file, one store at a time).
 *
 * If NULL is passed in to the next_record parameter, then the first record in the
 * collection will be returned. Any other value will return the next record in the
 * collection.
 *
 * The store_info parameter should be of some type of cmapl_store_def_t struct,
 * depending on the version of the file. This struct will be freed if the collection
 * is closed.
 *
 * The header parameter pass back to the user the header information associated with
 * this particular record and the file the record belongs to. This includes all metadata
 * and preset information.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_next_collection_record(
                        cmapl_collection_handle_t * collection,
			cmapl_order_type_t order_type,
                        cmapl_record_handle_t ** next_record,
			void * store_info,
			cmapl_header_handle_t ** header);

/**
 * @name cmapl_next_perfstat_file
 *
 * @param[in]           collection      Handle of collection to retrieve perfstat files from
 * @param[out]		next_file_path	Full path to the next perfstat file
 * @param[out]          store_info	Metadata of the store the perfstat file is in
 * @pre                                 Pass in NULL to next_file_path to get the
 *					first perfstat file path in the collection
 * @return              cmapl_error_t   Error flag
 *
 * This method takes in a handle of a collection, and retrieves the full file path
 * of the next perfstat file in the collection. The perfstat files are retrieved in
 * chronological timestamp order, one store at a time.
 *
 * If NULL is passed in to the next_file_path parameter, then the first perfstat file's
 * path in the collection will be returned. Any other value will return the next file's
 * path in the collection.
 *
 * The store_info parameter should be of some type of cmapl_store_def_t struct,
 * depending on the version of the file. This struct will be freed if the collection
 * is closed.
 *
 * The next_file_path returned to the user do not need to be freed by the user. It will
 * be freed when the collection handle is closed.
 *
 * The return value will be of type cmapl_error_t indicating either success or failure.
 */
extern cmapl_error_t
cmapl_next_perfstat_file(
                        cmapl_collection_handle_t * collection,
                        char ** next_file_path,
			void * store_info);

#endif /* _CMAPL_COLLECTION_H_ */
