/**
 * The CMAPL internal collection-level function declarations. These functions
 * are version specific declarations. Different version data files will
 * use functions that will have different implementations. The external
 * functions will call one of these functions depending on the version
 * of the data file.
 *
 * The descriptions of the external CMAPL functions related to each of
 * these internal functions are described in cmapl_collection.h.
 *
 * Syntax of functions:
 *
 *      <function_name>_v<major_version>_<minor_version>
 *
 *      function_name   - the external function name
 *      major_version   - the major version number of the data file
 *      minor_version   - the minor version number of the data file
 *
 *      For Example:    cmapl_open_collection_v6_0 stands for the
 *                      cmapl_open_collection function used on
 *                      a version 6.0 data file.
 *
 */

#ifndef _CMAPL_COLLECTION_INT_H_
#define _CMAPL_COLLECTION_INT_H_

cmapl_error_t
cmapl_close_collection_v6_0(
                        cmapl_collection_handle_t * collection);

cmapl_error_t
cmapl_open_collection_v6_0(
                        cmapl_collection_def_t_v6_0 * collection_def,
                        cmapl_collection_handle_t ** collection);

cmapl_error_t
cmapl_first_collection_record_v6_0(
                        cmapl_collection_handle_t * collection,
			cmapl_order_type_t order_type,
                        cmapl_record_handle_t ** first_record,
			cmapl_store_def_t_v6_0 * store_info,
			cmapl_header_handle_t ** header);

cmapl_error_t
cmapl_get_collection_info_v6_0(
			cmapl_collection_handle_t * collection,
			cmapl_collection_def_t_v6_0 * collection_info);

cmapl_error_t
cmapl_next_collection_record_v6_0(
                        cmapl_collection_handle_t * collection,
			cmapl_order_type_t order_type,
                        cmapl_record_handle_t ** next_record,
			cmapl_store_def_t_v6_0 * store_info,
			cmapl_header_handle_t ** header);

cmapl_error_t
cmapl_next_perfstat_file_v6_0(
			cmapl_collection_handle_t * collection,
			char ** next_file_path,
			cmapl_store_def_t_v6_0 * store_info);

///////////////////////////////////////////////////////////////////////////
// The following functions are internal-only functions, meaning they have no
// general external version of them.
///////////////////////////////////////////////////////////////////////////

/**
* @name cmapl_add_file_to_collection
*
* @param[in]           collection                Handle of collection
* @param[in]           file_path                Path to the generic file to add
* @return              cmapl_error_t            Error flag
*
* This method takes a path to a generic archive file, and adds it to the collection organized
* by the node, process, and store it belongs to. For each store, the files are sorted
* in chronological order.
*
* The return value will be of type cmapl_error_t indicating either success or failure.
*/
cmapl_error_t
cmapl_add_file_to_collection_v6_0(
			cmapl_collection_handle_t * collection,
			char * file_path);

/**
* @name cmapl_add_files_to_collection_recurs
*
* @param[in]		collection               Handle of collection
* @param[in]		dir_path		Directory path to look for files to add
* @param[in]		depth			Maximum depth in a directory to search for files
* @return		cmapl_error_t           Error flag
*
* This method takes a path to a directory and finds all the archive files deep
* within the directory. The depth parameter specifies how deep to look for files.
* Each file that's found will be added to the collection and organized by node,
* process, and store. This is a recursive function and will recurse on each
* subdirectory it runs into, decrementing the depth by 1.
*
* The return value will be of type cmapl_error_t indicating either success or failure.
*/
cmapl_error_t
cmapl_add_files_to_collection_recurs_v6_0(
			cmapl_collection_handle_t * collection,
			char * dir_path,
			uint32_t depth);

/**
* @name cmapl_create_store
*
* @param[in]		collection		Handle of the collection to add the store to
* @param[in]		store_info		Metadata of the new store to create
* @param[in]		created_store		Pointer to the newly created store
* @return		cmapl_error_t           Error flag
*
* This method takes a new store info and creates a new internal store structure in the
* collection. The user can then call cmapl_insert_file_to_store() on this new store
* to add files to it. The new store will be passed back through the parameter created_store.
*
* The return value will be of type cmapl_error_t indicating either success or failure.
*/
cmapl_error_t
cmapl_create_store_v6_0(
			cmapl_collection_handle_t * collection,
			cmapl_store_def_t_v6_0 store_info,
			cmapl_store_int_t_v6_0 ** created_store);

/**
* @name cmapl_gather_all_stores
*
* @param[in]		collection               Handle of collection
* @return		cmapl_error_t           Error flag
*
* This method finds all the stores within the collection hierarchy, and groups them
* all into a single internal array, allowing for easy iteration of all the stores.
*
* The return value will be of type cmapl_error_t indicating either success or failure.
*/
cmapl_error_t
cmapl_gather_all_stores_v6_0(
			cmapl_collection_handle_t * collection);

/**
* @name cmapl_get_earliest_record
*
* @param[in]		collection              Handle of collection
* @param[out]		next_record		Handle of the next earliest record
* @param[out]		store_info		Metadata of the store the record belongs to
* @param[out]		header			Metadata of the file the record belongs to
* @return		cmapl_error_t           Error flag
*
* This method finds the next earliest record to retrieve from the collection from
* all the stores and their currently opened files. If a record is found, it is returned
* along with its store information. If there are no more records available, then NULL
* is returned as the record handle.
*
* The return value will be of type cmapl_error_t indicating either success or failure.
*/
cmapl_error_t
cmapl_get_earliest_record_v6_0(
			cmapl_collection_handle_t * collection,
			cmapl_record_handle_t ** next_record,
			cmapl_store_def_t_v6_0 * store_info,
			cmapl_header_handle_t ** header);
/**
* @name cmapl_insert_file_to_store
*
* @param[in]		store			Pointer of the store to add the file
* @param[in]		file_info		Meta data info of the file
* @return		cmapl_error_t           Error flag
*
* This method takes a file and adds a new internal file definition sturcture in the
* store it belongs to. The file will be added to the store in order of creation timestamp,
* from earliest to latest.
*
* The return value will be of type cmapl_error_t indicating either success or failure.
*/
cmapl_error_t
cmapl_insert_file_to_store_v6_0(
			cmapl_store_int_t_v6_0 * store,
			cmapl_file_def_t_v6_0 file_info);

bool
cmapl_is_store_equal_v6_0(
			cmapl_store_def_t_v6_0 store_info,
			cmapl_store_int_t_v6_0 * store);

#endif /* _CMAPL_COLLECTION_INT_H_ */
