/**
 * The CMAPL corefile-level function definitions. These functions provide all the
 * functionality to parse and read archiver core files.
 *
 */

#ifdef _CMAPL_USER_SPACE_
#include "cmapl_common_defs.h"
#include "cmapl_int_defs.h"
#include "cmapl_read.h"
#include "cmapl_read_int.h"
#include "cmapl_write.h"
#include "cmapl_write_int.h"
#include "cmapl_directory.h"
#include "cmapl_directory_int.h"
#include "cmapl_corefile.h"
#include "cmapl_corefile_int.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

cmapl_error_t
cmapl_close_corefile(
                        cmapl_corefile_handle_t * corefile)
{
        if (corefile == NULL) {
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(corefile, 6, 0)) {
                return cmapl_close_corefile_v6_0(corefile);
        } else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_close_corefile_v6_0(
                        cmapl_corefile_handle_t * corefile)
{
        cmapl_error_t err = CMAPL_SUCCESS;

        // remove the previously opened file, if there was one
        err = cmapl_remove_current_file_v6_0(corefile);
        if (err != CMAPL_SUCCESS) {
                (void)cmapl_remove_current_directory_v6_0(corefile);
                (void)CMAPL_RMDIR(corefile->i6_0.temp_dir);
                cmapl_free_corefile(corefile);
                return err;
        }

        // remove the previously opened directory, if there was one
        err = cmapl_remove_current_directory_v6_0(corefile);
        if (err != CMAPL_SUCCESS) {
                (void)CMAPL_RMDIR(corefile->i6_0.temp_dir);
                cmapl_free_corefile(corefile);
                return err;
        }

        // remove temp directory
        int e = CMAPL_RMDIR(corefile->i6_0.temp_dir);
        if (e != 0) {
                cmapl_free_corefile(corefile);
                return CMAPL_FILE_NOT_FOUND;
        }

        cmapl_free_corefile(corefile);

        return err;
}

cmapl_error_t
cmapl_first_datafile(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_file_handle_t ** datafile)
{
        if (corefile == NULL) {
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(corefile, 6, 0)) {
                return cmapl_first_datafile_v6_0(corefile, datafile);
        } else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_datafile_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_file_handle_t ** datafile)
{
        cmapl_error_t err;

        (*datafile) = NULL;

        // remove the previously opened file, if there was one
        err = cmapl_remove_current_file_v6_0(corefile);
        if (err != CMAPL_SUCCESS) {
                return err;
        }

        // set file index to the first file
        corefile->i6_0.file_index = 0;
        if (corefile->i6_0.num_files == 0) {
                return err;
        }

        // extract the first file temporarily
        err = cmapl_extract_file_v6_0(corefile, 0);
        if (err != CMAPL_SUCCESS) {
                return err;
        }

        // open the first file and get the file handle
        const char* meta_path;
        const char* data_path;
        const char* file_name = corefile->i6_0.file_names[0];
        err = cmapl_open_data_file(file_name, "rb", datafile, &meta_path, &data_path);
        if (err != CMAPL_SUCCESS) {
                return err;
        }

        corefile->i6_0.current_file = *datafile;

        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_first_directory(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_directory_handle_t ** dir)
{
        if (corefile == NULL) {
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(corefile, 6, 0)) {
                return cmapl_first_directory_v6_0(corefile, dir);
        } else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_directory_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_directory_handle_t ** dir)
{
        // TODO: not implemented until corefile contains directories
        return CMAPL_SUCCESS;
}


cmapl_error_t
cmapl_next_datafile(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_file_handle_t ** datafile)
{
        if (corefile == NULL) {
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(corefile, 6, 0)) {
                return cmapl_next_datafile_v6_0(corefile, datafile);
        } else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_datafile_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_file_handle_t ** datafile)
{
        cmapl_error_t err;

        (*datafile) = NULL;

        // remove the previously opened file, if there was one
        // TODO: might want to delete all the files when the archive is closed or when quota limit is reached,
        //       we can allow the user to set some quota limit?
        err = cmapl_remove_current_file_v6_0(corefile);
        if (err != CMAPL_SUCCESS) {
                return err;
        }

        // set file index to the next file
        corefile->i6_0.file_index++;
        if (corefile->i6_0.file_index >= corefile->i6_0.num_files) {
                return err;
        }

        // extract the next file temporarily
        err = cmapl_extract_file_v6_0(corefile, corefile->i6_0.file_index);
        if (err != CMAPL_SUCCESS) {
                return err;
        }

        // open the next file and get the file handle
        const char* meta_path;
        const char* data_path;
        const char* file_name = corefile->i6_0.file_names[corefile->i6_0.file_index];
        err = cmapl_open_data_file(file_name, "rb", datafile, &meta_path, &data_path);
        if (err != CMAPL_SUCCESS) {
                return err;
        }
        corefile->i6_0.current_file = *datafile;

        return err;
}

cmapl_error_t
cmapl_next_directory(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_directory_handle_t ** dir)
{
        if (corefile == NULL) {
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(corefile, 6, 0)) {
                return cmapl_next_directory_v6_0(corefile, dir);
        } else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_directory_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        cmapl_directory_handle_t ** dir)
{
        // TODO: not implemented until corefile contains directories
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_open_corefile(
                        char * file_name,
                        cmapl_corefile_handle_t ** corefile)
{
        if (file_name == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        return cmapl_open_corefile_v6_0(file_name, corefile);
}

cmapl_error_t
cmapl_open_corefile_v6_0(
                        char * file_name,
                        cmapl_corefile_handle_t ** corefile)
{
        cmapl_error_t err = CMAPL_SUCCESS;
        (*corefile) = (cmapl_corefile_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_corefile_handle_t));
        if (*corefile == NULL) {
            return CMAPL_NULL_POINTER;
        }

        (*corefile)->major_v = 6;
        (*corefile)->minor_v = 0;
        (*corefile)->i6_0.num_files = 0;
        (*corefile)->i6_0.num_dirs = 0;
        (*corefile)->i6_0.file_index = 0;
        (*corefile)->i6_0.dir_index = 0;
        (*corefile)->i6_0.current_file = NULL;
        (*corefile)->i6_0.current_dir = NULL;

        (*corefile)->e6_0.major_v = 6;
        (*corefile)->e6_0.minor_v = 0;
        (*corefile)->e6_0.file_path = CMAPL_STRDUP(file_name);

        SRes res;

        // open the corefile with lzma
        if (InFile_Open(&((*corefile)->i6_0.archiveStream.file), file_name))
        {
            cmapl_free_corefile((*corefile));
            return CMAPL_FILE_NOT_FOUND;
        }
        // initialize the corefile to be parsed later
        FileInStream_CreateVTable(&((*corefile)->i6_0.archiveStream));
        LookToRead_CreateVTable(&((*corefile)->i6_0.lookStream), False);

        (*corefile)->i6_0.lookStream.realStream = &((*corefile)->i6_0.archiveStream.s);
        LookToRead_Init(&((*corefile)->i6_0.lookStream));

        (*corefile)->i6_0.allocImp.Alloc = SzAlloc;
        (*corefile)->i6_0.allocImp.Free = SzFree;
        (*corefile)->i6_0.allocTempImp.Alloc = SzAllocTemp;
        (*corefile)->i6_0.allocTempImp.Free = SzFreeTemp;
        CrcGenerateTable();

        // parse the corefile for all the archive files
        SzArEx_Init(&((*corefile)->i6_0.db));
        res = SzArEx_Open(&((*corefile)->i6_0.db), &((*corefile)->i6_0.lookStream.s),
                &((*corefile)->i6_0.allocImp), &((*corefile)->i6_0.allocTempImp));
        if (res == SZ_OK) {
            // arrays should be the size of the number of files in the corefile
            (*corefile)->i6_0.file_names = (const char**)CMAPL_ZALLOC((uint64_t)(*corefile)->i6_0.db.db.NumFiles * sizeof(const char*));
            (*corefile)->i6_0.dir_names = (const char**)CMAPL_ZALLOC((uint64_t)(*corefile)->i6_0.db.db.NumFiles * sizeof(const char*));

            // create temporary directory to extract to
            // TODO: might want this to include the process ID and user ID in the name in the future
#ifdef _WIN32
            // initialize random number length and seed
            int r_len = 12;
            srand(time(NULL));

            // gets the temporary directory in Windows
            char temp_dir[MAX_PATH];
            GetTempPath(MAX_PATH, temp_dir);

            // loops until a unique directory name is found
            char temp_name[MAX_PATH + r_len + 1];
            do {
                // create a random directory name
                int r = rand();
                CMAPL_SNPRINTF(temp_name, MAX_PATH + r_len + 1, "%s%d\0", temp_dir, r);
            } while (!CreateDirectory(temp_name, NULL));

            (*corefile)->i6_0.temp_dir = CMAPL_STRDUP(temp_name);
#else
            char name[] = "/tmp/cmapl_core_tempXXXXXX";
            char* temp_name = mkdtemp(name);
            (*corefile)->i6_0.temp_dir = CMAPL_STRDUP(temp_name);
#endif

            UInt32 i;
            for (i = 0; i < (*corefile)->i6_0.db.db.NumFiles; i++) {
                const CSzFileItem *f = (*corefile)->i6_0.db.db.Files + i;

                if (f->IsDir) {
                    // save all the directories in the corefile that contain archive files
                    err = cmapl_add_root_directory_v6_0((*corefile), f->Name);
                    if (err != CMAPL_SUCCESS) {
                        cmapl_free_corefile((*corefile));
                        return err;
                    }
                } else {
                    // only keep track of .meta files
                    if (NULL == CMAPL_STRSTR(f->Name, CMAPL_META_EXT)) {
                        continue;
                    }

                    // check if the file is at root level in the corefile
                    if (NULL == CMAPL_STRSTR(f->Name, "/")) {
                        // get generic archive file name
                        uint32_t file_name_len = (uint32_t)CMAPL_STRLEN(f->Name) - CMAPL_META_EXT_LEN;
                        uint32_t temp_len = (uint32_t)CMAPL_STRLEN((*corefile)->i6_0.temp_dir) + 1;
                        char* gen_file_name = (char*)CMAPL_ZALLOC(temp_len + file_name_len + 1);
                        if (gen_file_name == NULL) {
                            cmapl_free_corefile((*corefile));
                            return err;
                        }
                        CMAPL_SNPRINTF(gen_file_name, temp_len + file_name_len + 1, "%s/%s", (*corefile)->i6_0.temp_dir, f->Name);

                        (*corefile)->i6_0.file_names[(*corefile)->i6_0.file_index] = gen_file_name;
                        (*corefile)->i6_0.num_files++;
                        (*corefile)->i6_0.file_index++;
                    } else {
                        // file is in a directory so save the root directory name
                        err = cmapl_add_root_directory_v6_0((*corefile), f->Name);
                        if (err != CMAPL_SUCCESS) {
                            cmapl_free_corefile((*corefile));
                            return err;
                        }
                    }
                }
            }
        } else {
            cmapl_free_corefile((*corefile));
            return CMAPL_PARSING_ERROR;
        }

        // return error if corefile is empty??
        if ((*corefile)->i6_0.num_files == 0 && (*corefile)->i6_0.num_dirs == 0) {
            cmapl_free_corefile((*corefile));
            return CMAPL_FILE_NOT_FOUND;
        }

        //reset the file and directory indexes
        (*corefile)->i6_0.file_index = 0;
        (*corefile)->i6_0.dir_index = 0;

        return CMAPL_SUCCESS;
}

//////////////////////////////////////////////
// The following are internal helper functions
//////////////////////////////////////////////

cmapl_error_t
cmapl_extract_file_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        uint32_t file_index)
{
        // get file name without the temp directory prefix
        const char* file_name = CMAPL_STRRCHR(corefile->i6_0.file_names[file_index], '/');
        file_name++;

        UInt32 i;
        SRes res;
        UInt32 blockIndex = 0xFFFFFFFF;
        Byte *outBuffer = 0;
        size_t outBufferSize = 0;

        // extract each file associated with this archive file
        for (i = 0; i < corefile->i6_0.db.db.NumFiles; i++) {
                size_t offset;
                size_t outSizeProcessed;
                const CSzFileItem *f = corefile->i6_0.db.db.Files + i;

                // checks if this is a data or metadata file that belongs to this archive file
                if (CMAPL_STRSTR(f->Name, file_name) == NULL) {
                        continue;
                }

                // initialize the extraction of this archive file
                res = SzAr_Extract(&(corefile->i6_0.db), &(corefile->i6_0.lookStream.s), i,
                        &blockIndex, &outBuffer, &outBufferSize,
                        &offset, &outSizeProcessed,
                        &(corefile->i6_0.allocImp), &(corefile->i6_0.allocTempImp));
                if (res != SZ_OK) {
                        return CMAPL_PARSING_ERROR;
                }

                CSzFile outFile;
                size_t processedSize;

                // open the file to extract data to
                uint32_t temp_len = (uint32_t)CMAPL_STRLEN(corefile->i6_0.temp_dir) + 1;
                uint32_t file_len = (uint32_t)CMAPL_STRLEN(f->Name);
                char* full_name = (char*)CMAPL_ZALLOC(temp_len + file_len + 1);
                if (full_name == NULL) {
                        return CMAPL_NULL_POINTER;
                }
                CMAPL_SNPRINTF(full_name, temp_len + file_len + 1, "%s/%s", corefile->i6_0.temp_dir, f->Name);
                if (OutFile_Open(&outFile, full_name)) {
                        CMAPL_FREE(full_name);
                        IAlloc_Free(&corefile->i6_0.allocImp, outBuffer);
                        return CMAPL_FILE_NOT_FOUND;
                }

                // write extracted data to the archive file
                processedSize = outSizeProcessed;
                if (File_Write(&outFile, outBuffer + offset, &processedSize) != 0 ||
                        processedSize != outSizeProcessed) {
                        CMAPL_FREE(full_name);
                        IAlloc_Free(&corefile->i6_0.allocImp, outBuffer);
                        return CMAPL_WRITE_FAILED;
                }

                // close the extracted file
                if (File_Close(&outFile)) {
                        CMAPL_FREE(full_name);
                        IAlloc_Free(&corefile->i6_0.allocImp, outBuffer);
                        return CMAPL_CLOSE_FILE_FAILED;
                }

                CMAPL_FREE(full_name);
        }

        IAlloc_Free(&corefile->i6_0.allocImp, outBuffer);
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_remove_current_file_v6_0(
                        cmapl_corefile_handle_t * corefile)
{
        if (corefile == NULL) {
                return CMAPL_NULL_POINTER;
        }

        cmapl_error_t err = CMAPL_SUCCESS;
        if (corefile->i6_0.current_file != NULL) {
                err = cmapl_close_data_file(corefile->i6_0.current_file);
                if (err != CMAPL_SUCCESS) {
                        return err;
                }
                corefile->i6_0.current_file = NULL;

                // delete extracted files
                err = cmapl_unlink(corefile->i6_0.file_names[corefile->i6_0.file_index]);
                if (err != CMAPL_SUCCESS) {
                        return err;
                }
        }

        return err;
}

cmapl_error_t
cmapl_extract_directory_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        uint32_t dir_index)
{
        // TODO: not implemented until corefile contains directories
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_remove_current_directory_v6_0(
                        cmapl_corefile_handle_t * corefile)
{
        // TODO: not implemented until corefile contains directories
        return CMAPL_SUCCESS;
}

// TODO: This function will be used after directories are added to the corefiles. It'll
//       be used to store and root directories that are in a corefile.
cmapl_error_t
cmapl_add_root_directory_v6_0(
                        cmapl_corefile_handle_t * corefile,
                        char* dir_path)
{
        if (corefile == NULL || dir_path == NULL) {
                return CMAPL_NULL_POINTER;
        }

        char* first = CMAPL_STRCHR(dir_path, '/');
        uint32_t str_size = 0;
        if (first == NULL) {
                str_size = (uint32_t)CMAPL_STRLEN(dir_path) + 1;
        } else {
                str_size = (uint32_t)(first - dir_path + 1);
        }

        uint32_t temp_size = (uint32_t)CMAPL_STRLEN(corefile->i6_0.temp_dir) + 1;

        char* root_dir = (char*)CMAPL_ZALLOC(sizeof(char) * (temp_size + str_size + 1));
        if (root_dir == NULL) {
                return CMAPL_NULL_POINTER;
        }
        CMAPL_SNPRINTF(root_dir, temp_size + str_size + 1, "%s/%s", corefile->i6_0.temp_dir, dir_path);
        root_dir[temp_size + str_size - 1] = '/';
        root_dir[temp_size + str_size] = '\0';

        // check if directory name has been saved yet
        uint32_t i;
        for (i = 0; i < corefile->i6_0.num_dirs; i++) {
                if (CMAPL_STRCMP(corefile->i6_0.dir_names[i], root_dir) == 0) {
                        CMAPL_FREE(root_dir);
                        return CMAPL_SUCCESS;
                }
        }

        corefile->i6_0.dir_names[corefile->i6_0.dir_index] = root_dir;
        corefile->i6_0.num_dirs++;
        corefile->i6_0.dir_index++;

        return CMAPL_SUCCESS;
}

#endif /* _CMAPL_USER_SPACE_ */
