/**
 * The CMAPL directory-level function definitions. These functions provide all the
 * functionality to look-up and access archive files within directories.
 *
 */

#include "cmapl_common_defs.h"
#include "cmapl_int_defs.h"
#include "cmapl_read.h"
#include "cmapl_read_int.h"
#include "cmapl_write.h"
#include "cmapl_write_int.h"
#include "cmapl_directory.h"
#include "cmapl_directory_int.h"

cmapl_error_t
cmapl_close_directory(
                        cmapl_directory_handle_t * dir)
{
        if (dir == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (!cmapl_is_version(dir, 6, 0)){
                return CMAPL_INVALID_VERSION;
        }

	cmapl_free_directory(dir);
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_find_and_open_data_file(
                        cmapl_directory_handle_t * dir,
                        uint64_t timestamp,
                        cmapl_file_handle_t ** data_file)
{
        if (dir == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (!cmapl_is_archive_directory_v6_0(dir)) {
                return CMAPL_INVALID_FILE_TYPE;
        }

        if (cmapl_is_version(dir, 6, 0)){
                return cmapl_find_and_open_data_file_v6_0(
                                        dir,
                                        timestamp,
                                        data_file);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_find_and_open_data_file_v6_0(
                        cmapl_directory_handle_t * dir,
                        uint64_t timestamp,
                        cmapl_file_handle_t ** data_file)
{
	cmapl_error_t error = CMAPL_SUCCESS;

	if (timestamp < dir->e6_0.start_time || timestamp > dir->e6_0.end_time){
		error = CMAPL_TIMESTAMP_OUT_OF_RANGE;
		goto err;
	}

	// TODO: should update to binary search
	uint32_t file_index;
	for (file_index = 0; file_index < dir->i6_0.num_files; file_index++){
		// check if timestamp falls within some file's time range
		if (timestamp >= dir->i6_0.file_infos[file_index].start_time &&
			timestamp <= dir->i6_0.file_infos[file_index].end_time)
		{
			const char* file_path = dir->i6_0.file_infos[file_index].file_path;

			const char* meta_path;
			const char* data_path;
			error = cmapl_open_data_file(file_path, "rb", data_file, &meta_path, &data_path);
			if (error != CMAPL_SUCCESS){
				goto err;
			}

			dir->i6_0.file_index = file_index + 1;
			return error;
		}
	}
        error = CMAPL_FILE_NOT_FOUND;

err:
	*data_file = NULL;
        return error;
}

cmapl_error_t
cmapl_find_and_open_newest_data_file(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file)
{
        if (dir == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (!cmapl_is_archive_directory_v6_0(dir)) {
                return CMAPL_INVALID_FILE_TYPE;
        }

        if (cmapl_is_version(dir, 6, 0)){
                return cmapl_find_and_open_newest_data_file_v6_0(
                                                dir,
                                                data_file);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_find_and_open_newest_data_file_v6_0(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file)
{
        cmapl_error_t error = CMAPL_SUCCESS;

	// refresh directory for any new files
        error = cmapl_refresh_directory_v6_0(dir);
        if (error != CMAPL_SUCCESS){
                goto err;
        }

        // check again if any files exist
        if (dir->i6_0.num_files == 0){
                goto err;
        }

        const char* file_path = dir->i6_0.file_infos[dir->i6_0.num_files - 1].file_path;

        cmapl_file_handle_t* fh;
        const char* meta_path;
        const char* data_path;
        error = cmapl_open_data_file(file_path, "rb", &fh, &meta_path, &data_path);
        if (error != CMAPL_SUCCESS){
		goto err;
        }

	dir->i6_0.file_index = dir->i6_0.num_files;
        *data_file = fh;
        return error;

err:
        *data_file = NULL;
        return error;
}

cmapl_error_t
cmapl_find_and_open_oldest_data_file(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file)
{
        if (dir == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (!cmapl_is_archive_directory_v6_0(dir)) {
                return CMAPL_INVALID_FILE_TYPE;
        }

        if (cmapl_is_version(dir, 6, 0)){
                return cmapl_find_and_open_oldest_data_file_v6_0(
                                                dir,
                                                data_file);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_find_and_open_oldest_data_file_v6_0(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file)
{
	cmapl_error_t error = CMAPL_SUCCESS;

	if (dir->i6_0.num_files == 0){
		// refresh directory for any new files
		error = cmapl_refresh_directory_v6_0(dir);
		if (error != CMAPL_SUCCESS){
			goto err;
		}

		// check again if any files exist
		if (dir->i6_0.num_files == 0){
			goto err;
		}
	}

	const char* file_path = dir->i6_0.file_infos[0].file_path;

	cmapl_file_handle_t* fh;
	const char* meta_path;
	const char* data_path;
	error = cmapl_open_data_file(file_path, "rb", &fh, &meta_path, &data_path);
	if (error != CMAPL_SUCCESS){
		// file failed to open, possibly because it no longer exists
                if (error == CMAPL_NULL_POINTER){
                        // refresh directory for any new files
                        error = cmapl_refresh_directory_v6_0(dir);
                        if (error != CMAPL_SUCCESS){
                                goto err;
                        }

                        // retry opening oldest file with updated directory
                        file_path = dir->i6_0.file_infos[0].file_path;
                        error = cmapl_open_data_file(file_path, "rb", &fh, &meta_path, &data_path);
                        if (error != CMAPL_SUCCESS){
                                goto err;
                        }
                }
	}

	dir->i6_0.file_index = 1;
	*data_file = fh;
        return error;

err:
	*data_file = NULL;
        return error;
}

cmapl_error_t
cmapl_find_and_open_next_data_file(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file)
{
        if (dir == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (!cmapl_is_archive_directory_v6_0(dir)) {
                return CMAPL_INVALID_FILE_TYPE;
        }

        if (cmapl_is_version(dir, 6, 0)){
                return cmapl_find_and_open_next_data_file_v6_0(
                                                dir,
                                                data_file);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_find_and_open_next_data_file_v6_0(
                        cmapl_directory_handle_t * dir,
                        cmapl_file_handle_t ** data_file)
{
	cmapl_error_t error = CMAPL_SUCCESS;
	uint32_t new_index = dir->i6_0.file_index;

	// check if there are no more files to open
        if (dir->i6_0.num_files <= new_index){
		// refresh directory for any new files
                error = cmapl_refresh_directory_v6_0(dir);
                if (error != CMAPL_SUCCESS){
                        goto err;
                }

		// check again if there are any new files to open
		new_index = dir->i6_0.file_index;
                if (dir->i6_0.num_files <= new_index){
			error = CMAPL_FILE_NOT_FOUND;
                        goto err;
                }
        }

        const char* file_path = dir->i6_0.file_infos[new_index].file_path;

        cmapl_file_handle_t* fh;
        const char* meta_path;
        const char* data_path;
        error = cmapl_open_data_file(file_path, "rb", &fh, &meta_path, &data_path);
        if (error != CMAPL_SUCCESS){
		// file failed to open, possibly because it no longer exists
		if (error == CMAPL_NULL_POINTER){
			// refresh directory for any new files
			error = cmapl_refresh_directory_v6_0(dir);
			if (error != CMAPL_SUCCESS){
				goto err;
			}

			// retry opening next file with updated directory
			new_index = dir->i6_0.file_index;
			file_path = dir->i6_0.file_infos[new_index].file_path;
			error = cmapl_open_data_file(file_path, "rb", &fh, &meta_path, &data_path);
			if (error != CMAPL_SUCCESS){
				goto err;
			}
		}
        }

        *data_file = fh;
	dir->i6_0.file_index = new_index + 1;
        return error;

err:
        *data_file = NULL;
        return error;
}

cmapl_error_t
cmapl_get_file_infos(
                        cmapl_directory_handle_t * dir,
                        uint32_t * num_files,
                        void ** file_infos)
{
	if (dir == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(dir, 6, 0)){
                return cmapl_get_file_infos_v6_0(
                                        dir,
					num_files,
                                        (cmapl_file_def_t_v6_0**)file_infos);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_file_infos_v6_0(
                        cmapl_directory_handle_t * dir,
                        uint32_t * num_files,
                        cmapl_file_def_t_v6_0 ** file_infos)
{
	cmapl_error_t error = CMAPL_SUCCESS;

	// refresh directory first
	error = cmapl_refresh_directory_v6_0(dir);
        if (error != CMAPL_SUCCESS){
		*num_files = 0;
		*file_infos = NULL;
                return error;
	}

	// assign parameters
	*num_files = dir->i6_0.num_files;
	*file_infos = dir->i6_0.file_infos;
	return error;
}

cmapl_error_t
cmapl_open_directory(
                        void * dir_def,
                        cmapl_directory_handle_t ** dir,
						_Bool not_used)
{
        if (dir_def == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(dir_def, 6, 0)){
                return cmapl_open_directory_v6_0(
                                        (cmapl_directory_def_t_v6_0*)dir_def,
                                        dir,
										not_used);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_open_directory_v6_0(
                        cmapl_directory_def_t_v6_0 * dir_def,
                        cmapl_directory_handle_t ** dir,
						_Bool not_used)
{
	// validate directory definition struct
	cmapl_error_t error = cmapl_validate_directory_def_v6_0(dir_def);
	if (error != CMAPL_SUCCESS){
		return error;
	}

	// allocate new directory handle
	(*dir) = (cmapl_directory_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_directory_handle_t));
	if ((*dir) == NULL){
		return CMAPL_NULL_POINTER;
	}

	// version info
	(*dir)->major_v = 6;
	(*dir)->minor_v = 0;

	// copy all external directory metadata
	error = cmapl_copy_directory_def_v6_0(dir_def, &((*dir)->e6_0));
	if (error != CMAPL_SUCCESS){
		goto err;
	}

	// initialize internal data
        (*dir)->i6_0.num_files = 0;             // number of files in the directory
        (*dir)->i6_0.file_infos = NULL;         // all the metadata of each file
        (*dir)->i6_0.file_index = 0;            // iterator index of all the files

	// adjust directory path to not end with '/'
	uint32_t dir_path_len = (uint32_t)CMAPL_STRLEN((*dir)->e6_0.directory_path);
        if ((*dir)->e6_0.directory_path[dir_path_len - 1] == '/'){
                (*dir)->e6_0.directory_path[dir_path_len - 1] = '\0';
        }

	// get the initial file list of the directory with the refresh funtion
	error = cmapl_refresh_directory_v6_0(*dir);
	if (error != CMAPL_SUCCESS){
		goto err;
	}

        return CMAPL_SUCCESS;

err:
	cmapl_free_directory(*dir);
	*dir = NULL;
	return error;
}

//////////////////////////////////////////////
// The following are internal helper functions
//////////////////////////////////////////////

cmapl_error_t
cmapl_get_file_path_v6_0(
			cmapl_directory_handle_t * dir,
			char * file_name,
			char ** file_path)
{
	if (dir == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        // initialize path lengths
	uint32_t dir_path_len = (uint32_t)CMAPL_STRLEN(dir->e6_0.directory_path);
        uint32_t file_path_len = dir_path_len + 1;

        // increase file path length based on file type
        if (dir->i6_0.file_type & CMAPL_PERFSTAT_FILE) {
                // add the whole file length if perfstat file
                file_path_len += (uint32_t)CMAPL_STRLEN(file_name);
        } else if (dir->i6_0.file_type & CMAPL_ARCHIVE_FILE) {
                char* ccma_ext_ptr = CMAPL_STRSTR(file_name, CMAPL_CCMA_EXT);
                if (ccma_ext_ptr == NULL) {
                        return CMAPL_INVALID_FILE_TYPE;
                }
                // add the generic archiver file length if archiver file
                // add length up to the ".ccma" extension plus the length of the extension
                uint32_t file_len = (uint32_t)(ccma_ext_ptr - file_name) + CMAPL_CCMA_EXT_LEN;
                file_path_len += file_len;
        } else {
                return CMAPL_INVALID_FILE_TYPE;
        }

        // get the full path of the file
        *file_path = (char*)CMAPL_ZALLOC(file_path_len + 1);
        if (*file_path == NULL){
                return CMAPL_NULL_POINTER;
        }
        CMAPL_SNPRINTF((*file_path), file_path_len + 1, "%s/%s", dir->e6_0.directory_path, file_name);

	return CMAPL_SUCCESS;
}

cmapl_file_type_t
cmapl_get_file_type_v6_0(
                        char * file_name)
{
        if (NULL != CMAPL_STRSTR(file_name, CMAPL_META_EXT)) {
                return CMAPL_ARCHIVE_FILE;
        } else if (NULL != CMAPL_STRSTR(file_name, CMAPL_PERFSTAT_EXT)) {
                return CMAPL_PERFSTAT_FILE;
        } else {
                return CMAPL_UNKNOWN_FILE_TYPE;
        }
}

cmapl_error_t
cmapl_increase_file_array_size_v6_0(
			cmapl_directory_handle_t * dir)
{
	if (dir == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	// increase array sizes if there's not enough space for this file
        uint32_t new_size = 2 * dir->i6_0.array_size;

        cmapl_file_def_t_v6_0* new_file_infos = (cmapl_file_def_t_v6_0*)CMAPL_ZALLOC(
							new_size * sizeof(cmapl_file_def_t_v6_0));
        if (new_file_infos == NULL){
		return CMAPL_NULL_POINTER;
        }
        CMAPL_MEMCPY(new_file_infos, dir->i6_0.file_infos, dir->i6_0.array_size * sizeof(char*));
        CMAPL_FREE(dir->i6_0.file_infos);

        dir->i6_0.file_infos = new_file_infos;
        dir->i6_0.array_size = new_size;

	return CMAPL_SUCCESS;
}

bool
cmapl_is_archive_directory_v6_0(
                        cmapl_directory_handle_t * dir)
{
        if (dir->i6_0.file_type == CMAPL_ARCHIVE_FILE) {
                return true;
        }
        return false;
}

cmapl_error_t
cmapl_refresh_directory_v6_0(
                        cmapl_directory_handle_t * dir)
{
	if (dir == NULL){
		return CMAPL_INVALID_ARGUMENTS;
	}

	cmapl_error_t error = CMAPL_SUCCESS;

	// check if there are any existing file infos
	bool is_last_file = false;
	uint32_t current_index = 0;
	if (dir->i6_0.file_infos != NULL){
		// save the file index of the current file
		if (dir->i6_0.num_files > dir->i6_0.file_index){
			current_index = dir->i6_0.file_infos[dir->i6_0.file_index].file_index;
		} else if (dir->i6_0.num_files == dir->i6_0.file_index){
			// if index is past the last file, save index of the last file read
			current_index = dir->i6_0.file_infos[dir->i6_0.file_index - 1].file_index;
			is_last_file = true;
		}

		// free old data
		uint32_t num_files = dir->i6_0.num_files;
		uint32_t i;
		for (i = 0; i < num_files; i++){
			CMAPL_FREE(dir->i6_0.file_infos[i].file_path);
			CMAPL_FREE(dir->i6_0.file_infos[i].meta_path);
			CMAPL_FREE(dir->i6_0.file_infos[i].data_path);
		}
		CMAPL_FREE(dir->i6_0.file_infos);
	}

	// initialize internal arrays
	dir->i6_0.num_files = 0;
	dir->i6_0.file_index = 0;
        dir->i6_0.array_size = CMAPL_INITIAL_NUM_FILES;
        dir->i6_0.file_infos = (cmapl_file_def_t_v6_0*)CMAPL_ZALLOC(
                                        CMAPL_INITIAL_NUM_FILES * sizeof(cmapl_file_def_t_v6_0));

	// open directory
        CMAPL_DIRENT* entry;
        CMAPL_DIR* dh = CMAPL_OPENDIR(dir->e6_0.directory_path);
        if (dh == NULL){
                return CMAPL_NULL_POINTER;
        }

	// loop through all the files in the directory
        char* file_path = NULL;
        uint32_t file_index = 0;
        bool first_loop = true;
        while(1){
                // get next file in directory
                entry = CMAPL_READDIR(dh);
                if (!entry) {
                        break;
                }

                cmapl_file_type_t file_type_flag = cmapl_get_file_type_v6_0(entry->d_name);

                // only proceed if the entry is a .meta or .perfstat file, and isn't the current file being written to;
                // the data files that are paired with the meta files are assumed to be in the directory; the validation
                // check to make sure if the data file exists is made when the file is attempted to be opened
                if ((entry->d_type != DT_REG && entry->d_type != DT_UNKNOWN) ||
                        ((file_type_flag & (CMAPL_ARCHIVE_FILE | CMAPL_PERFSTAT_FILE)) == 0) ||
                        NULL != CMAPL_STRSTR(entry->d_name, CMAPL_EMPTY_FILE) ||
                        NULL != CMAPL_STRSTR(entry->d_name, CMAPL_CURRENT_FILE))
                {
                        continue;
                }

                // checks to make sure directory only contains either archiver or perfstat files, but not both
                if (first_loop) {
                        dir->i6_0.file_type = file_type_flag;
                        first_loop = false;
                }
                if ((dir->i6_0.file_type & file_type_flag) == 0) {
                        error = CMAPL_INCONSISTENT_FILE_TYPES;
                        goto err;
                }

                // get the file name
                char* file_name = entry->d_name;

		// TODO: Can do this when you open file. Better? What if you open a file multiple times?
		// get the whole path of each file
		error = cmapl_get_file_path_v6_0(dir, file_name, &file_path);
		if (error != CMAPL_SUCCESS){
			goto err;
                }

		// get file index and start and end time of the file from the filename
		uint32_t f_index;
		uint64_t longest_period;
                uint64_t utc_end_time;
                uint64_t utc_start_time;
                error = cmapl_parse_filename(file_name, &f_index, &longest_period, &utc_start_time, &utc_end_time);
		// skip file if the file format is incorrect
                if (error == CMAPL_FILENAME_FORMAT_ERROR){
			CMAPL_FREE(file_path);
                        continue;
                } else if (error != CMAPL_SUCCESS){
                        goto err;
                }

		// check if archive file contains samples at the specified period (ignore perfstat files)
                if ((file_type_flag & CMAPL_ARCHIVE_FILE) && longest_period < dir->e6_0.period) {
		        CMAPL_FREE(file_path);
                        continue;
                }

                // check if file is not within the specified time range
                if (utc_end_time < dir->e6_0.start_time || utc_start_time > dir->e6_0.end_time){
                        CMAPL_FREE(file_path);
                        continue;
                }

		// increase array sizes if there's not enough space for this file
                if (file_index >= dir->i6_0.array_size){
			error = cmapl_increase_file_array_size_v6_0(dir);
			if (error != CMAPL_SUCCESS){
				goto err;
			}
                }

		// add file information to file info's array
		error = cmapl_update_file_info_v6_0(dir, file_index, file_path, f_index, utc_start_time, utc_end_time);
		if (error != CMAPL_SUCCESS){
		        goto err;
		}
                file_index++;
	}

	// update actual number of data files found
        dir->i6_0.num_files = file_index;
        dir->i6_0.file_index = 0;

        // check if no files were found
        if (file_index == 0){
                int rt_err = CMAPL_CLOSEDIR(dh);
                if (rt_err != 0){
                        return CMAPL_CLOSE_DIRECTORY_FAILED;
                }
                return CMAPL_FILE_NOT_FOUND;
        }

	// sort arrays by start time (insertion sort)
	error = cmapl_sort_files_v6_0(dir);
	if (error != CMAPL_SUCCESS){
		return error;
	}

	int rt_err = CMAPL_CLOSEDIR(dh);
        if (rt_err != 0){
                return CMAPL_CLOSE_DIRECTORY_FAILED;
        }

	// update the directory's file index to point to the same current
	// file if it still exists; otherwise, point to the first file
	if (current_index != 0){
		uint32_t i;
		for (i = 0; i < file_index; i++){
			if(current_index == dir->i6_0.file_infos[i].file_index){
				dir->i6_0.file_index = i;

				// reset index to the file after the previous last file
				if (is_last_file){
					dir->i6_0.file_index++;
				}
				goto finished_updating;
			}
		}
		// same current file no longer exists; start index at first file
		dir->i6_0.file_index = 0;
	} else {
		// file list was empty before so start index at first file
		dir->i6_0.file_index = 0;
	}
finished_updating:

	return CMAPL_SUCCESS;

err:
        CMAPL_FREE(file_path);
	(void)CMAPL_CLOSEDIR(dh);
        return error;
}

cmapl_error_t
cmapl_sort_files_v6_0(
			cmapl_directory_handle_t * dir)
{
	if (dir == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        // sort archive files
	uint32_t num_files = dir->i6_0.num_files;
	uint32_t i;
	for (i = 1; i < num_files; i++){
                uint32_t j;
                for (j = i; j > 0 && dir->i6_0.file_infos[j].start_time < dir->i6_0.file_infos[j-1].start_time; j--){
                        // swap current index with previous index
                        cmapl_file_def_t_v6_0 temp_file_info = dir->i6_0.file_infos[j-1];
                        dir->i6_0.file_infos[j-1] = dir->i6_0.file_infos[j];
                        dir->i6_0.file_infos[j] = temp_file_info;
                }
        }

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_update_file_info_v6_0(
                        cmapl_directory_handle_t * dir,
			uint32_t file_index,
			char * file_path,
			uint32_t index,
			uint64_t start_time,
			uint64_t end_time)
{
	if (dir == NULL){
		return CMAPL_INVALID_ARGUMENTS;
	}

	// add file information to file_infos array
        dir->i6_0.file_infos[file_index].file_path = file_path;
        dir->i6_0.file_infos[file_index].file_index = index;
        dir->i6_0.file_infos[file_index].start_time = start_time;
        dir->i6_0.file_infos[file_index].end_time = end_time;
        dir->i6_0.file_infos[file_index].file_type = dir->i6_0.file_type;
        dir->i6_0.file_infos[file_index].meta_path = NULL;
        dir->i6_0.file_infos[file_index].data_path = NULL;

        if (dir->i6_0.file_type & CMAPL_ARCHIVE_FILE) {
                // add meta and data file paths
                char* metadata_file_path;
                CMAPL_ZALLOC_META_NAME(file_path, metadata_file_path);
                if (metadata_file_path == NULL){
                        return CMAPL_NULL_POINTER;
                }
                dir->i6_0.file_infos[file_index].meta_path = metadata_file_path;

                char* data_file_path;
                CMAPL_ZALLOC_GZ_NAME(file_path, data_file_path);
                if (data_file_path == NULL){
                        return CMAPL_NULL_POINTER;
                }
                dir->i6_0.file_infos[file_index].data_path = data_file_path;
        }

	return CMAPL_SUCCESS;
}
