/**
 * The CMAPL read-side function definitions. These functions provide all the
 * functionality to read, parse, and query an archive files.
 *
 */

#include "cmapl_common_defs.h"
#include "cmapl_int_defs.h"
#include "cmapl_read.h"
#include "cmapl_read_int.h"
#include "cmapl_write.h"
#include "cmapl_write_int.h"
#include <libxml/parser.h>
#include <libxml/tree.h>

#ifdef _CMAPL_FIJI_
// global fiji variables
bool cmapl_is_read_fail = false;
bool cmapl_is_seek_fail = false;
bool cmapl_is_write_fail = false;
#endif /* _CMAPL_FIJI_ */

// @TODO: rename all handle parameter names to <object>_h instead of just <object>
cmapl_error_t
cmapl_find_counter_by_id(
                        cmapl_object_handle_t * object,
                        uint32_t counter_id,
                        cmapl_counter_handle_t ** counter)
{
        if (object == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_find_counter_by_id_v6_0(object, counter_id, counter, true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_find_counter_by_id_v6_0(
                        cmapl_object_handle_t * object,
                        uint32_t counter_id,
                        cmapl_counter_handle_t ** counter,
			bool update_iter)
{
        cmapl_counter_handle_t* cnt = object->i6_0.cnt_list;
        while(cnt != NULL){
                if(cnt->e6_0.cnt_id == counter_id){
                        *counter = cnt;
			if (update_iter) {
				object->i6_0.current_cnt = cnt;
			}
                        return CMAPL_SUCCESS;
                }
                cnt = cnt->i6_0.next_cnt_by_id;
        }
	(*counter) = NULL;
	if (update_iter) {
		object->i6_0.current_cnt = NULL;
	}
        return CMAPL_COUNTER_NOT_FOUND;

}

cmapl_error_t
cmapl_find_counter_by_name(
                        cmapl_object_handle_t * object,
                        const char * counter_name,
                        cmapl_counter_handle_t ** counter)
{
        if (object == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_find_counter_by_name_v6_0(object, counter_name, counter, true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_find_counter_by_name_v6_0(
                        cmapl_object_handle_t * object,
                        const char * counter_name,
                        cmapl_counter_handle_t ** counter,
			bool update_iter)
{
        cmapl_counter_handle_t* cnt = object->i6_0.cnt_list;
        while(cnt != NULL){
                if(CMAPL_STRCMP(cnt->e6_0.name, counter_name) == 0){
                        *counter = cnt;
			if (update_iter) {
				object->i6_0.current_cnt = cnt;
			}
                        return CMAPL_SUCCESS;
                }
                cnt = cnt->i6_0.next_cnt_by_id;
        }
	(*counter) = NULL;
	if (update_iter) {
		object->i6_0.current_cnt = NULL;
	}
        return CMAPL_COUNTER_NOT_FOUND;

}

cmapl_error_t
cmapl_find_counter_data(
                        cmapl_instance_handle_t * instance,
			const char * counter_name,
                        cmapl_counter_handle_t ** counter,
                        cmapl_counter_data_handle_t ** counter_data)
{
        if (instance == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(instance, 6, 0)){
                return cmapl_find_counter_data_v6_0(
                                        instance,
					counter_name,
                                        counter,
                                        counter_data,
					true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_find_counter_data_v6_0(
                        cmapl_instance_handle_t * instance,
			const char * counter_name,
                        cmapl_counter_handle_t ** counter,
                        cmapl_counter_data_handle_t ** counter_data,
			bool update_iter)
{
	cmapl_error_t err = CMAPL_SUCCESS;

	// if the iterator is not supposed to be updated, save the next counter data index
	// so it can be set again later
	uint32_t saved_index = 0;
	if (!update_iter) {
		saved_index = instance->i6_0.next_counter_index;
	}
	err = cmapl_first_counter_data_v6_0(instance, counter, counter_data, true);
	if (err != CMAPL_SUCCESS){
		return err;
	}

	while ((*counter) != NULL){
		if (CMAPL_STRCMP((*counter)->e6_0.name, counter_name) == 0){
			// set saved counter data index
			if (!update_iter) {
				instance->i6_0.next_counter_index = saved_index;
			}
			return err;
		}
		err = cmapl_next_counter_data_v6_0(instance, counter, counter_data, true);
		if (err != CMAPL_SUCCESS){
			return err;
		}
	}

	// set saved counter data index
	if (!update_iter) {
		instance->i6_0.next_counter_index = saved_index;
	}
	(*counter) = NULL;
	(*counter_data) = NULL;
        return CMAPL_COUNTER_NOT_FOUND;
}

cmapl_error_t
cmapl_find_instance(
                        cmapl_timestamp_handle_t * timestamp,
                        const char ** key_cnt_names,
			void * key_cnt_values,
                        uint32_t num_key_cnts,
                        cmapl_instance_handle_t ** instance)
{
	if (timestamp == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(timestamp, 6, 0)){
                return cmapl_find_instance_v6_0(
                                        timestamp,
                                        key_cnt_names,
					(cmapl_counter_data_t_v6_0*)key_cnt_values,
					num_key_cnts,
                                        instance,
					true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_find_instance_v6_0(
                        cmapl_timestamp_handle_t * timestamp,
                        const char ** key_cnt_names,
			cmapl_counter_data_t_v6_0 * key_cnt_values,
                        uint32_t num_key_cnts,
                        cmapl_instance_handle_t ** instance,
			bool update_iter)
{
	cmapl_error_t err = CMAPL_INSTANCE_NOT_FOUND;
	cmapl_query_handle_t* query = timestamp->i6_0.query;
	cmapl_object_handle_t* object = query->i6_0.object;
	if (num_key_cnts != object->i6_0.num_key_cnts_sampled){
		return CMAPL_INVALID_ARGUMENTS;
	}

	// if the iterator is not supposed to be updated, save the next instance index
        // so it can be set again later
	uint64_t saved_index = 0;
	if (!update_iter) {
		saved_index = timestamp->i6_0.next_instance;
	}

	err = cmapl_first_instance_v6_0(timestamp, instance, true);
	if (err != CMAPL_SUCCESS){
		return err;
	}

	// get all key counter handles from names
	cmapl_counter_handle_t** key_cnts = (cmapl_counter_handle_t**)CMAPL_ZALLOC(num_key_cnts * sizeof(cmapl_counter_handle_t*));
	if (key_cnts == NULL){
		return CMAPL_NULL_POINTER;
	}
	uint32_t i;
	for (i = 0; i < num_key_cnts; i++){
		cmapl_counter_handle_t* key_cnt;
		err = cmapl_find_counter_by_name_v6_0(object, key_cnt_names[i], &key_cnt, false);
		if (err != CMAPL_SUCCESS){
			goto clean_up;
		}
		key_cnts[i] = key_cnt;
	}

	while ((*instance) != NULL){
		bool is_instance_equal = false;
		char* instance_buffer = (*instance)->i6_0.instance_buffer;

		// get the period associated with the first instance of the whole instance
		cmapl_period_t* period = query->i6_0.periods[0];
		err = cmapl_is_instance_equal_v6_0(key_cnts, key_cnt_values, num_key_cnts, period,
								instance_buffer, &is_instance_equal);
		if (err != CMAPL_SUCCESS){
			goto clean_up;
		}
		if (is_instance_equal == true){
			err = CMAPL_SUCCESS;
			goto clean_up;
		}
		err = cmapl_next_instance_v6_0(timestamp, instance, true);
		if (err != CMAPL_SUCCESS){
			goto clean_up;
		}
	}

        // no instance found that matches the specified values
        err = CMAPL_INSTANCE_NOT_FOUND;

clean_up:
	// set saved instance index
	if (!update_iter) {
                timestamp->i6_0.next_instance = saved_index;
        }
	CMAPL_FREE(key_cnts);
	return err;
}

cmapl_error_t
cmapl_find_object(
                        cmapl_header_handle_t * header,
                        const char * object_name,
                        cmapl_object_handle_t ** object)
{
        if (header == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(header, 6, 0)){
                return cmapl_find_object_v6_0(header, object_name, object, true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_find_object_v6_0(
                        cmapl_header_handle_t * header,
                        const char * object_name,
                        cmapl_object_handle_t ** object,
			bool update_iter)
{
	cmapl_object_handle_t* obj = header->i6_0.object_list;
        while (obj != NULL){
		if (CMAPL_STRCMP(obj->e6_0.name, object_name) == 0){
			(*object) = obj;
			if (update_iter) {
				header->i6_0.current_object = obj;
			}
			return CMAPL_SUCCESS;
		}
		obj = obj->i6_0.next_object;
        }
	(*object) = NULL;
	if (update_iter) {
		header->i6_0.current_object = NULL;
	}
        return CMAPL_OBJECT_NOT_FOUND;
}

cmapl_error_t
cmapl_first_counter(
                        cmapl_object_handle_t * object,
                        cmapl_counter_handle_t ** counter)
{
	if (object == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_first_counter_v6_0(object, counter, true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_counter_v6_0(
                        cmapl_object_handle_t * object,
                        cmapl_counter_handle_t ** counter,
			bool update_iter)
{
	(*counter) = object->i6_0.cnt_list;
	if (update_iter) {
		object->i6_0.current_cnt = (*counter);
	}
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_first_counter_by_period(
                        cmapl_object_handle_t * object,
			uint64_t period,
                        cmapl_counter_handle_t ** counter)
{
        if (object == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_first_counter_by_period_v6_0(object, period, counter, true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_counter_by_period_v6_0(
                        cmapl_object_handle_t * object,
			uint64_t period,
                        cmapl_counter_handle_t ** counter,
			bool update_iter)
{
        if (object->i6_0.num_cnts_sampled == 0){
                (*counter) = NULL;
                return CMAPL_SUCCESS;
        }

	cmapl_period_t* p = object->i6_0.period_list;
        while (p != NULL){
                if (p->period == period){
                        cmapl_counter_handle_t* current_cnt = p->cnt_list;
                        if (current_cnt == NULL){
				(*counter) = NULL;
                                return CMAPL_SUCCESS;
                        }
                        (*counter) = current_cnt;
			if (update_iter) {
				p->current_cnt = current_cnt;
			}
                        return CMAPL_SUCCESS;
                }
		p = p->next_period;
        }
        (*counter) = NULL;;
        return CMAPL_PERIOD_NOT_FOUND;
}

cmapl_error_t
cmapl_first_counter_data(
                        cmapl_instance_handle_t * instance,
			cmapl_counter_handle_t ** counter,
                        cmapl_counter_data_handle_t ** counter_data)
{
	if (instance == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(instance, 6, 0)){
                return cmapl_first_counter_data_v6_0(
						instance,
						counter,
						counter_data,
						true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_counter_data_v6_0(
                        cmapl_instance_handle_t * instance,
			cmapl_counter_handle_t ** counter,
                        cmapl_counter_data_handle_t ** counter_data,
			bool update_iter)
{
	if (instance->e6_0.num_key_cnts == 0){
		(*counter) = NULL;
		(*counter_data) = NULL;
		return CMAPL_SUCCESS;
	}

	cmapl_error_t err = CMAPL_SUCCESS;
        cmapl_object_handle_t* object = NULL;
        if (instance->i6_0.timestamp != NULL) {
                // whole instance
                object = instance->i6_0.timestamp->i6_0.query->i6_0.object;
        } else {
                // partial instance
                object = instance->i6_0.record->i6_0.object;
        }

	// free last counter data
	cmapl_free_counter_data(instance->i6_0.current_counter_data);
	instance->i6_0.current_counter_data = NULL;
	if (update_iter) {
		instance->i6_0.next_counter_index = 0;
	}

        (*counter_data) = (cmapl_counter_data_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_counter_data_handle_t));
        if ((*counter_data) == NULL){
                return CMAPL_NULL_POINTER;
        }
	(*counter_data)->major_v = 6;
	(*counter_data)->minor_v = 0;

	// get first key counter
        cmapl_counter_handle_t* first_counter;
	err = cmapl_find_counter_by_name_v6_0(object, instance->e6_0.key_cnt_names[0], &first_counter, false);
	if (err != CMAPL_SUCCESS){
		cmapl_free_counter_data(*counter_data);
		return err;
	}

	(*counter_data)->e6_0 = instance->e6_0.key_cnt_values[0];
	uint32_t num_values = (*counter_data)->e6_0.num_values;
	(*counter_data)->e6_0.values = (cmapl_cnt_value_t*)CMAPL_ZALLOC(num_values * sizeof(cmapl_cnt_value_t));
	if ((*counter_data)->e6_0.values == NULL){
		cmapl_free_counter_data(*counter_data);
                return CMAPL_NULL_POINTER;
	}

	uint32_t i;
	for (i = 0; i < num_values; i++){
		(*counter_data)->e6_0.values[i] = instance->e6_0.key_cnt_values[0].values[i];
		if ((*counter_data)->e6_0.value_type == CMAPL_CNT_STRING){
			(*counter_data)->e6_0.values[i].val_str = CMAPL_STRDUP(instance->e6_0.key_cnt_values[0].values[i].val_str);
			if ((*counter_data)->e6_0.values[i].val_str == NULL){
				cmapl_free_counter_data(*counter_data);
				return CMAPL_NULL_POINTER;
			}
		}
	}
	(*counter) = first_counter;
	(*counter_data)->i6_0.instance = instance;
	(*counter_data)->i6_0.counter = first_counter;

	instance->i6_0.current_counter_data = *counter_data;
	if (update_iter) {
		instance->i6_0.next_counter_index = 1;
	}
	return err;
}

cmapl_error_t
cmapl_first_instance(
                        cmapl_timestamp_handle_t * timestamp,
                        cmapl_instance_handle_t ** instance)
{
	if (timestamp == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(timestamp, 6, 0)){
                return cmapl_first_instance_v6_0(
                                                timestamp,
                                                instance,
						true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_instance_v6_0(
                        cmapl_timestamp_handle_t * timestamp,
                        cmapl_instance_handle_t ** instance,
			bool update_iter)
{
	// free previous current instance in timestamp
	cmapl_free_instance(timestamp->i6_0.current_instance);
	timestamp->i6_0.current_instance = NULL;

	cmapl_error_t err = CMAPL_SUCCESS;
	if (timestamp->i6_0.num_instances == 0){
		(*instance) = NULL;
		return err;
	}

	uint64_t instance_size = timestamp->i6_0.size / timestamp->i6_0.num_instances;
	if (update_iter) {
		timestamp->i6_0.next_instance = 0;
	}

	(*instance) = (cmapl_instance_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_instance_handle_t));
	if ((*instance) == NULL){
		return CMAPL_NULL_POINTER;
	}
	(*instance)->major_v = 6;
	(*instance)->minor_v = 0;

	(*instance)->i6_0.instance_buffer = timestamp->i6_0.instances_buffer;
	(*instance)->i6_0.size = instance_size;
	(*instance)->i6_0.timestamp = timestamp;
        (*instance)->i6_0.record = NULL;
	if (update_iter) {
		timestamp->i6_0.next_instance++;
	}

	// get the key counter information of the instance
	err = cmapl_get_instance_key_counters_v6_0(*instance);
        if(err != CMAPL_SUCCESS){
                CMAPL_FREE(*instance);
                return err;
        }

	timestamp->i6_0.current_instance = *instance;

	return err;
}

cmapl_error_t
cmapl_first_object(
                        cmapl_header_handle_t * header,
                        cmapl_object_handle_t ** object)
{
	if (header == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(header, 6, 0)){
                return cmapl_first_object_v6_0(header, object, true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_instance_from_record(
                        cmapl_record_handle_t * record,
                        cmapl_instance_handle_t ** instance)
{
	if (record == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(record, 6, 0)){
                return cmapl_first_instance_from_record_v6_0(record, instance);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_object_v6_0(
                        cmapl_header_handle_t * header,
                        cmapl_object_handle_t ** object,
			bool update_iter)
{
	(*object) = header->i6_0.object_list;
	if (update_iter) {
		header->i6_0.current_object = (*object);
	}
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_first_timestamp(
                        cmapl_query_handle_t * query,
                        cmapl_timestamp_handle_t ** timestamp)
{
	if (query == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	if (cmapl_is_version(query, 6, 0)){
                return cmapl_first_timestamp_v6_0(
                                                query,
                                                timestamp,
						true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_record(
                    cmapl_file_handle_t * data_file,
                    cmapl_record_handle_t ** record)
{
	if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_first_record_v6_0(data_file, record);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_first_timestamp_v6_0(
                        cmapl_query_handle_t * query,
                        cmapl_timestamp_handle_t ** timestamp,
			bool update_iter)
{
	// if the iterator is not supposed to be updated, save the next timestamp index
        // so it can be set again later
        uint32_t saved_index = 0;
        if (!update_iter) {
                saved_index = query->i6_0.next_timestamp;
        }

	query->i6_0.next_timestamp = 0;
	cmapl_error_t err = cmapl_next_timestamp_v6_0(query, timestamp, update_iter);

	// set saved timestamp index
	if (!update_iter) {
                query->i6_0.next_timestamp = saved_index;
        }

	return err;
}

cmapl_error_t
cmapl_get_counter_data_value(
                        cmapl_counter_data_handle_t * counter_data,
                        void * counter_data_value)
{
	if (counter_data == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(counter_data, 6, 0)){
                return cmapl_get_counter_data_value_v6_0(
                                        counter_data,
                                        (cmapl_counter_data_t_v6_0*)counter_data_value);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_counter_data_value_v6_0(
                        cmapl_counter_data_handle_t * counter_data,
                        cmapl_counter_data_t_v6_0 * counter_data_value)
{
	(*counter_data_value) = counter_data->e6_0;
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_counter_info(
                        cmapl_counter_handle_t * counter,
                        void * counter_info)
{
	if (counter == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(counter, 6, 0)){
                return cmapl_get_counter_info_v6_0(
					counter,
					(cmapl_counter_def_t_v6_0*)counter_info);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_counter_info_v6_0(
                        cmapl_counter_handle_t * counter,
                        cmapl_counter_def_t_v6_0 * counter_info)
{
	(*counter_info) = counter->e6_0;
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_counter_type(
                        cmapl_counter_handle_t * counter,
                        cmapl_cnt_size_t * size,
                        cmapl_cnt_type_t * type,
                        cmapl_cnt_unit_t * unit)
{
        if (counter == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(counter, 6, 0)){
                return cmapl_get_counter_type_v6_0(
                                        counter,
                                        size,
                                        type,
                                        unit);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_counter_type_v6_0(
                        cmapl_counter_handle_t * counter,
                        cmapl_cnt_size_t * size,
                        cmapl_cnt_type_t * type,
                        cmapl_cnt_unit_t * unit)
{
        // get the ctype of the counter
        uint32_t ctype = counter->e6_0.ctype;

        // mask the ctype to get each type value
        *size = CMAPL_CNT_ITEM_SIZE(ctype);
        *type = CMAPL_CNT_MATH(ctype);
        *unit = CMAPL_CNT_UNIT_MASK(ctype);

        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_file_info(
                        cmapl_file_handle_t * data_file,
                        void * file_info)
{
	if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_get_file_info_v6_0(
                                        data_file,
                                        (cmapl_file_def_t_v6_0*)file_info);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_file_info_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_file_def_t_v6_0 * file_info)
{
	// check if the file handle is valid
        if (data_file->i6_0.data_file == NULL){
                return CMAPL_FILE_NOT_FOUND;
        }

	*file_info = data_file->e6_0;
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_header(
                        cmapl_file_handle_t * data_file,
                        cmapl_header_handle_t ** header)
{
        if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_get_header_v6_0(
                                        data_file,
                                        header);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_header_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_header_handle_t ** header)
{
	// check if the file handle is valid
        if (data_file->i6_0.data_file == NULL){
                return CMAPL_FILE_NOT_FOUND;
        }

	(*header) = data_file->i6_0.header;
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_header_info(
                        cmapl_header_handle_t * header,
                        void * header_info)
{
	if (header == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(header, 6, 0)){
                return cmapl_get_header_info_v6_0(
                                        header,
                                        (cmapl_header_def_t_v6_0*)header_info);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_header_info_v6_0(
                        cmapl_header_handle_t * header,
                        cmapl_header_def_t_v6_0 * header_info)
{
	(*header_info) = header->e6_0;
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_instance_data(
                        cmapl_instance_handle_t * instance,
                        void * instance_data)
{
	if (instance == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(instance, 6, 0)){
                return cmapl_get_instance_data_v6_0(
                                        instance,
                                        (cmapl_instance_data_t_v6_0*)instance_data);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_instance_data_v6_0(
                        cmapl_instance_handle_t * instance,
                        cmapl_instance_data_t_v6_0 * instance_data)
{
	(*instance_data) = instance->e6_0;
	return CMAPL_SUCCESS;
}

uint32_t
cmapl_get_major_version(void)
{
        return CMAPL_MAJOR_VERSION;
}

uint32_t
cmapl_get_minor_version(void)
{
        return CMAPL_MINOR_VERSION;
}

cmapl_error_t
cmapl_get_object_info(
                        cmapl_object_handle_t * object,
                        void * object_info)
{
	if (object == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_get_object_info_v6_0(
                                        object,
                                        (cmapl_object_def_t_v6_0*)object_info);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_object_info_v6_0(
                        cmapl_object_handle_t * object,
                        cmapl_object_def_t_v6_0 * object_info)
{
	(*object_info) = object->e6_0;
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_object_periods(
                        cmapl_object_handle_t * object,
                        uint64_t ** periods,
                        uint32_t * num_periods)
{
	if (object == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_get_object_periods_v6_0(object, periods, num_periods);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_object_periods_v6_0(
                        cmapl_object_handle_t * object,
                        uint64_t ** periods,
                        uint32_t * num_periods)
{
	uint32_t num_p = object->i6_0.num_periods;
	(*num_periods) = num_p;
	if (num_p == 0){
		(*periods) = NULL;
		return CMAPL_SUCCESS;
	}

	(*periods) = (uint64_t*)CMAPL_ZALLOC(num_p * sizeof(uint64_t));
	if ((*periods) == NULL){
		(*periods) = NULL;
		return CMAPL_NULL_POINTER;
	}
	cmapl_period_t* period = object->i6_0.period_list;
	uint32_t i;
	for (i = 0; i < num_p; i++){
		if (period == NULL){
			CMAPL_FREE(*periods);
			return CMAPL_NULL_POINTER;
		}
		(*periods)[i] = period->period;
		period = period->next_period;
	}

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_record_info(
                        cmapl_record_handle_t * record,
                        void * record_info)
{
	if (record == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(record, 6, 0)){
                return cmapl_get_record_info_v6_0(
                                        record,
                                        (cmapl_record_header_data_t_v6_0*)record_info);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_record_info_v6_0(
                        cmapl_record_handle_t * record,
                        cmapl_record_header_data_t_v6_0 * record_info)
{
	(*record_info) = record->e6_0;
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_system_info(
                        cmapl_header_handle_t * header,
                        void * system_info)
{
	if (header == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(header, 6, 0)){
                return cmapl_get_system_info_v6_0(
                                        header,
                                        (cmapl_system_def_t_v6_0*)system_info);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_system_info_v6_0(
                        cmapl_header_handle_t * header,
                        cmapl_system_def_t_v6_0 * system_info)
{
	(*system_info) = header->i6_0.system.e6_0;
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_timestamp(
                        cmapl_timestamp_handle_t * timestamp,
                        uint64_t * time)
{
	if (timestamp == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(timestamp, 6, 0)){
                return cmapl_get_timestamp_v6_0(
                                        timestamp,
                                        time);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_timestamp_v6_0(
                        cmapl_timestamp_handle_t * timestamp,
                        uint64_t * time)
{
	*time = timestamp->i6_0.timestamp + timestamp->i6_0.query->i6_0.data_file->i6_0.header->i6_0.system.e6_0.base_time;
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_next_counter(
                        cmapl_object_handle_t * object,
                        cmapl_counter_handle_t ** counter)
{
	if (object == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_next_counter_v6_0(object, counter, true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_counter_v6_0(
                        cmapl_object_handle_t * object,
                        cmapl_counter_handle_t ** counter,
			bool update_iter)
{
	cmapl_counter_handle_t* current_cnt = object->i6_0.current_cnt;
	if (current_cnt == NULL){
		return CMAPL_SUCCESS;
	}
	current_cnt = current_cnt->i6_0.next_cnt_by_id;
	(*counter) = current_cnt;
	if (update_iter) {
		object->i6_0.current_cnt = current_cnt;
	}
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_next_counter_by_period(
                        cmapl_object_handle_t * object,
			uint64_t period,
                        cmapl_counter_handle_t ** counter)
{
        if (object == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_next_counter_by_period_v6_0(object, period, counter, true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_counter_by_period_v6_0(
                        cmapl_object_handle_t * object,
                        uint64_t period,
                        cmapl_counter_handle_t ** counter,
			bool update_iter)
{
	cmapl_period_t* p = object->i6_0.period_list;
	uint32_t period_index;
	cmapl_error_t err = cmapl_get_period_index(period, &period_index);
	if (err != CMAPL_SUCCESS){
		(*counter) = NULL;
		return err;
	}
	while (p != NULL){
		if (p->period == period){
			cmapl_counter_handle_t* current_cnt = p->current_cnt;
			if (current_cnt == NULL){
				return CMAPL_SUCCESS;
			}
			current_cnt = current_cnt->i6_0.next_cnt_by_period[period_index];
			(*counter) = current_cnt;
			if (update_iter) {
				p->current_cnt = current_cnt;
			}
			return CMAPL_SUCCESS;
		}
		p = p->next_period;
	}
	(*counter) = NULL;
	return CMAPL_PERIOD_NOT_FOUND;
}

cmapl_error_t
cmapl_next_counter_data(
                        cmapl_instance_handle_t * instance,
			cmapl_counter_handle_t ** counter,
                        cmapl_counter_data_handle_t ** counter_data)
{
	if (instance == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(instance, 6, 0)){
                return cmapl_next_counter_data_v6_0(
						instance,
						counter,
						counter_data,
						true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_counter_data_v6_0(
                        cmapl_instance_handle_t * instance,
			cmapl_counter_handle_t ** counter,
                        cmapl_counter_data_handle_t ** counter_data,
			bool update_iter)
{
	if (instance->e6_0.num_key_cnts == 0){
                (*counter) = NULL;
                (*counter_data) = NULL;
                return CMAPL_SUCCESS;
        }

        cmapl_error_t err = CMAPL_SUCCESS;

        cmapl_object_handle_t* object;
        cmapl_counter_handle_t** counters;
        uint32_t num_cnts;
        if (instance->i6_0.timestamp != NULL) {
                // whole instance
                object = instance->i6_0.timestamp->i6_0.query->i6_0.object;
                counters = instance->i6_0.timestamp->i6_0.query->i6_0.counters;
                num_cnts = instance->i6_0.timestamp->i6_0.query->e6_0.num_cnts;
        } else {
                // partial instance
                object = instance->i6_0.record->i6_0.object;
                counters = instance->i6_0.record->i6_0.counters;
                num_cnts = instance->i6_0.record->i6_0.num_cnts;
        }

        // free last counter data
        cmapl_free_counter_data(instance->i6_0.current_counter_data);
        instance->i6_0.current_counter_data = NULL;

	uint32_t next_counter_index = instance->i6_0.next_counter_index;
	if (next_counter_index >= (object->i6_0.num_key_cnts_sampled + num_cnts)){
		(*counter) = NULL;
                (*counter_data) = NULL;
                return err;
	}

        (*counter_data) = (cmapl_counter_data_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_counter_data_handle_t));
        if ((*counter_data) == NULL){
                return CMAPL_NULL_POINTER;
        }
	(*counter_data)->major_v = 6;
	(*counter_data)->minor_v = 0;

        // check if next counter is key counter
	cmapl_counter_data_t_v6_0 counter_data_value;
	cmapl_counter_handle_t* next_counter;
	if (next_counter_index < object->i6_0.num_key_cnts_sampled){
		err = cmapl_find_counter_by_name_v6_0(object, instance->e6_0.key_cnt_names[next_counter_index], &next_counter, false);
		if (err != CMAPL_SUCCESS){
			goto error_case;
		}
		counter_data_value = instance->e6_0.key_cnt_values[next_counter_index];
		(*counter_data)->e6_0 = counter_data_value;
		uint32_t num_values = (*counter_data)->e6_0.num_values;
		(*counter_data)->e6_0.values = (cmapl_cnt_value_t*)CMAPL_ZALLOC(num_values * sizeof(cmapl_cnt_value_t));
		if ((*counter_data)->e6_0.values == NULL){
			err = CMAPL_NULL_POINTER;
			goto error_case;
		}

		uint32_t i;
		for (i = 0; i < num_values; i++){
			(*counter_data)->e6_0.values[i] = counter_data_value.values[i];
			if ((*counter_data)->e6_0.value_type == CMAPL_CNT_STRING){
				(*counter_data)->e6_0.values[i].val_str = CMAPL_STRDUP(counter_data_value.values[i].val_str);
				if ((*counter_data)->e6_0.values[i].val_str == NULL){
					err = CMAPL_NULL_POINTER;
					goto error_case;
				}
			}
		}
	} else{
		next_counter = counters[next_counter_index - object->i6_0.num_key_cnts_sampled];
		cmapl_offset_t counter_offset = 0;
		err = cmapl_get_counter_offset_v6_0(instance, next_counter, &counter_offset);
		if (err != CMAPL_SUCCESS){
			goto error_case;
                }
                char* counter_buffer = instance->i6_0.instance_buffer + counter_offset;
                err = cmapl_get_counter_data_v6_0(next_counter, counter_buffer, &counter_data_value);
                if (err != CMAPL_SUCCESS){
			goto error_case;
                }
                (*counter_data)->e6_0 = counter_data_value;
	}
        (*counter) = next_counter;
        (*counter_data)->i6_0.instance = instance;
        (*counter_data)->i6_0.counter = next_counter;

        instance->i6_0.current_counter_data = *counter_data;
	if (update_iter) {
		instance->i6_0.next_counter_index++;
	}
        return err;

error_case:
	cmapl_free_counter_data(*counter_data);
	return err;
}

cmapl_error_t
cmapl_next_instance(
                        cmapl_timestamp_handle_t * timestamp,
                        cmapl_instance_handle_t ** instance)
{
	if (timestamp == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(timestamp, 6, 0)){
                return cmapl_next_instance_v6_0(
                                                timestamp,
                                                instance,
						true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_instance_v6_0(
                        cmapl_timestamp_handle_t * timestamp,
                        cmapl_instance_handle_t ** instance,
			bool update_iter)
{
	// free previous current instance in timestamp
        cmapl_free_instance(timestamp->i6_0.current_instance);
        timestamp->i6_0.current_instance = NULL;

	cmapl_error_t err = CMAPL_SUCCESS;
	uint32_t num_instances = timestamp->i6_0.num_instances;
        if (num_instances == 0 || timestamp->i6_0.next_instance >= num_instances){
                (*instance) = NULL;
                return CMAPL_SUCCESS;
        }

        uint64_t instance_size = timestamp->i6_0.size / num_instances;
        (*instance) = (cmapl_instance_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_instance_handle_t));
        if ((*instance) == NULL){
                return CMAPL_NULL_POINTER;
        }
	(*instance)->major_v = 6;
        (*instance)->minor_v = 0;

	cmapl_offset_t next_instance_offset = instance_size * timestamp->i6_0.next_instance;
        (*instance)->i6_0.instance_buffer = timestamp->i6_0.instances_buffer + next_instance_offset;
        (*instance)->i6_0.size = instance_size;
        (*instance)->i6_0.timestamp = timestamp;
	if (update_iter) {
		timestamp->i6_0.next_instance++;
	}

	// get the key counter information of the instance
        err = cmapl_get_instance_key_counters_v6_0(*instance);
        if(err != CMAPL_SUCCESS){
                CMAPL_FREE(*instance);
                return err;
        }

	timestamp->i6_0.current_instance = *instance;

        return err;
}

cmapl_error_t
cmapl_next_instance_from_record(
                        cmapl_record_handle_t * record,
                        cmapl_instance_handle_t ** instance)
{
	if (record == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(record, 6, 0)){
                return cmapl_next_instance_from_record_v6_0(record, instance);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_object(
                        cmapl_header_handle_t * header,
                        cmapl_object_handle_t ** object)
{
	if (header == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(header, 6, 0)){
                return cmapl_next_object_v6_0(header, object, true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_object_v6_0(
                        cmapl_header_handle_t * header,
                        cmapl_object_handle_t ** object,
			bool update_iter)
{
	if (header->i6_0.current_object == NULL){
		return CMAPL_SUCCESS;
	}
        (*object) = header->i6_0.current_object->i6_0.next_object;
	if (update_iter) {
		header->i6_0.current_object = (*object);
	}
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_next_record(
                    cmapl_file_handle_t * data_file,
                    cmapl_record_handle_t ** record)
{
	if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_next_record_v6_0(data_file, record);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_timestamp(
                        cmapl_query_handle_t * query,
                        cmapl_timestamp_handle_t ** timestamp)
{
	if (query == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	if (cmapl_is_version(query, 6, 0)){
                return cmapl_next_timestamp_v6_0(
                                                query,
                                                timestamp,
						true);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_next_timestamp_v6_0(
                        cmapl_query_handle_t * query,
                        cmapl_timestamp_handle_t ** timestamp,
			bool update_iter)
{
	// free previous current_timestamp in query
        cmapl_free_timestamp(query->i6_0.current_timestamp);
        query->i6_0.current_timestamp = NULL;

	cmapl_error_t err = CMAPL_SUCCESS;
	if (query->i6_0.num_timestamps == 0 || query->i6_0.next_timestamp >= query->i6_0.num_timestamps){
		(*timestamp) = NULL;
		return err;
	}

        (*timestamp) = (cmapl_timestamp_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_timestamp_handle_t));
        if (*timestamp == NULL){
                return CMAPL_NULL_POINTER;
        }
        (*timestamp)->major_v = 6;
        (*timestamp)->minor_v = 0;
        (*timestamp)->i6_0.query = query;

	// if the iterator is not supposed to be updated, save the next timestamp index
	// so it can be set again later
	uint32_t saved_index = 0;
	if (!update_iter){
		saved_index = query->i6_0.next_timestamp;
	}

	/* find the next time stamp and return it even if the record is empty */
        /* Do not skip over empty records */
        (*timestamp)->i6_0.timestamp = query->i6_0.timestamps[(query->i6_0.next_timestamp)++];
        err = cmapl_get_instances_v6_0(*timestamp);
 
         // set saved timestamp index
        if (!update_iter){
		query->i6_0.next_timestamp = saved_index;
        }

        if (err == CMAPL_RECORD_EMPTY) {
                query->i6_0.current_timestamp = *timestamp;
                return CMAPL_SUCCESS;
        } else if (err != CMAPL_SUCCESS){
		cmapl_free_timestamp(*timestamp);
                (*timestamp) = NULL;
                return err;
        }
	query->i6_0.current_timestamp = *timestamp;
 
        return err;
}

cmapl_error_t
cmapl_open_data_file(
                        const char * file_path,
                        const char * access_type,
                        cmapl_file_handle_t ** data_file,
			const char ** meta_path,
			const char ** data_path)
{
	cmapl_error_t err = CMAPL_SUCCESS;
	uint32_t access_flag = 0;
        if (CMAPL_STRCMP(access_type, "rb") == 0 || CMAPL_STRCMP(access_type, "r") == 0){
		access_flag = CMAPL_READ_ONLY;
        } else if (CMAPL_STRCMP(access_type, "rw") == 0 || CMAPL_STRCMP(access_type, "wr") == 0){
		access_flag = CMAPL_READ_WRITE;
	} else{
		return CMAPL_INVALID_ARGUMENTS;
	}

	char* metadata_file_path;
        CMAPL_ZALLOC_META_NAME(file_path, metadata_file_path);
	if (metadata_file_path == NULL){
                return CMAPL_NULL_POINTER;
        }

        char* data_file_path;
        CMAPL_ZALLOC_GZ_NAME(file_path, data_file_path);
	if (data_file_path == NULL){
                CMAPL_FREE(metadata_file_path);
                return CMAPL_NULL_POINTER;
        }

	*data_file = (cmapl_file_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_file_handle_t));
        if (*data_file == NULL){
                CMAPL_FREE(metadata_file_path);
                CMAPL_FREE(data_file_path);
                return CMAPL_NULL_POINTER;
        }
        (*data_file)->major_v = 6;
        (*data_file)->minor_v = 0;
        (*data_file)->i6_0.metadata_file_path = metadata_file_path;
        (*data_file)->i6_0.data_file_path = data_file_path;
        (*data_file)->i6_0.access_flag = access_flag;

	(*data_file)->e6_0.file_path = CMAPL_STRDUP(file_path);
	if ((*data_file)->e6_0.file_path == NULL){
		err = CMAPL_NULL_POINTER;
		(void)cmapl_free_file(*data_file);
		return err;
	}

	(*data_file)->i6_0.metadata_file = CMAPL_FOPEN(metadata_file_path, access_type);
        if ((*data_file)->i6_0.metadata_file == NULL){
		err = CMAPL_FILE_NOT_FOUND;
		(void)cmapl_free_file(*data_file);
                return err;
        }
        (*data_file)->i6_0.data_file = CMAPL_GZOPEN(data_file_path, access_type);
        if ((*data_file)->i6_0.data_file == NULL){
		err = CMAPL_FILE_NOT_FOUND;
		(void)cmapl_free_file(*data_file);
                return err;
        }

        *meta_path = (const char*)metadata_file_path;
        *data_path = (const char*)data_file_path;

	// update external data
	(*data_file)->e6_0.file_path = CMAPL_STRDUP(file_path);
	(*data_file)->e6_0.meta_path = CMAPL_STRDUP(metadata_file_path);
	(*data_file)->e6_0.data_path = CMAPL_STRDUP(data_file_path);

        xmlDoc *doc = NULL;
        xmlNode *root_element = NULL;
        doc = xmlReadFile(metadata_file_path, NULL, XML_PARSE_NOBLANKS);
        if (doc == NULL) {
		err = CMAPL_PARSING_ERROR;
		(void)cmapl_free_file(*data_file);
		return err;
        }
        root_element = xmlDocGetRootElement(doc);

	// get version number
	char* xml_str = (char*)xmlGetProp(root_element, (const xmlChar*)"version");
	char* version = CMAPL_STRTOK(xml_str, ".");
	if (version == NULL){
		xmlFree(xml_str);
		err = CMAPL_NULL_POINTER;
		goto error_case;
	}
	(*data_file)->major_v = CMAPL_ATOI(version);
	version = CMAPL_STRTOK(NULL, ".");
	if (version == NULL){
		xmlFree(xml_str);
                err = CMAPL_NULL_POINTER;
                goto error_case;
        }
	(*data_file)->minor_v = CMAPL_ATOI(version);
	xmlFree(xml_str);

	if ((*data_file)->major_v == 6 && (*data_file)->minor_v == 0){
		err = cmapl_parse_xml_header_v6_0(root_element, *data_file);
		if (err != CMAPL_SUCCESS){
			goto error_case;
		}
		err = cmapl_parse_object_index_v6_0(*data_file);
		if (err != CMAPL_SUCCESS){
			goto error_case;
		}
		err = cmapl_set_counter_value_offsets_v6_0((*data_file)->i6_0.header);
		if (err != CMAPL_SUCCESS){
			goto error_case;
                }

	}else{
		err = CMAPL_INVALID_VERSION;
		goto error_case;
	}

        xmlFreeDoc(doc);
	// TODO: xmlCleanupParser is supposed to be called only once when program exits.
	// It is commented out pending fix for burt 568484.
        // xmlCleanupParser();

        return err;

error_case:
	xmlFreeDoc(doc);
	// xmlCleanupParser();
	(void)cmapl_free_file(*data_file);
	return err;
}

cmapl_error_t
cmapl_query(
                        cmapl_file_handle_t * data_file,
                        void * query_def,
                        cmapl_query_handle_t ** query)
{
	if (data_file == NULL || query_def == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_query_v6_0(
					data_file,
					(cmapl_query_def_t_v6_0*)query_def,
					query);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_query_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_query_def_t_v6_0 * query_def,
                        cmapl_query_handle_t ** query)
{
	// check if the file handle is valid
        if (data_file->i6_0.data_file == NULL){
                return CMAPL_FILE_NOT_FOUND;
        }

	// validate query definition struct
	cmapl_error_t err = cmapl_validate_query_def_v6_0(query_def);
	if (err != CMAPL_SUCCESS){
		return err;
	}

	// make sure the query time range is within the file's time range
	if (query_def->start_time > data_file->e6_0.end_time ||
		query_def->end_time < data_file->e6_0.start_time)
	{
		(*query) = NULL;
		return CMAPL_SUCCESS;
	}

	cmapl_header_handle_t* header = data_file->i6_0.header;
	cmapl_object_handle_t* object = NULL;
	err = cmapl_find_object_v6_0(header, query_def->object_name, &object, false);
	if (err != CMAPL_SUCCESS){
		return err;
	}

	// free previous query handle
	cmapl_free_query(data_file->i6_0.query);
	data_file->i6_0.query = NULL;

	*query = (cmapl_query_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_query_handle_t));
	if (*query == NULL){
		return CMAPL_NULL_POINTER;
	}

	(*query)->major_v = 6;
	(*query)->minor_v = 0;

	// copy all external query metadata
	err = cmapl_copy_query_def_v6_0(query_def, &((*query)->e6_0));
	if (err != CMAPL_SUCCESS){
		goto error_case;
	}

	(*query)->i6_0.data_file = data_file;
        (*query)->i6_0.object = object;

	// find all key counter and regular counter handles
	err = cmapl_get_query_counters_v6_0(*query, query_def);
	if (err != CMAPL_SUCCESS){
                goto error_case;
        }

	// get all unique counters' periods
	err = cmapl_get_counters_periods_v6_0(*query);
        if (err != CMAPL_SUCCESS){
                goto error_case;
        }

	// calculate all the valid timestamps
	err = cmapl_get_timestamps_v6_0(*query);
	if (err != CMAPL_SUCCESS){
		goto error_case;
	}

	data_file->i6_0.query = *query;

	return err;

error_case:
	cmapl_free_query((*query));
	return err;
}

#ifdef _CMAPL_FIJI_
cmapl_error_t
cmapl_disable_fiji(
                        cmapl_error_t error)
{
        if (error == CMAPL_READ_FAILED){
                cmapl_is_read_fail = false;
        } else if (error == CMAPL_SEEK_FAILED){
                cmapl_is_seek_fail = false;
        } else if (error == CMAPL_WRITE_FAILED){
                cmapl_is_write_fail = false;
        } else{
                return CMAPL_INVALID_ARGUMENTS;
        }
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_enable_fiji(
                        cmapl_error_t error)
{
	if (error == CMAPL_READ_FAILED){
		cmapl_is_read_fail = true;
	} else if (error == CMAPL_SEEK_FAILED){
		cmapl_is_seek_fail = true;
	} else if (error == CMAPL_WRITE_FAILED){
		cmapl_is_write_fail = true;
	} else{
		return CMAPL_INVALID_ARGUMENTS;
	}
	return CMAPL_SUCCESS;
}
#endif /* _CMAPL_FIJI_ */

//////////////////////////////////////////////
// The following are internal helper functions
//////////////////////////////////////////////


cmapl_error_t
cmapl_add_instance_v6_0(
			cmapl_timestamp_handle_t * timestamp,
			uint32_t instance_index,
			cmapl_record_handle_t ** records,
			uint32_t num_records)
{
	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_query_handle_t* query = timestamp->i6_0.query;
	cmapl_object_handle_t* object = query->i6_0.object;

	// get all key counters' info and value for this instance of the first record
	cmapl_record_handle_t* first_record = records[0];
	uint32_t instance_size = cmapl_get_instance_size_v6_0(first_record);
        char* instance_buffer = first_record->i6_0.data_buffer + (instance_size * instance_index);
	uint32_t num_key_cnts = object->i6_0.num_key_cnts_sampled;
	cmapl_counter_handle_t** key_cnts =
		(cmapl_counter_handle_t**)CMAPL_ZALLOC(num_key_cnts * sizeof(cmapl_counter_handle_t*));
	if (key_cnts == NULL){
		return CMAPL_NULL_POINTER;
	}
	cmapl_counter_data_t_v6_0* key_cnts_vals =
		(cmapl_counter_data_t_v6_0*)CMAPL_ZALLOC(num_key_cnts * sizeof(cmapl_counter_data_t_v6_0));
	if (key_cnts_vals == NULL){
		CMAPL_FREE(key_cnts);
                return CMAPL_NULL_POINTER;
        }
	// get first period, which also happens to be associated with the first record
	cmapl_period_t* period = query->i6_0.periods[0];
	cmapl_counter_handle_t* key_cnt = object->i6_0.key_cnt_list;
	uint32_t i = 0;
	while (key_cnt != NULL){
		key_cnts[i] = key_cnt;
		char* counter_buffer = instance_buffer + key_cnt->i6_0.record_offsets[period->period_index];
                err = cmapl_get_counter_data_v6_0(key_cnt, counter_buffer, &(key_cnts_vals[i]));
		if (err != CMAPL_SUCCESS){
			goto clean_up;
		}
		i++;
                key_cnt = key_cnt->i6_0.next_cnt_by_key;
	}

	char* whole_instance_buffer = (char*)CMAPL_ZALLOC(instance_size * sizeof(char));
	if (whole_instance_buffer == NULL){
		err = CMAPL_NULL_POINTER;
		goto clean_up;
	}
	CMAPL_MEMCPY(whole_instance_buffer, instance_buffer, instance_size * sizeof(char));
	uint32_t whole_instance_size = instance_size;

	// check if this instance exists in all the other records
	uint32_t record_index;
	for (record_index = 1; record_index < num_records; record_index++){
		cmapl_record_handle_t* record = records[record_index];
		// since the records were retrieved in the same order as the periods,
		// thus the record index matches the period index
		period = query->i6_0.periods[record_index];
		instance_size = cmapl_get_instance_size_v6_0(record);
		bool is_equal = false;

		// search through all the instances of the record
		for (instance_index = 0; instance_index < record->i6_0.num_instances; instance_index++){
			instance_buffer = record->i6_0.data_buffer + (instance_size * instance_index);
			err = cmapl_is_instance_equal_v6_0(key_cnts, key_cnts_vals, num_key_cnts,
								period, instance_buffer, &is_equal);
			if (err != CMAPL_SUCCESS){
				goto error_case;
			}
			if (is_equal == true){
				instance_buffer = record->i6_0.data_buffer + (instance_size * instance_index);
				char* new_whole_instance = (char*)CMAPL_ZALLOC((whole_instance_size + instance_size) * sizeof(char));
				if (new_whole_instance == NULL){
					err = CMAPL_NULL_POINTER;
					goto error_case;
				}
				CMAPL_MEMCPY(new_whole_instance, whole_instance_buffer, whole_instance_size * sizeof(char));
				CMAPL_MEMCPY(new_whole_instance + whole_instance_size, instance_buffer, instance_size * sizeof(char));
				CMAPL_FREE(whole_instance_buffer);
				whole_instance_buffer = new_whole_instance;
				whole_instance_size += instance_size;
				break;
			}
		}
		if (is_equal == false){
			goto error_case;
		}
	}

	if (timestamp->i6_0.num_instances == 0){
		timestamp->i6_0.instances_buffer = whole_instance_buffer;
		timestamp->i6_0.size = whole_instance_size;
		timestamp->i6_0.num_instances = 1;
	} else{
		char* new_instances_buffer = (char*)CMAPL_ZALLOC(((uint32_t)timestamp->i6_0.size + whole_instance_size) * sizeof(char));
		if (new_instances_buffer == NULL){
                        err = CMAPL_NULL_POINTER;
			goto error_case;
                }
		CMAPL_MEMCPY(new_instances_buffer, timestamp->i6_0.instances_buffer, (uint32_t)timestamp->i6_0.size * sizeof(char));
		CMAPL_MEMCPY(new_instances_buffer + timestamp->i6_0.size, whole_instance_buffer, whole_instance_size * sizeof(char));
		CMAPL_FREE(timestamp->i6_0.instances_buffer);
		CMAPL_FREE(whole_instance_buffer);
		timestamp->i6_0.instances_buffer = new_instances_buffer;
		timestamp->i6_0.size += whole_instance_size;
		timestamp->i6_0.num_instances++;
	}

	goto clean_up;

error_case:
	CMAPL_FREE(whole_instance_buffer);

clean_up:
	for (i = 0; i < num_key_cnts; i++){
                if (key_cnts_vals[i].value_type == CMAPL_CNT_STRING){
                        uint32_t j;
                        for (j = 0; j < key_cnts_vals[i].num_values; j++){
                                CMAPL_FREE(key_cnts_vals[i].values[j].val_str);
                        }
                }
                CMAPL_FREE(key_cnts_vals[i].values);
        }
	CMAPL_FREE(key_cnts);
        CMAPL_FREE(key_cnts_vals);
        return err;
}

cmapl_error_t
cmapl_add_record_data_v6_0(
			cmapl_record_handle_t * record,
			cmapl_offset_t record_offset,
			cmapl_file_handle_t * data_file)
{
	cmapl_record_header_data_t_v6_0* record_info = (cmapl_record_header_data_t_v6_0*)CMAPL_ZALLOC(sizeof(cmapl_record_header_data_t_v6_0));
	if (record_info == NULL){
		return CMAPL_NULL_POINTER;
	}
	record_info->major_v = 6;
	record_info->minor_v = 0;

	long num = CMAPL_GZSEEK(data_file->i6_0.data_file, (long)record_offset, SEEK_SET);
        if (num == -1){
		CMAPL_FREE(record_info);
                return CMAPL_SEEK_FAILED;
        }
	size_t version_size = (size_t)&(record_info->msecs_since_base) - (size_t)&(record_info->major_v);
	size_t record_info_size = sizeof(cmapl_record_header_data_t_v6_0) - version_size;
        num = CMAPL_GZREAD(((char*)record_info) + version_size, (unsigned int)record_info_size, data_file->i6_0.data_file);
        if (num != record_info_size){
		CMAPL_FREE(record_info);
                return CMAPL_READ_FAILED;
        }

        uint32_t r_size = record_info->size;
        uint64_t r_period = record_info->period;
        uint32_t r_oid = record_info->obj_id;

        // validation checks on the integrity of the record
        if (r_size > CMAPL_MAX_RECORD_SIZE ||
                r_oid > CMAPL_DEFAULT_NUM_OBJECTS ||
                !cmapl_is_period_valid(r_period))
        {
                CMAPL_FREE(record_info);
                return CMAPL_CORRUPT_ARCHIVE;
        }

        cmapl_object_handle_t* object = data_file->i6_0.header->i6_0.objects[r_oid];
        // XML header preset should contain the record's object
        if (data_file->i6_0.header->i6_0.objects[r_oid] == NULL) {
                CMAPL_FREE(record_info);
                return CMAPL_CORRUPT_ARCHIVE;
        }


        uint64_t instance_size = 0;

	if (object == NULL) {
		CMAPL_FREE(record_info);
		return CMAPL_INVALID_RECORD;
	}

        cmapl_period_t* period = object->i6_0.period_list;
        while (period != NULL) {
                if (r_period == period->period) {
                        instance_size = period->cnts_size;
                        break;
                }
                period = period->next_period;
        }

        // validation checks to make sure the object is sampled at the specified period
        // and that the record size is a multiple of the instance size
	if (period == NULL ||
                instance_size == 0 ||
                r_size % (uint32_t)instance_size != 0)
	{
		CMAPL_FREE(record_info);
		return CMAPL_INVALID_RECORD;
	}

	// either add new record data, or append record data
	uint32_t num_instance = r_size/(uint32_t)instance_size;
	if (record->i6_0.num_instances == 0){
		record->e6_0 = *record_info;
		uint32_t buffer_size = r_size * sizeof(char);
		record->i6_0.data_buffer = (char*)CMAPL_ZALLOC(buffer_size);
		if (record->i6_0.data_buffer == NULL){
			CMAPL_FREE(record_info);
			return CMAPL_NULL_POINTER;
		}
		num = CMAPL_GZREAD(record->i6_0.data_buffer, buffer_size, data_file->i6_0.data_file);
		if (num != buffer_size){
			CMAPL_FREE(record_info);
			return CMAPL_READ_FAILED;
		}
		record->i6_0.num_instances = num_instance;
                record->i6_0.object = object;
                record->i6_0.period = period;
                record->i6_0.instance_offset = 0;
                record->i6_0.current_instance = NULL;

                // populate counter handle list with counters sampled in this record
                if (period->num_cnts != 0) {
                        record->i6_0.counters = (cmapl_counter_handle_t**)CMAPL_ZALLOC(period->num_cnts * sizeof(cmapl_counter_handle_t*));
                        if (record->i6_0.counters == NULL) {
                                CMAPL_FREE(record_info);
                                return CMAPL_NULL_POINTER;
                        }
                        cmapl_counter_handle_t* p_counter = period->cnt_list;
                        uint32_t cnt_index = 0;
                        while (p_counter != NULL) {
                                CMAPL_ASSERT(cnt_index < period->num_cnts);
                                record->i6_0.counters[cnt_index] = p_counter;
                                p_counter = p_counter->i6_0.next_cnt_by_period[period->period_index];
                                cnt_index++;
                        }

                        record->i6_0.num_cnts = period->num_cnts;
                }
	} else {
		uint32_t total_size = record->e6_0.size + r_size;
		uint32_t new_buffer_size = r_size * sizeof(char);
		uint32_t total_buffer_size = total_size * sizeof(char);
		char* total_buffer = (char*)CMAPL_ZALLOC(total_buffer_size);
		if (total_buffer == NULL){
			CMAPL_FREE(record_info);
			return CMAPL_NULL_POINTER;
		}
		CMAPL_MEMCPY(total_buffer, record->i6_0.data_buffer, record->e6_0.size * sizeof(char));
		num = CMAPL_GZREAD(total_buffer + record->e6_0.size, new_buffer_size, data_file->i6_0.data_file);
		if (num != new_buffer_size){
			CMAPL_FREE(record_info);
			CMAPL_FREE(total_buffer);
			return CMAPL_READ_FAILED;
		}
		CMAPL_FREE(record->i6_0.data_buffer);
		record->i6_0.data_buffer = total_buffer;
		record->e6_0.size += r_size;
		record->i6_0.num_instances += num_instance;
	}
	CMAPL_FREE(record_info);
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_first_instance_from_record_v6_0(
                        cmapl_record_handle_t * record,
                        cmapl_instance_handle_t ** instance)
{
        cmapl_error_t err = CMAPL_SUCCESS;

        // have the offset point to the first instance
        record->i6_0.instance_offset = 0;

        err = cmapl_next_instance_from_record_v6_0(record, instance);
	return err;
}

cmapl_error_t
cmapl_first_record_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_record_handle_t ** record)
{
	cmapl_error_t err = CMAPL_SUCCESS;

        if (data_file->i6_0.current_record != NULL) {
                cmapl_free_record(data_file->i6_0.current_record);
                data_file->i6_0.current_record = NULL;
        }

	(*record) = (cmapl_record_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_record_handle_t));
	if (*record == NULL){
		return CMAPL_NULL_POINTER;
	}
	(*record)->major_v = 6;
	(*record)->minor_v = 0;

        // reset offset to point to the beginning of the file
	data_file->i6_0.current_offset = 0;
        (*record)->i6_0.offset = 0;

	err = cmapl_add_record_data_v6_0(*record, data_file->i6_0.current_offset, data_file);
	if (err != CMAPL_SUCCESS){
		cmapl_free_record(*record);
                *record = NULL;
		return err;
	}

        data_file->i6_0.current_record = *record;

        // in the record buffer, the header does not contain the version
        // values so it is subtracted from the size
        const uint32_t version_size = sizeof(data_file->i6_0.current_record->e6_0.major_v) + sizeof(data_file->i6_0.current_record->e6_0.minor_v);
        // increment offset to the next record in the data file
        data_file->i6_0.current_offset += sizeof(cmapl_record_header_data_t_v6_0) - version_size;
        data_file->i6_0.current_offset += data_file->i6_0.current_record->e6_0.size;

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_counter_data_v6_0(
			cmapl_counter_handle_t * counter,
			char* counter_buffer,
			cmapl_counter_data_t_v6_0 * counter_data)
{
	uint32_t cnt_size;
	uint32_t num_values;
	if (counter->e6_0.num_x_items == 0) {
		num_values = 1;
	} else if (counter->e6_0.num_y_items == 0) {
		num_values = counter->e6_0.num_x_items;
	} else {
		num_values = counter->e6_0.num_x_items * counter->e6_0.num_y_items;
	}
	cnt_size = counter->e6_0.size / num_values;

	(*counter_data).values = (cmapl_cnt_value_t*)CMAPL_ZALLOC(num_values * sizeof(cmapl_cnt_value_t));
	if ((*counter_data).values == NULL){
		return CMAPL_NULL_POINTER;
	}
	(*counter_data).num_values = num_values;

        // get counter type
        uint32_t type = CM_CNT_TYPE_MASK & counter->e6_0.ctype;

	uint32_t i;
	for (i = 0; i < num_values; i++){
		if (cnt_size == 4 && type != CM_CNT_STRING){
			(*counter_data).value_type = CMAPL_CNT_4BYTES;
			uint32_t* cv = (uint32_t*)counter_buffer;
			(*counter_data).values[i].val_4B = *cv;
		} else if (cnt_size == 8 && type != CM_CNT_STRING){
			(*counter_data).value_type = CMAPL_CNT_8BYTES;
			uint64_t* cv = (uint64_t*)counter_buffer;
	                (*counter_data).values[i].val_8B = *cv;
		} else if (type == CM_CNT_STRING) {
			(*counter_data).value_type = CMAPL_CNT_STRING;
			char* cv = (char*)CMAPL_ZALLOC((cnt_size + 1) * sizeof(char));
			if (cv == NULL){
				CMAPL_FREE((*counter_data).values);
				return CMAPL_NULL_POINTER;
			}
			CMAPL_STRNCPY(cv, counter_buffer, cnt_size * sizeof(char));
			size_t cv_length = CMAPL_STRLEN(cv);
			if (cv_length == cnt_size){
				cv[cnt_size] = '\0';
			}
			(*counter_data).values[i].val_str = cv;
		}
		counter_buffer += cnt_size;
	}

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_counter_offset_v6_0(
			cmapl_instance_handle_t * instance,
			cmapl_counter_handle_t * counter,
			cmapl_offset_t * counter_offset)
{
	cmapl_period_t** periods = NULL;
	uint32_t num_periods = 0;

        if (instance->i6_0.timestamp != NULL) {
                // whole instance
                periods = instance->i6_0.timestamp->i6_0.query->i6_0.periods;
                num_periods = instance->i6_0.timestamp->i6_0.query->i6_0.num_periods;
        } else {
                // partial instance
                periods = &(instance->i6_0.record->i6_0.period);
                num_periods = 1;
        }

	cmapl_offset_t offset = 0;

        // return counter offset value right away if the counter is a key counter or if
        // we're trying to get the counter data from a single record (partial instance)
        if (counter->e6_0.is_key == true || instance->i6_0.record != NULL){
		*counter_offset = counter->i6_0.record_offsets[periods[0]->period_index];
		return CMAPL_SUCCESS;
	}

	uint32_t i;
	for (i = 0; i < num_periods; i++){
		cmapl_period_t* period = periods[i];
		// config period is special case; guaranteed to have counter value
		if (period->period == counter->e6_0.period || period->period == CMAPL_CONFIG_PERIOD){
			offset += counter->i6_0.record_offsets[period->period_index];
			if (offset == CMAPL_NO_RECORD){
				*counter_offset = 0;
				return CMAPL_RECORD_NOT_FOUND;
			}
			*counter_offset = offset;
			return CMAPL_SUCCESS;
		}
		offset += period->cnts_size;
	}

	return CMAPL_PERIOD_NOT_FOUND;
}

cmapl_error_t
cmapl_get_counters_periods_v6_0(
			cmapl_query_handle_t * query)
{
	cmapl_period_t** periods = (cmapl_period_t**)CMAPL_ZALLOC(CMAPL_MAX_NUM_PERIODS * sizeof(cmapl_period_t*));
	if (periods == NULL){
		return CMAPL_NULL_POINTER;
	}
	uint32_t period_index = 0;
	uint32_t num_cnts = query->e6_0.num_cnts;
	uint32_t i;
        for(i = 0; i < num_cnts; i++){
                // check if timestamp is a multiple of this counter's period
                cmapl_counter_handle_t* counter = query->i6_0.counters[i];
		uint32_t j;
		cmapl_period_t** cnt_periods = counter->i6_0.periods;

		// only get config period if query period is config
		if (query->e6_0.period == CMAPL_CONFIG_PERIOD){
			cmapl_period_t* period = cnt_periods[CMAPL_MAX_NUM_PERIODS-1];
			if (period == NULL){
				continue;
			} else {
				periods[0] = period;
				period_index = 1;
				goto resize;
			}
		}

		for (j = 0; j < CMAPL_MAX_NUM_PERIODS; j++){
			uint32_t k;
			bool is_new_period = true;
			cmapl_period_t* period = cnt_periods[j];
			if (period == NULL || period->period != counter->e6_0.period){
				continue;
			}
			for (k = 0; k < period_index; k++){
				if (periods[k] == period){
					is_new_period = false;
					break;
				}
			}
			if (is_new_period == true){
				periods[period_index] = period;
				period_index++;
			}

			break;
		}
        }

resize:
	if (period_index == 0){
		CMAPL_FREE(periods);
		return CMAPL_PERIOD_NOT_FOUND;
	}

	query->i6_0.periods = (cmapl_period_t**)CMAPL_ZALLOC(period_index * sizeof(cmapl_period_t*));
	if (query->i6_0.periods == NULL){
		CMAPL_FREE(periods);
                return CMAPL_NULL_POINTER;
        }
	CMAPL_MEMCPY(query->i6_0.periods, periods, period_index * sizeof(cmapl_period_t*));
	CMAPL_FREE(periods);
	query->i6_0.num_periods = period_index;

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_instance_key_counters_v6_0(
			cmapl_instance_handle_t * instance)
{
	cmapl_error_t err = CMAPL_SUCCESS;

        // get the object and period associated with this instance, whether the
        // instance is a partial instance or whole instance
        cmapl_object_handle_t* object = NULL;
        cmapl_period_t* period = NULL;
        if (instance->i6_0.timestamp != NULL) {
                // whole instance
	        object = instance->i6_0.timestamp->i6_0.query->i6_0.object;
                period = instance->i6_0.timestamp->i6_0.query->i6_0.periods[0];
        } else {
                // partial instance
                object = instance->i6_0.record->i6_0.object;
                period = instance->i6_0.record->i6_0.period;
        }

	uint32_t num_key_cnts = object->i6_0.num_key_cnts_sampled;
	const char** key_cnt_names = (const char**)CMAPL_ZALLOC(num_key_cnts * sizeof(cmapl_counter_handle_t*));
	if (key_cnt_names == NULL){
                return CMAPL_NULL_POINTER;
        }
	cmapl_counter_data_t_v6_0* key_cnt_values = (cmapl_counter_data_t_v6_0*)CMAPL_ZALLOC(num_key_cnts * sizeof(cmapl_counter_data_t_v6_0));
	if (key_cnt_values == NULL){
		CMAPL_FREE(key_cnt_names);
                return CMAPL_NULL_POINTER;
        }
	uint32_t i = 0;
	cmapl_counter_handle_t* key_cnt = object->i6_0.key_cnt_list;
	while (key_cnt != NULL){
		key_cnt_names[i] = CMAPL_STRDUP(key_cnt->e6_0.name);
		if (key_cnt_names[i] == NULL){
			err = CMAPL_NULL_POINTER;
			goto error_case;
		}
		char* counter_buffer = instance->i6_0.instance_buffer + key_cnt->i6_0.record_offsets[period->period_index];
		err = cmapl_get_counter_data_v6_0(key_cnt, counter_buffer, &(key_cnt_values[i]));
		if (err != CMAPL_SUCCESS){
			goto error_case;
		}
		key_cnt = key_cnt->i6_0.next_cnt_by_key;
		i++;
	}
	instance->e6_0.key_cnt_names = key_cnt_names;
	instance->e6_0.key_cnt_values = key_cnt_values;
	instance->e6_0.num_key_cnts = num_key_cnts;
	return err;

error_case:
	for (i = 0; i < num_key_cnts; i++){
		CMAPL_FREE(key_cnt_names[i]);
		if (key_cnt_values[i].value_type == CMAPL_CNT_STRING){
			uint32_t j;
			for (j = 0; j < key_cnt_values[i].num_values; j++){
				CMAPL_FREE(key_cnt_values[i].values[j].val_str);
			}
		}
		CMAPL_FREE(key_cnt_values[i].values);
	}
	CMAPL_FREE(key_cnt_names);
	CMAPL_FREE(key_cnt_values);
	return err;
}

uint32_t
cmapl_get_instance_size_v6_0(cmapl_record_handle_t * record)
{
        return record->e6_0.size / record->i6_0.num_instances;
}

cmapl_error_t
cmapl_get_instances_v6_0(
			cmapl_timestamp_handle_t * timestamp)
{
	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_query_handle_t* query = timestamp->i6_0.query;
	// same number of records as valid periods
	uint32_t num_records = query->i6_0.num_periods;
	if (num_records == 0){
                return CMAPL_RECORD_NOT_FOUND;
        }
	cmapl_record_handle_t** records = (cmapl_record_handle_t**)CMAPL_ZALLOC(num_records * sizeof(cmapl_record_handle_t*));
	if (records == NULL){
		return CMAPL_NULL_POINTER;
	}

	// get every record at this timestamp, requested by this query
	cmapl_record_handle_t* record;
	cmapl_period_t* period;
	uint32_t i;
	for (i = 0; i < num_records; i++){
		period = query->i6_0.periods[i];
		record = NULL;
		err = cmapl_get_record_v6_0(timestamp, period, &record);
		if (err != CMAPL_SUCCESS){
			goto clean_up;
		}
		records[i] = record;
                if (record != NULL && record->e6_0.size == 0){
                        timestamp->i6_0.instances_buffer = NULL;
                        timestamp->i6_0.size = 0;
                        timestamp->i6_0.num_instances = 0;
                        err = CMAPL_RECORD_EMPTY;
                        goto clean_up;
                }
	}

	// get first period which is associated with the first record
	period = query->i6_0.periods[0];
	record = records[0];
	uint32_t num_instances = record->i6_0.num_instances;
	uint32_t instance_size = cmapl_get_instance_size_v6_0(record);
	for (i = 0; i < num_instances; i++){
		char* instance_buffer = record->i6_0.data_buffer + (instance_size * i);
		bool is_equal = true;
		err = cmapl_is_instance_equal_v6_0(query->i6_0.key_counters, query->e6_0.key_cnt_values,
						query->e6_0.num_key_cnts, period, instance_buffer, &is_equal);
		if (err != CMAPL_SUCCESS){
			goto clean_up;
		}
		if (is_equal == true){
			err = cmapl_add_instance_v6_0(timestamp, i, records, num_records);
			if (err != CMAPL_SUCCESS){
				goto clean_up;
			}
		}
	}

        // check if no instances were found that matches the query
        if (timestamp->i6_0.num_instances == 0) {
                err = CMAPL_INSTANCE_NOT_FOUND;
        }

clean_up:
	for (i = 0; i < num_records; i++){
		cmapl_free_record(records[i]);
	}
	CMAPL_FREE(records);
	return err;
}

cmapl_error_t
cmapl_get_query_counters_v6_0(
                        cmapl_query_handle_t * query_h,
                        cmapl_query_def_t_v6_0 * query_def)
{
	// check if arguments are NULL
	if (query_h == NULL || query_def == NULL){
		return CMAPL_NULL_POINTER;
	}

	// check version of the query handle
	if (!cmapl_is_version(query_h, 6, 0)){
		return CMAPL_INVALID_VERSION;
	}

	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_object_handle_t * object = query_h->i6_0.object;

	// find all key counter handles
        uint32_t num_key_cnts = query_def->num_key_cnts;
        query_h->i6_0.key_counters = (cmapl_counter_handle_t**)CMAPL_ZALLOC(num_key_cnts * sizeof(cmapl_counter_handle_t*));
        if (query_h->i6_0.key_counters == NULL){
                return CMAPL_NULL_POINTER;;
        }
        uint32_t i;
        for(i = 0; i < num_key_cnts; i++){
                cmapl_counter_handle_t* key_counter = NULL;
                err = cmapl_find_counter_by_name_v6_0(object, query_def->key_cnt_names[i], &key_counter, false);
                if (err != CMAPL_SUCCESS){
                        return err;
                }
                query_h->i6_0.key_counters[i] = key_counter;
        }

        // copy regular counter names to the query handle; find each counter handle associated with each counter name
        // if no counters specified, get all the counters
        if (query_def->num_cnts != 0){
                query_h->e6_0.cnt_names = (const char**)CMAPL_ZALLOC(query_def->num_cnts * sizeof(const char*));
                query_h->i6_0.counters = (cmapl_counter_handle_t**)CMAPL_ZALLOC(query_def->num_cnts * sizeof(cmapl_counter_handle_t));
                if (query_h->e6_0.cnt_names == NULL || query_h->i6_0.counters == NULL){
                        return CMAPL_NULL_POINTER;
                }
                for(i = 0; i < query_def->num_cnts; i++){
                        query_h->e6_0.cnt_names[i] = CMAPL_STRDUP(query_def->cnt_names[i]);
                        if (query_h->e6_0.cnt_names[i] == NULL){
				return CMAPL_NULL_POINTER;
                        }
                        cmapl_counter_handle_t* counter = NULL;
                        err = cmapl_find_counter_by_name_v6_0(object, query_def->cnt_names[i], &counter, false);
                        if (err != CMAPL_SUCCESS){
				return CMAPL_NULL_POINTER;
                        }
                        query_h->i6_0.counters[i] = counter;
                }
	} else {
                uint32_t num_cnts = object->i6_0.num_cnts_sampled - object->i6_0.num_key_cnts_sampled;
                if (num_cnts == 0){
                        return CMAPL_INVALID_QUERY;
                }
                query_h->e6_0.num_cnts = num_cnts;
                query_h->e6_0.cnt_names = (const char**)CMAPL_ZALLOC(num_cnts * sizeof(const char*));
                query_h->i6_0.counters = (cmapl_counter_handle_t**)CMAPL_ZALLOC(num_cnts * sizeof(cmapl_counter_handle_t));
                if (query_h->e6_0.cnt_names == NULL || query_h->i6_0.counters == NULL){
			return CMAPL_NULL_POINTER;
                }
                cmapl_counter_handle_t* counter = object->i6_0.cnt_list;
                uint32_t cnt_count = 0;
                // get all non-key counters
                while (counter != NULL){
                        if (counter->e6_0.is_key != true){
                                query_h->e6_0.cnt_names[cnt_count] = CMAPL_STRDUP(counter->e6_0.name);
                                if (query_h->e6_0.cnt_names[cnt_count] == NULL){
					return CMAPL_NULL_POINTER;
                                }
                                query_h->i6_0.counters[cnt_count] = counter;
                                cnt_count++;
                        }
                        counter = counter->i6_0.next_cnt_by_id;
                }
        }

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_record_v6_0(
			cmapl_timestamp_handle_t * timestamp,
			cmapl_period_t * period,
			cmapl_record_handle_t ** record)
{
	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_file_handle_t* data_file = timestamp->i6_0.query->i6_0.data_file;

	(*record) = (cmapl_record_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_record_handle_t));
	if (*record == NULL){
		return CMAPL_NULL_POINTER;
	}
	(*record)->major_v = 6;
	(*record)->minor_v = 0;

	uint64_t record_index = 0;
	if (period->period == CMAPL_CONFIG_PERIOD){
		record_index = (timestamp->i6_0.timestamp - data_file->i6_0.start_time)/1000;
		if (record_index != 0){
			cmapl_free_record(*record);
			return CMAPL_RECORD_NOT_FOUND;
		}
	} else{
		record_index = CMAPL_GET_BUCKET(timestamp->i6_0.timestamp, data_file->i6_0.start_time, period->period);
	}

	cmapl_offset_t record_offset = period->record_offsets[record_index];
	if (CMAPL_IS_ITERATIVE(record_offset)){
		cmapl_iterative_record_offsets_t iter_record_offsets = period->iterative_record_offsets[record_index];
		uint32_t i;
		for (i = 0; i < iter_record_offsets.num_offsets; i++){
			record_offset = iter_record_offsets.offsets[i];
			err = cmapl_add_record_data_v6_0(*record, record_offset, data_file);
			if (err != CMAPL_SUCCESS){
				cmapl_free_record(*record);
				return err;
			}
		}
	} else {
		err = cmapl_add_record_data_v6_0(*record, record_offset, data_file);
		if (err != CMAPL_SUCCESS){
			cmapl_free_record(*record);
			return err;
		}
	}

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_timestamps_v6_0(
			cmapl_query_handle_t * query)
{
	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_file_handle_t* data_file = query->i6_0.data_file;

        // total number of timestamps will equal the number of buckets from the timestamp
        // of the first sample to the last sample of the file
        uint32_t num_timestamps = (uint32_t)(CMAPL_GET_BUCKET(data_file->i6_0.end_time, data_file->i6_0.start_time, query->e6_0.period) + 1);
        query->i6_0.timestamps = (uint64_t*)CMAPL_ZALLOC(num_timestamps * sizeof(uint64_t));
	if (query->i6_0.timestamps == NULL){
		return CMAPL_NULL_POINTER;
	}

	// get first timestamp that's greater than the start of the file in msecs since boot
        // that is aligned with the query period. if the query period is config, then just
        // use the start time of the file in msecs since boot since there is only one record
        // per file for config period records
        uint64_t current_timestamp = data_file->i6_0.start_time + (query->e6_0.period - (data_file->i6_0.start_time % query->e6_0.period));
        if (query->e6_0.period == CMAPL_CONFIG_PERIOD) {
                current_timestamp = data_file->i6_0.start_time;
        }

	uint32_t num_valid_timestamps = 0;
	uint32_t i;
        for(i = 0; i < num_timestamps; i++){
		bool is_valid = false;
		err = cmapl_is_timestamp_valid_v6_0(current_timestamp, query, &is_valid);
		if (err != CMAPL_SUCCESS){
			return err;
		}
		if (is_valid){
			query->i6_0.timestamps[num_valid_timestamps] = current_timestamp;
			num_valid_timestamps++;
		}
                current_timestamp += query->e6_0.period;
        }
	query->i6_0.num_timestamps = num_valid_timestamps;
	uint64_t* valid_timestamps = (uint64_t*)CMAPL_ZALLOC(num_valid_timestamps * sizeof(uint64_t));
	if (valid_timestamps == NULL){
		return CMAPL_NULL_POINTER;
	}
	CMAPL_MEMCPY(valid_timestamps, query->i6_0.timestamps, num_valid_timestamps * sizeof(uint64_t));
	CMAPL_FREE(query->i6_0.timestamps);
	query->i6_0.timestamps = valid_timestamps;

	return err;
}

cmapl_error_t
cmapl_is_timestamp_valid_v6_0(
			uint64_t timestamp,
			cmapl_query_handle_t * query,
			bool * is_valid)
{
	cmapl_file_handle_t* data_file = query->i6_0.data_file;

	if (timestamp < data_file->i6_0.start_time ||
                timestamp < query->e6_0.start_time - data_file->i6_0.header->i6_0.system.e6_0.base_time ||
                timestamp > query->e6_0.end_time - data_file->i6_0.header->i6_0.system.e6_0.base_time)
        {
		goto not_valid;
	}

	uint32_t i;
	uint32_t num_periods = query->i6_0.num_periods;
	for(i = 0; i < num_periods; i++){
		// check if timestamp is a multiple of each unique period
		cmapl_period_t* period = query->i6_0.periods[i];
		uint64_t period_value = period->period;
		// account for CONFIG counters
		if (period_value == CMAPL_CONFIG_PERIOD){
			period_value = 1000;
		}

		// check if a record was sampled at this timestamp for this period
		uint64_t record_index = CMAPL_GET_BUCKET(timestamp, data_file->i6_0.start_time, period_value);
                if (record_index >= period->num_offsets) {
                        goto not_valid;
                }

		// CONFIG counters are only in one record
                if (period->period == CMAPL_CONFIG_PERIOD && record_index != 0){
                        goto not_valid;
                }

		if (period->record_offsets[record_index] == CMAPL_NO_RECORD){
			goto not_valid;
		}
	}
	*is_valid = true;
	return CMAPL_SUCCESS;

not_valid:
	*is_valid = false;
	return CMAPL_SUCCESS;
}

//lint -esym(429,key_cnts) "custodial pointer" message. The function doesn't allocate memory for "key_cnts".
//lint -esym(429,key_cnts_vals) "custodial pointer" message. The function doesn't allocate memory for "key_cnts_vals".
cmapl_error_t
cmapl_is_instance_equal_v6_0(
			cmapl_counter_handle_t ** key_cnts,
			cmapl_counter_data_t_v6_0 * key_cnts_vals,
			uint32_t num_key_cnts,
			cmapl_period_t * period,
			char * instance_buffer,
			bool * is_equal)
{
	cmapl_error_t err = CMAPL_SUCCESS;
	*is_equal = true;
	if (num_key_cnts == 0){
		return err;
	}

	uint32_t i;
	for (i = 0; i < num_key_cnts; i++){
		cmapl_counter_handle_t* key_cnt = key_cnts[i];
		char* counter_buffer = instance_buffer + key_cnt->i6_0.record_offsets[period->period_index];
		cmapl_counter_data_t_v6_0 key_cnt_data = key_cnts_vals[i];
		uint32_t num_values = key_cnt_data.num_values;
		uint32_t j;
		if (key_cnt_data.value_type == CMAPL_CNT_4BYTES){
			for (j = 0; j < num_values; j++){
				uint32_t* val_4B = (uint32_t*)counter_buffer;
				if (*val_4B != key_cnt_data.values[j].val_4B){
					goto not_valid;
				}
				counter_buffer = counter_buffer + sizeof(uint32_t);
			}
		} else if (key_cnt_data.value_type == CMAPL_CNT_8BYTES){
			for (j = 0; j < num_values; j++){
				uint64_t* val_8B = (uint64_t*)counter_buffer;
				if (*val_8B != key_cnt_data.values[j].val_8B){
					goto not_valid;
				}
				counter_buffer = counter_buffer + sizeof(uint64_t);
			}
		} else if (key_cnt_data.value_type == CMAPL_CNT_STRING){
			uint32_t cnt_size = key_cnt->e6_0.size / num_values;
			for (j = 0; j < num_values; j++){
				if (CMAPL_STRNCMP(counter_buffer, key_cnt_data.values[j].val_str, cnt_size) != 0){
					goto not_valid;
				}
				counter_buffer = counter_buffer + cnt_size;
			}
		} else{
			err = CMAPL_COUNTER_TYPE_NOT_FOUND;
			goto not_valid;
		}
	}

	return err;

not_valid:
	*is_equal = false;
	return err;
}
//lint +esym(429,key_cnts)
//lint +esym(429,key_cnts_vals)

cmapl_error_t
cmapl_next_instance_from_record_v6_0(
                        cmapl_record_handle_t * record,
                        cmapl_instance_handle_t ** instance)
{
        cmapl_error_t err = CMAPL_SUCCESS;

	// free previous current instance in record
        if (record->i6_0.current_instance) {
                cmapl_free_instance(record->i6_0.current_instance);
                record->i6_0.current_instance = NULL;
        }

        uint32_t num_instances = record->i6_0.num_instances;
        if (num_instances == 0 || record->i6_0.instance_offset >= num_instances){
                (*instance) = NULL;
                return CMAPL_SUCCESS;
        }

        uint64_t instance_size = cmapl_get_instance_size_v6_0(record);
        (*instance) = (cmapl_instance_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_instance_handle_t));
        if ((*instance) == NULL){
                return CMAPL_NULL_POINTER;
        }

        cmapl_offset_t next_instance_offset = instance_size * record->i6_0.instance_offset;

        // populate the next instance
	(*instance)->major_v = 6;
        (*instance)->minor_v = 0;
        (*instance)->i6_0.instance_buffer = record->i6_0.data_buffer + next_instance_offset;
        (*instance)->i6_0.size = instance_size;
        (*instance)->i6_0.timestamp = NULL;
        (*instance)->i6_0.record = record;
	record->i6_0.instance_offset++;

        // populating the instance handle with its key counter information
        // that uniquely identify this instance
        err = cmapl_get_instance_key_counters_v6_0(*instance);
        if(err != CMAPL_SUCCESS){
                CMAPL_FREE(*instance);
                return err;
        }

	record->i6_0.current_instance = *instance;

        return err;
}

cmapl_error_t
cmapl_next_record_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_record_handle_t ** record)
{
	cmapl_error_t err = CMAPL_SUCCESS;

        if (data_file->i6_0.current_record != NULL) {
                cmapl_free_record(data_file->i6_0.current_record);
                data_file->i6_0.current_record = NULL;
        }

	(*record) = (cmapl_record_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_record_handle_t));
	if (*record == NULL){
		return CMAPL_NULL_POINTER;
	}
	(*record)->major_v = 6;
	(*record)->minor_v = 0;

	err = cmapl_add_record_data_v6_0(*record, data_file->i6_0.current_offset, data_file);
	if (err != CMAPL_SUCCESS){
		cmapl_free_record(*record);
                *record = NULL;
                if (err == CMAPL_READ_FAILED) {
                        // read the last record
                        return CMAPL_SUCCESS;
                }
		return err;
	}

        data_file->i6_0.current_record = *record;
        (*record)->i6_0.offset = data_file->i6_0.current_offset;

        // increment offset to the next record in the data file
        uint32_t version_size = sizeof(data_file->i6_0.current_record->e6_0.major_v) + sizeof(data_file->i6_0.current_record->e6_0.minor_v);
        data_file->i6_0.current_offset += sizeof(cmapl_record_header_data_t_v6_0) - version_size;
        data_file->i6_0.current_offset += data_file->i6_0.current_record->e6_0.size;

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_parse_attribute_def_v6_0(
			xmlNode * attr_def_node,
			cmapl_header_handle_t * header)
{
	if (header == NULL){
                return CMAPL_NULL_POINTER;
        }

	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_attribute_def_t_v6_0 attr_def;
	CMAPL_BZERO(&attr_def, sizeof(attr_def));
	attr_def.major_v = 6;
	attr_def.minor_v = 0;

	xmlNode *current_node = NULL;
	char* xml_str;
        for (current_node = attr_def_node->children; current_node; current_node = current_node->next){
                const char* node_name = (const char*)current_node->name;
                if (CMAPL_STRCMP("name", node_name) == 0){
                        attr_def.name = (char*)xmlNodeGetContent(current_node);
                } else if (CMAPL_STRCMP("value", node_name) == 0){
			xml_str = (char*)xmlNodeGetContent(current_node);
                        attr_def.value = CMAPL_ATOI(xml_str);
			xmlFree(xml_str);
                } else if (CMAPL_STRCMP("description", node_name) == 0){
                        attr_def.desc = (char*)xmlNodeGetContent(current_node);
                } else {
                        err = CMAPL_PARSING_ERROR;
			goto clean_up;
                }
        }

	err = cmapl_add_attr_def(header, &attr_def);
clean_up:
	xmlFree((xmlChar*)attr_def.name);
	xmlFree((xmlChar*)attr_def.desc);
        return err;
}

cmapl_error_t
cmapl_parse_counter_v6_0(
                        xmlNode * counter_node,
                        cmapl_object_handle_t * object)
{
	if (object == NULL){
                return CMAPL_NULL_POINTER;
        }

	cmapl_error_t err = CMAPL_SUCCESS;
	uint64_t periods[CMAPL_MAX_NUM_PERIODS];
	CMAPL_BZERO(periods, sizeof(periods));

	cmapl_counter_def_t_v6_0 counter_def;
	CMAPL_BZERO(&counter_def, sizeof(counter_def));
	counter_def.major_v = 6;
	counter_def.minor_v = 0;
	counter_def.num_x_items = 0;
	counter_def.num_y_items = 0;
	char* xml_str;
	if (CMAPL_STRCMP("counter", (const char*)counter_node->name) == 0){
		xml_str = (char*)xmlGetProp(counter_node, (const xmlChar*)"is_key");
                counter_def.is_key = CMAPL_ATOI(xml_str);
		xmlFree(xml_str);
		xml_str = (char*)xmlGetProp(counter_node, (const xmlChar*)"ctype");
                counter_def.ctype = CMAPL_ATOI(xml_str);
		xmlFree(xml_str);
        }else{
                return CMAPL_PARSING_ERROR;
        }

	xmlNode *current_node = NULL;
        for (current_node = counter_node->children; current_node; current_node = current_node->next){
                const char* node_name = (const char*)current_node->name;
                if (CMAPL_STRCMP("name", node_name) == 0){
                        counter_def.name = (char*)xmlNodeGetContent(current_node);
                } else if (CMAPL_STRCMP("description", node_name) == 0){
                        counter_def.desc = (char*)xmlNodeGetContent(current_node);
                } else if (CMAPL_STRCMP("id", node_name) == 0){
			xml_str = (char*)xmlNodeGetContent(current_node);
                        counter_def.cnt_id = CMAPL_ATOI(xml_str);
			xmlFree(xml_str);
                } else if (CMAPL_STRCMP("base_id", node_name) == 0){
			xml_str = (char*)xmlNodeGetContent(current_node);
                        counter_def.base_id = CMAPL_ATOI(xml_str);
			xmlFree(xml_str);
                } else if (CMAPL_STRCMP("size", node_name) == 0){
			xml_str = (char*)xmlNodeGetContent(current_node);
                        counter_def.size = CMAPL_ATOI(xml_str);
			xmlFree(xml_str);
                } else if (CMAPL_STRCMP("period", node_name) == 0){
			char* strtoll_str = NULL;
			xml_str = (char*)xmlNodeGetContent(current_node);
                        counter_def.period = CMAPL_STRTOULL(xml_str, &strtoll_str, 10);
			if (CMAPL_STRCMP(strtoll_str, "\0") != 0){
				xmlFree((xmlChar*)counter_def.name);
				xmlFree((xmlChar*)counter_def.desc);
				xmlFree((xmlChar*)counter_def.priv);
				xmlFree(xml_str);
                                return CMAPL_PARSING_ERROR;
                        }
			xmlFree(xml_str);

			uint32_t period_index = 0;
			err = cmapl_get_period_index(counter_def.period, &period_index);
			if (err != CMAPL_SUCCESS){
				xmlFree((xmlChar*)counter_def.name);
                                xmlFree((xmlChar*)counter_def.desc);
                                xmlFree((xmlChar*)counter_def.priv);
				return err;
			}
			periods[period_index] = counter_def.period;
                } else if (CMAPL_STRCMP("privilege", node_name) == 0){
                        counter_def.priv = (char*)xmlNodeGetContent(current_node);
                } else if (CMAPL_STRCMP("x_name", node_name) == 0){
                        counter_def.num_x_items++;
                } else if (CMAPL_STRCMP("y_name", node_name) == 0){
                        counter_def.num_y_items++;
                } else {
			xmlFree((xmlChar*)counter_def.name);
			xmlFree((xmlChar*)counter_def.desc);
			xmlFree((xmlChar*)counter_def.priv);
			return CMAPL_PARSING_ERROR;
		}
        }

	// parse x_names and y_names if array counter
	const char* x_names[counter_def.num_x_items];
	const char* y_names[counter_def.num_y_items];
	uint32_t num_x_names = 0;
        uint32_t num_y_names = 0;
	if (counter_def.num_x_items != 0 || counter_def.num_y_items != 0){
		current_node = NULL;
		for (current_node = counter_node->children; current_node; current_node = current_node->next){
			const char* node_name = (const char*)current_node->name;
			if (CMAPL_STRCMP("x_name", node_name) == 0){
				x_names[num_x_names] = (char*)xmlNodeGetContent(current_node);
				num_x_names++;
			} else if (CMAPL_STRCMP("y_name", node_name) == 0){
				y_names[num_y_names] = (char*)xmlNodeGetContent(current_node);
				num_y_names++;
			}
		}
		counter_def.x_names = x_names;
		counter_def.y_names = y_names;
	}

	cmapl_counter_handle_t* counter;
	uint32_t i;
	for (i = 0; i < CMAPL_MAX_NUM_PERIODS; i++){
		if (periods[i] != 0){
			counter_def.period = periods[i];
			err = cmapl_add_counter(object, &counter_def, &counter);
			if (err != CMAPL_SUCCESS){
				return err;
			}
		}
	}

	// free xml strings
        for (i = 0; i < num_x_names; i++){
                xmlFree((xmlChar*)x_names[i]);
        }
        for (i = 0; i < num_y_names; i++){
                xmlFree((xmlChar*)y_names[i]);
        }
        xmlFree((xmlChar*)counter_def.name);
        xmlFree((xmlChar*)counter_def.desc);
        xmlFree((xmlChar*)counter_def.priv);
	return err;
}

cmapl_error_t
cmapl_parse_iterative_record_offsets_v6_0(
			cmapl_file_handle_t * data_file,
			cmapl_period_t * period,
			uint32_t index)
{
	if (data_file == NULL || period == NULL){
		return CMAPL_NULL_POINTER;
	}

	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_header_handle_t* header = data_file->i6_0.header;
	cmapl_offset_t iterative_record_offset = CMAPL_GET_OFFSET(period->record_offsets[index]);
	iterative_record_offset += header->i6_0.section_defs[CMAPL_OBJECT_INDEX].section_offset;
	int num = CMAPL_FSEEK(data_file->i6_0.metadata_file, (long int)iterative_record_offset, SEEK_SET);
	if (num != 0){
		return CMAPL_SEEK_FAILED;
	}
	uint32_t num_iterative_records = 0;
	num = (int)CMAPL_FREAD(&num_iterative_records, sizeof(uint32_t), data_file->i6_0.metadata_file);
	if (num != sizeof(uint32_t)){
		return CMAPL_READ_FAILED;
	}

        if (num_iterative_records > CMAPL_MAX_NUM_ITERATIVE_RECORDS) {
                return CMAPL_INVALID_NUM_PARTIAL_RECORDS;
        }

	cmapl_iterative_record_offsets_t* iter_record = &(period->iterative_record_offsets[index]);
	iter_record->num_offsets = num_iterative_records;
	size_t iter_record_offsets_size = num_iterative_records * sizeof(cmapl_offset_t);
	iter_record->offsets = (cmapl_offset_t*)CMAPL_ZALLOC(iter_record_offsets_size);
	if (iter_record->offsets == NULL){
		return CMAPL_NULL_POINTER;
	}
	iter_record->offsets_size = num_iterative_records;
	num = (int)CMAPL_FREAD(iter_record->offsets, iter_record_offsets_size, data_file->i6_0.metadata_file);
	if (num != iter_record_offsets_size){
                return CMAPL_READ_FAILED;
        }

	return err;
}

cmapl_error_t
cmapl_parse_object_v6_0(
                        xmlNode * object_node,
                        cmapl_header_handle_t * header)
{
	if (header == NULL){
                return CMAPL_NULL_POINTER;
        }

	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_object_def_t_v6_0 object_def;
	CMAPL_BZERO(&object_def, sizeof(object_def));
	object_def.major_v = 6;
	object_def.minor_v = 0;

	char* xml_str;
	xmlNode *current_node = NULL;
        for (current_node = object_node->children; current_node; current_node = current_node->next){
                const char* node_name = (const char*)current_node->name;
                if (CMAPL_STRCMP("name", node_name) == 0){
			object_def.name = (char*)xmlNodeGetContent(current_node);
                } else if (CMAPL_STRCMP("id", node_name) == 0){
			xml_str = (char*)xmlNodeGetContent(current_node);
			object_def.obj_id = CMAPL_ATOI(xml_str);
			xmlFree(xml_str);
		}
        }

	cmapl_object_handle_t* object;
	err = cmapl_add_object(header, &object_def, &object);
	if (err != CMAPL_SUCCESS){
		return err;
	}
	xmlFree((xmlChar*)object_def.name);

	for (current_node = object_node->children; current_node; current_node = current_node->next){
                const char* node_name = (const char*)current_node->name;
                if (CMAPL_STRCMP("counter", node_name) == 0){
			err = cmapl_parse_counter_v6_0(current_node, object);
			if (err != CMAPL_SUCCESS){
				return err;
			}
                }
        }

	return err;
}

cmapl_error_t
cmapl_parse_object_index_v6_0(
			cmapl_file_handle_t * data_file)
{
	if (data_file == NULL){
		return CMAPL_NULL_POINTER;
	}

	// read in object index's object period table into memory
	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_header_handle_t* header = data_file->i6_0.header;
	cmapl_offset_t object_index_offset = header->i6_0.section_defs[CMAPL_OBJECT_INDEX].section_offset;
	long num = CMAPL_FSEEK(data_file->i6_0.metadata_file, (long)object_index_offset, SEEK_SET);
	if (num != 0){
                return CMAPL_SEEK_FAILED;
        }
	size_t object_index_size = (size_t)header->i6_0.num_objects_sampled * (size_t)CMAPL_MAX_NUM_PERIODS * sizeof(cmapl_offset_t);
	header->i6_0.object_index = (cmapl_offset_t*)CMAPL_ZALLOC(object_index_size);
	if (header->i6_0.object_index == NULL){
		return CMAPL_NULL_POINTER;
	}
	num = (long)CMAPL_FREAD(header->i6_0.object_index, object_index_size, data_file->i6_0.metadata_file);
	if (num != object_index_size){
                return CMAPL_READ_FAILED;
        }

	cmapl_object_handle_t* object = header->i6_0.object_list;
	uint32_t obj_count = 0;
	while(object != NULL){
		cmapl_period_t* period = object->i6_0.period_list;
		while(period != NULL){
			err = cmapl_parse_record_offsets_v6_0(data_file, period, obj_count);
			if (err != CMAPL_SUCCESS){
				return err;
			}
			period = period->next_period;
		}
		object = object->i6_0.next_object;
		obj_count++;
	}

	return err;
}

cmapl_error_t
cmapl_parse_objects_v6_0(
			xmlNode * preset_node,
			cmapl_header_handle_t * header)
{
	if (header == NULL){
		return CMAPL_NULL_POINTER;
	}

	cmapl_error_t err = CMAPL_SUCCESS;

	xmlNode *current_node = NULL;
        for (current_node = preset_node->children; current_node; current_node = current_node->next){
                const char* node_name = (const char*)current_node->name;
                if (CMAPL_STRCMP("object", node_name) == 0){
			err = cmapl_parse_object_v6_0(current_node, header);
			if (err != CMAPL_SUCCESS){
				return err;
			}
                } else {
			return CMAPL_PARSING_ERROR;
		}
        }
	return err;
}

cmapl_error_t
cmapl_parse_record_offsets_v6_0(
			cmapl_file_handle_t * data_file,
			cmapl_period_t * period,
			uint32_t obj_count)
{
	if (data_file == NULL || period == NULL){
		return CMAPL_NULL_POINTER;
	}

	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_header_handle_t * header = data_file->i6_0.header;
	cmapl_offset_t period_offset = 0;
	err = cmapl_get_object_period_offset(header, obj_count, period->period, &period_offset);
	if (err != CMAPL_SUCCESS){
		return err;
	}
	period_offset += header->i6_0.section_defs[CMAPL_OBJECT_INDEX].section_offset;
	long num = CMAPL_FSEEK(data_file->i6_0.metadata_file, (long)period_offset, SEEK_SET);
	if (num != 0){
                return CMAPL_SEEK_FAILED;
        }
	unsigned long record_offsets_size = period->num_offsets * sizeof(cmapl_offset_t);
	num = (long)CMAPL_FREAD(period->record_offsets, record_offsets_size, data_file->i6_0.metadata_file);
	if (num != record_offsets_size){
                return CMAPL_READ_FAILED;
        }
	uint32_t index;
	for (index = 0; index < period->num_offsets; index++){
		if (CMAPL_IS_ITERATIVE(period->record_offsets[index])){
			err = cmapl_parse_iterative_record_offsets_v6_0(data_file, period, index);
			if (err != CMAPL_SUCCESS){
				return err;
			}
		}
	}

	return err;
}

cmapl_error_t
cmapl_parse_section_def_v6_0(
                        xmlNode * section_def_node,
                        cmapl_header_handle_t * header)
{
	// get section index
	uint32_t section_index = (uint32_t)(-1);
        xmlNode *current_node = NULL;
	char* xml_str;
        for (current_node = section_def_node->children; current_node; current_node = current_node->next){
                const char* node_name = (const char*)current_node->name;
                if (CMAPL_STRCMP("section_type", node_name) == 0){
			xml_str = (char*)xmlNodeGetContent(current_node);
                        section_index = CMAPL_ATOI(xml_str);
			xmlFree(xml_str);
			break;
                }
        }
	if (section_index == (uint32_t)(-1)){
		return CMAPL_PARSING_ERROR;
	}

	// fill in section def info
	for (current_node = section_def_node->children; current_node; current_node = current_node->next){
                const char* node_name = (const char*)current_node->name;
                if (CMAPL_STRCMP("section_type", node_name) == 0){
			xml_str = (char*)xmlNodeGetContent(current_node);
                        header->i6_0.section_defs[section_index].section_type = CMAPL_ATOI(xml_str);
			xmlFree(xml_str);
                } else if (CMAPL_STRCMP("section_version", node_name) == 0){
			xml_str = (char*)xmlNodeGetContent(current_node);
			header->i6_0.section_defs[section_index].section_version = CMAPL_ATOI(xml_str);
			xmlFree(xml_str);
                } else if (CMAPL_STRCMP("section_offset", node_name) == 0){
			char* strtoll_str = NULL;
			xml_str = (char*)xmlNodeGetContent(current_node);
			header->i6_0.section_defs[section_index].section_offset = header->i6_0.header_offset +
				CMAPL_STRTOULL(xml_str, &strtoll_str, 16);
			if (CMAPL_STRCMP(strtoll_str, "\0") != 0){
				xmlFree(xml_str);
				return CMAPL_PARSING_ERROR;
			}
			xmlFree(xml_str);
                } else {
                        return CMAPL_PARSING_ERROR;
                }
        }
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_parse_system_v6_0(
			xmlNode * system_node,
			cmapl_file_handle_t * data_file,
			cmapl_header_def_t_v6_0 * header_def,
			cmapl_header_handle_t ** header)
{
	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_system_def_t_v6_0 system_def;
	CMAPL_BZERO(&system_def, sizeof(system_def));
	system_def.major_v = 6;
	system_def.minor_v = 0;

	char* xml_str;
	xmlNode *current_node = NULL;
	for (current_node = system_node->children; current_node; current_node = current_node->next){
		const char* node_name = (const char*)current_node->name;
		if (CMAPL_STRCMP("version", node_name) == 0){
			system_def.ontap_version = (char*)xmlNodeGetContent(current_node);
		} else if (CMAPL_STRCMP("base_time", node_name) == 0){
			char* strtoll_str = NULL;
			xml_str = (char*)xmlNodeGetContent(current_node);
                        system_def.base_time = CMAPL_STRTOULL(xml_str, &strtoll_str, 10);
			if (CMAPL_STRCMP(strtoll_str, "\0") != 0){
				xmlFree(xml_str);
				return CMAPL_PARSING_ERROR;
			}
			xmlFree(xml_str);
                } else if (CMAPL_STRCMP("node_uuid", node_name) == 0){
			system_def.node_uuid = (char*)xmlNodeGetContent(current_node);
		} else if (CMAPL_STRCMP("hostname", node_name) == 0){
			system_def.hostname = (char*)xmlNodeGetContent(current_node);
                } else if (CMAPL_STRCMP("model", node_name) == 0){
			system_def.model = (char*)xmlNodeGetContent(current_node);
                } else if (CMAPL_STRCMP("serial_number", node_name) == 0){
			system_def.serial_number = (char*)xmlNodeGetContent(current_node);
                } else if (CMAPL_STRCMP("system_id", node_name) == 0){
			system_def.system_id = (char*)xmlNodeGetContent(current_node);
                } else {
			err = CMAPL_PARSING_ERROR;
			goto clean_up;
		}
	}

	err = cmapl_create_header(data_file, header_def, &system_def, header);
clean_up:
	xmlFree((xmlChar*)system_def.ontap_version);
	xmlFree((xmlChar*)system_def.node_uuid);
	xmlFree((xmlChar*)system_def.hostname);
	xmlFree((xmlChar*)system_def.model);
	xmlFree((xmlChar*)system_def.serial_number);
	xmlFree((xmlChar*)system_def.system_id);
	return err;
}

cmapl_error_t
cmapl_parse_xml_header_v6_0(
			xmlNode * root_node,
			cmapl_file_handle_t * data_file)
{
	if (root_node == NULL || data_file == NULL){
		return CMAPL_NULL_POINTER;
	}
	cmapl_header_handle_t* header = NULL;
	char* strtoll_str = NULL;
	cmapl_error_t err = CMAPL_SUCCESS;

	// get archiver attributes
	cmapl_offset_t header_offset = 0;
	cmapl_header_def_t_v6_0 header_def;
	CMAPL_BZERO(&header_def, sizeof(header_def));
	header_def.major_v = 6;
	header_def.minor_v = 0;
	char* xml_str;
	if (CMAPL_STRCMP("archive", (const char*)root_node->name) == 0){
		xml_str = (char*)xmlGetProp(root_node, (const xmlChar*)"generation");
		header_def.generation_id = CMAPL_ATOI(xml_str);
		xmlFree(xml_str);
		header_def.preset_version = (char*)xmlGetProp(root_node, (const xmlChar*)"preset_version");
		xml_str = (char*)xmlGetProp(root_node, (const xmlChar*)"duration");
		data_file->i6_0.duration = CMAPL_STRTOULL(xml_str, &strtoll_str, 16);
		if (CMAPL_STRCMP(strtoll_str, "\0") != 0){
			xmlFree(xml_str);
                        err = CMAPL_PARSING_ERROR;
			goto clean_up;
                }
		xmlFree(xml_str);
		xml_str = (char*)xmlGetProp(root_node, (const xmlChar*)"header_offset");
		header_offset = CMAPL_STRTOULL(xml_str, &strtoll_str, 10);
		if (CMAPL_STRCMP(strtoll_str, "\0") != 0){
			xmlFree(xml_str);
			err = CMAPL_PARSING_ERROR;
                        goto clean_up;
		}
		xmlFree(xml_str);
		xml_str = (char*)xmlGetProp(root_node, (const xmlChar*)"start_time");
		data_file->i6_0.start_time = CMAPL_STRTOULL(xml_str, &strtoll_str, 16);
                if (CMAPL_STRCMP(strtoll_str, "\0") != 0){
			xmlFree(xml_str);
			err = CMAPL_PARSING_ERROR;
                        goto clean_up;
                }
		xmlFree(xml_str);
		xml_str = (char*)xmlGetProp(root_node, (const xmlChar*)"end_time");
		data_file->i6_0.end_time = CMAPL_STRTOULL(xml_str, &strtoll_str, 16);
                if (CMAPL_STRCMP(strtoll_str, "\0") != 0){
			xmlFree(xml_str);
			err = CMAPL_PARSING_ERROR;
                        goto clean_up;
                }
		xmlFree(xml_str);
	}else{
		return CMAPL_PARSING_ERROR;
	}

	// check if start and end times are valid
	if (data_file->i6_0.end_time <= data_file->i6_0.start_time) {
		err = CMAPL_INVALID_TIMESTAMP;
		goto clean_up;
	}

	// parse each element and create internal handles to the whole file
	xmlNode *current_node = NULL;
	for (current_node = root_node->children; current_node; current_node = current_node->next) {
		const char* node_name = (const char*)current_node->name;
		if (CMAPL_STRCMP("system", node_name) == 0){
			err = cmapl_parse_system_v6_0(current_node, data_file, &header_def, &header);
			if (err != CMAPL_SUCCESS){
				goto clean_up;
			}
			header->i6_0.header_offset = header_offset;
		} else if (CMAPL_STRCMP("attribute_def", node_name) == 0){
			if(header == NULL){
				err = CMAPL_PARSING_ERROR;
				goto clean_up;
			}
			err = cmapl_parse_attribute_def_v6_0(current_node, header);
			if (err != CMAPL_SUCCESS){
				goto clean_up;
                        }
		} else if (CMAPL_STRCMP("section_def", node_name) == 0){
			if(header == NULL){
				err = CMAPL_PARSING_ERROR;
                                goto clean_up;
                        }
			err = cmapl_parse_section_def_v6_0(current_node, header);
			if (err != CMAPL_SUCCESS){
				goto clean_up;
                        }
                } else if (CMAPL_STRCMP("preset", node_name) == 0){
			if(header == NULL){
				err = CMAPL_PARSING_ERROR;
                                goto clean_up;
                        }
			err = cmapl_parse_objects_v6_0(current_node, header);
			if (err != CMAPL_SUCCESS){
				goto clean_up;
                        }
		} else {
			return CMAPL_PARSING_ERROR;
		}
	}

	// update actual UTC start and end time
	data_file->e6_0.start_time = data_file->i6_0.start_time + data_file->i6_0.header->i6_0.system.e6_0.base_time;
	data_file->e6_0.end_time = data_file->i6_0.end_time + data_file->i6_0.header->i6_0.system.e6_0.base_time;

clean_up:
	xmlFree((xmlChar*)header_def.preset_version);
	return err;
}

cmapl_error_t
cmapl_set_counter_value_offsets_v6_0(
			cmapl_header_handle_t * header)
{
	if (header == NULL){
		return CMAPL_INVALID_ARGUMENTS;
	}

	cmapl_object_handle_t* object = header->i6_0.object_list;
	while (object != NULL){
		cmapl_counter_handle_t* counter = object->i6_0.cnt_list;
		while (counter != NULL){
			if (counter->e6_0.is_key == true){
				cmapl_period_t* period = object->i6_0.period_list;
				while (period != NULL){
					counter->i6_0.record_offsets[period->period_index] = period->cnts_size;
					period->cnts_size += counter->e6_0.size;
					period = period->next_period;
				}
			} else{
				uint32_t i;
				for (i = 0; i < CMAPL_MAX_NUM_PERIODS; i++){
					cmapl_period_t* period = counter->i6_0.periods[i];
					if (period != NULL){
						counter->i6_0.record_offsets[period->period_index] = period->cnts_size;
						period->cnts_size += counter->e6_0.size;
					}
				}
			}
			counter = counter->i6_0.next_cnt_by_id;
		}
		object = object->i6_0.next_object;
	}
	return CMAPL_SUCCESS;
}
