/**
 * The CMAPL test module. This program contains a bunch of unit tests
 * to make sure the CMAPL is functioning properly. Some of these tests
 * include:
 *
 *	- Testing the write-side library
 *		- Creating a data file
 *		- Creating the XML header
 *		- Writing records to a data file
 *		- Writing iterative records
 *		- Writing the object index
 *	- Testing the read-side library
 *		- Opening and parsing a directory
 *		- Opening and parsing a data file
 *		- Compare what's read matches what was written
 *		- Reading the object index
 *		- Making a query on a data file
 *		- Getting the sample results from a query
 *		- Closing a data file and directory
 *	- Comparing a data file against the default preset
 *		- Check if every counter that should be sampled
 *                in the default preset is actually sampled
 *		- Check if all counters and objects are sampled
 *                at the correct period
 *		- Check if at least 90% of the expected samples
 *                show up
 *
 * TODO: Modularize tests more where ever possible.
 */

#include "cmapl_common_defs.h"
#include "cmapl_int_defs.h"
#include "cmapl_write.h"
#include "cmapl_write_int.h"
#include "cmapl_read.h"
#include "cmapl_read_int.h"
#include "cmapl_directory.h"
#include "cmapl_directory_int.h"
#include <errno.h>
#include <unistd.h>
#if defined(__linux__)
#include "../../../countergen/bedrock/export/common/headers/cm_preset.auto.h"
#else
#include <countergen/cm_preset.auto.h>
#endif

/**
 * @name        cmapl_create_test_record_buffer
 *
 * @param[in]	counters	List of counter handles for some object
 * @param[in]   num_counters	Number of counters for some object
 * @param[in]   period		The period of the record to create
 * @param[in]   msecs		The msecs since base of the record to create
 * @param[in]   num_instances	Number of instances of this record
 * @param[in]   record_index	Represents the nth iterative record to create
 * @param[out]  record_header	The header of the record
 *
 * @return      char*		The buffer of the whole record
 *
 * This function creates a record buffer given a list of counter handles, the period,
 * the number of instances, and the record index if it is an iterative record. The
 * whole record buffer will be created and the record header will be passed back
 * to the user through the record_header parameter.
 */
static char*
cmapl_create_test_record_buffer(
	cmapl_counter_handle_t ** counters,
	uint32_t num_counters,
	uint64_t period,
	uint64_t msecs,
	uint32_t num_instances,
	uint32_t record_index,
	cmapl_record_header_data_t * record_header)
{
	// need at least one counter and instance to create a record
	if (num_counters == 0 || num_instances == 0){
                return NULL;
        }

	// test values for 4b/8b integers and strings
	uint32_t num32 = 1;
	uint64_t num64 = 1;
	uint32_t numstr = 1;

	uint32_t period_index = 0;
	cmapl_error_t err = cmapl_get_period_index(period, &period_index);
	if (err != CMAPL_SUCCESS){
		return NULL;
	}

	// find the number of key counters and calculate the instance size
	uint32_t instance_size = 0;
	uint32_t numkeycnt = 0;
	uint32_t i;
	for (i = 0; i < num_counters; i++){
		cmapl_counter_handle_t* counter = counters[i];
		if (counter->e6_0.is_key == true || counter->i6_0.periods[period_index] != NULL){
			instance_size += counter->e6_0.size;
			if (counter->e6_0.is_key == true){
				numkeycnt++;
			}
		}
	}
	numkeycnt = num_instances*numkeycnt*record_index + 1;

	// get object associated with these counters
	cmapl_object_handle_t* object = counters[0]->i6_0.object;

	// create the record buffer
	uint32_t buffer_size = instance_size * num_instances;
	char* record_buffer = (char*)CMAPL_ZALLOC(buffer_size * sizeof(char));
	if (record_buffer == NULL){
		return NULL;
	}
	char* current_position = record_buffer;
	for (i = 0; i < num_instances; i++){
		// create each instance in the record
		uint32_t j;
		for (j = 0; j < num_counters; j++){
			// add each counter data to an instance
			cmapl_counter_handle_t* counter = counters[j];
			uint32_t cnt_size;
			uint32_t num_values;
			if (counter->e6_0.num_x_items == 0) {
				num_values = 1;
			} else if (counter->e6_0.num_y_items == 0) {
				num_values = counter->e6_0.num_x_items;
			} else {
				num_values = counter->e6_0.num_x_items * counter->e6_0.num_y_items;
			}
			cnt_size = counter->e6_0.size / num_values;

			uint32_t k;
			for (k = 0; k < num_values; k++){
				// fill in counter data value depending on the counter size
				if (current_position == NULL){
					CMAPL_FREE(record_buffer);
					return NULL;
				}
				if (counter->e6_0.is_key == true){
					if (cnt_size == 4){
						CMAPL_MEMCPY(current_position, &numkeycnt, sizeof(uint32_t));
						numkeycnt++;
						current_position += sizeof(uint32_t);
					} else if (cnt_size == 8){
						uint64_t temp_keycnt = numkeycnt;
						CMAPL_MEMCPY(current_position, &temp_keycnt, sizeof(uint64_t));
						numkeycnt++;
						current_position += sizeof(uint64_t);
					} else if (cnt_size > 8){
						char str[cnt_size];
						sprintf(str, "%d",numkeycnt);
						strcat(str, "_str_key_cnt");
						CMAPL_MEMCPY(current_position, str, cnt_size * sizeof(char));
						numkeycnt++;
						current_position += (cnt_size * sizeof(char));
					}
				} else if (counter->i6_0.periods[period_index] != NULL){
                                        if (cnt_size == 4){
                                                CMAPL_MEMCPY(current_position, &num32, sizeof(uint32_t));
                                                num32++;
                                                current_position += sizeof(uint32_t);
                                        } else if (cnt_size == 8){
                                                CMAPL_MEMCPY(current_position, &num64, sizeof(uint64_t));
                                                num64++;
                                                current_position += sizeof(uint64_t);
                                        } else if (cnt_size > 8){
                                                char str[cnt_size];
					        sprintf(str, "%d",numstr);
                                                strcat(str, "_str_cnt");
                                                CMAPL_MEMCPY(current_position, str, cnt_size * sizeof(char));
                                                numstr++;
                                                current_position += (cnt_size * sizeof(char));
                                        }
				}
			}
		}
	}

	// fill in record header information
        cmapl_set_latest_version(record_header);
        (*record_header).size = buffer_size;
        (*record_header).period = period;
        (*record_header).msecs_since_base = msecs;
        (*record_header).obj_id = object->e6_0.obj_id;

	return record_buffer;
}

/**
 * @name	cmapl_get_period_value
 *
 * @param[in]	period		A period string from the preset header
 *
 * @return	uint64_t	The integer value of the string in milliseconds
 *
 * This function converts a period string from the preset header to an
 * actual integer value of the period in milliseconds.
 */
static uint64_t
cmapl_get_period_value(
	const char * period)
{
	if (CMAPL_STRCMP(period, CMAPL_CONFIG_PRESET_STR) == 0){
		return CMAPL_CONFIG_PERIOD;
	}

	uint64_t p_val = 0;
	uint32_t p_len = (uint32_t)CMAPL_STRLEN(period);
	char* p = (char*)CMAPL_ZALLOC(p_len);
	if (p == NULL){
		return 0;
	}
	CMAPL_STRNCPY(p, period, p_len - 1);
	char* strtoll_str = NULL;
	p_val = CMAPL_STRTOULL(p, &strtoll_str, 10);
	if (CMAPL_STRCMP(strtoll_str, "\0") != 0){
		CMAPL_FREE(p);
		return 0;
	}
	CMAPL_FREE(p);

	char t = period[p_len - 1];
	switch(t)
	{
		case 's':
			p_val *= 1000;
			break;
		case 'm':
			p_val *= 60*1000;
			break;
		case 'h':
			p_val *= 60*60*1000;
			break;
		case 'd':
			p_val *= 24*60*60*1000;
			break;
		case 'w':
			p_val *= 7*24*60*60*1000;
			break;
		default:
			return 0;
	}

	return p_val;
}

// TODO: Is it possible to make these checks more extensible when
//       new enum values are added?
static bool
cmapl_is_valid_counter_types(
	cmapl_cnt_size_t c_size,
	cmapl_cnt_type_t c_type,
	cmapl_cnt_unit_t c_unit)
{
	switch(c_size){
		case CM_CNT_4B:
		case CM_CNT_8B:
		case CM_CNT_16B:
		case CM_CNT_32B:
		case CM_CNT_64B:
		case CM_CNT_96B:
		case CM_CNT_128B:
		case CM_CNT_160B:
		case CM_CNT_252B:
			break;
		default:
			return false;
	}

	switch(c_type){
                case CM_CNT_RAW:
                case CM_CNT_RATE:
                case CM_CNT_DELTA:
                case CM_CNT_PERCT:
                case CM_CNT_AVERAGE:
                case CM_CNT_STRING:
                        break;
                default:
                        return false;
        }

	switch(c_unit){
                case CM_NO_SUFFIX:
                case CM_PER_SEC:
		case CM_PERCT:
		case CM_B_PER_SEC:
		case CM_KB_PER_SEC:
		case CM_MSECS:
		case CM_USECS:
		case CM_SECS:
		case CM_MB_PER_SEC:
		case CM_MB:
		case CM_CNT_NODISP:
                        break;
                default:
                        return false;
        }

	return true;
}

/**
 * @name        cmapl_is_counter_valid
 *
 * @param[in]	data_file	Handle of the data file
 * @param[in]   object		Handle of the object
 * @param[in]   counter_name	Name of the counter to check
 * @param[in]   period		The period of the counter
 * @param[in]   start_time	The time of the first sample to search for
 * @param[in]   end_time	The time of the last sample to search for
 *
 * @return	bool		Flag that specifies if this counter is valid
 *
 * This function checks if a certain counter is sampled in a data file between a
 * given time range a valid number of times depending on it's period. It calculates
 * the expected number of times this counter should have been sampled, and it checks
 * if the actual number of times it was sampled is at least 90% of the expected. If
 * it was sampled an appropriate amount of time, then the function returns true.
 * Otherwise it will return false.
 */
static bool
cmapl_is_counter_valid(
	cmapl_file_handle_t * data_file,
	cmapl_object_handle_t * object,
	char * counter_name,
	uint64_t period,
	uint64_t start_time,
	uint64_t end_time)
{
	cmapl_error_t error = CMAPL_SUCCESS;

	cmapl_counter_handle_t* counter;
	error = cmapl_find_counter_by_name(object, counter_name, &counter);
	if (error != CMAPL_SUCCESS){
		fprintf(stderr, "ERROR: TEST FAILED FINDING OBJECT %s COUNTER %s: %d!\n", object->e6_0.name, counter_name, error);
		return false;
	}

	// if counter is a key counter, it is implicitly valid if all the
	// other counters are valid.
	if (counter->e6_0.is_key){
		return true;
	}

	// check if counter types are valid
	cmapl_cnt_size_t c_size;
	cmapl_cnt_type_t c_type;
	cmapl_cnt_unit_t c_unit;
	error = cmapl_get_counter_type(counter, &c_size, &c_type, &c_unit);
	if (error != CMAPL_SUCCESS){
		fprintf(stderr, "ERROR: TEST FAILED GETTING COUNTER TYPES FOR OBJECT %s COUNTER %s: %d!\n", object->e6_0.name, counter_name, error);
		return false;
	}
	bool is_ctype_valid = cmapl_is_valid_counter_types(c_size, c_type, c_unit);
	if (!is_ctype_valid){
		fprintf(stderr, "ERROR: TEST COUNTER TYPE NOT VALID FOR OBJECT %s COUNTER %s!\n", object->e6_0.name, counter_name);
                return false;
        }

	uint32_t period_index = 0;
	error = cmapl_get_period_index(period, &period_index);
	if(error != CMAPL_SUCCESS || counter->i6_0.periods[period_index] == NULL){
		fprintf(stderr, "ERROR: TEST GET PERIOD INDEX FAILED FOR OBJECT %s COUNTER %s: %d!\n", object->e6_0.name, counter_name, error);
		return false;
	}

	const char* object_name = object->e6_0.name;

	// query for this counter at the given period
	cmapl_query_def_t query_def;
        cmapl_set_latest_version(&query_def);
        query_def.object_name = object_name;
        query_def.num_cnts = 1;
        const char* cnt_names[] = {counter_name};
        query_def.cnt_names = cnt_names;
        query_def.start_time = start_time;
        query_def.end_time = end_time;
        query_def.period = period;
        query_def.num_key_cnts = 0;

        cmapl_query_handle_t* query;
        error = cmapl_query(data_file, &query_def, &query);
        if (error != CMAPL_SUCCESS){
		fprintf(stderr, "ERROR: TEST FAILED MAKING QUERY FOR OBJECT %s COUNTER %s: %d!\n", object->e6_0.name, counter_name, error);
                return false;
        }

	// calculate number of results expected based on the duration and period
	float expected_results = 0.0;
	uint64_t duration = end_time - start_time;
	if (period == CMAPL_CONFIG_PERIOD){
		expected_results = 1.0;
	} else{
		expected_results = (1.0 * duration) / period;
	}

	// get the actual number of samples for this counter; the number of samples
	// is equal to the number of timestamps found in the query
	uint64_t actual_results = query->i6_0.num_timestamps;

	// verify if actual number of times sampled is at least 90%
	// of expected number of times sampled
	uint64_t valid_amount = (unsigned) (long) (expected_results * 9.0 / 10.0);
	if (actual_results < valid_amount){
		fprintf(stderr, "ERROR: OBJECT %s COUNTER %s PERIOD %lums NOT SAMPLED ENOUGH OVER %lums (%lu < floor(90pct of %f))!\n",
			object->e6_0.name, counter_name, period, duration, actual_results, expected_results);
		return false;
	}

	return true;
}

/**
 * @name        cmapl_preset_validation
 *
 * @param[in]	data_file	The file handle of the archive file to validate
 * @param[in]	preset		The preset to validate
 *
 * @return	cmapl_error_t	Error flag
 *
 * This function compares a data file against a certain preset defined
 * in the preset header. It'll check if every counter that's supposed to
 * be sampled in this data file is actually sampled. And it checks that
 * each counter contains at least 90% of the samples that are expected to
 * be there.
 */
static cmapl_error_t
cmapl_preset_validation(
	cmapl_file_handle_t * data_file,
	const char * preset)
{
	cmapl_error_t error = CMAPL_SUCCESS;
	bool file_is_valid = true;

	cmapl_header_handle_t * header;
        error = cmapl_get_header(data_file, &header);
        if (error != CMAPL_SUCCESS){
		fprintf(stderr, "ERROR: TEST GET HEADER FAILED : %d!\n", error);
                goto preset_fail;
        }

	// get the duration of the file; this gives an idea of what counters
	// were sampled in this time range
	uint64_t start_time = 0;
	error = cmapl_get_oldest_record_time(data_file, &start_time);
	if (error != CMAPL_SUCCESS){
		fprintf(stderr, "ERROR: TEST GET OLDEST RECORD TIME FAILED : %d!\n", error);
		goto preset_fail;
	}

	uint64_t end_time = 0;
	error = cmapl_get_newest_record_time(data_file, &end_time);
	if (error != CMAPL_SUCCESS){
		fprintf(stderr, "ERROR: TEST GET NEWEST RECORD TIME FAILED : %d!\n", error);
		goto preset_fail;
	}
	uint64_t duration = end_time - start_time;

	// find the preset that was requested
	uint32_t preset_num;
	for (preset_num = 0; preset_num < cm_preset_import_defs_count; preset_num++){
		const char* preset_name = cm_preset_import_defs[preset_num].preset;
		if(CMAPL_STRCMP(preset, preset_name) == 0){
			break;
		}
	}
	if (preset_num == cm_preset_import_defs_count){
		error = CMAPL_INVALID_ARGUMENTS;
		fprintf(stderr, "CMAPL PRESET DOESNT EXIST : %d!\n", error);
		goto preset_fail;
	}

	// check every object if its counters are sampled correctly
	const cm_cdf_preset_import_t* preset_def = cm_preset_import_defs + preset_num;
	uint32_t num_details = preset_def->num_details;
	uint32_t i;
	for (i = 0; i < num_details; i++){
		const cm_cdf_preset_detail_import_t* detail = preset_def->details + i;
		char* object_name = (char*)detail->object;

		// remove "*:constituent" endings from object names
		char* o_name = CMAPL_STRSTR(object_name, ":constituent");
		uint32_t o_name_len = 1;
		if (o_name != NULL){
			o_name_len = (uint32_t)(o_name - object_name) + 1;
		}
		char temp[o_name_len];
		if (o_name != NULL){
			int num = CMAPL_SSCANF(object_name, "%s:constituent", temp);
			if (num != 1){
				error = CMAPL_PARSING_ERROR;
				fprintf(stderr, "ERROR: TEST SSCANF FAILED : %d!\n", error);
				goto preset_fail;
			}
			temp[o_name_len-1] = '\0';
			o_name = temp;
		} else{
			o_name = object_name;
		}

		cmapl_object_handle_t * object;
		error = cmapl_find_object(header, o_name, &object);
		if (error != CMAPL_SUCCESS){
			fprintf(stderr, "ERROR: TEST FIND OBJECT FAILED : %d!\n", error);
			goto preset_fail;
		}

		// verify counters for each of this object's periods
		uint32_t num_periods = detail->num_periods;
		uint32_t p;
		for (p = 0; p < num_periods; p++){
			const char* period = detail->periods[p];
			uint64_t p_val = cmapl_get_period_value(period);
			if (p_val != CMAPL_CONFIG_PERIOD && (p_val == 0 || p_val > duration)){
				continue;
			}
			uint32_t p_index = 0;
			error = cmapl_get_period_index(p_val, &p_index);
			char* counters = CMAPL_STRDUP(detail->counter_sets[p]);
			if (counters == NULL){
				error = CMAPL_NULL_POINTER;
				fprintf(stderr, "ERROR: TEST STRDUP FAILED : %d!\n", error);
				goto preset_fail;
			}
			char* counter_name = CMAPL_STRTOK(counters, ",");
			// verify each counter sampled at this period
			while (counter_name != NULL){
				bool is_valid = cmapl_is_counter_valid(data_file, object, counter_name, p_val, start_time, end_time);
				if (!is_valid){
					if (file_is_valid){
						fprintf(stderr, "The following counters are not sampled frequently enough:\n");
						file_is_valid = false;
					}
//					fprintf(stderr, "object name: %s, period: %"PRIu64", counter_name: %s\n", o_name, p_val, counter_name);
				}
				counter_name = CMAPL_STRTOK(NULL, ",");
			}
			CMAPL_FREE(counters);
		}
	}

	if(file_is_valid){
		printf("CMAPL PRESET VALIDATION PASSED!\n");
	} else {
preset_fail:
		fprintf(stderr, "CMAPL PRESET VALIDATION FAILED!\n");
		printf("CMAPL PRESET VALIDATION FAILED!\n");
	}

	return error;
}

static cmapl_error_t
cmapl_test_write_side(cmapl_file_handle_t ** data_file)
{
	// create data file
	cmapl_file_handle_t * file;
	const char* meta_path;
	const char* data_path;

	// format of an archive file is
	// "<file_index>_<longest_period_sampled_in_file_secs>_<UTC_start_time_msecs>_<duration_msecs>.ccma"
	cmapl_error_t error = cmapl_create_data_file("000001_000060_72000000_7200000.ccma", 6, 0, 60*60*1000, &file, &meta_path, &data_path);
	if (error != CMAPL_SUCCESS){
		fprintf(stderr, "Create file error %d\n", error);
	}

	// initialize XML header information
	cmapl_header_def_t header_def;
	cmapl_set_latest_version(&header_def);
	header_def.generation_id = 1;
	header_def.preset_version = "5.5";

	// initialize ontap information
	cmapl_system_def_t system_def;
	cmapl_set_latest_version(&system_def);
	system_def.ontap_version = "NetApp Release stats_fresh_1407687_1107191437 7-Mode: Tue Jul 12 01:20:33 EDT 2011";
	system_def.base_time = 0;
	system_def.node_uuid = "57ba303e-c127-11e1-a474-09e16d0761d3";
	system_def.hostname = "crane";
	system_def.model = "FAS3170";
	system_def.serial_number = "30025170";
	system_def.system_id = "0151700070";

	// create XML header
	cmapl_header_handle_t * header;
	error = cmapl_create_header(file, &header_def, &system_def, &header);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "CREATE HEADER FAILED\n");
        }

	// add attribute definitions to XML header
	cmapl_attribute_def_t attr_def;
	cmapl_set_latest_version(&attr_def);
	attr_def.name = "ctype";
	attr_def.value = 4;
	attr_def.desc = "CM_CNT_4B";
	error = cmapl_add_attr_def(header, &attr_def);
	if (error != CMAPL_SUCCESS){
		fprintf(stderr, "ADD ATTR FAILED\n");
	}

	attr_def.name = "ctype";
	attr_def.value = 8;
	attr_def.desc = "CM_CNT_8B";
	error = cmapl_add_attr_def(header, &attr_def);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD ATTR FAILED\n");
        }

	// add object to XML header
	cmapl_object_def_t object_def;
	cmapl_set_latest_version(&object_def);
	object_def.name = "aggregate";
	object_def.obj_id = 8;

	cmapl_object_handle_t * object;
	error = cmapl_add_object(header, &object_def, &object);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD OBJECT FAILED\n");
        }

	// add counter to first object
	cmapl_counter_def_t counter_def;
	cmapl_set_latest_version(&counter_def);
	counter_def.ctype = 18948348;
	counter_def.size = 64;
	counter_def.priv = "CM_DIAG";
	counter_def.is_key = 1;
	counter_def.period = 30*1000;
	counter_def.name = "parent_host";
	counter_def.cnt_id = 7;
	counter_def.desc = "Name of parent host";
	counter_def.base_id = 0;
	counter_def.num_x_items = 0;
	counter_def.num_y_items = 0;

	cmapl_counter_handle_t * counter1;
	error = cmapl_add_counter(object, &counter_def, &counter1);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD COUNTER FAILED\n");
        }

	// add another counter to first object
	counter_def.ctype = 18940412;
	counter_def.size = 8;
	counter_def.priv = "CM_DIAG";
	counter_def.is_key = 0;
	counter_def.period = 60*1000;
	counter_def.name = "total_transfers";
	counter_def.cnt_id = 5;
	counter_def.desc = "Total number of transfers per second serviced by the aggregate";
	counter_def.base_id = 0;
	counter_def.num_x_items = 0;
	counter_def.num_y_items = 0;

	cmapl_counter_handle_t * counter2;
	error = cmapl_add_counter(object, &counter_def, &counter2);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD COUNTER FAILED\n");
        }

	// add another counter to first object
	counter_def.ctype = 18940412;
        counter_def.size = 8;
        counter_def.priv = "CM_DIAG";
        counter_def.is_key = 0;
        counter_def.period = 30*1000;
        counter_def.name = "num_transfers";
        counter_def.cnt_id = 3;
        counter_def.desc = "Number of transfers per second serviced by the aggregate";
        counter_def.base_id = 0;
        counter_def.num_x_items = 0;
        counter_def.num_y_items = 0;

	cmapl_counter_handle_t * counter3;
        error = cmapl_add_counter(object, &counter_def, &counter3);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD COUNTER FAILED\n");
        }

	// add another counter to first object
	counter_def.ctype = 18948348;
        counter_def.size = 64;
        counter_def.priv = "CM_DIAG";
        counter_def.is_key = 1;
        counter_def.period = 30*1000;
        counter_def.name = "instance_name";
        counter_def.cnt_id = 10;
        counter_def.desc = "Instance name of aggregate";
        counter_def.base_id = 0;
        counter_def.num_x_items = 0;
        counter_def.num_y_items = 0;

	cmapl_counter_handle_t * counter4;
        error = cmapl_add_counter(object, &counter_def, &counter4);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD COUNTER FAILED\n");
        }

	// add another counter to first object
	counter_def.ctype = 18940412;
        counter_def.size = 32;
        counter_def.priv = "CM_DIAG";
        counter_def.is_key = 0;
        counter_def.period = 60*1000;
        counter_def.name = "test_ops";
        counter_def.cnt_id = 4;
        counter_def.desc = "Ops";
        counter_def.base_id = 0;
	const char* c[] = {"0 - <1ms", "1 - <2ms", "2 - <4ms", "4 - <6ms"};
        counter_def.x_names = c;
        counter_def.num_x_items = 4;
        counter_def.num_y_items = 0 ;

        cmapl_counter_handle_t * counter15;
        error = cmapl_add_counter(object, &counter_def, &counter15);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD COUNTER FAILED\n");
        }

	// add another counter to first object
	counter_def.ctype = 18940412;
        counter_def.size = 8;
        counter_def.priv = "CM_DIAG";
        counter_def.is_key = 0;
        counter_def.period = 60*1000;
        counter_def.name = "num_transfers";
        counter_def.cnt_id = 3;
        counter_def.desc = "Number of transfers per second serviced by the aggregate";
        counter_def.base_id = 0;
        counter_def.num_x_items = 0;
        counter_def.num_y_items = 0;

        cmapl_counter_handle_t * counter13;
        error = cmapl_add_counter(object, &counter_def, &counter13);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD COUNTER FAILED: %d\n", error);
        }

	cmapl_counter_handle_t* obj1_cnts[5] = {counter3, counter15, counter2, counter1, counter4};

	// add another object to XML header
	object_def.name = "cifs";
	object_def.obj_id = 14;

	cmapl_object_handle_t * object2;
	error = cmapl_add_object(header, &object_def, &object2);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD OBJECT FAILED\n");
        }

	// add an array counter to second object
	counter_def.ctype = 18948348;
	counter_def.size = 32;
	counter_def.priv = "CM_DIAG";
	counter_def.is_key = 1;
	counter_def.period = 10*1000;
	counter_def.name = "cifs_read_latency_hist";
	counter_def.cnt_id = 1;
	counter_def.desc = "Histogram of latency for CIFS operations in milliseconds";
	counter_def.base_id = 0;
	const char* a[] = {"0 - <1ms", "1 - <2ms", "2 - <4ms", "4 - <6ms"};
	counter_def.x_names = a;
	counter_def.num_x_items = 4;
	counter_def.num_y_items = 0 ;

	cmapl_counter_handle_t * counter5;
	error = cmapl_add_counter(object2, &counter_def, &counter5);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD COUNTER FAILED\n");
        }

	// add another array counter to second object (has 2 arrays)
	counter_def.ctype = 18948348;
	counter_def.size = 64;
        counter_def.priv = "CM_DIAG";
	counter_def.is_key = 0;
	counter_def.period = CMAPL_CONFIG_PERIOD;
	counter_def.name = "cifs_latency_hist";
	counter_def.cnt_id = 7;
	counter_def.desc = "Histogram of latency for CIFS operations in milliseconds";
	counter_def.base_id = 0;
	const char* b[] = {"0 - <1ms", "1 - <2ms", "2 - <4ms", "4 - <6ms"};
	counter_def.x_names = b;
	counter_def.num_x_items = 4;
	counter_def.y_names = b;
	counter_def.num_y_items = 4;

	cmapl_counter_handle_t * counter6;
	error = cmapl_add_counter(object2, &counter_def, &counter6);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ADD COUNTER FAILED\n");
        }

	cmapl_counter_handle_t* obj2_cnts[2] = {counter5, counter6};

	// write XML header
	error = cmapl_finalize_header(file);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "FINALIZE HEADER FAILED\n");
        }

	// write some records
	cmapl_record_header_data_t record_header;
        char* binary_record = cmapl_create_test_record_buffer(obj2_cnts, 2, CMAPL_CONFIG_PERIOD, (20*60*60*1000), 2, 0, &record_header);
        error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD2 FAILED = %d\n", error);
        }
        CMAPL_FREE(binary_record);

        binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 30*1000, 20*60*60*1000, 2, 0, &record_header);
        error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD1 FAILED\n");
        }
        CMAPL_FREE(binary_record);

	binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 30*1000, 20*60*60*1000 + 60*1000, 2, 0, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD3 FAILED\n");
        }
	CMAPL_FREE(binary_record);
	binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 30*1000, 20*60*60*1000 + 60*1000, 2, 1, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD4 FAILED\n");
        }
	CMAPL_FREE(binary_record);

	binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 60*1000, 20*60*60*1000 + 60*1000, 2, 0, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD5 FAILED\n");
        }
	CMAPL_FREE(binary_record);
	binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 60*1000, 20*60*60*1000 + 60*1000, 2, 1, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD6 FAILED\n");
        }
	CMAPL_FREE(binary_record);

#ifdef _CMAPL_FIJI_
	// test write failing fiji
	error = cmapl_enable_fiji(CMAPL_WRITE_FAILED);
        binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 30*1000, 20*60*60*1000 + 90*1000, 2, 1, &record_header);
        error = cmapl_write_record(file, &record_header, binary_record);
        CMAPL_FREE(binary_record);
	CMAPL_ASSERT(_CMAPL_FIJI_ && error == CMAPL_WRITE_FAILED);
        error = cmapl_disable_fiji(CMAPL_WRITE_FAILED);
#endif /* _CMAPL_FIJI_ */

	binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 30*1000, 20*60*60*1000 + 2*60*1000, 2, 0, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD7 FAILED\n");
        }
        CMAPL_FREE(binary_record);
        binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 30*1000, 20*60*60*1000 + 2*60*1000, 2, 1, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD8 FAILED\n");
        }
        CMAPL_FREE(binary_record);

        binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 60*1000, 20*60*60*1000 + 2*60*1000, 2, 0, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD9 FAILED\n");
        }
        CMAPL_FREE(binary_record);
        binary_record = cmapl_create_test_record_buffer(obj1_cnts, 5, 60*1000, 20*60*60*1000 + 2*60*1000, 2, 1, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD10 FAILED\n");
        }
        CMAPL_FREE(binary_record);

	binary_record = cmapl_create_test_record_buffer(obj2_cnts, 2, 10*1000, 20*60*60*1000 + 2*60*60*1000, 2, 0, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD11 FAILED\n");
        }
	CMAPL_FREE(binary_record);
	binary_record = cmapl_create_test_record_buffer(obj2_cnts, 2, 10*1000, 20*60*60*1000 + 2*60*60*1000, 2, 1, &record_header);
	error = cmapl_write_record(file, &record_header, binary_record);
        if (error != CMAPL_SUCCESS){
                fprintf(stderr, "WRITE RECORD12 FAILED\n");
        }
	CMAPL_FREE(binary_record);

	// close data file (indexes created)
	error = cmapl_close_data_file(file);
	if (error == CMAPL_SUCCESS){
                printf("CMAPL WRITE TEST PASSED!\n");
        } else {
                fprintf(stderr, "CMAPL WRITE TEST FAILED WITH ERROR: %d!\n", error);
        }

	// pass the file handle to the parameter
        *data_file = file;

	return error;
}

static cmapl_error_t
cmapl_test_read_side(cmapl_file_handle_t * file)
{
	cmapl_error_t error = CMAPL_SUCCESS;

        cmapl_file_handle_t * file2;
	const char* meta_path;
	const char* data_path;

#ifdef _CMAPL_FIJI_
	// test read failing fiji
	error = cmapl_enable_fiji(CMAPL_READ_FAILED);
	error = cmapl_open_data_file("000001_000060_72000000_7200000.ccma", "rb", &file2, &meta_path, &data_path);
	CMAPL_ASSERT(_CMAPL_FIJI_ && error == CMAPL_READ_FAILED);
	error = cmapl_disable_fiji(CMAPL_READ_FAILED);

	// test seek failing fiji
	error = cmapl_enable_fiji(CMAPL_SEEK_FAILED);
        error = cmapl_open_data_file("000001_000060_72000000_7200000.ccma", "rb", &file2, &meta_path, &data_path);
	CMAPL_ASSERT(_CMAPL_FIJI_ && error == CMAPL_SEEK_FAILED);
        error = cmapl_disable_fiji(CMAPL_SEEK_FAILED);
#endif /* _CMAPL_FIJI_ */

        error = cmapl_open_data_file("000001_000060_72000000_7200000.ccma", "rb", &file2, &meta_path, &data_path);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ERROR: TEST OPEN FILE ERROR: %d!\n", error);;
        }

	// TODO: VALIDATE XML HEADER

	// TEST: VALIDATE OBJECT INDEX
	uint32_t index;
	CMAPL_ASSERT(file->i6_0.header->i6_0.num_objects_sampled == file2->i6_0.header->i6_0.num_objects_sampled);
	for (index = 0; index < file->i6_0.header->i6_0.num_objects_sampled; index++){
                uint32_t period_index;
                for (period_index = 0; period_index < CMAPL_MAX_NUM_PERIODS; period_index++){
                        CMAPL_ASSERT(file->i6_0.header->i6_0.object_index[(index*CMAPL_MAX_NUM_PERIODS)+period_index] ==
                                        file2->i6_0.header->i6_0.object_index[(index*CMAPL_MAX_NUM_PERIODS)+period_index]);
                }
	}
	cmapl_object_handle_t* current_object = file->i6_0.header->i6_0.object_list;
	cmapl_object_handle_t* current_object2 = file2->i6_0.header->i6_0.object_list;
	while (current_object != NULL && current_object2 != NULL){
		cmapl_period_t* period = current_object->i6_0.period_list;
		cmapl_period_t* period2 = current_object2->i6_0.period_list;
		while (period != NULL && period2 != NULL){
			CMAPL_ASSERT(period->period == period2->period);
			CMAPL_ASSERT(period->num_offsets == period2->num_offsets);
			uint32_t i;
			for (i = 0; i < period2->num_offsets; i++){
				CMAPL_ASSERT(period->record_offsets[i] == period2->record_offsets[i]);
				if (CMAPL_IS_ITERATIVE(period->record_offsets[i])){
					cmapl_iterative_record_offsets_t iter_rec_offsets = period->iterative_record_offsets[i];
					cmapl_iterative_record_offsets_t iter_rec_offsets2 = period2->iterative_record_offsets[i];
					CMAPL_ASSERT(iter_rec_offsets.num_offsets == iter_rec_offsets2.num_offsets);
					uint32_t k;
					for (k = 0; k < iter_rec_offsets.num_offsets; k++){
						CMAPL_ASSERT(iter_rec_offsets.offsets[k] == iter_rec_offsets2.offsets[k]);
					}
				}
			}
			period = period->next_period;
			period2 = period2->next_period;
		}
		CMAPL_ASSERT(period == NULL && period2 == NULL);
		current_object = current_object->i6_0.next_object;
		current_object2 = current_object2->i6_0.next_object;
	}
	CMAPL_ASSERT(current_object == NULL && current_object2 == NULL);

	// close read data file
        error = cmapl_close_data_file(file2);
	file2->i6_0.metadata_file = NULL;
        file2->i6_0.data_file = NULL;

        error = cmapl_free_file(file2);
	if (error != CMAPL_SUCCESS){
		fprintf(stderr, "ERROR: TEST FAILED FREEING FILE: %d!\n", error);
	}
	file2 = NULL;

	// TEST: DIRECTORY
	cmapl_directory_def_t dir_def;
	cmapl_set_latest_version(&dir_def);
	dir_def.directory_path = ".";
	dir_def.start_time = 72000000;
	dir_def.end_time = 72001000;
	dir_def.period = 1000;

	cmapl_directory_handle_t* dir;
	error = cmapl_open_directory(&dir_def, &dir);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ERROR: TEST FAILED OPENING DIRECTORY: %d!\n", error);
        }

	// find all the files
	uint32_t num_files;
	cmapl_file_def_t_v6_0* file_infos;
	error = cmapl_get_file_infos(dir, &num_files, (void**)&file_infos);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ERROR: TEST FAILED GETTING FILES FROM DIRECTORY: %d!\n", error);
        }
	uint32_t file_index;
	for (file_index = 0; file_index < num_files; file_index++){
                printf("meta file path = %s\n", file_infos[file_index].meta_path);
		printf("data file path = %s\n", file_infos[file_index].data_path);
		printf("start time = %"PRIu64"; end time = %"PRIu64"\n", file_infos[file_index].start_time,
				file_infos[file_index].end_time);
        }

	// read file from directory
	error = cmapl_find_and_open_oldest_data_file(dir, &file2);
	if (error != CMAPL_SUCCESS){
		fprintf(stderr, "ERROR: TEST FAILED SCANNING DIRECTORY: %d!\n", error);
	}

	// TEST: QUERY
	cmapl_query_def_t query_def;
	cmapl_set_latest_version(&query_def);
	query_def.object_name = "aggregate";
	query_def.num_cnts = 3;
	const char* cnt_names[] = {"test_ops", "total_transfers", "num_transfers"};
	query_def.cnt_names = cnt_names;
	query_def.start_time = 72000000;
	query_def.end_time = 72000000 + 3*60*1000;
	query_def.period = 60000;
	query_def.num_key_cnts = 0;

	cmapl_query_handle_t* query;
	error = cmapl_query(file2, &query_def, &query);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ERROR: TEST FAILED MAKING QUERY: %d!\n", error);
        }

	cmapl_timestamp_handle_t* timestamp;
	error = cmapl_first_timestamp(query, &timestamp);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "ERROR: TEST FAILED GETTING FIRST TIMESTAMP: %d!\n", error);
        }

	while (timestamp != NULL){
		cmapl_instance_handle_t* instance;;
	        error = cmapl_first_instance(timestamp, &instance);
	        if (error != CMAPL_SUCCESS){
	                fprintf(stderr, "ERROR: TEST FAILED GETTING FIRST INSTANCE: %d!\n", error);
	        }

		while (instance != NULL){
			uint32_t num_key_cnts = instance->e6_0.num_key_cnts;
			uint32_t i;
			for (i = 0; i < num_key_cnts; i++){
				printf("key cnt name = %s\t", instance->e6_0.key_cnt_names[i]);
				if (instance->e6_0.key_cnt_values[i].value_type == CMAPL_CNT_4BYTES){
					printf("key cnt value = %d\n", instance->e6_0.key_cnt_values[i].values[0].val_4B);
				} else if (instance->e6_0.key_cnt_values[i].value_type == CMAPL_CNT_8BYTES){
					printf("key cnt value = %"PRIu64"\n", instance->e6_0.key_cnt_values[i].values[0].val_8B);
				} else if (instance->e6_0.key_cnt_values[i].value_type == CMAPL_CNT_STRING){
					printf("key cnt value = %s\n", instance->e6_0.key_cnt_values[i].values[0].val_str);
				}
			}
			cmapl_counter_handle_t* ret_counter;
			cmapl_counter_data_handle_t* counter_data;
			error = cmapl_first_counter_data(instance, &ret_counter, &counter_data);
			if (error != CMAPL_SUCCESS){
				fprintf(stderr, "ERROR: TEST FAILED GETTING FIRST COUNTER DATA: %d!\n", error);
			}

			while (counter_data != NULL){
				printf("counter name = %s\t", ret_counter->e6_0.name);
				if (counter_data->e6_0.value_type == CMAPL_CNT_4BYTES){
					printf("counter value = %d\n", counter_data->e6_0.values[0].val_4B);
				} else if (counter_data->e6_0.value_type == CMAPL_CNT_8BYTES){
					if (CMAPL_STRCMP(ret_counter->e6_0.name, "test_ops") == 0){
						printf("counter value = %"PRIu64" ", counter_data->e6_0.values[0].val_8B);
						printf("counter value = %"PRIu64" ", counter_data->e6_0.values[1].val_8B);
						printf("counter value = %"PRIu64" ", counter_data->e6_0.values[2].val_8B);
						printf("counter value = %"PRIu64"\n", counter_data->e6_0.values[3].val_8B);
					}else{
						printf("counter value = %"PRIu64"\n", counter_data->e6_0.values[0].val_8B);
					}
				} else if (counter_data->e6_0.value_type == CMAPL_CNT_STRING){
					printf("counter value = %s\n", counter_data->e6_0.values[0].val_str);
				}
				error = cmapl_next_counter_data(instance, &ret_counter, &counter_data);
				if (error != CMAPL_SUCCESS){
					fprintf(stderr, "ERROR: TEST FAILED GETTING NEXT COUNTER DATA: %d!\n", error);
				}
			}

			error = cmapl_next_instance(timestamp, &instance);
			if (error != CMAPL_SUCCESS){
				fprintf(stderr, "ERROR: TEST FAILED GETTING NEXT INSTANCE: %d!\n", error);
			}
		}
		error = cmapl_next_timestamp(query, &timestamp);
		if (error != CMAPL_SUCCESS){
			fprintf(stderr, "ERROR: TEST FAILED GETTING NEXT TIMESTAMP: %d!\n", error);
		}
	}

	// close read data file
        error = cmapl_close_data_file(file2);
	if (error != CMAPL_SUCCESS){
                fprintf(stderr, "CLOSE FILE2 FAILED");
        }

	// close directory handle
        error = cmapl_close_directory(dir);

	if (error == CMAPL_SUCCESS){
		printf("CMAPL READ TEST PASSED!\n");
	} else {
		fprintf(stderr, "CMAPL READ TEST FAILED WITH ERROR: %d!\n", error);
	}

	// file handle clean up
	if (file2) {
		file2->i6_0.metadata_file = NULL;
		file2->i6_0.data_file = NULL;
		error = cmapl_free_file(file2);
		if (error != CMAPL_SUCCESS){
			fprintf(stderr, "ERROR: TEST FAILED FREEING FILE: %d!\n", error);
		}
	}

	return error;
}

static cmapl_error_t
cmapl_test_linear_read(cmapl_file_handle_t * data_file)
{
        // validate if records are missing in the object index that are in the linear read
        cmapl_record_handle_t* record = NULL;
        cmapl_error_t err = cmapl_first_record_v6_0(data_file, &record);
        if (err == CMAPL_INVALID_RECORD) {
                fprintf(stderr, "FIRST RECORD IS CORRUPTED\n");
        } else if (err != CMAPL_SUCCESS) {
                return err;
        }

        bool is_valid = true;
        while (record != NULL) {
                // check if this record is in the object index
                uint64_t r_timestamp = record->e6_0.msecs_since_base;
                uint64_t r_period = record->e6_0.period;
                uint32_t index = (uint32_t)(CMAPL_GET_BUCKET(r_timestamp, data_file->i6_0.start_time, r_period));

                cmapl_period_t* period = record->i6_0.period;

                if (period->record_offsets[index] == CMAPL_NO_RECORD) {
                        cmapl_object_handle_t* object = period->object;
                        fprintf(stderr, "RECORD FOR OBJECT %s AND PERIOD %"PRIu64" AT TIMESTAMP %"PRIu64" NOT FOUND IN OBJECT INDEX.\n",
                                    object->e6_0.name, r_period, r_timestamp);
                        is_valid = false;
                }

                // check if the offsets match
                if (period->record_offsets[index] != record->i6_0.offset) {
                        fprintf(stderr, "The object index is pointing to the wrong record.\n");
                        is_valid = false;
                }

                err = cmapl_next_record_v6_0(data_file, &record);
                if (err == CMAPL_INVALID_RECORD) {
                        fprintf(stderr, "NEXT RECORD IS CORRUPTED\n");
                        is_valid = false;
                        goto test_end;
                } else if (err != CMAPL_SUCCESS) {
                        fprintf(stderr, "COULDN'T GET NEXT RECORD\n");
                        is_valid = false;
                        goto test_end;
                }
        }

test_end:
        if (is_valid) {
                printf("CMAPL LINEAR RECORD SEARCH VALIDATION TESTS PASSED!\n");
                return CMAPL_SUCCESS;
        } else {
                printf("CMAPL LINEAR RECORD SEARCH VALIDATION TESTS FAILED!\n");
                return CMAPL_RECORD_NOT_FOUND;
        }
}

static cmapl_error_t
cmapl_validate_record_counter_values(cmapl_record_handle_t* record)
{
        cmapl_error_t err = CMAPL_SUCCESS;

        cmapl_instance_handle_t* instance;
        err = cmapl_first_instance_from_record_v6_0(record, &instance);
        if (err != CMAPL_SUCCESS) {
                return err;
        }
        // TODO: only validating first instance of each record right now
        //while (instance != NULL) {
        if (instance != NULL) {
                cmapl_counter_handle_t* counter;
                cmapl_counter_data_handle_t* counter_data;
                err = cmapl_first_counter_data(instance, &counter, &counter_data);
                if (err != CMAPL_SUCCESS) {
                        return err;
                }

                // validate each counter value in the instance
                while (counter_data != NULL) {
                        // validate monotonically increasing counter values
                        if ((counter_data->e6_0.value_type == CMAPL_CNT_4BYTES ||
                                counter_data->e6_0.value_type == CMAPL_CNT_8BYTES) &&
                                counter_data->e6_0.num_values == 1)
                        {
                                uint64_t counter_value = 0;
                                if (counter_data->e6_0.value_type == CMAPL_CNT_4BYTES) {
                                        counter_value = counter_data->e6_0.values[0].val_4B;
                                } else {
                                        counter_value = counter_data->e6_0.values[0].val_8B;
                                }

                                uint32_t cnt_type = CM_CNT_TYPE_MASK & counter->e6_0.ctype;

                                // check if counter value increased from the last record of the same period
                                // skip raw counters since they are non-monotonically increasing
                                if (cnt_type != CM_CNT_RAW) {
                                        if (counter->i6_0.last_value > counter_value) {
                                                fprintf (stderr, "Value for counter %s of object %s is not monotonically increasing.\n",
                                                        counter->e6_0.name, counter->i6_0.object->e6_0.name);
                                                fprintf (stderr, "Previous value was %"PRIu64" but current value is %"PRIu64".\n",
                                                        counter->i6_0.last_value, counter_value);
                                                return CMAPL_INVALID_COUNTER_DATA;
                                        } else {
                                                counter->i6_0.last_value = counter_value;
                                        }
                                }
                        }

                        // validate instance_name counter does not have empty value
                        if (CMAPL_STRCMP("instance_name", counter->e6_0.name) == 0) {
                                if (counter_data->e6_0.values[0].val_str[0] == 0) {
                                        fprintf (stderr, "The instance_name counter value is empty for object %s\n",
                                                counter->i6_0.object->e6_0.name);
                                        return CMAPL_INVALID_COUNTER_DATA;
                                }
                        }

                        err = cmapl_next_counter_data(instance, &counter, &counter_data);
                        if (err != CMAPL_SUCCESS) {
                                return err;
                        }
                }

                // TODO: only validating first instance of each record right now
                //err = cmapl_next_instance_from_record_v6_0(record, &instance);
                //if (err != CMAPL_SUCCESS) {
                //        return err;
                //}
        }

        return err;
}

static void
cmapl_reset_counter_values(cmapl_period_t * period) {
        // reset previous counter values back to 0 for the next period's records validation
        cmapl_counter_handle_t* counter = period->cnt_list;
        while (counter != NULL) {
            counter->i6_0.last_value = 0;
            counter = counter->i6_0.next_cnt_by_period[period->period_index];
        }

        // reset previous key counter values back to 0 as well
        counter = period->object->i6_0.key_cnt_list;
        while (counter != NULL) {
                counter->i6_0.last_value = 0;
                counter = counter->i6_0.next_cnt_by_key;
        }
}

static cmapl_error_t
cmapl_validate_counter_values(cmapl_file_handle_t * data_file)
{
        cmapl_error_t err = CMAPL_SUCCESS;
        cmapl_header_handle_t* header = data_file->i6_0.header;

        cmapl_object_handle_t* object = header->i6_0.object_list;
        while(object != NULL) {
                cmapl_period_t* period = object->i6_0.period_list;
                while(period != NULL) {
                        // iterate through all records for this object and period in chronological order
                        uint32_t i;
                        for (i=0; i < period->num_offsets; i++) {
                                cmapl_offset_t offset = period->record_offsets[i];

                                // skip any missing records
                                if (offset == CMAPL_NO_RECORD) {
                                        continue;
                                }

                                // get record data
                                data_file->i6_0.current_offset = offset;
                                cmapl_record_handle_t* record = NULL;
                                err = cmapl_next_record_v6_0(data_file, &record);
                                if (err != CMAPL_SUCCESS) {
                                        return err;
                                }

                                // validate the counter values in the record
                                err = cmapl_validate_record_counter_values(record);
                                if (err != CMAPL_SUCCESS) {
                                        return err;
                                }
                        }
                        // reset previous counter values back to 0 for the next period's records validation
                        cmapl_reset_counter_values(period);

                        period = period->next_period;
                }

                object = object->i6_0.next_object;
        }

        printf("CMAPL COUNTER VALUE VALIDATION TESTS PASSED!\n");
        return err;
}

static void
cmapl_test_print_usage(void)
{
        printf("Usage:\n\n");
        printf("cmapl_test -d <directory_path> -p <preset> -v -r -w -l -c\n\n");
        printf("\t-d:\tPath of the directory to look for archive files. Requires -v option.\n");
	printf("\t-p:\tThe preset to validate an archive file with. Default value is \"default\". Requires -v and -d options.\n");
	printf("\t-v:\tFlag to enable preset validation test on an archive file. Preset specified by -p option. Requires -d option.\n");
        printf("\t-r:\tFlag to enable read side CMAPL tests. Requires -w option.\n");
        printf("\t-w:\tFlag to enable write side CMAPL tests. Requires -r option.\n");
        printf("\t-l:\tFlag to enable linear record search tests. Requires -d option.\n");
        printf("\t-c:\tFlag to enable counter value validation tests. Requires -d option.\n");
}

int main(int argc, char **argv){
	cmapl_error_t error = CMAPL_SUCCESS;

	int c;
        opterr = 0;
	char* d_value = NULL;
	char* p_value = NULL;
	bool v_value = false;
	bool r_value = false;
	bool w_value = false;
        bool l_value = false;
        bool c_value = false;
	int vlc_count = 0;	// ensure only one of v, l and c is specified.

        while ((c = getopt(argc, argv, "d:p:vrwlc")) != -1){
                switch (c){
                        case 'd':
                                d_value  = optarg;
                                break;
                        case 'p':
                                p_value  = optarg;
                                break;
                        case 'v':
                                v_value  = true;
				vlc_count++;
                                break;
                        case 'r':
                                r_value  = true;
                                break;
			case 'w':
                                w_value  = true;
                                break;
                        case 'l':
                                l_value = true;
				vlc_count++;
                                break;
                        case 'c':
                                c_value = true;
				vlc_count++;
                                break;
                        case '?':
                                fprintf (stderr, "Unrecognized option -%c\n", optopt);
                                cmapl_test_print_usage();
                                return 1;
                        default:
                                cmapl_test_print_usage();
                                return 1;
                }
        }

	// options error check
	if (r_value != w_value){
		fprintf (stderr, "Options -r and -w must be specified together.\n");
		cmapl_test_print_usage();
		return 1;
	} else if (l_value && d_value == NULL) {
                fprintf (stderr, "Option -l must be specifed with option -d.\n");
                cmapl_test_print_usage();
                return 1;
        } else if (c_value && d_value == NULL) {
                fprintf (stderr, "Option -c must be specifed with option -d.\n");
                cmapl_test_print_usage();
                return 1;
        } else if (v_value && d_value == NULL) {
		fprintf (stderr, "Option -v must be specified with option -d.\n");
                cmapl_test_print_usage();
                return 1;
	} else if (p_value != NULL && (!v_value || d_value == NULL)){
		fprintf (stderr, "Option -p must be specifed with options -v and -d.\n");
                cmapl_test_print_usage();
                return 1;
	} else if (vlc_count > 1) {
		fprintf (stderr, "Only one of -v, -l and -c can be specifed at a time.\n");
		cmapl_test_print_usage();
		return 1;
	} else if (argc == 1){
		fprintf (stderr, "Not enough arguments specified.\n");
		cmapl_test_print_usage();
		return 1;
	}

	printf("CMAPL TEST STARTED\n");

	// check if read and write tests should be run
	if (r_value && w_value){
                // get CMAPL version
                printf("CMAPL version: %d.%d\n", cmapl_get_major_version(), cmapl_get_minor_version());

		// Test write-side library
		cmapl_file_handle_t* file;
		error = cmapl_test_write_side(&file);
		if (error != CMAPL_SUCCESS){
			return error;
		}

		// Test read-side library
		error = cmapl_test_read_side(file);
		if (error != CMAPL_SUCCESS){
			return error;
		}

		// Testing file clean up
		file->i6_0.metadata_file = NULL;
		file->i6_0.data_file = NULL;
		error = cmapl_free_file(file);
		if (error != CMAPL_SUCCESS){
			fprintf(stderr, "ERROR: TEST FAILED FREEING FILE: %d!\n", error);
			return error;
		}
	}

	// check if directory should be opened for tests
	if (d_value != NULL){
		// open specified archive file directory and validate the oldest
		// file using a time range that all the files will fall in
		cmapl_directory_def_t dir_def;
		cmapl_set_latest_version(&dir_def);
		dir_def.directory_path = d_value;
		dir_def.start_time = 0;
		dir_def.end_time = 9999999999999ULL;
		dir_def.period = 1000;

		cmapl_directory_handle_t* dir;
		error = cmapl_open_directory(&dir_def, &dir);
		if (error != CMAPL_SUCCESS){
			fprintf(stderr, "CMAPL FAILED TO OPEN DIRECTORY: %d!\n", error);
			return 1;
		}

		// check if validation tests should be run
                if (vlc_count) {
			// get newest file
			cmapl_file_handle_t * data_file;
			error = cmapl_find_and_open_newest_data_file(dir, &data_file);
			if (error != CMAPL_SUCCESS){
				error = cmapl_close_directory(dir);
				if (error != CMAPL_SUCCESS){
					fprintf(stderr, "CMAPL FAILED TO CLOSE DIRECTORY!\n");
				}
				fprintf(stderr, "CMAPL FAILED TO FIND FILE IN DIRECTORY!\n");
				return error;
			}

			if (l_value) {
                                error = cmapl_test_linear_read(data_file);
                        } else if (c_value) {
                                error = cmapl_validate_counter_values(data_file);
			} else if (p_value != NULL){
				error = cmapl_preset_validation(data_file, p_value);
			} else { // validate the file with the default preset
				error = cmapl_preset_validation(data_file, "default");
			}

			if (error != CMAPL_SUCCESS){
				return error;
			}

			// close the file handle
			error = cmapl_close_data_file(data_file);
			if (error != CMAPL_SUCCESS){
				fprintf(stderr, "CMAPL FAILED TO CLOSE FILE!\n");
				return error;
			}

			// free file handle
			data_file->i6_0.metadata_file = NULL;
			data_file->i6_0.data_file = NULL;
			error = cmapl_free_file(data_file);
			if (error != CMAPL_SUCCESS){
				fprintf(stderr, "ERROR: TEST FAILED FREEING FILE: %d!\n", error);
				return error;
			}

			// close directory handle
			error = cmapl_close_directory(dir);
			if (error != CMAPL_SUCCESS){
				fprintf(stderr, "CMAPL FAILED TO CLOSE DIRECTORY!\n");
                                return error;
			}
		}
	}

	printf("CMAPL TEST FINISHED\n");

	return 0;
}
