/**
 * The CMAPL write-side function definitions. These functions provide all the
 * functionality to create, update, and modify archive files.
 *
 */

#include "cmapl_common_defs.h"
#include "cmapl_int_defs.h"
#include "cmapl_write.h"
#include "cmapl_write_int.h"

#ifdef _KERNEL
// some code taken from ///ontap/prod/common/net/ipsec/ike/logger.c
int
cmapl_vfprintf(
			CMAPL_FILE *fp,
			const char *format,
			va_list ap)
{
        int size = 2048;
        char* buffer = (char*)CMAPL_ZALLOC(size*sizeof(char));
	if (buffer == NULL){
		return -1;
	}
        size_t count;
	int res;

        count = vsnprintf(buffer, size, format, ap);

	// if the buffer is not big enough to fit the string
	// keep doubling the size of the buffer until it is big
	// enough
        while(count == size-1){
		CMAPL_FREE(buffer);
                size = size * 2;
                buffer = (char*)CMAPL_ZALLOC(size * sizeof(char));
		if (buffer == NULL){
			return -1;
		}
                count = vsnprintf(buffer, size, format, ap);
        }
        res = (int)fio_fwrite(buffer, sizeof(char), count, fp);
	CMAPL_FREE(buffer);
	return res;
}

int
cmapl_fprintf(
			CMAPL_FILE *fp,
			const char *format,
			...)
{
	va_list     ap;
	int         count;

	// this is for variable input parameters
	va_start(ap, format);
	count = cmapl_vfprintf(fp, format, ap);
	va_end(ap);
	return count;
}
#endif /* _KERNEL */

cmapl_error_t
cmapl_add_attr_def(
			cmapl_header_handle_t * header,
			void * attr_def)
{
	if (header == NULL || attr_def == NULL){
		return CMAPL_INVALID_ARGUMENTS;
	}

	// user specified version must match file version
	if (!cmapl_is_version(attr_def, header->major_v, header->minor_v)){
		return CMAPL_VERSION_MISMATCH;
	}

	if (cmapl_is_version(attr_def, 6, 0)){
                return cmapl_add_attr_def_v6_0(header, (cmapl_attribute_def_t_v6_0*)attr_def);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_add_attr_def_v6_0(
			cmapl_header_handle_t * header,
			cmapl_attribute_def_t_v6_0 * attr_def)
{
	// validate attribute definition struct
	cmapl_error_t error = cmapl_validate_attr_def_v6_0(attr_def);
	if (error != CMAPL_SUCCESS){
		return error;
	}

        // add the attribute def to the XML header at the next available spot
        uint32_t attr_index = header->i6_0.num_attr_defs;

	// double size of attribute def list in XML header if not enough space
	if (attr_index >= header->i6_0.attr_array_size){
		uint32_t size = header->i6_0.attr_array_size *2;
		cmapl_attribute_def_t_v6_0** attrs;
		attrs = (cmapl_attribute_def_t_v6_0**)CMAPL_ZALLOC(size * sizeof(cmapl_attribute_def_t_v6_0*));

		if (attrs == NULL){
			return CMAPL_NULL_POINTER;
		}

		CMAPL_MEMCPY(attrs, header->i6_0.attr_defs,
			header->i6_0.attr_array_size * sizeof(cmapl_attribute_def_t_v6_0*));
		CMAPL_FREE(header->i6_0.attr_defs);
		header->i6_0.attr_defs = attrs;
		header->i6_0.attr_array_size = size;
	}

	header->i6_0.attr_defs[attr_index] = (cmapl_attribute_def_t_v6_0*)CMAPL_ZALLOC(sizeof(cmapl_attribute_def_t_v6_0));
	if (header->i6_0.attr_defs[attr_index] == NULL){
		return CMAPL_NULL_POINTER;
	}

	// new attr defs will always be added at the next available index
	error = cmapl_copy_attr_def_v6_0(attr_def, header->i6_0.attr_defs[attr_index]);
	if (error != CMAPL_SUCCESS){
		return error;
	}

	header->i6_0.num_attr_defs++;

        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_add_counter(
                        cmapl_object_handle_t * object,
                        void * counter_def,
                        cmapl_counter_handle_t ** counter)
{
        if (object == NULL || counter_def == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        // user specified version must match file version
        if (!cmapl_is_version(counter_def, object->major_v, object->minor_v)){
                return CMAPL_VERSION_MISMATCH;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_add_counter_v6_0(
                                object,
                                (cmapl_counter_def_t_v6_0*)counter_def,
                                counter);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_add_counter_v6_0(
                        cmapl_object_handle_t * object,
                        cmapl_counter_def_t_v6_0 * counter_def,
                        cmapl_counter_handle_t ** counter)
{
	cmapl_error_t err;
	cmapl_period_t* period;
        cmapl_period_t* prev_period;
	cmapl_file_handle_t* data_file = object->i6_0.header->i6_0.data_file;

	// validate counter definition struct
        err = cmapl_validate_counter_def_v6_0(counter_def);
        if (err != CMAPL_SUCCESS){
                return err;
        }

	// check if counter exists by counter id
	bool counter_exists = false;
	cmapl_counter_handle_t* next_counter = object->i6_0.cnt_list;
	while(next_counter != NULL){
		if (next_counter->e6_0.cnt_id == counter_def->cnt_id){
			counter_exists = true;
			break;
		}
		next_counter = next_counter->i6_0.next_cnt_by_id;
	}
	if (next_counter != NULL){
		if (CMAPL_STRCMP(next_counter->e6_0.name, counter_def->name) == 0 &&
                        CMAPL_STRCMP(next_counter->e6_0.priv, counter_def->priv) == 0 &&
                        next_counter->e6_0.ctype == counter_def->ctype &&
                        next_counter->e6_0.size == counter_def->size &&
                        next_counter->e6_0.base_id == counter_def->base_id &&
                        CMAPL_STRCMP(next_counter->e6_0.desc, counter_def->desc) == 0)
                {
			if (counter_def->is_key == true && next_counter->e6_0.is_key == true){
				if (counter_def->period < next_counter->e6_0.period){
					next_counter->e6_0.period = counter_def->period;
				}
			} else if (counter_def->is_key != true && next_counter->e6_0.is_key != true){
				if (counter_def->period == next_counter->e6_0.period){
					// same identical counter was added with the same period
                                        // TODO: error?
                                } else {
					if (counter_def->period < next_counter->e6_0.period){
						next_counter->e6_0.period = counter_def->period;
					}
					(*counter) = next_counter;
					goto add_period;
				}
			} else{
				return CMAPL_COUNTER_MISMATCH;
			}
                }else{
			return CMAPL_COUNTER_MISMATCH;
		}
                (*counter) = next_counter;
                return CMAPL_SUCCESS;
	}

        // create counter handle and pass in values
        *counter = (cmapl_counter_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_counter_handle_t));
        if (*counter == NULL){
                return CMAPL_NULL_POINTER;
        }
        (*counter)->major_v = 6;
        (*counter)->minor_v = 0;

	// copy all external counter metadata
	err = cmapl_copy_counter_def_v6_0(counter_def, &((*counter)->e6_0));
	if (err != CMAPL_SUCCESS){
		cmapl_free_counter((*counter));
		return err;
	}

	// set parent object
	(*counter)->i6_0.object = object;

        // set linked list pointers to NULL
        (*counter)->i6_0.next_cnt_by_id = NULL;
	(*counter)->i6_0.next_cnt_by_key = NULL;
	uint32_t i;
	for (i = 0; i < CMAPL_MAX_NUM_PERIODS; i++){
		(*counter)->i6_0.next_cnt_by_period[i] = NULL;
		(*counter)->i6_0.periods[i] = NULL;
	}

	if (counter_def->is_key){
                // add counter to key counter list sorted by counter id
		err = cmapl_add_counter_to_list(&(object->i6_0.key_cnt_list), *counter, CMAPL_CNT_KEY, 0);
		if (err != CMAPL_SUCCESS){
			cmapl_free_counter((*counter));
			return err;
		}
		object->i6_0.num_key_cnts_sampled++;
        }

add_period:
        // find the correct object's period list
        period = object->i6_0.period_list;
        prev_period = NULL;
        while(period != NULL){
		// periods are sorted from fastest to slowest
                if (period->period >= counter_def->period){
			break;
                }
		prev_period = period;
                period = period->next_period;
	}

	// check if period was found
	// add counter to period counter list in order of counter id; don't add if counter is a key counter
	if (period != NULL && period->period == counter_def->period && counter_def->is_key == false){
		err = cmapl_add_counter_to_list(&(period->cnt_list), *counter, CMAPL_CNT_PERIOD, counter_def->period);
		if (err != CMAPL_SUCCESS){
			cmapl_free_counter((*counter));
			return err;
		}
                period->num_cnts++;
	}

	uint32_t period_index = 0;
        err = cmapl_get_period_index(counter_def->period, &period_index);
        if (err != CMAPL_SUCCESS){
                cmapl_free_counter((*counter));
		return err;
	}

	// check if no period list matched this counter's period
	// if key counter, create new period list but don't add key counter to period's counter list
        if (period == NULL || period->period > counter_def->period){
                period = (cmapl_period_t*)CMAPL_ZALLOC(sizeof(cmapl_period_t));
                if (period == NULL){
			// only free if this is a new counter; existing counters
			// will be freed when the file is closed
			if (!counter_exists){
				cmapl_free_counter((*counter));
			}
                        return CMAPL_NULL_POINTER;
                }

		period->period_index = period_index;
		period->object = object;
                period->period = counter_def->period;
                period->next_period = NULL;
                period->num_cnts = 0;
		if(counter_def->is_key == false){
			period->cnt_list = *counter;
                        period->num_cnts++;
		}

		// get the number of record offsets there should be
		uint32_t num_offsets = 0;
                if (data_file->i6_0.start_time == CMAPL_MAX_INT) {
                        err = cmapl_get_num_record_offsets_v6_0(0, data_file->i6_0.duration, period->period, &num_offsets);
                } else {
                        err = cmapl_get_num_record_offsets_v6_0(data_file->i6_0.start_time, data_file->i6_0.end_time,
                                                                                            period->period, &num_offsets);
                }
		if (err != CMAPL_SUCCESS){
			CMAPL_FREE(period);
			cmapl_free_counter((*counter));
			return err;
		}
		period->num_offsets = num_offsets;
		period->record_offsets =(cmapl_offset_t*)CMAPL_ZALLOC(num_offsets * sizeof(cmapl_offset_t));
		period->iterative_record_offsets =
			(cmapl_iterative_record_offsets_t*)CMAPL_ZALLOC(num_offsets * sizeof(cmapl_iterative_record_offsets_t));
		// set all time offsets to CMAPL_MAX_INT initially with the lead bit set to zero since 0 is legal
		for(i = 0; i < num_offsets; i++){
			period->record_offsets[i] = CMAPL_NO_RECORD;
		}

		// check if this is the first period in the object's period list
		// add the period into the list in sorted order
		if (object->i6_0.period_list == NULL){
			object->i6_0.period_list = period;
		}else{
			cmapl_period_t* temp_period;
			if (prev_period == NULL){
				temp_period = object->i6_0.period_list;
				object->i6_0.period_list = period;
			} else {
				temp_period = prev_period->next_period;
				prev_period->next_period = period;
			}
			period->next_period = temp_period;
		}
		object->i6_0.num_periods++;
	}
	(*counter)->i6_0.periods[period_index] = period;

        // add counter to object's counter linked list used for
        // easy iteration of counters by counter id; only add
	// if counter hasn't been added yet
	if (!counter_exists){
		err = cmapl_add_counter_to_list(&(object->i6_0.cnt_list), *counter, CMAPL_CNT_ID, 0);
		if (err != CMAPL_SUCCESS){
			cmapl_free_counter((*counter));
			return err;
		}
		object->i6_0.num_cnts_sampled++;
	}

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_add_object(
                        cmapl_header_handle_t * header,
                        void * object_def,
                        cmapl_object_handle_t ** object)
{
        if (header == NULL || object_def == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        // user specified version must match file version
        if (!cmapl_is_version(object_def, header->major_v, header->minor_v)){
                return CMAPL_VERSION_MISMATCH;
        }

        if (cmapl_is_version(object_def, 6, 0)){
                return cmapl_add_object_v6_0(
                                header,
                                (cmapl_object_def_t_v6_0*) object_def,
                                object);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_add_object_v6_0(
                        cmapl_header_handle_t * header,
                        cmapl_object_def_t_v6_0 * object_def,
                        cmapl_object_handle_t ** object)
{
	// validate object definition struct
        cmapl_error_t err = cmapl_validate_object_def_v6_0(object_def);
        if (err != CMAPL_SUCCESS){
                return err;
        }

        // check if object exists already
        if (header->i6_0.objects[object_def->obj_id] != NULL){
                *object = header->i6_0.objects[object_def->obj_id];
                return CMAPL_SUCCESS;
        }

        // create object handle and pass in object values
        *object = (cmapl_object_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_object_handle_t));

        if (*object == NULL){
                return CMAPL_NULL_POINTER;
        }

        (*object)->major_v = 6;
        (*object)->minor_v = 0;

	// copy all external object metadata
	err = cmapl_copy_object_def_v6_0(object_def, &((*object)->e6_0));
	if (err != CMAPL_SUCCESS){
		cmapl_free_object((*object));
		return err;
	}

        // zero out memory for pointers
        (*object)->i6_0.key_cnt_list = NULL;
        (*object)->i6_0.period_list = NULL;
        (*object)->i6_0.cnt_list = NULL;

        (*object)->i6_0.iterator_index = 0;
        (*object)->i6_0.num_cnts_sampled = 0;
        (*object)->i6_0.num_periods = 0;

	// set parent XML header
        (*object)->i6_0.header = header;

        // Add object to the XML header's object array.
        // Keep doubling the size of the XML header's object array
        // if it's not big enough
        while ((*object)->e6_0.obj_id >= header->i6_0.objects_array_size){
                uint32_t size = header->i6_0.objects_array_size * 2;
                cmapl_object_handle_t** obj_list;
                obj_list = (cmapl_object_handle_t**)CMAPL_ZALLOC(size * sizeof(cmapl_object_handle_t*));

                if (obj_list == NULL){
                        cmapl_free_object((*object));
                        return CMAPL_NULL_POINTER;
                }

                CMAPL_MEMCPY(obj_list, header->i6_0.objects,
                        header->i6_0.objects_array_size * sizeof(cmapl_object_handle_t*));
                CMAPL_FREE(header->i6_0.objects);
#ifdef _KERNEL
                RETURN_ADDRESS(header->i6_0.pc);
                header->i6_0.pid = sk_my_pid();
#endif /* _KERNEL */
                header->i6_0.objects = obj_list;
                header->i6_0.objects_array_size = size;
        }
        header->i6_0.num_objects_sampled++;
        header->i6_0.objects[(*object)->e6_0.obj_id] = *object;
	err = cmapl_add_object_to_list(&(header->i6_0.object_list), *object);
	if (err != CMAPL_SUCCESS){
		cmapl_free_object((*object));
		return err;
	}
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_close_data_file(
			cmapl_file_handle_t * data_file)
{
	if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	if (cmapl_is_version(data_file, 6, 0)){
		return cmapl_close_data_file_v6_0(data_file);
	}else {
		return CMAPL_INVALID_VERSION;
	}
}

cmapl_error_t
cmapl_close_data_file_v6_0(
                        cmapl_file_handle_t * data_file)
{
	// check if the file handle is valid
	if (data_file->i6_0.data_file == NULL){
		return CMAPL_FILE_NOT_FOUND;
	}

	cmapl_error_t err = CMAPL_SUCCESS;
        CMAPL_FILE* metadata_file = data_file->i6_0.metadata_file;
        cmapl_header_handle_t* header = data_file->i6_0.header;
	if (header == NULL){
		err = CMAPL_NULL_POINTER;
		goto clean_up;
	}
	int num = 0;

	if (data_file->i6_0.access_flag & CMAPL_READ_ONLY){
                goto clean_up;
        }

	// write the object index
	err = cmapl_write_object_index_v6_0(data_file);
	if (err != CMAPL_SUCCESS){
		// remove the offset to the object index in the XML header
		header->i6_0.section_defs[CMAPL_OBJECT_INDEX].header_offset = 0;
		goto clean_up;
	}

	// update each offset in the XML header that points to each section in the archive file
	uint32_t i;
	for(i = CMAPL_OBJECT_INDEX; i < TOTAL_NUM_SECTIONS; i++){
		cmapl_offset_t header_offset = header->i6_0.section_defs[i].header_offset;
		num = CMAPL_FSEEK(metadata_file, (long int)header_offset, SEEK_SET);
		if (num != 0){
			err = CMAPL_SEEK_FAILED;
			goto clean_up;
		}
		cmapl_offset_t section_offset = header->i6_0.section_defs[i].section_offset;
		if (section_offset != 0){
			num = CMAPL_FPRINTF(metadata_file, "%16"PRIX64, section_offset - header->i6_0.header_offset);
			if (num < 0){
				err = CMAPL_WRITE_FAILED;
				goto clean_up;
			}
		}
	}

	// update the actual duration, start time, and end time of the file
        cmapl_offset_t duration_header_offset = header->i6_0.duration_header_offset;
        num = CMAPL_FSEEK(metadata_file, (long int)duration_header_offset, SEEK_SET);
        if (num != 0){
                err = CMAPL_SEEK_FAILED;
                goto clean_up;
        }
        uint64_t duration = data_file->i6_0.end_time - data_file->i6_0.start_time;
        num = CMAPL_FPRINTF(metadata_file, "%16"PRIX64, duration);
	if (num < 0){
		err = CMAPL_WRITE_FAILED;
		goto clean_up;
	}
        data_file->i6_0.duration = duration;

	cmapl_offset_t start_time_header_offset = header->i6_0.start_time_header_offset;
        num = CMAPL_FSEEK(metadata_file, (long int)start_time_header_offset, SEEK_SET);
        if (num != 0){
                err = CMAPL_SEEK_FAILED;
                goto clean_up;
        }
	num = CMAPL_FPRINTF(metadata_file, "%16"PRIX64, data_file->i6_0.start_time);
	if (num < 0){
                err = CMAPL_WRITE_FAILED;
                goto clean_up;
        }

	cmapl_offset_t end_time_header_offset = header->i6_0.end_time_header_offset;
        num = CMAPL_FSEEK(metadata_file, (long int)end_time_header_offset, SEEK_SET);
        if (num != 0){
                err = CMAPL_SEEK_FAILED;
                goto clean_up;
        }
        num = CMAPL_FPRINTF(metadata_file, "%16"PRIX64, data_file->i6_0.end_time);
	if (num < 0){
                err = CMAPL_WRITE_FAILED;
                goto clean_up;
        }

clean_up:

#ifdef _CMAPL_TEST_
	CMAPL_FCLOSE(data_file->i6_0.metadata_file);
	num = CMAPL_GZCLOSE(data_file->i6_0.data_file);
	if (num != 0){
                return CMAPL_CLOSE_FILE_FAILED;
        }
	data_file->i6_0.metadata_file = NULL;
        data_file->i6_0.data_file = NULL;
#else
        err = cmapl_free_file(data_file);
#endif
        return err;
}

cmapl_error_t
cmapl_copy_counter(
			cmapl_object_handle_t * object,
			cmapl_counter_handle_t * counter)
{
	// Unimplemented
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_copy_header(
			cmapl_file_handle_t * data_file,
			cmapl_header_handle_t * header)
{
	// Unimplemented
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_copy_object(
			cmapl_header_handle_t * header,
			cmapl_object_handle_t * object)
{
	// Unimplemented
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_create_data_file(
			const char * file_path,
			uint32_t major_v,
			uint32_t minor_v,
			uint64_t duration,
			cmapl_file_handle_t ** data_file,
			const char ** meta_path,
			const char ** data_path)
{
	if (major_v == 6 && minor_v == 0){
		return cmapl_create_data_file_v6_0(
						file_path,
						major_v,
						minor_v,
						duration,
						data_file,
						meta_path,
						data_path);
	}else {
		return CMAPL_INVALID_VERSION;
	}
}

cmapl_error_t
cmapl_create_data_file_v6_0(
			const char * file_path,
			uint32_t major_v,
			uint32_t minor_v,
			uint64_t duration,
			cmapl_file_handle_t ** data_file,
			const char ** meta_path,
			const char ** data_path)
{
	// check if version numbers are valid
	if (!((major_v == 6 && minor_v == 0) || (major_v == 5 && minor_v == 0)))
	{
		return CMAPL_INVALID_ARGUMENTS;
	}

	// check if duration is valid
	if (duration == 0){
		return CMAPL_INVALID_ARGUMENTS;
	}

	// create and open metadata and data file
	char* metadata_file_path;
	CMAPL_ZALLOC_META_NAME(file_path, metadata_file_path);
	if (metadata_file_path == NULL){
		return CMAPL_NULL_POINTER;
	}

	char* data_file_path;
	CMAPL_ZALLOC_GZ_NAME(file_path, data_file_path);
	if (data_file_path == NULL){
		CMAPL_FREE(metadata_file_path);
                return CMAPL_NULL_POINTER;
        }

	// create data file handle
        *data_file = (cmapl_file_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_file_handle_t));
        if (*data_file == NULL){
                CMAPL_FREE(metadata_file_path);
                CMAPL_FREE(data_file_path);
                return CMAPL_NULL_POINTER;
        }

        (*data_file)->major_v = major_v;
        (*data_file)->minor_v = minor_v;
        (*data_file)->i6_0.metadata_file = NULL;
        (*data_file)->i6_0.data_file = NULL;
        (*data_file)->i6_0.metadata_file_path = metadata_file_path;
        (*data_file)->i6_0.data_file_path = data_file_path;
        (*data_file)->i6_0.access_flag = CMAPL_WRITE_ONLY;
        (*data_file)->i6_0.metadata_written_count = 0;
        (*data_file)->i6_0.data_written_count = 0;
        (*data_file)->i6_0.iterator_offset = 0;
        (*data_file)->i6_0.current_offset = 0;
        (*data_file)->i6_0.duration = duration;
        (*data_file)->i6_0.start_time = CMAPL_MAX_INT;
        (*data_file)->i6_0.end_time = 0;
	(*data_file)->i6_0.header = NULL;
	(*data_file)->i6_0.query = NULL;

	(*data_file)->i6_0.metadata_file = CMAPL_FOPEN(metadata_file_path, "w");
	if ((*data_file)->i6_0.metadata_file == NULL){
		(void)cmapl_free_file(*data_file);
		return CMAPL_NULL_POINTER;
	}

	(*data_file)->i6_0.data_file = CMAPL_GZOPEN(data_file_path, "wb");
	if ((*data_file)->i6_0.data_file == NULL){
		(void)cmapl_free_file(*data_file);
                return CMAPL_NULL_POINTER;
        }

	(*data_file)->e6_0.file_path = CMAPL_STRDUP(file_path);
	if ((*data_file)->e6_0.file_path == NULL){
		(void)cmapl_free_file(*data_file);
		return CMAPL_NULL_POINTER;
	}

	*meta_path = (const char*)metadata_file_path;
	*data_path = (const char*)data_file_path;

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_create_header(
			cmapl_file_handle_t * data_file,
                        void * header_def,
                        void * system_def,
                        cmapl_header_handle_t ** header)
{
	if (data_file == NULL ||
                header_def == NULL ||
		system_def == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	// user specified version must match file version
	if (!cmapl_is_version(header_def, data_file->major_v, data_file->minor_v)||
		!cmapl_is_version(system_def, data_file->major_v, data_file->minor_v))
	{
		return CMAPL_VERSION_MISMATCH;
        }

	if (cmapl_is_version(data_file, 6, 0)){
		return cmapl_create_header_v6_0(
	                        data_file,
	                        (cmapl_header_def_t_v6_0*)header_def,
	                        (cmapl_system_def_t_v6_0*)system_def,
				header);
	}else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_create_header_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_header_def_t_v6_0 * header_def,
                        cmapl_system_def_t_v6_0 * system_def,
                        cmapl_header_handle_t ** header)
{
	// check if the file handle is valid
        if (data_file->i6_0.data_file == NULL){
                return CMAPL_FILE_NOT_FOUND;
        }

	// validate header definition struct
        cmapl_error_t err = cmapl_validate_header_def_v6_0(header_def);
        if (err != CMAPL_SUCCESS){
                return err;
        }

	// validate system definition struct
        err = cmapl_validate_system_def_v6_0(system_def);
        if (err != CMAPL_SUCCESS){
                return err;
        }

	// create XML header handle and pass in header values
	*header = (cmapl_header_handle_t*)CMAPL_ZALLOC(sizeof(cmapl_header_handle_t));
	if (*header == NULL){
		return CMAPL_NULL_POINTER;
	}

	(*header)->major_v = 6;
	(*header)->minor_v = 0;

	// copy all external header metadata
	err = cmapl_copy_header_def_v6_0(header_def, &((*header)->e6_0));
	if (err != CMAPL_SUCCESS){
		goto header_failed;
	}

	(*header)->i6_0.system.major_v = 6;
	(*header)->i6_0.system.minor_v = 0;

	// copy all external system metadata
        err = cmapl_copy_system_def_v6_0(system_def, &((*header)->i6_0.system.e6_0));
        if (err != CMAPL_SUCCESS){
                goto header_failed;
        }

	// set parent data file
        (*header)->i6_0.data_file = data_file;

	// allocate memory for list of objects
	(*header)->i6_0.objects = (cmapl_object_handle_t**)CMAPL_ZALLOC(CMAPL_DEFAULT_NUM_OBJECTS * sizeof(cmapl_object_handle_t*));
	if ((*header)->i6_0.objects == NULL){
		goto header_failed;
        }
	(*header)->i6_0.objects_array_size = CMAPL_DEFAULT_NUM_OBJECTS;
	(*header)->i6_0.num_objects_sampled = 0;
	(*header)->i6_0.object_list = NULL;

	// allocate memory for attribute definitions
	(*header)->i6_0.attr_defs =
			(cmapl_attribute_def_t_v6_0**)CMAPL_ZALLOC(CMAPL_DEFAULT_NUM_ATTR_DEFS * sizeof(cmapl_attribute_def_t_v6_0*));
	if ((*header)->i6_0.attr_defs == NULL){
		goto header_failed;
        }
	(*header)->i6_0.attr_array_size = CMAPL_DEFAULT_NUM_ATTR_DEFS;
	(*header)->i6_0.num_attr_defs = 0;

	(*header)->i6_0.header_offset = 0;
	// initialize indexing definitions
	(*header)->i6_0.section_defs = (cmapl_section_def_t_v6_0*)CMAPL_ZALLOC(TOTAL_NUM_SECTIONS * sizeof(cmapl_section_def_t_v6_0));
	if ((*header)->i6_0.section_defs == NULL){
		goto header_failed;
        }
	uint32_t i;
	for (i = CMAPL_OBJECT_INDEX; i < TOTAL_NUM_SECTIONS; i++){
		(*header)->i6_0.section_defs[i].section_type = i;
		(*header)->i6_0.section_defs[i].section_version = CMAPL_CURRENT_SECTION_VERSION;
		(*header)->i6_0.section_defs[i].section_offset = 0;
		(*header)->i6_0.section_defs[i].header_offset = 0;
	}

	// set header to data file
	data_file->i6_0.header = *header;

	return CMAPL_SUCCESS;

header_failed:
	cmapl_free_header((*header));
	*header = NULL;
	return CMAPL_NULL_POINTER;
}

cmapl_error_t
cmapl_delete_counter(
			cmapl_object_handle_t * object,
			cmapl_counter_handle_t * counter)
{
	if (object == NULL || counter == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	// user specified version must match file version
	if (!cmapl_is_version(counter, object->major_v, object->minor_v)){
                return CMAPL_VERSION_MISMATCH;
        }

        if (cmapl_is_version(counter, 6, 0)){
		return cmapl_delete_counter_v6_0(object, counter);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_delete_counter_v6_0(
                        cmapl_object_handle_t * object,
                        cmapl_counter_handle_t * counter)
{
	// Unimplemented
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_delete_object(
			cmapl_header_handle_t * header,
			cmapl_object_handle_t * object)
{
	if (header == NULL || object == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	// user specified version must match file version
	if (!cmapl_is_version(object, header->major_v, header->minor_v)){
		return CMAPL_VERSION_MISMATCH;
        }

        if (cmapl_is_version(object, 6, 0)){
                return cmapl_delete_object_v6_0(header, object);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_delete_object_v6_0(
                        cmapl_header_handle_t * header,
                        cmapl_object_handle_t * object)
{
	// Unimplemented
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_finalize_header(
			cmapl_file_handle_t * data_file)
{
	if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_finalize_header_v6_0(data_file);
        }else {
		return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_finalize_header_v6_0(
                        cmapl_file_handle_t * data_file)
{
	// check if the file handle is valid
        if (data_file->i6_0.data_file == NULL){
                return CMAPL_FILE_NOT_FOUND;
        }

	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_header_handle_t * header = data_file->i6_0.header;
	CMAPL_FILE * metadata_file = data_file->i6_0.metadata_file;
	int num = 0;

	uint64_t count = data_file->i6_0.metadata_written_count;

	// write the XML DTD
	num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_XML_STR_V6_0);
	if (num < 0){
		return CMAPL_WRITE_FAILED;
	}
	count += num;
	num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_DTD_STR_V6_0);
	if (num < 0){
                return CMAPL_WRITE_FAILED;
        }
        count += num;

	// begin writing the XML header
	header->i6_0.header_offset = count;

	uint32_t num_strs = 1;
	const char* archive_strs[num_strs];
	archive_strs[0] = header->e6_0.preset_version;
	char* archive_encoded_strs[num_strs];
	bool archive_encoded_flags[num_strs];
	err = cmapl_xml_encode(num_strs, archive_strs, archive_encoded_strs, archive_encoded_flags);
	if (err != CMAPL_SUCCESS){
		cmapl_free_xml_encoded_strs(num_strs, archive_encoded_strs, archive_encoded_flags);
		return err;
	}
	num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_ARCHIVE_BEGIN_STR_V6_0,
				data_file->major_v,
				data_file->minor_v,
				header->e6_0.generation_id,
				archive_encoded_strs[0],
				header->i6_0.header_offset);
	if (num < 0){
                return CMAPL_WRITE_FAILED;
        }
        count += num;
	cmapl_free_xml_encoded_strs(num_strs, archive_encoded_strs, archive_encoded_flags);

	header->i6_0.duration_header_offset = count;
	num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_ARCHIVE_BEGIN_DURATION_STR_V6_0,
				data_file->i6_0.duration);
	if (num < 0){
                return CMAPL_WRITE_FAILED;
        }
        count += num;
	header->i6_0.start_time_header_offset = count;
        num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_ARCHIVE_BEGIN_START_TIME_STR_V6_0,
                                data_file->i6_0.start_time);
	if (num < 0){
                return CMAPL_WRITE_FAILED;
        }
        count += num;
	header->i6_0.end_time_header_offset = count;
        num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_ARCHIVE_BEGIN_END_TIME_STR_V6_0,
                                data_file->i6_0.end_time);
	if (num < 0){
                return CMAPL_WRITE_FAILED;
        }
        count += num;

	num_strs = 6;
        const char* system_strs[num_strs];
        system_strs[0] = header->i6_0.system.e6_0.ontap_version;
	system_strs[1] = header->i6_0.system.e6_0.node_uuid;
	system_strs[2] = header->i6_0.system.e6_0.hostname;
	system_strs[3] = header->i6_0.system.e6_0.model;
	system_strs[4] = header->i6_0.system.e6_0.serial_number;
	system_strs[5] = header->i6_0.system.e6_0.system_id;
        char* system_encoded_strs[num_strs];
        bool system_encoded_flags[num_strs];
        err = cmapl_xml_encode(num_strs, system_strs, system_encoded_strs, system_encoded_flags);
	if (err != CMAPL_SUCCESS){
                cmapl_free_xml_encoded_strs(num_strs, system_encoded_strs, system_encoded_flags);
                return err;
        }
	num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_SYSTEM_STR_V6_0,
				system_encoded_strs[0],
				header->i6_0.system.e6_0.base_time,
				system_encoded_strs[1],
				system_encoded_strs[2],
				system_encoded_strs[3],
				system_encoded_strs[4],
				system_encoded_strs[5]);
	if (num < 0){
                return CMAPL_WRITE_FAILED;
        }
        count += num;
	cmapl_free_xml_encoded_strs(num_strs, system_encoded_strs, system_encoded_flags);

	// write each attribute def of the XML header
	uint32_t attr_index;
	for(attr_index = 0; attr_index < header->i6_0.num_attr_defs; attr_index++){
		cmapl_attribute_def_t_v6_0 * attr_def = header->i6_0.attr_defs[attr_index];

		num_strs = 2;
		const char* attr_strs[num_strs];

		if (NULL == attr_def) {
			break;
		}
		attr_strs[0] = attr_def->name;
		attr_strs[1] = attr_def->desc;
		char* attr_encoded_strs[num_strs];
		bool attr_encoded_flags[num_strs];
		err = cmapl_xml_encode(num_strs, attr_strs, attr_encoded_strs, attr_encoded_flags);
		if (err != CMAPL_SUCCESS){
			cmapl_free_xml_encoded_strs(num_strs, attr_encoded_strs, attr_encoded_flags);
			return err;
		}
		num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_ATTR_DEF_STR_V6_0,
					attr_encoded_strs[0],
					attr_def->value,
					attr_encoded_strs[1]);
		if (num < 0){
			return CMAPL_WRITE_FAILED;
		}
		count += num;
		cmapl_free_xml_encoded_strs(num_strs, attr_encoded_strs, attr_encoded_flags);
	}

	// write each index definition of the XML header
	uint32_t section_id;
	for(section_id = CMAPL_OBJECT_INDEX; section_id < TOTAL_NUM_SECTIONS; section_id++){
		cmapl_section_def_t_v6_0 section_def = header->i6_0.section_defs[section_id];
		num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_SECTION_DEF_BEGIN_STR_V6_0,
					section_def.section_type,
					section_def.section_version);
		if (num < 0){
                        return CMAPL_WRITE_FAILED;
                }
                count += num;
		header->i6_0.section_defs[section_id].header_offset = count;
		num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_SECTION_DEF_END_STR_V6_0,
					section_def.section_offset);
		if (num < 0){
                        return CMAPL_WRITE_FAILED;
                }
                count += num;
        }

	// begin writing the preset of the XML header
	num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_PRESET_BEGIN_STR_V6_0);
	if (num < 0){
                return CMAPL_WRITE_FAILED;
        }
        count += num;

	// write each object in the preset
	uint32_t obj_index;
	for(obj_index = 0; obj_index < header->i6_0.objects_array_size; obj_index++){
		cmapl_object_handle_t * object =  header->i6_0.objects[obj_index];
		if (object != NULL){
			num_strs = 1;
			const char* obj_strs[num_strs];
			obj_strs[0] = object->e6_0.name;
			char* obj_encoded_strs[num_strs];
			bool obj_encoded_flags[num_strs];
			err = cmapl_xml_encode(num_strs, obj_strs, obj_encoded_strs, obj_encoded_flags);
			if (err != CMAPL_SUCCESS){
				cmapl_free_xml_encoded_strs(num_strs, obj_encoded_strs, obj_encoded_flags);
				return err;
			}
			num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_OBJECT_BEGIN_STR_V6_0,
						obj_encoded_strs[0],
						object->e6_0.obj_id);
			if (num < 0){
				return CMAPL_WRITE_FAILED;
			}
			count += num;
			cmapl_free_xml_encoded_strs(num_strs, obj_encoded_strs, obj_encoded_flags);

			// write each counter in the object
			cmapl_counter_handle_t* counter = object->i6_0.cnt_list;
			while (counter != NULL){
				num_strs = 3;
				const char* cnt_strs[num_strs];
				cnt_strs[0] = counter->e6_0.name;
				cnt_strs[1] = counter->e6_0.desc;
				cnt_strs[2] = counter->e6_0.priv;
				char* cnt_encoded_strs[num_strs];
				bool cnt_encoded_flags[num_strs];
				err = cmapl_xml_encode(num_strs, cnt_strs, cnt_encoded_strs, cnt_encoded_flags);
				if (err != CMAPL_SUCCESS){
					cmapl_free_xml_encoded_strs(num_strs, cnt_encoded_strs, cnt_encoded_flags);
					return err;
				}
				num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_COUNTER_BEGIN_STR_V6_0,
							counter->e6_0.is_key,
							counter->e6_0.ctype,
							cnt_encoded_strs[0],
							cnt_encoded_strs[1],
							counter->e6_0.cnt_id,
							counter->e6_0.base_id,
							counter->e6_0.size,
							cnt_encoded_strs[2]);
				if (num < 0){
					return CMAPL_WRITE_FAILED;
				}
				count += num;
				cmapl_free_xml_encoded_strs(num_strs, cnt_encoded_strs, cnt_encoded_flags);

				uint32_t p;
				for (p = 0; p < CMAPL_MAX_NUM_PERIODS; p++){
					if (counter->i6_0.periods[p] != NULL){
						num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_COUNTER_PERIOD_STR_V6_0,
								counter->i6_0.periods[p]->period);
						if (num < 0){
							return CMAPL_WRITE_FAILED;
						}
						count += num;
					}
				}

				// check for array counters
				if (counter->e6_0.x_names != NULL){
					uint32_t x_index;
					for(x_index = 0; x_index < counter->e6_0.num_x_items; x_index++){
						num_strs = 1;
						const char* x_strs[num_strs];
						x_strs[0] = counter->e6_0.x_names[x_index];
						char* x_encoded_strs[num_strs];
						bool x_encoded_flags[num_strs];
						err = cmapl_xml_encode(num_strs, x_strs, x_encoded_strs, x_encoded_flags);
						if (err != CMAPL_SUCCESS){
							cmapl_free_xml_encoded_strs(num_strs, x_encoded_strs, x_encoded_flags);
							return err;
						}
						num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_COUNTER_X_ITEM_STR_V6_0,
								x_encoded_strs[0]);
						if (num < 0){
                                                        return CMAPL_WRITE_FAILED;
                                                }
                                                count += num;
						cmapl_free_xml_encoded_strs(num_strs, x_encoded_strs, x_encoded_flags);
					}
				}
				if (counter->e6_0.y_names != NULL){
					uint32_t y_index;
					for(y_index = 0; y_index < counter->e6_0.num_y_items; y_index++){
						num_strs = 1;
                                                const char* y_strs[num_strs];
                                                y_strs[0] = counter->e6_0.y_names[y_index];
                                                char* y_encoded_strs[num_strs];
                                                bool y_encoded_flags[num_strs];
                                                err = cmapl_xml_encode(num_strs, y_strs, y_encoded_strs, y_encoded_flags);
						if (err != CMAPL_SUCCESS){
                                                        cmapl_free_xml_encoded_strs(num_strs, y_encoded_strs, y_encoded_flags);
                                                        return err;
                                                }
						num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_COUNTER_Y_ITEM_STR_V6_0,
                                                                y_encoded_strs[0]);
						if (num < 0){
                                                        return CMAPL_WRITE_FAILED;
                                                }
                                                count += num;
						cmapl_free_xml_encoded_strs(num_strs, y_encoded_strs, y_encoded_flags);
					}
				}

				num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_COUNTER_END_STR_V6_0);
				if (num < 0){
					return CMAPL_WRITE_FAILED;
                                }
                                count += num;
				counter = counter->i6_0.next_cnt_by_id;
			}
			num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_OBJECT_END_STR_V6_0);
			if (num < 0){
                                return CMAPL_WRITE_FAILED;
                        }
                        count += num;
		}
	}
	num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_PRESET_END_STR_V6_0);
	if (num < 0){
		return CMAPL_WRITE_FAILED;
	}
	count += num;
	num = CMAPL_FPRINTF(metadata_file, CMAPL_XML_HEADER_ARCHIVE_END_STR_V6_0);
	if (num < 0){
                return CMAPL_WRITE_FAILED;
        }
        count += num;

	// write null character to make libxml2 to think file has ended
	char eof_c[1];
	eof_c[0] = 0;
	num = (int)CMAPL_FWRITE(eof_c, 1, metadata_file);
	if (num != 1){
		return CMAPL_WRITE_FAILED;
	}
	count += num;

	data_file->i6_0.metadata_written_count = count;

        // flush all written data to metafile first before writting the data file
        num = CMAPL_FFLUSH(metadata_file);
        if (num != 0) {
                return CMAPL_WRITE_FAILED;
        }

        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_ftell(
                        cmapl_file_handle_t * data_file,
			uint64_t * count)
{
	if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	if (cmapl_is_version(data_file, 6, 0)){
		// check if the file handle is valid
		if (data_file->i6_0.data_file == NULL){
			return CMAPL_FILE_NOT_FOUND;
		}
                *count = data_file->i6_0.metadata_written_count + data_file->i6_0.data_written_count;
		return CMAPL_SUCCESS;
        }else {
		return CMAPL_INVALID_VERSION;
	}
}

cmapl_error_t
cmapl_get_newest_record_time(
                        cmapl_file_handle_t * data_file,
                        uint64_t * newest_time)
{
        if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_get_newest_record_time_v6_0(
                                        data_file,
                                        newest_time);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_newest_record_time_v6_0(
                        cmapl_file_handle_t * data_file,
                        uint64_t * newest_time)
{
	// check if the file handle is valid
        if (data_file->i6_0.data_file == NULL){
                return CMAPL_FILE_NOT_FOUND;
        }
	uint64_t base_time = (data_file->i6_0.header) ? data_file->i6_0.header->i6_0.system.e6_0.base_time : 0;

        *newest_time = data_file->i6_0.end_time + base_time;
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_oldest_record_time(
                        cmapl_file_handle_t * data_file,
                        uint64_t * oldest_time)
{
        if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_get_oldest_record_time_v6_0(
                                        data_file,
                                        oldest_time);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_get_oldest_record_time_v6_0(
                        cmapl_file_handle_t * data_file,
                        uint64_t * oldest_time)
{
	// check if the file handle is valid
        if (data_file->i6_0.data_file == NULL){
                return CMAPL_FILE_NOT_FOUND;
        }
	uint64_t base_time = (data_file->i6_0.header) ? data_file->i6_0.header->i6_0.system.e6_0.base_time : 0;
        *oldest_time = data_file->i6_0.start_time + base_time;
        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_unlink(
                        const char * file_path)
{
	if (file_path == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	cmapl_error_t error = CMAPL_SUCCESS;
	char* metadata_file_path;
        CMAPL_ZALLOC_META_NAME(file_path, metadata_file_path);
	if (metadata_file_path == NULL){
		return CMAPL_NULL_POINTER;
	}

        char* data_file_path;
        CMAPL_ZALLOC_GZ_NAME(file_path, data_file_path);
	if (data_file_path == NULL){
		CMAPL_FREE(metadata_file_path);
                return CMAPL_NULL_POINTER;
        }

	// remove both meta and data file
	int num = CMAPL_UNLINK(metadata_file_path);
        if (num != 0){
                error = CMAPL_UNLINK_ERROR;
                goto cleanup;
        }
        num = CMAPL_UNLINK(data_file_path);
        if (num != 0){
                error = CMAPL_UNLINK_ERROR;
                goto cleanup;
        }

cleanup:
        CMAPL_FREE(metadata_file_path);
        CMAPL_FREE(data_file_path);

        return error;
}

/*
// TODO: remove function
cmapl_error_t
cmapl_write_records(
			cmapl_file_handle_t * data_file,
			char * records_buffer,
			size_t buffer_size)
{
	if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_write_records_v6_0(data_file, records_buffer, buffer_size);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

// TODO: remove function
cmapl_error_t
cmapl_write_records_v6_0(
                        cmapl_file_handle_t * data_file,
                        char * records_buffer,
                        size_t buffer_size)
{
	CMAPL_FILE * gzdata_file = data_file->i6_0.data_file;

	size_t num = CMAPL_FWRITE(records_buffer, buffer_size, gzdata_file);
	if (num != buffer_size){
		return CMAPL_WRITE_FAILED;
	}

	return CMAPL_SUCCESS;
}
*/

cmapl_error_t
cmapl_update_object_index_v6_0(
			cmapl_file_handle_t* data_file,
			uint64_t r_period_timestamp,
			uint64_t r_period,
			uint32_t r_oid)
{
	cmapl_header_handle_t * header = data_file->i6_0.header;
        cmapl_object_handle_t * object = header->i6_0.objects[r_oid];
        if (object == NULL){
                return CMAPL_OBJECT_NOT_FOUND;
        }
        cmapl_period_t * period = object->i6_0.period_list;
        while(period != NULL){
                if (period->period == r_period){
                        break;
                }
		period = period->next_period;
        }
        if (period == NULL){
                return CMAPL_PERIOD_NOT_FOUND;
        }

        // set the timestamp of the first record being written; start_time will always point
        // to the oldest sample date
	if (data_file->i6_0.start_time == CMAPL_MAX_INT){
                cmapl_error_t err = cmapl_set_start_time(data_file, r_period_timestamp);
                if (err != CMAPL_SUCCESS) {
                    return err;
                }
	}

        // throw away anything older than start_date once it is set
        if (r_period_timestamp < data_file->i6_0.start_time) {
		// If coarse check fails then account for period adjustment and try again.
		uint64_t r_period_start_time =
			CMAPL_ROUND_TIME(data_file->i6_0.start_time, r_period);

		if (r_period_timestamp < r_period_start_time) {
	                return CMAPL_INVALID_TIMESTAMP;
		}
        }

        // if period is CONFIG, set index to 0 since there's only one record per file
        uint32_t index = 0;
        if (r_period != CMAPL_CONFIG_PERIOD){
                index = (uint32_t)(CMAPL_GET_BUCKET(r_period_timestamp, data_file->i6_0.start_time, r_period));
        }

	// check if index is within bounds
	// if not, then keep doubling the number of record and iterative record offsets
	while (index >= period->num_offsets){
		uint32_t size = period->num_offsets * 2;
                cmapl_offset_t* record_offsets;
		record_offsets = (cmapl_offset_t*)CMAPL_ZALLOC(size * sizeof(cmapl_offset_t));
		if (record_offsets == NULL){
                        return CMAPL_NULL_POINTER;
                }
		cmapl_iterative_record_offsets_t* iterative_record_offsets;
		iterative_record_offsets = (cmapl_iterative_record_offsets_t*)CMAPL_ZALLOC(size * sizeof(cmapl_iterative_record_offsets_t));
		if (iterative_record_offsets == NULL){
			CMAPL_FREE(record_offsets);
                        return CMAPL_NULL_POINTER;
                }
                CMAPL_MEMCPY(record_offsets, period->record_offsets, period->num_offsets * sizeof(cmapl_offset_t));
                CMAPL_FREE(period->record_offsets);
		period->record_offsets = record_offsets;
		CMAPL_MEMCPY(iterative_record_offsets, period->iterative_record_offsets, period->num_offsets * sizeof(cmapl_iterative_record_offsets_t));
                CMAPL_FREE(period->iterative_record_offsets);
                period->iterative_record_offsets = iterative_record_offsets;
		// set new offsets to the default max int
		uint32_t i;
		for(i = period->num_offsets; i < size; i++){
                        period->record_offsets[i] = CMAPL_NO_RECORD;
                }
                period->num_offsets = size;
	}
        // check if another record for this oid, period, and timestamp exists
	// if no record exists, the value will be max int with the leading bit set to 0
        if (period->record_offsets[index] == CMAPL_NO_RECORD){
                period->record_offsets[index] =
                        data_file->i6_0.data_written_count;
        }else{
                CMAPL_SET_ITERATIVE(period->record_offsets[index]);
		cmapl_iterative_record_offsets_t* iter_record = &(period->iterative_record_offsets[index]);
                if(iter_record->num_offsets == 0){
                        iter_record->num_offsets = 1;
                        iter_record->offsets = (cmapl_offset_t*)CMAPL_ZALLOC(CMAPL_DEFAULT_NUM_ITER_RECS * sizeof(cmapl_offset_t));
			if (iter_record->offsets == NULL){
				return CMAPL_NULL_POINTER;
			}
                        iter_record->offsets[0] = CMAPL_GET_OFFSET(period->record_offsets[index]);
			iter_record->offsets_size = CMAPL_DEFAULT_NUM_ITER_RECS;
                }
		if (iter_record->num_offsets >= iter_record->offsets_size){
			uint32_t size = iter_record->offsets_size * 2;
			cmapl_offset_t* offsets;
			offsets = (cmapl_offset_t*)CMAPL_ZALLOC(size * sizeof(cmapl_offset_t));
			if (offsets == NULL){
				return CMAPL_NULL_POINTER;
			}
			CMAPL_MEMCPY(offsets, iter_record->offsets, iter_record->offsets_size * sizeof(cmapl_offset_t));
			CMAPL_FREE(iter_record->offsets);
			iter_record->offsets = offsets;
			iter_record->offsets_size = size;
		}
                iter_record->offsets[iter_record->num_offsets] = data_file->i6_0.data_written_count;
                iter_record->num_offsets++;
        }
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_write_iterative_record_offsets_v6_0(
			cmapl_file_handle_t * data_file,
			cmapl_period_t * period,
			uint32_t offset_index,
			uint64_t * written_count,
			bool write_flag)
{
	uint64_t count = *written_count;
	CMAPL_FILE* metadata_file = data_file->i6_0.metadata_file;
        cmapl_header_handle_t* header = data_file->i6_0.header;
        cmapl_offset_t object_index_offset = header->i6_0.section_defs[CMAPL_OBJECT_INDEX].section_offset;
	if(CMAPL_IS_ITERATIVE(period->record_offsets[offset_index])){
		cmapl_iterative_record_offsets_t record_offset = period->iterative_record_offsets[offset_index];
                uint32_t num_offsets_size = sizeof(uint32_t);
		uint32_t offsets_size = (uint32_t)record_offset.num_offsets * (uint32_t)sizeof(cmapl_offset_t);
		if (write_flag == true){
			uint32_t num = (uint32_t)CMAPL_FWRITE((char*)&record_offset.num_offsets, num_offsets_size, metadata_file);
                        if (num != num_offsets_size){
                                return CMAPL_WRITE_FAILED;
                        }
			num = (uint32_t)CMAPL_FWRITE((char*)record_offset.offsets, offsets_size, metadata_file);
                        if (num != offsets_size){
                                return CMAPL_WRITE_FAILED;
                        }
		}else{
			period->record_offsets[offset_index] = count - object_index_offset;
			CMAPL_SET_ITERATIVE(period->record_offsets[offset_index]);
			count += (num_offsets_size + offsets_size);
		}
	}
	*written_count = count;
	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_write_object_index_v6_0(
                        cmapl_file_handle_t * data_file)
{
	cmapl_error_t err = CMAPL_SUCCESS;
        uint64_t count = data_file->i6_0.metadata_written_count;
        CMAPL_FILE* metadata_file = data_file->i6_0.metadata_file;
        cmapl_header_handle_t* header = data_file->i6_0.header;

        // first we calculate the offsets for the object index
        header->i6_0.section_defs[CMAPL_OBJECT_INDEX].section_offset = count;
        uint32_t object_index_size = header->i6_0.num_objects_sampled * CMAPL_MAX_NUM_PERIODS * (uint32_t)sizeof(cmapl_offset_t);
        header->i6_0.object_index = (cmapl_offset_t*)CMAPL_ZALLOC(object_index_size);
	if (header->i6_0.object_index == NULL){
		return CMAPL_NULL_POINTER;
	}
        count += object_index_size;

        uint32_t obj_count = 0;
	cmapl_object_handle_t* object = header->i6_0.object_list;
	while(object != NULL){
		cmapl_period_t* period = object->i6_0.period_list;
                while(period != NULL){
			err = cmapl_write_period_offsets_v6_0(data_file, period, obj_count, &count, false);
			if (err != CMAPL_SUCCESS){
				goto write_failed;
			}
                        uint32_t offset_index;
                        for(offset_index = 0; offset_index < period->num_offsets; offset_index++){
				err = cmapl_write_iterative_record_offsets_v6_0(data_file, period, offset_index, &count, false);
				if(err != CMAPL_SUCCESS){
					goto write_failed;
				}
                        }
                        period = period->next_period;
		}
		object = object->i6_0.next_object;
		obj_count++;
        }

	// now we actually write the object index
        uint32_t num = (uint32_t)CMAPL_FWRITE((char*)header->i6_0.object_index, object_index_size, metadata_file);
        if (num != object_index_size){
		err = CMAPL_WRITE_FAILED;
		goto write_failed;
        }

	obj_count = 0;
        object = header->i6_0.object_list;
        while(object != NULL){
                cmapl_period_t* period = object->i6_0.period_list;
                while(period != NULL){
			err = cmapl_write_period_offsets_v6_0(data_file, period, obj_count, &count, true);
                        if (err != CMAPL_SUCCESS){
				goto write_failed;
                        }
                        uint32_t offset_index;
                        for(offset_index = 0; offset_index < period->num_offsets; offset_index++){
				err = cmapl_write_iterative_record_offsets_v6_0(data_file, period, offset_index, &count, true);
                                if(err != CMAPL_SUCCESS){
					goto write_failed;
                                }
                        }
                        period = period->next_period;
                }
                object = object->i6_0.next_object;
                obj_count++;
        }

	data_file->i6_0.metadata_written_count = count;
	return CMAPL_SUCCESS;

write_failed:
	CMAPL_FREE(header->i6_0.object_index);
        header->i6_0.object_index = NULL;
        return err;
}

//lint -esym(429,iterative_record_offsets) "custodial pointer" message. The pointer "iterative_record_offsets" is assigned to the period struct so it doesn't need to be freed.
cmapl_error_t
cmapl_write_period_offsets_v6_0(
			cmapl_file_handle_t * data_file,
			cmapl_period_t * period,
			uint32_t obj_count,
			uint64_t * written_count,
			bool write_flag)
{
	cmapl_error_t err = CMAPL_SUCCESS;
	cmapl_header_handle_t* header = data_file->i6_0.header;
	CMAPL_FILE* metadata_file = data_file->i6_0.metadata_file;
	uint64_t count = *written_count;
	cmapl_offset_t object_index_offset = header->i6_0.section_defs[CMAPL_OBJECT_INDEX].section_offset;

	// check if start and end times are valid
        if (data_file->i6_0.end_time <= data_file->i6_0.start_time){
                return CMAPL_INVALID_TIMESTAMP;
        }

	// get the final number of records offsets to be written to the object index; this final number is
        // the size of the index stripped of unused buckets, hence (end/period - start/period + fudge)
	uint32_t num_final_offsets = 0;
	err = cmapl_get_num_record_offsets_v6_0(data_file->i6_0.start_time, data_file->i6_0.end_time,
		period->period, &num_final_offsets);
	if (err != CMAPL_SUCCESS){
		return err;
	}

	// resize the record offsets arrays to the new final size if it isn't already
	if (num_final_offsets != period->num_offsets){
		cmapl_offset_t* record_offsets;
		record_offsets = (cmapl_offset_t*)CMAPL_ZALLOC(num_final_offsets * sizeof(cmapl_offset_t));
		if (record_offsets == NULL){
			return CMAPL_NULL_POINTER;
		}
		cmapl_iterative_record_offsets_t* iterative_record_offsets;
		iterative_record_offsets = (cmapl_iterative_record_offsets_t*)CMAPL_ZALLOC(num_final_offsets *
										sizeof(cmapl_iterative_record_offsets_t));
		if (iterative_record_offsets == NULL){
			CMAPL_FREE(record_offsets);
			return CMAPL_NULL_POINTER;
		}

                // set any new trailing offset values to the default offset value
                if (num_final_offsets > period->num_offsets){
                        uint32_t i;
                        for(i = period->num_offsets; i < num_final_offsets; i++){
                            record_offsets[i] = CMAPL_NO_RECORD;
                        }
                }

                // do not copy past the end of new array
                uint32_t copy_length = (num_final_offsets < period->num_offsets) ? num_final_offsets: period->num_offsets;

                // copy over all the old offset values to the new final sized arrays
                CMAPL_MEMCPY(record_offsets, period->record_offsets,
                        copy_length * sizeof(cmapl_offset_t));
                CMAPL_MEMCPY(iterative_record_offsets, period->iterative_record_offsets,
                        copy_length * sizeof(cmapl_iterative_record_offsets_t));

		CMAPL_FREE(period->record_offsets);
		period->record_offsets = record_offsets;
                CMAPL_FREE(period->iterative_record_offsets);
                period->iterative_record_offsets = iterative_record_offsets;
                period->num_offsets = num_final_offsets;
	}
	uint32_t record_offsets_size = (uint32_t)(num_final_offsets * sizeof(cmapl_offset_t));
	if (write_flag == true){
		uint32_t num = (uint32_t)CMAPL_FWRITE((char*)period->record_offsets, record_offsets_size, metadata_file);
                if (num != record_offsets_size){
                        return CMAPL_WRITE_FAILED;
                }
	}else{
		err = cmapl_set_object_period_offset(header, obj_count, period->period, count - object_index_offset);
                if (err != CMAPL_SUCCESS){
                        return err;
                }
		count += record_offsets_size;
	}

	*written_count = count;
	return CMAPL_SUCCESS;
}
//lint +esym(429,iterative_record_offsets)

cmapl_error_t
cmapl_write_record(
                        cmapl_file_handle_t * data_file,
                        void * record_info,
                        char * record_buffer)
{
        if (data_file == NULL || record_info == NULL || record_buffer == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

	// user specified version must match file version
        if (!cmapl_is_version(record_info, data_file->major_v, data_file->minor_v)){
		return CMAPL_VERSION_MISMATCH;
	}

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_write_record_v6_0(
                                                data_file,
                                                (cmapl_record_header_data_t_v6_0*)record_info,
                                                record_buffer);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_write_record_v6_0(
                        cmapl_file_handle_t * data_file,
                        cmapl_record_header_data_t_v6_0* record_info,
                        char * record_buffer)
{
	// check if the file handle is valid
        if (data_file->i6_0.data_file == NULL){
                return CMAPL_FILE_NOT_FOUND;
        }

        CMAPL_GZFILE gzdata_file = data_file->i6_0.data_file;

	uint64_t r_msecs = record_info->msecs_since_base;
	uint32_t r_size = record_info->size;
	uint32_t r_oid = record_info->obj_id;
	uint64_t r_period = record_info->period;
	// calculate the period timestamp
	uint64_t r_period_timestamp = 0;
	if (r_period == CMAPL_CONFIG_PERIOD){
		r_period_timestamp = r_msecs;
	}else{
		r_period_timestamp = CMAPL_ROUND_TIME(r_msecs, r_period);
	}
	cmapl_error_t err = cmapl_update_object_index_v6_0(data_file, r_period_timestamp, r_period, r_oid);
	if (err != CMAPL_SUCCESS){
		return err;
	}

	// updating end time
	if (data_file->i6_0.end_time < r_period_timestamp){
		data_file->i6_0.end_time = r_period_timestamp;
	}

	// don't write out the version information in the record header data structure
	size_t version_size = (size_t)&record_info->msecs_since_base - (size_t)&record_info->major_v;
        size_t num = CMAPL_GZWRITE((char*)record_info + version_size,
		(uint32_t)sizeof(cmapl_record_header_data_t_v6_0) - (uint32_t)version_size, gzdata_file);
        if (num != sizeof(cmapl_record_header_data_t_v6_0) - version_size){
                return CMAPL_WRITE_FAILED;
        }
	data_file->i6_0.data_written_count += (sizeof(cmapl_record_header_data_t_v6_0) - version_size);

        num = CMAPL_GZWRITE(record_buffer, r_size, gzdata_file);
        if (num != r_size){
                return CMAPL_WRITE_FAILED;
        }
	data_file->i6_0.data_written_count += r_size;

        return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_get_num_record_offsets_v6_0(
                        uint64_t start_time,
			uint64_t end_time,
                        uint64_t period,
                        uint32_t * num_offsets)
{
        // CONFIG period always has just one offset per file
        *num_offsets = 1;
        if (period != CMAPL_CONFIG_PERIOD){
		// calculate number of record offsets there should be, rounded up to account for
		// the extreme case that the period is sampled at the ends of the time range;
                *num_offsets = (uint32_t)(CMAPL_GET_BUCKET(end_time, start_time, period) + 1);
        }

	return CMAPL_SUCCESS;
}

cmapl_error_t
cmapl_set_start_time(
                        cmapl_file_handle_t * data_file,
                        uint64_t start_time)
{
        if (data_file == NULL){
                return CMAPL_INVALID_ARGUMENTS;
        }

        if (cmapl_is_version(data_file, 6, 0)){
                return cmapl_set_start_time_v6_0(
                                        data_file,
                                        start_time);
        }else {
                return CMAPL_INVALID_VERSION;
        }
}

cmapl_error_t
cmapl_set_start_time_v6_0(
                        cmapl_file_handle_t * data_file,
                        uint64_t start_time)
{
        // start time should only be set once
        if (data_file->i6_0.start_time != CMAPL_MAX_INT) {
                return CMAPL_INVALID_TIMESTAMP;
        }

        data_file->i6_0.start_time = start_time;
        return CMAPL_SUCCESS;
}
