﻿using NUnit.Framework;
using odct;
using odct.Clusters;
using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;

namespace odct.tests {
    [TestFixture]
    class ClusterNodeTests {

        public Credentials credentials;
        public ClusterNode clusterNode;

        [TestFixtureSetUp]
        public void init() {
            credentials = new Credentials("10.26.9.242", "admin", "netapp11");
            credentials.Verify(new MessengerStub());
            clusterNode = new ClusterNode(new MessengerStub(), new ClusterStub_8_3(), "atlas-02", credentials, "");
            clusterNode.nodeMgmtIp = "10.26.9.242";
            CmaplWrapper.LoadDllsIfNecessary();
        }

        [Test]
        [Category("Integration")]
        public void canGetClusterNodeInfo() {

            Assert.AreEqual("1874006229", clusterNode.FilerSystemId);
            Assert.AreEqual("800000098897", clusterNode.FilerSerialNumber);
            Assert.AreEqual("atlas-01", clusterNode.PartnerName);

        }

        [Test]
        [Category("BrokenTests")]
        // now works from nunit-x86.exe
        public void canParseArchiveFile() {

            const string infileName = "043311_003600_1462991400454_0899000.ccma";

            var outfileFullPath = Path.GetTempFileName();
            var outfileName = Path.GetFileName(outfileFullPath);
            var outfilePath = Path.GetTempPath();

            string filepath = OdctSettingsTest.GetRelativeUnitTestDataDirectory("cmode84");

            ClusterNode.ParseParameters parseParameters = new ClusterNode.ParseParameters() {
                objectName = "system",
                numObjects = 1,
                countersString = "nfs_ops",
                numCounters = "1"
            };

            var expected = "TIMESTAMP,instance_name,system:nfs_ops\n"
                + "1462991448454,gorgon-03,11695900\n"
                + "1462991748454,gorgon-03,11704051\n"
                + "1462992048454,gorgon-03,11712085\n";

            clusterNode.parseArchiveFile(infileName, outfileFullPath, filepath, 300000, parseParameters);

            var actual = File.ReadAllText(outfileFullPath);
            Console.WriteLine(actual);

            Assert.AreEqual(expected, actual);
        }

        [Test]
        [Category("Integration")]
        public void canGetFileList() {

            List<string> results = clusterNode.getArchiveBaseList();

            Assert.Less(1500, results.Count);
            Assert.Greater(3500, results.Count);

            Regex regex = new Regex("\\d{6}_\\d{6}_\\d{13}_\\d{7}");
            MatchCollection matches;

            foreach (string ccmaBase in results) {
                Assert.AreEqual(35, ccmaBase.Length);
                matches = regex.Matches(ccmaBase);
            }
        }

        [Test]
        [Category("Integration")]
        public void canCopySpiFileToLocal() {

            TempFolder tempFolder = new TempFolder();
            string localpath = Path.Combine(tempFolder.PathName, "atlas-02");
            if (!Directory.Exists(localpath)) {
                Directory.CreateDirectory(localpath);
            }
            List<string> results = clusterNode.getArchiveBaseList();
            string ccmaBase = clusterNode.copyCcmaFilesToLocal(results[0]);

            Assert.That(File.Exists(Path.Combine(localpath, ccmaBase + ".gz")));
            Assert.That(File.Exists(Path.Combine(localpath, ccmaBase + ".meta")));

            Directory.Delete(localpath, true);
            tempFolder.Cleanup();
        }

        [Test]
        [Category("BrokenTests")]
        public void work() {

            const string infileName = "043311_003600_1462991400454_0899000.ccma";

            var outfilepath = Path.GetTempFileName();
//            var outfilename = Path.GetFileName(outfilepath);
//            var outfilefolder = Path.GetTempPath();

            string infilefolder = OdctSettingsTest.GetRelativeUnitTestDataDirectory("cmode84");
            string infilepath = Path.Combine(infilefolder, infileName);

            string counters = CmaplWrapper.GetCounterNames(
                infilepath,
                "system",
                new Int32[] {5000, 60000}, 
                new MessengerStub(),
                "");

            ClusterNode.ParseParameters parseParameters = new ClusterNode.ParseParameters() {
                objectName = "system",
                numObjects = 1,
                countersString = counters,
                numCounters = "1"
            };

            var expected = "TIMESTAMP,instance_name,system:nfs_ops\n"
                + "1462991448454,gorgon-03,11695900\n"
                + "1462991748454,gorgon-03,11704051\n"
                + "1462992048454,gorgon-03,11712085\n";

            clusterNode.parseArchiveFile(infilepath, outfilepath, infilefolder, 60000, parseParameters);

            var actual = File.ReadAllText(outfilepath);
            Console.WriteLine(actual);

            Assert.AreEqual(expected, actual);
        }


    }
}
