using System;
using System.Collections.Generic;
using System.IO;
using NUnit.Framework;
using odct.Counters;
using odct.Settings;

namespace odct.tests.Counters {

    [TestFixture]
    [Category("Integration")]
    public class ArchiveParserCounterTests {
        ArchiveParser ap_config;
        ArchiveParser ap_freq1;
        ArchiveParser ap_freq2;
        ArchiveParser ap_freq3;
        ArchiveParser ap_freq4;
        List<string> systemCounterNames;
        List<string> volumeCounterNames;
        
        public static Dictionary<string, int> getTestPerfCounters(int freq) {
            return new Dictionary<string, int> {
                    { "system.system_model", freq },
                    { "system.ontap_version", freq },
                    { "system.serial_no", freq },
                    { "system.system_id", freq },
                    { "system.hostname", freq },
                    { "system.nfs_ops", freq },
                    { "system.cifs_ops", freq }, 
                    { "system.http_ops", freq }, 
                    { "system.fcp_ops", freq }, 
                    { "system.iscsi_ops", freq },
                    { "system.sys_read_latency_hist", freq },
                    { "system.sys_write_latency_hist", freq },
                    { "system.sys_latency_hist", freq },
                    { "system.net_data_recv", freq },
                    { "system.net_data_sent", freq },
                    { "system.disk_data_read", freq },
                    { "system.disk_data_written", freq },
                    { "system.cpu_busy", freq },
                    { "system.cpu_elapsed_time", freq },
                    { "system.avg_processor_busy", freq },
                    { "system.cpu_elapsed_time1", freq },
                    { "system.total_processor_busy", freq },
                    { "system.cpu_elapsed_time2", freq },
                    { "system.num_processors", freq },
                    { "volume.parent_aggr", freq },
                    { "volume.avg_latency", freq },
                    { "volume.total_ops", freq },
                    { "volume.read_data", freq },
                    { "volume.read_latency", freq },
                    { "volume.read_ops", freq },
                    { "volume.write_data", freq },
                    { "volume.write_latency", freq },
                    { "volume.write_ops", freq },
                    { "volume.other_latency", freq },
                    { "volume.other_ops", freq }};
        }
            
        [TestFixtureSetUp]
        public void init() {
            ZapiContextStub zc = new ZapiContextStub(null, null);
            var gzipSource = new ZapiArchiveFileSource(zc) {
                Filename = "stats_archive_data_1447684431_20151116_093351.gz",
                Path = OdctSettingsTest.GetRelativeUnitTestDataDirectory("ontap73")
            };


            var work = TestContext.CurrentContext.WorkDirectory;

            ap_config = new ArchiveParser_7_3 (
                "", new MessengerStub(), gzipSource, Path.Combine(work, @"\odcttest\parsertest"),
                getTestPerfCounters(0), false);

            ap_freq1 = new ArchiveParser_7_3 (
                "", new MessengerStub(), gzipSource, Path.Combine(work, @"\odcttest\parsertest"),
                getTestPerfCounters(1), false);

            ap_freq2 = new ArchiveParser_7_3 (
                "", new MessengerStub(), gzipSource, Path.Combine(work, @"\odcttest\parsertest"),
                getTestPerfCounters(2), false);

            ap_freq3 = new ArchiveParser_7_3 (
                "", new MessengerStub(), gzipSource, Path.Combine(work, @"\odcttest\parsertest"),
                getTestPerfCounters(3), false);

            ap_freq4 = new ArchiveParser_7_3 (
                "", new MessengerStub(), gzipSource, Path.Combine(work, @"\odcttest\parsertest"),
                getTestPerfCounters(4), false);

            List<string> systemCounterNamesBeforeHistograms = new List<string> {
                    "system_model",
                    "ontap_version",
                    "serial_no",
                    "system_id",
                    "hostname",
                    "nfs_ops",
                    "cifs_ops",
                    "http_ops",
                    "fcp_ops",
                    "iscsi_ops"};

            List<string> systemCounterNamesAfterHistograms = new List<string> {
                    "net_data_recv",
                    "net_data_sent",
                    "disk_data_read",
                    "disk_data_written",
                    "cpu_busy",
                    "cpu_elapsed_time",
                    "avg_processor_busy",
                    "cpu_elapsed_time1",
                    "total_processor_busy",
                    "cpu_elapsed_time2",
                    "num_processors"};

            List<string> systemXNames = new List<string> {
                    "0 - <1ms",
                    "1 - <2ms",
                    "2 - <4ms",
                    "4 - <6ms",
                    "6 - <8ms",
                    "8 - <10ms",
                    "10 - <12ms",
                    "12 - <16ms",
                    "16 - <20ms",
                    "20 - <30ms",
                    "30 - <40ms",
                    "40 - <50ms",
                    "50 - <60ms",
                    "60 - <70ms",
                    "70 - <80ms",
                    "80 - <90ms",
                    "90 - <100ms",
                    "100 - <120ms",
                    "120 - <140ms",
                    "140 - <160ms",
                    "160 - <180ms",
                    "180 - <200ms",
                    "200 - <400ms",
                    "400 - <600ms",
                    "600 - <800ms",
                    "800 - <1000ms",
                    "1000 - <1500ms",
                    "1500 - <2000ms",
                    "2000 - <2500ms",
                    "2500 - <3000ms",
                    "3000 - <3500ms",
                    "3500 - <4000ms",
                    "4000 - <8000ms",
                    "8000 - <12000ms",
                    "12000 - <16000ms",
                    ">16000ms"};


            List<string> systemHist1 = new List<string>();
            List<string> systemHist2 = new List<string>();
            List<string> systemHist3 = new List<string>();
            foreach (string s in systemXNames) {
                systemHist1.Add(string.Format("sys_read_latency_hist:{0}", s.ToString()));
            }
            foreach (string s in systemXNames) {
                systemHist2.Add(string.Format("sys_write_latency_hist:{0}", s.ToString()));
            }
            foreach (string s in systemXNames) {
                systemHist3.Add(string.Format("sys_latency_hist:{0}", s.ToString()));
            }

            systemCounterNames = new List<string>();
            systemCounterNames.AddRange(systemCounterNamesBeforeHistograms);
            systemCounterNames.AddRange(systemHist1);
            systemCounterNames.AddRange(systemHist2);
            systemCounterNames.AddRange(systemHist3);
            systemCounterNames.AddRange(systemCounterNamesAfterHistograms);

            volumeCounterNames = new List<string> {               
                    "parent_aggr",
                    "avg_latency",
                    "total_ops",
                    "read_data",
                    "read_latency",
                    "read_ops",
                    "write_data",
                    "write_latency",
                    "write_ops",
                    "other_latency",
                    "other_ops"};

        }

        [Test]
        public void CanGetSystemCounterNamesFromConfigCounter() {
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_config.getArchiveXmlAsString(), "system");
            List<string> actual = ap_config.getCounterNames(o.GetCounterDefinitionListSortedByCounterId());
            Assert.AreEqual(systemCounterNames, actual);
        }

        [Test]
        public void CanGetSystemCounterNamesFromFreq1Counter() {
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq1.getArchiveXmlAsString(), "system");
            List<string> actual = ap_freq1.getCounterNames(o.GetCounterDefinitionListSortedByCounterId());
            Assert.AreEqual(systemCounterNames, actual);
        }

        [Test]
        public void CanGetSystemCounterNamesFromFreq2Counter() {
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq2.getArchiveXmlAsString(), "system");
            List<string> actual = ap_freq2.getCounterNames(o.GetCounterDefinitionListSortedByCounterId());
            Assert.AreEqual(systemCounterNames, actual);
        }

        [Test]
        public void CanGetSystemCounterNamesFromFreq3Counter() {
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq3.getArchiveXmlAsString(), "system");
            List<string> actual = ap_freq3.getCounterNames(o.GetCounterDefinitionListSortedByCounterId());
            Assert.AreEqual(systemCounterNames, actual);
        }

        [Test]
        public void CanGetSystemCounterNamesFromFreq4Counter() {
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq4.getArchiveXmlAsString(), "system");
            List<string> actual = ap_freq4.getCounterNames(o.GetCounterDefinitionListSortedByCounterId());
            Assert.AreEqual(systemCounterNames, actual);
        }

        [Test]
        public void CanGetVolumeCounterNamesFromConfigCounter() {
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_config.getArchiveXmlAsString(), "volume");
            List<string> actual = ap_config.getCounterNames(o.GetCounterDefinitionListSortedByCounterId());
            Assert.AreEqual(volumeCounterNames, actual);
        }

        [Test]
        public void CanGetVolumeCounterValuesFromConfigCounter() {
            //          Record from 171611 to 178370
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_config.getArchiveXmlAsString(), "volume");
            int offset = 171747;
            List<string> actual
                = ap_config.getCounterValues(ref ap_config.allBytes, o.GetCounterDefinitionListSortedByCounterId(), new Flag(917504), ref offset, 171611);
            List<string> expected = new List<string> {               
                    "oracle",
                    "110268030733",
                    "378477517",
                    "9762668794",
                    "17852619895",
                    "1406492",
                    "2686316089344",
                    "55080806575",
                    "364491213",
                    "37334604263",
                    "12579812"};
            Assert.AreEqual(expected, actual);
        }

        [Test]
        public void CanGetSystemHistogramNamesFromFreq4Counter() {
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq4.getArchiveXmlAsString(), "system");
            List<string> expected = new List<string> {
                    "0 - <1ms",
                    "1 - <2ms",
                    "2 - <4ms",
                    "4 - <6ms",
                    "6 - <8ms",
                    "8 - <10ms",
                    "10 - <12ms",
                    "12 - <16ms",
                    "16 - <20ms",
                    "20 - <30ms",
                    "30 - <40ms",
                    "40 - <50ms",
                    "50 - <60ms",
                    "60 - <70ms",
                    "70 - <80ms",
                    "80 - <90ms",
                    "90 - <100ms",
                    "100 - <120ms",
                    "120 - <140ms",
                    "140 - <160ms",
                    "160 - <180ms",
                    "180 - <200ms",
                    "200 - <400ms",
                    "400 - <600ms",
                    "600 - <800ms",
                    "800 - <1000ms",
                    "1000 - <1500ms",
                    "1500 - <2000ms",
                    "2000 - <2500ms",
                    "2500 - <3000ms",
                    "3000 - <3500ms",
                    "3500 - <4000ms",
                    "4000 - <8000ms",
                    "8000 - <12000ms",
                    "12000 - <16000ms",
                    ">16000ms"};
/*            List<string> actual1
                = ap_freq4.getCounterNames(o.GetCounterDefinitionByCounterName("sys_read_latency_hist"), new Flag(3932160));
            List<string> actual2
                = ap_freq4.getCounterNames(o.GetCounterDefinitionByCounterName("sys_read_latency_hist"), new Flag(3932160));
            List<string> actual3
                = ap_freq4.getCounterNames(o.GetCounterDefinitionByCounterName("sys_read_latency_hist"), new Flag(3932160));
            Assert.AreEqual(expected, actual1);
            Assert.AreEqual(expected, actual2);
            Assert.AreEqual(expected, actual3);  */
        }

        [Test]
        public void CanGetSystemCountersFromConfigCounter() {
            //          Record from 132583 to 133258
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_config.getArchiveXmlAsString(), "system");
            int offset = 132719;
            List<string> actual
                = ap_config.getCounterValues(ref ap_config.allBytes, o.GetCounterDefinitionListSortedByCounterId(), new Flag(917504), ref offset, 132583);
            List<string> expected = new List<string> {               
                    "FAS3170",
                    "NetApp Release 7.3.4: Thu May 27 16:40:57 PDT 2010",
                    "70002447",
                    "0151703719",
                    "sylvester",
                    "76",
                    "5570",
                    "0",
                    "3556213778",
                    "0",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "259992591",
                    "17950649775",
                    "312955289092",
                    "70427909848",
                    "3098645562191",
                    "2678387511794",
                    "1399020298692",
                    "2678387511794",
                    "5596081194768",
                    "2678387511794",
                    "4"};
            Assert.AreEqual(expected, actual);
        }

        [Test]
        public void CanGetSystemCountersFromFreq1Counter() {
            //          Record from 180163 to 180418
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq1.getArchiveXmlAsString(), "system");
            int offset = 180299;
            List<string> actual
                = ap_freq1.getCounterValues(ref ap_freq1.allBytes, o.GetCounterDefinitionListSortedByCounterId(), new Flag(262144), ref offset, 180163);
            List<string> expected = new List<string> {               
                    "",
                    "",
                    "",
                    "",
                    "",
                    "76",
                    "5570",
                    "0",
                    "3556213790",
                    "0",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "259992591",
                    "17950649775",
                    "312955289180",
                    "70427909856",
                    "3098645683259",
                    "2678388511760",
                    "1399020342592",
                    "2678388511760",
                    "5596081370368",
                    "2678388511760",
                    ""};
            Assert.AreEqual(expected, actual);
        }

        [Test]
        public void CanGetSystemCountersFromFreq2Counter() {
            //          Record from 225315 to 225570
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq2.getArchiveXmlAsString(), "system");
            int offset = 225451;
            List<string> actual
                = ap_freq2.getCounterValues(ref ap_freq2.allBytes, o.GetCounterDefinitionListSortedByCounterId(), new Flag(786432), ref offset, 225315);
            List<string> expected = new List<string> {               
                    "",
                    "",
                    "",
                    "",
                    "",
                    "76",
                    "5570",
                    "0",
                    "3556213994",
                    "0",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "259992591",
                    "17950649775",
                    "312955291624",
                    "70427914056",
                    "3098645738583",
                    "2678392511538",
                    "1399020381032",
                    "2678392511538",
                    "5596081524128",
                    "2678392511538",
                    ""};
            Assert.AreEqual(expected, actual);
        }

        [Test]
        public void CanGetSystemCountersFromFreq3Counter() {
            //          Record from 844515 to 844770
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq3.getArchiveXmlAsString(), "system");
            int offset = 844651;
            List<string> actual
                = ap_freq3.getCounterValues(ref ap_freq3.allBytes, o.GetCounterDefinitionListSortedByCounterId(), new Flag(1835008), ref offset, 844515);
            List<string> expected = new List<string> {               
                    "",
                    "",
                    "",
                    "",
                    "",
                    "76",
                    "5570",
                    "0",
                    "3556216494",
                    "0",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
                    "259992874",
                    "17950670862",
                    "312955362480",
                    "70427988316",
                    "3098649230454",
                    "2678432518292",
                    "1399021648529",
                    "2678432518292",
                    "5596086594116",
                    "2678432518292",
                    ""};
            Assert.AreEqual(expected, actual);
        }

        [Test]
        public void CanGetSystemCountersFromFreq4Counter() {
            //          Record from 1774659 to 1775778
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq4.getArchiveXmlAsString(), "system");
            int offset = 1774795;
            List<string> actual
                = ap_freq4.getCounterValues(ref ap_freq4.allBytes, o.GetCounterDefinitionListSortedByCounterId(), new Flag(3932160), ref offset, 1774659);
            List<string> expected = new List<string> {               
                    "",
                    "",
                    "",
                    "",
                    "",
                    "76",
                    "5570",
                    "0",
                    "3556219935",
                    "0",
                    "779615367", // start of histogram 1
                    "7499875",
                    "5759637",
                    "1851939",
                    "1867830",
                    "1749792",
                    "2198288",
                    "5405621",
                    "5952876",
                    "24026227",
                    "40961486",
                    "42281584",
                    "28407488",
                    "17796760",
                    "13431484",
                    "12504627",
                    "12118489",
                    "21200895",
                    "16842007",
                    "11928063",
                    "6427272",
                    "2802393",
                    "2252117",
                    "12247",
                    "2036",
                    "1483",
                    "2236",
                    "437",
                    "219",
                    "58",
                    "30",
                    "20",
                    "14",
                    "6",
                    "6",
                    "8",  // end of histogram 1
                    "1337347446",  // start of histogram 2
                    "14197661",
                    "12592312",
                    "10478289",
                    "1543992",
                    "964297",
                    "648634",
                    "998163",
                    "739408",
                    "678685",
                    "226054",
                    "105722",
                    "52022",
                    "33257",
                    "23241",
                    "15567",
                    "11260",
                    "16303",
                    "9280",
                    "4291",
                    "2199",
                    "1305",
                    "5385",
                    "1576",
                    "506",
                    "1365",
                    "2984",
                    "3402",
                    "3658",
                    "5375",
                    "5475",
                    "5044",
                    "18266",
                    "704",
                    "20",
                    "12",  // end of histogram 2
                    "2116966842",  // start of histogram 3
                    "21697544",
                    "18351953",
                    "12330231",
                    "3411822",
                    "2714091",
                    "2846923",
                    "6403785",
                    "6692285",
                    "24704912",
                    "41187540",
                    "42387306",
                    "28459510",  // element 94
                    "17830017",
                    "13454725",
                    "12520194",
                    "12129749",
                    "21217198",
                    "16851287",  // 100
                    "11932354",
                    "6429471",
                    "2803698",
                    "2257502",
                    "13823",  // 105
                    "2542",
                    "2848",
                    "5220",
                    "3839",
                    "3877",  // 110
                    "5433",
                    "5505",
                    "5064",
                    "19431",
                    "710", // 115
                    "26",
                    "20",  // end of histogram 3
                    "259993339",
                    "17950691821",
                    "312955450620",
                    "70428084732",
                    "3098652961483",
                    "2678492519944",
                    "1399023124866",
                    "2678492519944",
                    "5596092499464",
                    "2678492519944",
                    ""};
            Assert.AreEqual(expected, actual);
        }

        [Test]
        public void CanGetSystemHistogramValuesFromFreq4Counter() {
            CounterDefinitionLookup o = new CounterDefinitionLookup(ap_freq4.getArchiveXmlAsString(), "system");
            CounterDefinition hist12 = o.GetCounterDefinitionByCounterId(12);
            CounterDefinition hist15 = o.GetCounterDefinitionByCounterId(15);
            CounterDefinition hist18 = o.GetCounterDefinitionByCounterId(18);
            List<string> expected1 = new List<string> {
                    "779615367","7499875","5759637","1851939","1867830","1749792",
                    "2198288","5405621","5952876","24026227","40961486","42281584",
                    "28407488","17796760","13431484","12504627","12118489","21200895",
                    "16842007","11928063","6427272","2802393","2252117","12247",
                    "2036","1483","2236","437","219","58",
                    "30","20","14","6","6","8"};
            List<string> expected2 = new List<string> {
                    "1337347446","14197661","12592312","10478289","1543992","964297",
                    "648634","998163","739408","678685","226054","105722",
                    "52022","33257","23241","15567","11260","16303",
                    "9280","4291","2199","1305","5385","1576",
                    "506","1365","2984","3402","3658","5375",
                    "5475","5044","18266","704","20","12"};
            List<string> expected3 = new List<string> {
                    "2116966842","21697544","18351953","12330231","3411822","2714091",
                    "2846923","6403785","6692285","24704912","41187540","42387306",
                    "28459510","17830017","13454725","12520194","12129749","21217198",
                    "16851287","11932354","6429471","2803698","2257502","13823",
                    "2542","2848","5220","3839","3877","5433",
                    "5505","5064","19431","710","26","20"};
            int offset = 1774795 + 40;
            List<string> actual1 = ap_freq4.getHistogramValues(ref ap_freq4.allBytes, hist12, new Flag(3932160), offset, true);
            offset = 1774795 + 328;
            List<string> actual2 = ap_freq4.getHistogramValues(ref ap_freq4.allBytes, hist15, new Flag(3932160), offset, true);
            offset = 1774795 + 616;
            List<string> actual3 = ap_freq4.getHistogramValues(ref ap_freq4.allBytes, hist18, new Flag(3932160), offset, true);
            Assert.AreEqual(expected1, actual1);
            Assert.AreEqual(expected2, actual2);
            Assert.AreEqual(expected3, actual3);
        }

    }
}
