﻿using System;
using System.Collections.Generic;
using System.IO;
using NUnit.Framework;
using odct;
using odct.Counters;
using odct.Settings;

namespace odct.tests.Counters {

    [TestFixture]
    [Category("Integration")]
    public class ArchiveParserTests {
        ArchiveParser ap;
        static ZapiContextStub zc = new ZapiContextStub(null, null);
        static ArchiveFileSource gzipSource =  new ZapiArchiveFileSource(zc) {
                Filename = "stats_archive_data_1447684431_20151116_093351.gz",
                Path = OdctSettingsTest.GetRelativeUnitTestDataDirectory("ontap73")
            };
        static string work = TestContext.CurrentContext.WorkDirectory;
        static string testDir = Path.Combine(work, @"\odcttest\parsertest");

        [TestFixtureSetUp]
        public void init() {
            var work = TestContext.CurrentContext.WorkDirectory;

            ZapiContextStub zc = new ZapiContextStub(null, null);
            var gzipSource = new ZapiArchiveFileSource(zc) {
                Filename = "stats_archive_data_1447684431_20151116_093351.gz",
                Path = OdctSettingsTest.GetRelativeUnitTestDataDirectory("ontap73")
            };

            ap = new ArchiveParserFactory(
                "",                     // system serial number, unneeded for test
                new MessengerStub(),
                gzipSource, 
                testDir, 
                "trt",
                false,                  // write xml to payload
                false)                  // isClusterMode
                .Build("7.3.x");

        }

        [Test]
        public void ArchiveParserFindsEndOfXml() {
            Assert.AreEqual(132581, ap.findXmlEnd());
        }

        [Test]
        [Ignore("Will not be used in production but was useful in development")]
        public void canCopyXmlToFile() {
            Assert.AreEqual(1, ap.writeXMLAsFile());
        }

        [Test]
        public void CanConvert5LittleEndianBytesToLong() {
            byte[] testArray = new byte[] { 0x3E, 0x0B, 0xE8, 0x8D, 0x34, 0x89, 0x0A, 0x00, 0x00, 0x00, 0xA4 };
            Assert.AreEqual(45251595752, testArray.getBytesAsUInt64(2, 5));
        }

        [Test]
        public void CanConvertUint64MaxValue() {
            byte[] testArray = new byte[] { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
            Assert.AreEqual(18446744073709551615, testArray.getBytesAsUInt64(0, 8));
        }

        [Test]
        public void CanConvertUint64MaxValueToString() {
            byte[] testArray = new byte[] { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
            Assert.AreEqual("18446744073709551615", testArray.getBytesAsUInt64(0, 8).ToString());
        }

        [Test]
        public void CanConvertUint64CounterValueToString() {
            byte[] testArray = new byte[] { 0x80, 0x5C, 0x06, 0xBC, 0xA7, 0xEC, 0xFF, 0xFF };
            Assert.AreEqual("18446722803891068032", testArray.getBytesAsUInt64(0, 8).ToString());
        }

        [Test]
        public void CanGetRecord1() {
            Assert.AreEqual(133258, ap.getRecordEnd(132583));
        }

        [Test]
        public void CanProcessRecords() {
            Assert.AreEqual(1, ap.processRecords());
        }

        [Test]
        public void CanGetArchiveAttributeVersion() {
            Assert.AreEqual("2", ap.getArchiveAttribute("version"));
        }

        [Test]
        public void CanGetArchiveAttributeGeneration() {
            Assert.AreEqual("1", ap.getArchiveAttribute("generation"));
        }

        [Test]
        public void CanGetArchiveAttributeUnixTime() {
            Assert.AreEqual("1447684431", ap.getArchiveAttribute("unix_timestamp"));
        }

        [Test]
        public void CanGetArchiveAttributeMSecsSinceBoot() {
            Assert.AreEqual("45251595752", ap.getArchiveAttribute("msecs_since_boot"));
        }

        [Test]
        public void CanGetArchiveDateTime() {
            Assert.AreEqual(new DateTime(2015, 11, 16, 14, 33, 51), ap.getArchiveDateTime());
        }

        [Test]
        [Category("Unit")]
        public void CanGetDateTimeFromUnixTime() {
            const long seconds = 2500000;
            Assert.AreEqual(new DateTime(1970, 1, 29, 22, 26, 40), ap.getDateTimeFomUnixTime(seconds));

        }

        [Test]
        public void CanConvertStringListToCsv() {
            List<String> testList = new List<String>() { "one", "two", "three" };
            const string expected = "one,two,three";
            Assert.AreEqual(expected, ArchiveParser.getStringListAsCsv(testList));
        }

        [Test]
        public void CanConvertStringListWithBlanksToCsv() {
            List<String> testList = new List<String>() { "", "","two", "", "three" };
            const string expected = ",,two,,three";
            Assert.AreEqual(expected, ArchiveParser.getStringListAsCsv(testList));
        }

        [Test]
        public void CanCalcCollectedFrequencyByDottedItem() {
            Dictionary<string, int> input = new Dictionary<string, int>();
            // CONFIG counters
            input["system.system_model"] = 0;
            input["system.ontap_version"] = 1;
            input["system.serial_no"] = 2;
            input["system.system_id"] = 3;
            input["system.hostname"] = 4;
            // Freq 1 counters
            input["system.nfs_ops"] = 0;
            input["system.cifs_ops"] = 1;
            input["system.http_ops"] = 2;
            input["system.fcp_ops"] = 3;
            input["system.iscsi_ops"] = 4;
            // Freq 2 counters
            input["volume.avg_latency"] = 0;
            input["volume.total_ops"] = 1;
            input["volume.read_data"] = 2;
            input["volume.read_ops"] = 3;
            input["volume.write_data"] = 4;
            // Freq 4 counters
            input["system.sys_read_latency_hist"] = 0;
            input["system.sys_write_latency_hist"] = 3;
            input["system.sys_latency_hist"] = 4;

            Dictionary<string, long> calculated = ap.calcCollectedFrequencyByDottedItem(input);
            // CONFIG counters
            Assert.AreEqual(Flag.CONFIG_FLAG, calculated["system.system_model"]);
            Assert.AreEqual(Flag.CONFIG_FLAG, calculated["system.ontap_version"]);
            Assert.AreEqual(Flag.CONFIG_FLAG, calculated["system.serial_no"]);
            Assert.AreEqual(Flag.CONFIG_FLAG, calculated["system.system_id"]);
            Assert.AreEqual(Flag.CONFIG_FLAG, calculated["system.hostname"]);
            // Freq 1 counters
            Assert.AreEqual(Flag.CONFIG_FLAG, calculated["system.nfs_ops"]);
            Assert.AreEqual(Flag.FREQUENCY_1_FLAG, calculated["system.cifs_ops"]);
            Assert.AreEqual(Flag.FREQUENCY_2_FLAG, calculated["system.http_ops"]);
            Assert.AreEqual(Flag.FREQUENCY_3_FLAG, calculated["system.fcp_ops"]);
            Assert.AreEqual(Flag.FREQUENCY_4_FLAG, calculated["system.iscsi_ops"]);
            // Freq 2 counters
            Assert.AreEqual(Flag.CONFIG_FLAG, calculated["volume.avg_latency"]);
            Assert.AreEqual(Flag.FREQUENCY_2_FLAG, calculated["volume.total_ops"]);
            Assert.AreEqual(Flag.FREQUENCY_2_FLAG, calculated["volume.read_data"]);
            Assert.AreEqual(Flag.FREQUENCY_3_FLAG, calculated["volume.read_ops"]);
            Assert.AreEqual(Flag.FREQUENCY_4_FLAG, calculated["volume.write_data"]);
            // Freq 4 counters
            Assert.AreEqual(Flag.CONFIG_FLAG, calculated["system.sys_read_latency_hist"]);
            Assert.AreEqual(Flag.FREQUENCY_4_FLAG, calculated["system.sys_write_latency_hist"]);
            Assert.AreEqual(Flag.FREQUENCY_4_FLAG, calculated["system.sys_latency_hist"]);
        }

        [Test]
        public void canCalcNeededFrequencyByOidWithSingleFrequency() {

            Dictionary<string, long> dottedItemsC = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems1 = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems2 = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems3 = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems4 = new Dictionary<string, long>();

            dottedItemsC["system.test1"] = Flag.CONFIG_FLAG;
            dottedItems1["system.test1"] = Flag.FREQUENCY_1_FLAG;
            dottedItems2["system.test1"] = Flag.FREQUENCY_2_FLAG;
            dottedItems3["system.test1"] = Flag.FREQUENCY_3_FLAG;
            dottedItems4["system.test1"] = Flag.FREQUENCY_4_FLAG;

            Dictionary<int, long> expectedC = new Dictionary<int, long>();
            Dictionary<int, long> expected1 = new Dictionary<int, long>();
            Dictionary<int, long> expected2 = new Dictionary<int, long>();
            Dictionary<int, long> expected3 = new Dictionary<int, long>();
            Dictionary<int, long> expected4 = new Dictionary<int, long>();

            expectedC[0] = Flag.CONFIG_FLAG;
            expected1[0] = Flag.FREQUENCY_1_FLAG;
            expected2[0] = Flag.FREQUENCY_2_FLAG;
            expected3[0] = Flag.FREQUENCY_3_FLAG;
            expected4[0] = Flag.FREQUENCY_4_FLAG;

            Assert.AreEqual(expectedC, ap.calcNeededFrequencyByOid(dottedItemsC));
            Assert.AreEqual(expected1, ap.calcNeededFrequencyByOid(dottedItems1));
            Assert.AreEqual(expected2, ap.calcNeededFrequencyByOid(dottedItems2));
            Assert.AreEqual(expected3, ap.calcNeededFrequencyByOid(dottedItems3));
            Assert.AreEqual(expected4, ap.calcNeededFrequencyByOid(dottedItems4));
        }

        [Test]
        public void canCalcNeededFrequencyByOidWithMultipleFrequencies() {

            Dictionary<string, long> dottedItems1 = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems2 = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems3 = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems4 = new Dictionary<string, long>();

            dottedItems1["system.testc"] = Flag.CONFIG_FLAG;
            dottedItems1["system.test1"] = Flag.FREQUENCY_1_FLAG;

            dottedItems2["system.testc"] = Flag.CONFIG_FLAG;
            dottedItems2["system.test1"] = Flag.FREQUENCY_1_FLAG;
            dottedItems2["system.test2"] = Flag.FREQUENCY_2_FLAG;

            dottedItems3["system.test1"] = Flag.FREQUENCY_1_FLAG;
            dottedItems3["system.test3"] = Flag.FREQUENCY_3_FLAG;
            dottedItems3["system.test4"] = Flag.FREQUENCY_4_FLAG;

            dottedItems4["system.testc"] = Flag.CONFIG_FLAG;
            dottedItems4["system.test1"] = Flag.FREQUENCY_2_FLAG;
            dottedItems4["system.test4"] = Flag.FREQUENCY_4_FLAG;

            Dictionary<int, long> expected1 = new Dictionary<int, long>();
            Dictionary<int, long> expected2 = new Dictionary<int, long>();
            Dictionary<int, long> expected3 = new Dictionary<int, long>();
            Dictionary<int, long> expected4 = new Dictionary<int, long>();

            expected1[0] = Flag.CONFIG_FLAG + Flag.FREQUENCY_1_FLAG;
            expected2[0] = Flag.CONFIG_FLAG + Flag.FREQUENCY_1_FLAG + Flag.FREQUENCY_2_FLAG;
            expected3[0] = Flag.FREQUENCY_1_FLAG + Flag.FREQUENCY_3_FLAG + Flag.FREQUENCY_4_FLAG;
            expected4[0] = Flag.CONFIG_FLAG + Flag.FREQUENCY_2_FLAG + Flag.FREQUENCY_4_FLAG;

            Assert.AreEqual(expected1, ap.calcNeededFrequencyByOid(dottedItems1));
            Assert.AreEqual(expected2, ap.calcNeededFrequencyByOid(dottedItems2));
            Assert.AreEqual(expected3, ap.calcNeededFrequencyByOid(dottedItems3));
            Assert.AreEqual(expected4, ap.calcNeededFrequencyByOid(dottedItems4));
        }

        [Test]
        public void canCalcNeededFrequencyByOidWithMultipleFrequenciesAndOids() {

            Dictionary<string, long> dottedItems1 = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems2 = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems3 = new Dictionary<string, long>();
            Dictionary<string, long> dottedItems4 = new Dictionary<string, long>();

            dottedItems1["system.testc"] = Flag.CONFIG_FLAG;
            dottedItems1["system.test1"] = Flag.FREQUENCY_1_FLAG;
            dottedItems1["volume.test3"] = Flag.CONFIG_FLAG;
            dottedItems1["volume.test4"] = Flag.FREQUENCY_1_FLAG;

            dottedItems2["system.testc"] = Flag.CONFIG_FLAG;
            dottedItems2["system.test1"] = Flag.FREQUENCY_1_FLAG;
            dottedItems2["system.test2"] = Flag.FREQUENCY_2_FLAG;
            dottedItems2["volume.test5"] = Flag.CONFIG_FLAG;
            dottedItems2["volume.test6"] = Flag.FREQUENCY_2_FLAG;

            dottedItems3["system.test1"] = Flag.FREQUENCY_1_FLAG;
            dottedItems3["system.test3"] = Flag.FREQUENCY_3_FLAG;
            dottedItems3["system.test4"] = Flag.FREQUENCY_4_FLAG;
            dottedItems3["volume.test5"] = Flag.FREQUENCY_4_FLAG;

            dottedItems4["system.testc"] = Flag.CONFIG_FLAG;
            dottedItems4["system.test1"] = Flag.FREQUENCY_2_FLAG;
            dottedItems4["system.test4"] = Flag.FREQUENCY_4_FLAG;
            dottedItems4["volume.test7"] = Flag.FREQUENCY_2_FLAG;
            dottedItems4["volume.test8"] = Flag.FREQUENCY_4_FLAG;

            Dictionary<int, long> expected1 = new Dictionary<int, long>();
            Dictionary<int, long> expected2 = new Dictionary<int, long>();
            Dictionary<int, long> expected3 = new Dictionary<int, long>();
            Dictionary<int, long> expected4 = new Dictionary<int, long>();

            expected1[0] = Flag.CONFIG_FLAG + Flag.FREQUENCY_1_FLAG;
            expected1[10] = Flag.CONFIG_FLAG + Flag.FREQUENCY_1_FLAG;

            expected2[0] = Flag.CONFIG_FLAG + Flag.FREQUENCY_1_FLAG + Flag.FREQUENCY_2_FLAG;
            expected2[10] = Flag.CONFIG_FLAG + Flag.FREQUENCY_2_FLAG;

            expected3[0] = Flag.FREQUENCY_1_FLAG + Flag.FREQUENCY_3_FLAG + Flag.FREQUENCY_4_FLAG;
            expected3[10] = Flag.FREQUENCY_4_FLAG;

            expected4[0] = Flag.CONFIG_FLAG + Flag.FREQUENCY_2_FLAG + Flag.FREQUENCY_4_FLAG;
            expected4[10] = Flag.FREQUENCY_2_FLAG + Flag.FREQUENCY_4_FLAG;

            Assert.AreEqual(expected1, ap.calcNeededFrequencyByOid(dottedItems1));
            Assert.AreEqual(expected2, ap.calcNeededFrequencyByOid(dottedItems2));
            Assert.AreEqual(expected3, ap.calcNeededFrequencyByOid(dottedItems3));
            Assert.AreEqual(expected4, ap.calcNeededFrequencyByOid(dottedItems4));
        }

        [Test]
        public void canGetArchiveUnixTimestampAsLong() {
            Assert.AreEqual(1447684431, ap.getArchiveUnixTimestamp());
        }

        [Test]
        public void canGetArchiveMsecsSinceBoot() {
            Assert.AreEqual(45251595752, ap.getArchiveMsecsSinceBoot());
        }

        [Test]
        public void canGetTimestampFromMsecsSinceBoot() {
            Assert.AreEqual(1447684431000, ap.getTimestampFromMsecsSinceBoot(45251595752));
            Assert.AreEqual(1447684431005, ap.getTimestampFromMsecsSinceBoot(45251595757));
            Assert.AreEqual(1447684436575, ap.getTimestampFromMsecsSinceBoot(45251601327));
        }

        [Test]
        public void canGet73ArchiveParser() {
            ap = new ArchiveParserFactory(
                "",
                new MessengerStub(),
                gzipSource,
                testDir,
                "trt",
                false,
                false)
                .Build("7.3.x");

            Assert.IsInstanceOf(typeof(ArchiveParser_7_3), ap);
        }

        [Test]
        [Category("Unit")]
        public void canGet81ArchiveParser() {
            ZapiArchiveFileSource gzipSource_8_1 
                =  new ZapiArchiveFileSource(zc) {
                Filename = "stats_archive_data_1468925789_20160719_105629.gz",
                Path = OdctSettingsTest.GetRelativeUnitTestDataDirectory("cmode81")
            };
            ap = new ArchiveParserFactory(
                "",
                new MessengerStub(),
                gzipSource_8_1,
                testDir,
                "trt",
                false,
                false)
                .Build("8.1.x");
            Assert.IsInstanceOf(typeof(ArchiveParser_8_1), ap);
        }

        [Test]
        [Category("Unit")]
        public void canGet82ArchiveParser() {
            ZapiArchiveFileSource gzipSource_8_2
                = new ZapiArchiveFileSource(zc) {
                    Filename = "stats_archive_data_1463486502_20160517_060142.gz",
                    Path = OdctSettingsTest.GetRelativeUnitTestDataDirectory("7mode82")
                };
            ap = new ArchiveParserFactory(
                "",
                new MessengerStub(),
                gzipSource_8_2,
                testDir,
                "trt",
                false,
                false)
                .Build("8.2.x");

            Assert.IsInstanceOf(typeof(ArchiveParser_8_2), ap);
        }
    }
}
