﻿using NUnit.Framework;
using System;
using System.IO;
using odct.Counters;

namespace odct.tests.Counters {
    [TestFixture]
    class CounterDefinitionCollectorTests {
        CounterDefinitionCollector ap;
        string xmlContent;

        [TestFixtureSetUp]
        public void init() {
            var metaXmlPath = Path.Combine(
                OdctSettingsTest.GetRelativeUnitTestDataDirectory("ontap73"),
                "stats_archive_data_1447684431_20151116_093351.xml");

            var work = TestContext.CurrentContext.WorkDirectory;
            xmlContent = File.ReadAllText(metaXmlPath);
            ap = new CounterDefinitionCollector(xmlContent);
        }

        [Test, Description("Get counter data for a given object and array of counternames")]
        [Category("Unit")]
        public void canGetCounterDetail() {
            var metaDataXmlDoc = ap.getMetaDataXml(xmlContent);
            var counterInfo = ap.getCounterInformationFromObjectNode(metaDataXmlDoc, "system", 0, "nfs_ops");
            Assert.AreEqual(360447, counterInfo.flag);
            Assert.AreEqual(131592, counterInfo.ctype);
            Assert.AreEqual("nfs_ops", counterInfo.counterName);
            Assert.AreEqual("system", counterInfo.objectName);
            Assert.AreEqual(5, counterInfo.id);
            Assert.AreEqual(8, counterInfo.size);
            Assert.AreEqual(0, counterInfo.base_id);
            Assert.AreEqual(null, counterInfo.x_name);
        }

        [Test, Description("Get all the counters for a given object if the second argument is null or not passed")]
        [Category("Unit")]
        public void canGetCountersByObjectName() {
            var counterInfo = ap.getCounterDefinitionByCounterId("system");

            Assert.AreEqual(24, counterInfo.Keys.Count);
        }

        [Test, Description("There can be counters without base counter, for those the basename should be null")]
        [Category("Unit")]
        public void canGetCounterWithoutBaseDictionary() {
            //string[]  = new string[] { "nfs_ops" };
            var counterInfo = ap.getCounterDefinitionByCounterId("system");

            Assert.AreEqual(true, counterInfo.ContainsKey(5));
        }

        [Test, Description("There can be counters without base counter, for those the basename should have value")]
        [Category("Unit")]
        public void canGetCounterWithBaseDictionary() {
            //string[]  = new string[] { "cpu_busy" };
            var counterInfo = ap.getCounterDefinitionByCounterId("system");
            Assert.LessOrEqual(2, counterInfo.Keys.Count);
        }

        [Test, Description("The return is a dictionary for each of the countes along with their base counters")]
        [Category("Unit")]
        public void canGetMultipleCountersDictionary() {
            //string[]  = new string[] { "nfs_ops", "cpu_busy" };
            var counterInfo = ap.getCounterDefinitionByCounterId("system");
            Assert.LessOrEqual(3, counterInfo.Keys.Count);
        }

        [Test, Description("The return is a dictionary for each of the countes")]
        [Category("Unit")]
        public void canGetMultipleCounterswithDuplicatesDictionary() {
            //string[]  = new string[] { "nfs_ops", "cpu_busy", "cpu_elapsed_time" };
            var counterInfo = ap.getCounterDefinitionByCounterId("system");
            Assert.LessOrEqual(3, counterInfo.Keys.Count);
        }

        [Test, Description("The return should be null for invalid xml")]
        [Category("Unit")]
        public void canhandleInvalidXml() {
            string xml = "<xml><tag></xml>";
            var ap1 = new CounterDefinitionCollector(xml);
            //var  = new string[] { "nfs_ops", "cpu_busy", "cpu_elapsed_time" };
            var counterInfo = ap1.getCounterDefinitionByCounterId("system_1");
            Assert.IsNull(counterInfo);
        }

        [Test, Description("Null is returned for invalid object names")]
        [Category("Unit")]
        public void canHandleInvalidObject() {
            //var want = new string[] { "nfs_ops", "cpu_busy", "cpu_elapsed_time" };
            var counterInfo = ap.getCounterDefinitionByCounterId("system_1");
            Assert.IsNull(counterInfo);
        }

        [Test, Description("No counters anymore. All are returned.")]
        [Category("Unit")]
        public void canHandleInvalidCounter() {
            //string[] want = new string[] { "nfs111" };
            var counterInfo = ap.getCounterDefinitionByCounterId("system");
            Assert.Less(0, counterInfo.Keys.Count);
        }

        [Test, Description("Valid and Invalid counetrs are treated as they are")]
        [Category("Unit")]
        public void canHandleMixOfValidInvalidCounter() {
            //string[] want = new string[] { "nfs_ops", "nfs111", "cpu_busy" };
            var counterInfo = ap.getCounterDefinitionByCounterId("system");
            Assert.LessOrEqual(3, counterInfo.Keys.Count);
        }

        [Test, Description("Can return all counters for object")]
        [Category("Unit")]
        public void cangetCounterListsByObjectName() {
            var metaDataXmlDoc = ap.getMetaDataXml(xmlContent);
            var counters = ap.getCountersByObjectName(metaDataXmlDoc, "system");
            Assert.AreEqual(24, counters.Count);
        }

        [Test, Description("Xnames should be populated for the array type counters")]
        [Category("Unit")]
        public void canGetXnamesForCountersWithXNames() {
            //string[] want = new string[] { "sys_latency_hist" };
            var counterInfo = ap.getCounterDefinitionByCounterId("system");
            Assert.LessOrEqual(1, counterInfo.Keys.Count);
            Assert.AreEqual(true, counterInfo.ContainsKey(18));
            Assert.IsInstanceOf(typeof(Array), counterInfo[18].x_name);
            Assert.AreEqual(36, counterInfo[18].x_name.Length);
        }

        [Test, Description("Xnames should be null for the non-array type counters")]
        [Category("Unit")]
        public void canGetXnamesForCountersWoutXnames() {
            //string[] want = new string[] { "net_data_recv" };
            var counterInfo = ap.getCounterDefinitionByCounterId("system");
            Assert.LessOrEqual(1, counterInfo.Keys.Count);
            Assert.AreEqual(true, counterInfo.ContainsKey(19));
            Assert.IsNull(counterInfo[19].x_name);
        }
    }
}
