﻿namespace odct.tests {
    using System;
    using System.IO;
    using NUnit.Framework;
    using odct.Settings;

    [TestFixture]
    class OdctSettingsTest {
        public static string GetRelativeUnitTestDataDirectory(string subdirectory = null) {
            string result = Path.Combine(TestContext.CurrentContext.TestDirectory, "testdata");
            if (!string.IsNullOrEmpty(subdirectory)) {
                result = Path.Combine(result, subdirectory);
            }
            return result;
        }

        public static void SetupUnitTestOdctSettings() {
            if (!OdctSettings.IsInit()) { // workaround no NUnit 2 OneTimeSetUp
                var settingsFile = Path.Combine(
                    GetRelativeUnitTestDataDirectory(),
                    "odct-settings.xml");

                OdctSettings.InitializeInstance(settingsFile);
            }
        }

        public static Credentials GetFirstCredential() {
            var credentials = OdctSettings.Instance.GetHostConnections();
            var credIterator = credentials.GetEnumerator();
            credIterator.MoveNext();
            return credIterator.Current;
        }

        [Test]
        [Category("Unit")]
        public void TestCounterList() {
            var c = OdctSettings.Instance.GetCounterDefinitions();
            Assert.That(c.ContainsKey("system.cifs_ops"));
            Assert.AreEqual(c["system.cifs_ops"], odct.Settings.OdctSettings.DefaultCounterFrequency);
        }

        [Test]
        [Category("Unit")]
        public void TestSynergyCounterList() {
            var c = OdctSettings.GetCounterDefinitions("synergy", "");
            Assert.That(!c.ContainsKey("system.cifs_ops"));
        }

        [Test]
        [Category("Unit")]
        public void TestSpecialCounterList() {
            var c = OdctSettings.GetCounterDefinitions("trt", "special1,special2:2,special3:4");
            Assert.That(c.ContainsKey("system.cifs_ops"));
            Assert.AreEqual(c["system.cifs_ops"], odct.Settings.OdctSettings.DefaultCounterFrequency);
            Assert.Contains("special1", c.Keys);
            Assert.AreEqual(4, c["special1"]);
            Assert.Contains("special2", c.Keys);
            Assert.AreEqual(2, c["special2"]);
            Assert.Contains("special3", c.Keys);
            Assert.AreEqual(4, c["special1"]);
        }

        [Test]
        [Category("Unit")]
        public void TestNoAppSpecialCounterList() {
            var c = OdctSettings.GetCounterDefinitions("", "special1,special2:2,special3:4");
            Assert.That(!c.ContainsKey("system.cifs_ops"));
            Assert.Contains("special1", c.Keys);
            Assert.AreEqual(4, c["special1"]);
            Assert.Contains("special2", c.Keys);
            Assert.AreEqual(2, c["special2"]);
            Assert.Contains("special3", c.Keys);
            Assert.AreEqual(4, c["special1"]);
        }

        [Test]
        [Category("Unit")]
        public void TestMinSpecialCounterList5() {
            var c = OdctSettings.GetCounterDefinitions("trt", "system.cifs_ops:5");
            Assert.That(c.ContainsKey("system.cifs_ops"));
            Assert.AreEqual(c["system.cifs_ops"], odct.Settings.OdctSettings.DefaultCounterFrequency);
        }

        [Test]
        [Category("Unit")]
        public void TestMinSpecialCounterListVsConfig() {
            Assert.AreEqual(0, odct.Settings.OdctSettings.ConfigCounterFrequency);
            var c = OdctSettings.GetCounterDefinitions("trt", "system.cifs_ops:0"); // ie 0 is config
            Assert.That(c.ContainsKey("system.cifs_ops"));
            Assert.AreEqual(c["system.cifs_ops"], odct.Settings.OdctSettings.DefaultCounterFrequency);
        }

        [Test]
        [Category("Unit")]
        public void TestMinSpecialCounterList1() {
            var c = OdctSettings.GetCounterDefinitions("trt", "system.cifs_ops:1");
            Assert.That(c.ContainsKey("system.cifs_ops"));
            Assert.AreEqual(c["system.cifs_ops"], 1);
        }

        [Test]
        [Category("Unit")]
        public void TestTimeUtil() {
            var tomorrow = new DateTime(2016, 2, 12);
            var now = TimeUtil.ConvertFromUnixTime(1455213228);
            Assert.That(TimeUtil.TimeSlewWithinRange(now, tomorrow, TimeSpan.FromDays(3)));
            Assert.That(!TimeUtil.TimeSlewWithinRange(now, tomorrow, TimeSpan.FromHours(6)));

        }
    }
}
