package Archive::Parser;

use Data::Dump qw{dump};
use POSIX;
use XML::Simple;
$XML::Simple::PREFERRED_PARSER = "XML::Parser";
use Digest::MD5 qw{md5};
use strict;

my $verbose = undef;

# Used to record the location of records within a data file for the next
# time it is read
my %dataMap;
#use CMPG::ExceptionHandler
#    qw/fatal output logit dlogit set_concurrent_output unset_concurrent_output/;

use base 'Exporter';
our @EXPORT = qw{W_SKIPPED_COUNTERS};

############################################################
# Error values
use constant W_SKIPPED_COUNTERS	=>	0x40000001;
############################################################
# These hashes are used to convert to cm-hourly-stats format
my %properties = 
(
	"CM_CNT_STRING" => "string",	"CM_CNT_AVERAGE"	=> "average",
	"CM_CNT_DELTA"	=> "delta",		"CM_CNT_NODISP"		=> "no-display",
	"CM_CNT_PERCT"	=> "percent",	"CM_CNT_RATE"		=> "rate",
	"CM_CNT_RAW"	=> "raw"
);

my %units =
(
	"CM_B_PER_SEC"	=> "b_per_sec",	"CM_KB_PER_SEC"	=> "kb_per_sec",
	"CM_MB_PER_SEC"	=> "mb_per_sec","CM_PER_SEC"	=> "per_sec",
	"CM_SECS"		=> "second",	"CM_USECS"		=> "microsec",
	"CM_MSECS"		=> "millisec",	"CM_PERCT"		=> "percent",
	"CM_NO_SUFFIX"	=> "none"
);
############################################################
# constructor
sub spawn
{
	my ( $class, $filename ) = @_;
	my $self = { 'handle' => undef };
	$self->{'delete_inflate'} = 0;
	bless $self;
	return $self->_initialise( $filename );
}
############################################################
sub type
{
	return 'archive';
}
############################################################
# initialisation
sub _initialise
{
	my ( $self, $filename ) = @_;
		
	if ( -e $filename )
	{
		$self->{'archive'} = $filename;
		$self->{'hash'} = md5( stat $filename, $filename );
		open $self->{'handle'}, $self->{'archive'} or 
			die "EIOERROR", __LINE__, $!;
	}
	else 
	{ 
		die "EINVALIDPARAM", __LINE__;
	}	
	
	$/ = "<\/archive>\n";	
	$self->{'xml_header'} = readline $self->{'handle'}
		or die "EIOERROR", __LINE__, $!;
	$self->{'header_size'} = tell $self->{'handle'};	
	
	eval
	{
		$self->{'header'} = XMLin( $self->{'xml_header'}, 
								NormaliseSpace => 2, 			# Remove whitespace
								KeyAttr => {"name"}, 			# name is a key attribute
								ForceArray => ["counter"] );	# Force counters into arrays
	};
	
	if ( $@ )
	{
		$self->{'error'} = "$@";
		die "EBADXML", __LINE__, "", $@;
	}
	
	$self->{'gmtOffset'} = unixTimestamp( $self->{'header'}{'ontap'}{'localtime'} ) - 
		unixTimestamp( $self->{'header'}{'ontap'}{'gmtime'} );

	die 'EINVALIDDATA', __LINE__, "Invalid GMT offset" 
		if 	$self->{'gmtOffset'} >= 86_400 or $self->{'gmtOffset'} <= -86_400;
		
	for ( @{$self->{'header'}{'attribute_def'}} )
	{
		if ( $_->{'name'} eq "ctype" )
		{
			$self->{'ctypes'}{$_->{'description'}} = $_->{'value'};
		}
		elsif ( $_->{'name'} eq "flag" )
		{
			$self->{'flags'}{$_->{'description'}}{'value'} = $_->{'value'};
			$self->{'flags'}{$_->{'description'}}{'frequency'} = $_->{'frequency'};
		}
	}

	return $self;
}
############################################################
sub allowDataAverage
{
	return 1;
}
############################################################
sub allowInterpolate
{
	return 1;
}
############################################################
# destructor
sub DESTROY
{
	my ( $self ) = @_;
	close $self->{'handle'} if tell $self->{'handle'} != -1;
	if ($self->{'delete_inflate'} == 1) {
 	    unlink $self->{'archive'};
	}
	undef $self;
}
############################################################
# Fill a hash with the frequency names and their values
sub frequencies
{
	my ( $self, $hash ) = @_;
	
	for ( keys %{$self->{'flags'}} )
	{
		$hash->{$_} = $self->{'flags'}{$_}{'frequency'}
	}
	
	return $hash;
}
############################################################
# Get the timestamp
sub timestamp
{
	my ( $self ) = @_;
	return $self->{'header'}{'unix_timestamp'};
}
############################################################
# Set the date format to unix timestamp or excel format
sub dateFormat
{
	my ( $self, $format ) = @_;
	return unless ( $format =~ /^(?:excel|unix)$/i );
	$self->{'dateFormat'} = lc( $format );
}
############################################################
# set a list of counters that we want collected as it can 
# speed up processing if we skip what we don't need
# They take the format type:counter:sub-counter
sub filterCounters
{
	my ( $self, $objects ) = @_;
	$self->{'filter'} = $objects;
	return 1;	
}
############################################################
# Get or set the file
sub file
{
	my ( $self, $filename ) = @_;
	
	if ( lc( $filename ) ne lc( $self->{'archive'} ) )
	{
		close $self->{'handle'} if fileno $self->{'handle'};
		delete $self->{'header'};
		delete $self->{'xml_header'};
		delete $self->{'header_size'};
		delete $self->{'error'};
		delete $self->{'flags'};
		delete $self->{'ctypes'};
		$self->_initialise( $filename );
	}
	return $self->{'archive'};
}
############################################################
# Dump out the XML format header
sub getHeader
{
	my $self = shift;
	return $self->{'xml_header'};
}

############################################################
# Dump out the performance meta-data in the same format as 
# cm-hourly-stats files
sub perfInfo
{
	my $self = shift;
	my $XML = $self->{'header'};
	my $perfInfo = "<perf-info>";
	
	for my $object ( @{$$XML{'preset'}{'object'}} )
	{
		$perfInfo .= "<object-info><objectname>" . $object->{'name'} . "</objectname>"; 
		$perfInfo .= "<counters>";
		
		for my $counter ( @{$object->{'counter'}} )
		{
			$perfInfo .= "<counter-info>";
			$perfInfo .= "<name>" . $counter->{'name'} . "</name>";
			$perfInfo .= "<desc>" . $counter->{'description'} . "</desc>";
			$perfInfo .= "<privilege-level>basic</privilege-level>";
			
			my $props;
			for my $prop ( keys %properties )
			{
				my $p = $self->hasProperty( $counter->{'ctype'}, $prop );
				$props .= $properties{$prop}."," if $p;
			}
			$props =~ s/,$//;
			$perfInfo .= "<properties>$props</properties>";
			
			my $unit = $self->getUnit( $counter->{'ctype'} );
			$perfInfo .= "<unit>" . $units{$unit} . "</unit>" if $unit;
			
			if ( $counter->{'base_id'} > 0 )
			{
				my $baseCounter = $self->getCounter( $object->{'id'}, $counter->{'base_id'} );
				$perfInfo .= "<base-counter>" . $baseCounter->{'name'} . "</base-counter>" if $baseCounter;
			}
			
			if ( defined $counter->{'x_name'} )
			{
				my $labels;
				for ( @{$counter->{'x_name'}} )
				{
					$labels .= "$_,";
				}
				$labels =~ s/,$//g;
				$labels =~ s/</&lt;/g;
				$labels =~ s/>/&gt;/g;
				$perfInfo .= "<type>array</type><labels><label-info>$labels</label-info></labels>";
			}
			
			# Not sure why but with this code we never get true for the following if statement for
			# the first counter, which happens to be system_model.
			# Random, odd and unexplainable
			my $dummy = sprintf "%#x", $counter->{'flag'} if $counter->{'name'} eq "system_model";
			for my $f ( keys %{$self->{'flags'}} )
			{
				if ( ( $counter->{'flag'} & $self->{'flags'}{$f}{'value'} ) == $self->{'flags'}{$f}{'value'} )
				{
					# We never get here for system_model without the sprintf above.
					$perfInfo .= "<frequency-type>" . $f . "</frequency-type>";
					$perfInfo .= "<frequency>" . $self->{'flags'}{$f}{'frequency'} . "</frequency>";
					
				}
			}
			
			$perfInfo .= "</counter-info>";
		}		
		$perfInfo .= "</counters></object-info>";
	}
	$perfInfo .= "</perf-info>";
	
	return $perfInfo;
}
############################################################
# Parse the binary portion of the data file
sub parse
{
	my ( $self, $frequency, $perfHash ) = @_;

	unless ( $frequency or ref $frequency ne "" )
	{
		die "EINVALIDPARAM", __LINE__;
	}
	
	my $result;
	seek $self->{'handle'}, 0, 0 or die "EIOERROR", __LINE__, $!;
		
	# Check if the value passed is valid
	if ( $frequency =~ /^\w*$/ )
	{
		$frequency = uc( $frequency );
		unless ( defined $self->{'flags'}{$frequency} )
		{
			dump $frequency, $self->{'flags'};
			die "EINVALIDPARAM", __LINE__, "$frequency is not valid";
		}
		else
		{
			$frequency = $self->{'flags'}{$frequency}{'value'};
		}
	}
	elsif ( $frequency =~ /^\d*$/ )
	{
		my $valid;
		grep { $valid++ if $self->{'flags'}{$_}{'value'} eq $frequency } keys %{$self->{'flags'}};
		unless ( $valid )
		{
			die "EINVALIDPARAM", __LINE__;
		}
	}
	# Go to the start of the binary data if we aren't there already
	seek $self->{'handle'}, $self->{'header_size'}, 0;

	# switch to binary mode
	# Windows uses two characters to represent a new line. If perl assumes a text file
	# then these two characters are wrapped into a single \n character and we lose data
	binmode $self->{'handle'};
	
	my $instance_name_size = $self->{'header'}{'instance_name'}{'size'};
	
	#logit("Version".$self->{'header'}{'version'});
	#$verbose = 2;
	# read instance headers in
	while( !eof $self->{'handle'} )
	{
		my $start = tell $self->{'handle'};
		my $sample_header;			
		read $self->{'handle'}, $sample_header, $self->{'header'}{'instance_header_def'}{'size'};
		#my ( $msecs_lo, $msecs_hi, $size, $oid, $num_instances, $flags ) = unpack( "L6", $sample_header );
		my ( $msecs, $size, $oid, $num_instances, $flags );
		
		if ( $ENV{PROCESSOR_ARCHITECTURE} ne 'x86' )
		{  
			( $msecs, $size, $oid, $num_instances, $flags ) = unpack( "QL4", $sample_header );
		}
		else
		{
			# Perl can handle 53-bit numbers, but no proper 64-bit support on 32-bit platforms
			# This means unpredictable results if uptime exceeds 142,000 years =)
			my ( $msecs_lo, $msecs_hi );
			( $msecs_lo, $msecs_hi, $size, $oid, $num_instances, $flags ) = unpack( "L6", $sample_header );
			$msecs = $msecs_lo + $msecs_hi * ( 1 + ~0 );
		}
		
		if($size == "")
		{
			next;
		}
		# If the requested frequency is not in this sample then skip to the next one
		# also skip if this object is not in the filter list
		#unless ( ( defined $self->{'filter'}{$obj->{'name'}} ) and ( $flags & $frequency ) )

		unless ( ( $flags & $frequency ) == $frequency )
		{
			seek $self->{'handle'}, $size + $start, 0;
			next;
		}

		# Locate the object definition for this object's ID
		my $obj = $self->getObject( $oid ) 
			or die "EINVALIDDATA", __LINE__, "Bad object ID";

		my $timestamp = $self->{'header'}{'unix_timestamp'} + 
						( ( $msecs - $self->{'header'}{'msecs_since_boot'} ) / 1000 ) +
						$self->{'gmtOffset'}; # Put timestamp into local time 
							
		if ( $self->{'dateFormat'} eq "excel" )
		{
			# Excel format is days since 01-01-1900
			# Need to convert secs since 01-01-1970 to days and
			# then add the days between 1900 and 1970
			$timestamp = ( $timestamp / 86_400 ) + 25_569;
		}


		for ( my $i = 0; $i < $num_instances; $i++ )
		{	
			my $binary;
			my $instance;	
			if ($obj->{'counter'}->[0]->{name} ne "instance_name") {
				read $self->{'handle'}, $binary, $instance_name_size;
				$instance = unpack( "Z$instance_name_size", $binary );
			}
			
			#logit("Size::[$instance_name_size]");
			#read $self->{'handle'}, $binary, $instance_name_size;
			#$instance = unpack( "Z$instance_name_size", $binary );
			
			
			#dump( $instance );# if $verbose == 2;
												
			for ( @{$obj->{'counter'}} )
			{
				if ($_->{name} eq "instance_name") {
					my $size = $_->{size};
					read $self->{'handle'}, $binary, $size;
					$instance = unpack( "Z$size", $binary );
					next;
				}
				# Check if this counter's frequency flag says it is in this record
				#unless ( ($_->{'flag'} & $frequency ) == $frequency )
				unless ( $_->{'flag'} & $flags )
				{
					#my $newpos = int( tell $self->{'handle'} ) + $_->{'size'};
					#seek $self->{'handle'}, $newpos, 0;

					next;
				}
				
				# Has a filter been set?
				if ( defined $self->{'filter'} )
				{
					# Do we want to pick up this counter?
					unless ( 	defined $self->{'filter'}{$obj->{'name'}}{$_->{'name'}} or 
								defined $self->{'filter'}{$obj->{'name'}}{"*"} )
					{
						my $newpos = int( tell $self->{'handle'} ) + $_->{'size'};
						seek $self->{'handle'}, $newpos, 0;
						next;
					}
				}
								
				# if x_name is present then we have array data
				my $isArray = defined $_->{'x_name'};
				my @data = undef;
								
				# Not sure why but there is some strange behaviour unless this code 
				# is executed.
				$self->{'ctypes'}{'CM_CNT_4B'} 	== $self->{'ctypes'}{'CM_CNT_4B'} ;				
				$self->{'ctypes'}{'CM_CNT_8B'} 	== $self->{'ctypes'}{'CM_CNT_8B'} ;				
				$self->{'ctypes'}{'STRING'} 	== $self->{'ctypes'}{'STRING'} ;				
				
				if ( ( $_->{'ctype'} & $self->{'ctypes'}{'CM_CNT_STRING'} ) == $self->{'ctypes'}{'CM_CNT_STRING'} )
				{
					read $self->{'handle'}, $binary, $_->{'size'};
					@data = unpack( "Z" . $_->{'size'}, $binary );
				}
				elsif ( ( $_->{'ctype'} & $self->{'ctypes'}{'CM_CNT_8B'} ) == $self->{'ctypes'}{'CM_CNT_8B'} )
				{
					read $self->{'handle'}, $binary, $_->{'size'};
									
					if ( $ENV{PROCESSOR_ARCHITECTURE} eq 'x86' )
					{
						@data = unpack( "(LL)".($_->{'size'}/8), $binary );
	
						my @new_data;
						for ( my $i = 0; $i < @data; $i ++ )
						{
							my ( $lo, $hi ) = ( $data[$i++], ( defined $data[$i] ? $data[$i] : undef ) );
							if ( $hi )
							{
								push @new_data, $lo + ( $hi * ( 1 + ~0 ) );
							}
							else
							{
								push @new_data, $lo;
							}
						}
						@data = @new_data;
					} 
					else 
					{
						@data = unpack( "Q".($_->{'size'}/8), $binary );
					}
					#	dump( $_->{'name'}, @data ) if $verbose > 1;

				}
				
				elsif ( ( $_->{'ctype'} & $self->{'ctypes'}{'CM_CNT_4B'} ) == $self->{'ctypes'}{'CM_CNT_4B'} )
				{
					read $self->{'handle'}, $binary, $_->{'size'};
					@data = unpack( "L".($_->{'size'}/4), $binary );
				}
				else
				{
				#	if ( $verbose )
				#	{
				#		dump( "skipped", $_ );
				#	}
				
					$result = W_SKIPPED_COUNTERS;
					seek $self->{'handle'}, $_->{'size'}, 1;
				}
								
				if ( @data )
				{
					if ( $isArray )
					{
						my $d;
						map { $d .= "$_," } @data;
						$d =~ s/,$//;
						$$perfHash{$obj->{'name'}}{$timestamp}{$instance}{$_->{'name'}} = $d;					
					}
					else
					{
						my $d = $data[0];
						# if we have a 64-bit number then we'll have two items
						$d = $data[0] + $data[1] * ( 1 + ~0 ) if ( scalar @data == 2 );
						$$perfHash{$obj->{'name'}}{$timestamp}{$instance}{$_->{'name'}} = $d;
					}
				}
			}
		}
		
		seek $self->{'handle'}, $start + $size, 0 if tell $self->{'handle'} != $start + $size;
	}
	
	return $result || 1;
}
############################################################
# Return the last error recorded
sub getError
{
	my $self = shift;
	return $self->{'error'};
}
############################################################
# Internal functions
############################################################
# Get an object given it's id or name
sub getObject
{
	my ( $self, $oid ) = @_;
	for ( @{$self->{'header'}{'preset'}{'object'}} )
	{
		return $_ if ( $_->{'id'} == $oid );
		return $_ if ( $_->{'name'} eq $oid );
	}
	return undef;
}
############################################################
# Get a counter given the object's id and counter id
sub getCounter
{
	my ( $self, $oid, $id ) = @_;
	my $obj = $self->getObject( $oid );
	for ( @{$obj->{'counter'}} )
	{
		return $_ if ( $_->{'id'} == $id );	
	}
	return undef;
}
############################################################
# Does this ctype contain this text description of a property
sub hasProperty
{
	my ( $self, $ctype, $desc ) = ( shift, int( shift ), shift );
	return $self->{'ctypes'}{$desc} if ( $self->{'ctypes'}{$desc} & $ctype ) == $self->{'ctypes'}{$desc};
	return undef;
}
############################################################
# Get the unit type from the ctype
sub getUnit
{
	my ( $self, $ctype ) = ( shift, int( shift ) );
	$ctype = $ctype & 0xf0000;
	for ( keys %{$self->{'ctypes'}} )
	{
		next if $_ =~ /_CNT_/;
		return $_ if ( $self->{'ctypes'}{$_} == $ctype );
	}
	return undef;
}
############################################################
# Get unix timestamp from string time
sub unixTimestamp
{
	my $string = shift;
	die 'EINVALIDPARAM', __LINE__, "Timestamp is not defined" unless $string;
	my ( $wday, $month, $mday, $time, $tz, $year ) = split /\s+/, $string;
	my ( $hour, $min, $sec ) = split /:/, $time;
	
	if ( $month =~ /\w{3}/ )
	{
		if ( $month =~ /jan/i ) 	{$month = 0;}
		elsif ( $month =~ /feb/i )	{$month = 1;}
		elsif ( $month =~ /mar/i )	{$month = 2;}
		elsif ( $month =~ /apr/i )	{$month = 3;}
		elsif ( $month =~ /may/i )	{$month = 4;}
		elsif ( $month =~ /jun/i )	{$month = 5;}
		elsif ( $month =~ /jul/i )	{$month = 6;}
		elsif ( $month =~ /aug/i )	{$month = 7;}
		elsif ( $month =~ /sep/i )	{$month = 8;}
		elsif ( $month =~ /oct/i )	{$month = 9;}
		elsif ( $month =~ /nov/i )	{$month = 10;}
		elsif ( $month =~ /dec/i )	{$month = 11;}
		die 'EINVALIDPARAM', __LINE__, "Unable to parse month '$month'" 
			if $month =~ /\w{3}/;
	}
	
	if ( $wday =~ /\w{3}/ )
	{
		if ( $wday =~ /mon/i ) 		{$wday = 0;}
		elsif ( $wday =~ /tue/i )	{$wday = 1;}
		elsif ( $wday =~ /wed/i )	{$wday = 2;}
		elsif ( $wday =~ /thu/i )	{$wday = 3;}
		elsif ( $wday =~ /fri/i )	{$wday = 4;}
		elsif ( $wday =~ /sat/i )	{$wday = 5;}
		elsif ( $wday =~ /sun/i )	{$wday = 6;}
		die 'EINVALIDPARAM', __LINE__, "Unable to parse weekday '$wday'" 
			if $wday =~ /\w{3}/;
	}
	
	unless ( defined $year )
	{
		$year = ( localtime( time() ) )[5] + 1900;
	}
	
	my $timestamp = mktime( $sec, $min, $hour, $mday, $month, $year - 1900, $wday );
	die 'EINVALIDPARAM', __LINE__, "Unable to make timestamp from '$string'" unless $timestamp;
		
	return $timestamp;
}

1;