﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using NetApp.Manage;
using odct.FilerConfig;

namespace odct.Clusters {
    public class Cluster {

        public string clusterContact { get; private set; }
        public string clusterLocation { get; private set; }
        public string clusterName { get; private set; }
        public string clusterUuid { get; private set; }
        public string clusterSerialNumber { get; private set; }
        public string clusterRdbUuid { get; private set; }
        public string clusterMgmtIp { get; private set; }
        public string clusterHomeNode { get; private set; }

        public int ontapGeneration { get; private set; }
        public int ontapMajor { get; private set; }
        public int ontapMinor { get; private set; }
        public string ontapMajorVersion { get; private set; }
        public string ontapFullVersion { get; private set; }

        public Credentials credentials { get; private set; }
        public NaServer naServer { get; private set; }
        public List<string> nodeNames { get; private set; }
        public Dictionary<string, ClusterNode> nodeNamesToNodes { get; private set; }

        public Messenger messenger { get; set; }
        private ClusterConfiguration clusterConfig;
        public string payloadPath { get; set; }
        public string clusterPath { get; set; }
        public string clusterConfigPath { get; set; }

        public Cluster() {

        }

        public Cluster(Credentials c, Messenger m, string payloadpath) {
            this.messenger = m;
            this.messenger.sendDebug("");
            this.messenger.sendDebug("Cluster Constructor started.");
            this.payloadPath = payloadpath;
            this.credentials = c;
            this.naServer = c.naServer;
            getClusterIdentityInfo();
            getClusterOntapVersion();
            getNodes();
            this.messenger.sendDebug(string.Format("Cluster {0} Constructor ended.", this.clusterName));
        }

        public void getClusterIdentityInfo() {
            this.messenger.sendDebug(string.Format("Cluster {0} getClusterIdentityInfo() started.", this.clusterName));

            NaElement api = new NaElement("cluster-identity-get");
            NaElement xo = naServer.InvokeElem(api);

            NaElement attributes = xo.GetChildByName("attributes");
            NaElement clusterIdentityInfo = attributes.GetChildByName("cluster-identity-info");
            this.clusterContact = clusterIdentityInfo.GetChildContent("cluster-contact");
            this.clusterLocation = clusterIdentityInfo.GetChildContent("cluster-location");
            this.clusterName = clusterIdentityInfo.GetChildContent("cluster-name");
            this.clusterSerialNumber = clusterIdentityInfo.GetChildContent("cluster-serial-number");
            this.clusterUuid = clusterIdentityInfo.GetChildContent("cluster-uuid");
            this.clusterRdbUuid = clusterIdentityInfo.GetChildContent("rdb-uuid");

            this.clusterPath = Path.Combine(this.payloadPath, this.clusterUuid);

            this.messenger.sendDebug(string.Format("Cluster {0} getClusterIdentityInfo() returning.", this.clusterName));
            this.messenger.sendDebug(string.Format("... cluster contact is {0}", this.clusterContact));
            this.messenger.sendDebug(string.Format("... cluster location is {0}", this.clusterLocation));
            this.messenger.sendDebug(string.Format("... cluster name is {0}", this.clusterName));
            this.messenger.sendDebug(string.Format("... cluster serial number is {0}", this.clusterSerialNumber));
            this.messenger.sendDebug(string.Format("... cluster uuid is {0}", this.clusterUuid));
            this.messenger.sendDebug(string.Format("... cluster rdb uuid is {0}", this.clusterRdbUuid));
            this.messenger.sendDebug(string.Format("... cluster path is {0}", this.clusterPath));
        }

        public void getClusterOntapVersion() {
            this.messenger.sendDebug(string.Format("Cluster {0} getClusterOntapVersion() started.", this.clusterName));

            NaElement api = new NaElement("system-get-version");
            NaElement xo = naServer.InvokeElem(api);

            NaElement versionTuple = xo.GetChildByName("version-tuple");
            NaElement systemVersionTuple = versionTuple.GetChildByName("system-version-tuple");
            this.ontapGeneration = systemVersionTuple.GetChildIntValue("generation", 0);
            this.ontapMajor = systemVersionTuple.GetChildIntValue("major", 0);
            this.ontapMinor = systemVersionTuple.GetChildIntValue("minor", 0);
            this.ontapMajorVersion = string.Format("{0}.{1}", this.ontapGeneration, this.ontapMajor);
            this.ontapFullVersion = string.Format("{0}.{1}.{2}", this.ontapGeneration, this.ontapMajor, this.ontapMinor);
            this.messenger.sendDebug(string.Format("Cluster {0} getClusterOntapVersion() returning.", this.clusterName));
            this.messenger.sendDebug(string.Format("... cluster ontap generation is {0}", this.ontapGeneration));
            this.messenger.sendDebug(string.Format("... cluster ontap major is {0}", this.ontapMajor));
            this.messenger.sendDebug(string.Format("... cluster ontap minor is {0}", this.ontapMinor));
            this.messenger.sendDebug(string.Format("... cluster ontap Major version is {0}", this.ontapMajorVersion));
        }

        public void getNodes() {
            this.messenger.sendDebug(string.Format("Cluster {0} getNodes() started.", this.clusterName));

            nodeNamesToNodes = new Dictionary<string, ClusterNode>();

            NaElement api = new NaElement("net-interface-get-iter");
            api.AddNewChild("max-records", "2000");

            NaElement xo = naServer.InvokeElem(api);

            NaElement attributesListElement = xo.GetChildByName("attributes-list");
            List<NaElement> netInterfaceInfoList = attributesListElement.GetChildren();

            foreach (NaElement netInterfaceInfo in netInterfaceInfoList) {
                string role = netInterfaceInfo.GetChildContent("role");
                if (role.Equals("node_mgmt")) {
                    string address = netInterfaceInfo.GetChildContent("address");
                    string operationalStatus = netInterfaceInfo.GetChildContent("operational-status");
                    string homeNode = netInterfaceInfo.GetChildContent("home-node");
                    ClusterNode clusterNode = new ClusterNode(this.messenger, this, homeNode, this.credentials, this.clusterPath);
                    clusterNode.nodeMgmtIp = address;
                    nodeNamesToNodes.Add(clusterNode.nodeName, clusterNode);
                }
                if (role.Equals("cluster_mgmt")) {
                    string address = netInterfaceInfo.GetChildContent("address");
                    string operationalStatus = netInterfaceInfo.GetChildContent("operational-status");
                    string homeNode = netInterfaceInfo.GetChildContent("home-node");
                    this.clusterMgmtIp = address;
                    this.clusterHomeNode = homeNode;
                }
            }
            this.messenger.sendDebug(string.Format("Cluster {0} getNodes() returned.", this.clusterName));
        }

        public CollectionResults CollectConfig(Collector collector) {

            CollectionResults cr = new CollectionResults();
            cr.count = 1;
            cr.servers.Add(this.credentials.Server);
            cr.collected = true;
            cr.configCollected = true;

            try
            {
                this.messenger.sendInfo(string.Format("Cluster {0} CollectConfig() started.", this.clusterName));
                this.CollectClusterConfiguration();
                this.SaveClusterConfig();

                foreach (ClusterNode clusterNode in this.nodeNamesToNodes.Values)
                {
                    if (clusterNode.isNodeHealthy)
                    {
                        clusterNode.CollectAndSaveConfig();
                    }
                }
            }
            catch (OdctClusterCounterCollectionFailedException e)
            {
                this.messenger.sendDebug("Cluster.CollectConfig() caught OdctClusterCounterCollectionFailedException during configuration collection.");
                cr.collected = false;
                cr.configCollected = false;
                cr.description = e.Message;
            }
            this.messenger.sendInfo(string.Format("Cluster {0} CollectConfig() returned.", this.clusterName));
            collector.clusterByClusterUuid[this.clusterUuid] = this;
            return cr;
        }

        public void CollectClusterConfiguration() {
            this.messenger.sendDebug(string.Format("Cluster {0} CollectClusterConfiguration() started.", this.clusterName));
            this.clusterConfig = new ClusterConfiguration(this);
            this.clusterConfig.Build(this.credentials, this.messenger);
            this.messenger.sendDebug(string.Format("Cluster {0} CollectClusterConfiguration() returned.", this.clusterName));
        }

        public void SaveClusterConfig() {
            this.messenger.sendDebug(string.Format("Cluster {0} SaveClusterConfig() started.", this.clusterName));
            this.clusterConfigPath = Path.Combine(this.clusterPath, "configuration");
            Directory.CreateDirectory(this.clusterConfigPath);

            foreach (var s in clusterConfig.Sections) {
                var xo = s.GetNaElement();
                if (xo != null) {
                    var xoPrettyString = xo.ToPrettyString("");
                    int len = 500;
                    this.messenger.sendDebug(string.Format("****** {0}.ToPrettyString first {1} characters ******", s.GetSectionTitle(), len));
                    this.messenger.sendDebug(xoPrettyString.Substring(0, Math.Min(len, xoPrettyString.Length - 1)));
                    this.messenger.sendDebug(string.Format("------ end {0}.ToPrettyString sample ------", s.GetSectionTitle()));
                    var filePath = Path.Combine(
                        clusterConfigPath,
                        string.Format("{0}.xml", s.GetSectionTitle()));

                    File.WriteAllText(filePath, xoPrettyString);
                }
            }
            this.messenger.sendDebug(string.Format("Cluster {0} SaveClusterConfig() returned.", this.clusterName));
        }


    }
}
