﻿using NetApp.Manage;
using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Text.RegularExpressions;
using System.Threading;
using odct.Counters;
using odct.Settings;

namespace odct.Clusters {
    public class ClusterNode {

        public Credentials credentials { get; set; }
        public string nodeName { get; set; }
        public string nodeMgmtIp { get; set; }
        public bool isNodeHealthy { get; set; }
        public string FilerSystemId { get; private set; }
        public string FilerSerialNumber { get; private set; }
        public string PartnerName { get; private set; }
        public string systemModel { get; private set; }

        public Messenger messenger { get; private set; }
        public int processingErrorCount = 0;
        public bool collectionFailed = false;
        public bool performanceCollectionFailed = false;

        private NaServer naServer;
        public Cluster cluster;

        private string clusterPath;
        private string nodePath;
        private string nodeConfigPath;
        private string nodeCountersPath;
        private ClusterNodeConfiguration clusterNodeConfig;
        private bool haveCopiedCcmaMetaToPayload = false;
        private static readonly object objLock = new object();
        private static readonly object progressLock = new object();

        public struct ParseParameters {
            public string objectName;
            public int numObjects;
            public string countersString;
            public string numCounters;
        }

        public ClusterNode() {

        }

        public ClusterNode(Messenger messenger, Cluster containingCluster, string nodeName, Credentials c, string clusterpath) {

            this.messenger = messenger;
            this.messenger.sendDebug("");
            this.messenger.sendDebug(string.Format("ClusterNode {0} Constructor() started.", nodeName));
            this.cluster = containingCluster;
            this.clusterPath = clusterpath;
            this.credentials = c;
            this.nodeName = nodeName;
            this.naServer = c.naServer;
            this.processingErrorCount = 0;
            this.isNodeHealthy = true;

            getNodeInfo();
            if (this.isNodeHealthy) {
                this.nodePath = Path.Combine(clusterPath, this.FilerSerialNumber);
                this.nodeCountersPath = Path.Combine(this.nodePath, "counters");
            }
            this.messenger.sendDebug(string.Format("ClusterNode {0} Constructor() ending.", this.nodeName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ...containingCluster name is {1}", this.nodeName, this.cluster.clusterName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ...clusterPath is {1}", this.nodeName, this.clusterPath));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ...nodePath is {1}", this.nodeName, this.nodePath));

        }

        public void getNodeInfo() {
            this.messenger.sendDebug(string.Format("ClusterNode {0} getNodeInfo() started.", this.nodeName));

            // get node info

            NaElement api = new NaElement("system-node-get");
            api.AddNewChild("node", nodeName);

            NaElement xo = naServer.InvokeElem(api);

            NaElement attributes = xo.GetChildByName("attributes");
            NaElement nodeDetailsInfoElement = attributes.GetChildByName("node-details-info");
            this.isNodeHealthy = "true".Equals(nodeDetailsInfoElement.GetChildContent("is-node-healthy")) ? true : false;
            if (!this.isNodeHealthy) { 
                this.collectionFailed = true;
            }
            this.FilerSystemId = nodeDetailsInfoElement.GetChildContent("node-system-id");
            this.FilerSerialNumber = nodeDetailsInfoElement.GetChildContent("node-serial-number");
            this.systemModel = nodeDetailsInfoElement.GetChildContent("node-model");

            // alternate get node info (seems to be required for ontap 8.3.2...)

            if ((this.isNodeHealthy) &&
                (this.FilerSystemId == null || this.FilerSerialNumber == null || this.systemModel == null)) {

                api = new NaElement("system-node-get-iter");
                api.AddNewChild("max-records", "999");

                NaElement queryElement = new NaElement("query");
                api.AddChildElement(queryElement);
                
                NaElement systemInfoElement = new NaElement("system-info");
                queryElement.AddChildElement(systemInfoElement);

                systemInfoElement.AddNewChild("system-name", nodeName);

                xo = credentials.naServer.InvokeElem(api);

                NaElement attributesList = xo.GetChildByName("attributes-list");
                NaElement systemInfo = attributesList.GetChildByName("system-info");
                string altFilerSystemId = systemInfo.GetChildContent("system-id");
                string altSystemSerialNumber = systemInfo.GetChildContent("system-serial-number");
                string altSystemModel = systemInfo.GetChildContent("system-model");

                if (this.FilerSystemId == null) {
                    this.FilerSystemId = altFilerSystemId;
                }

                if (this.FilerSerialNumber == null) {
                    this.FilerSerialNumber = altSystemSerialNumber;
                }

                if (this.systemModel == null) {
                    this.systemModel = altSystemModel;
                }

            }

            // get partner info

            api = new NaElement("cf-get-partner");
            api.AddNewChild("node", nodeName);

            xo = credentials.naServer.InvokeElem(api);

            this.PartnerName = xo.GetChildContent("partner");

            this.messenger.sendDebug(string.Format("ClusterNode {0} getNodeInfo() returning.", this.nodeName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ...FilerSystemId is {1}.", this.nodeName, this.FilerSystemId));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ...FilerSerialNumber is {1}.", this.nodeName, this.FilerSerialNumber));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ...PartnerName is {1}.", this.nodeName, this.PartnerName));
        }

        public void parseArchiveFiles() {
            this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveFiles() started.", this.nodeName));
            List<string> filebaseList = getArchiveBaseListWithRetry();

            int filesProcessed = 0;
            int maxFilesToProcess = OdctSettings.Instance.GetPerformanceFilesCollectionLimit();

            if (filebaseList != null) {

                foreach (string filebase in filebaseList) {
                    if (filesProcessed >= maxFilesToProcess)
                        break;
                    try {
                        string fileccma = copyCcmaFilesToLocal(filebase);
                        if (!haveCopiedCcmaMetaToPayload)
                            copyCcmaMetaToPayload(fileccma);
                        parseArchiveFile(fileccma, getLocalArchiveDirectory());
                        if (OdctSettings.Instance.GetDeleteCcmaCopiesAfterProcessing()) {
                            deleteCcmaFilesFromLocal(fileccma);
                        }
                        filesProcessed++;
                        this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveFiles() waiting for progressLock for ProgressIncrement.", this.nodeName));
                        lock (progressLock) {
                            Messenger.updateObject updateObject = Messenger.ProgressIncrement(1, this.nodeName);
                            this.messenger.update(updateObject);
                        }
                        this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveFiles() done with progressLock for ProgressIncrement.", this.nodeName));
                        if (OdctSettings.Instance.GetDeleteCcmaCopiesAfterProcessing()) {
                            Directory.Delete(getLocalArchiveDirectory());
                        }
                    } catch (OdctCmaplFailureException) {
                        // These will occasionally be encountered; retries have been added to reduce 
                        // the impact of these.  This would be the place to provide more tracking and 
                        // reporting of these errors, but they should probably not cause the collection
                        // to fail.
                        this.processingErrorCount++;
                    } catch (OdctSpiDownloadFailed) {
                        // Similar to above, these sometimes happen; retries have been added to reduce 
                        // the impact of these.  This would be the place to provide more tracking and 
                        // reporting of these errors, but they should probably not cause the collection
                        // to fail.
                        this.processingErrorCount++;
                    } catch (Exception e) {
                        this.processingErrorCount++;
                        messenger.sendError(string.Format("ClusterNode {0} parseArchiveFiles() encountered Exception.", this.nodeName));
                        messenger.sendError(string.Format("... on filebase: {0}", filebase));
                        messenger.sendError(e.Message);
                        messenger.sendError(e.StackTrace);
                        messenger.sendError(e.InnerException.StackTrace);
                        throw new OdctClusterCounterCollectionFailedException("Unexpected error during counter collection", e);
                    }
                }
            }
            this.messenger.sendDebug(string.Format("ClusterNode {0} ...parseArchiveFiles() processed {1} files.", this.nodeName, filesProcessed));
        }

        public void parse81ArchiveFiles() {
            this.messenger.sendDebug(string.Format("ClusterNode {0} parse81ArchiveFiles() started.", this.nodeName));
            int filesProcessed = 0;
            int maxFilesToProcess = OdctSettings.Instance.GetPerformanceFilesCollectionLimit();
            ArchiveFileSource currentSource = null;

            Cluster81SourceGenerator gen = new Cluster81SourceGenerator(this);
            try {

                foreach (Cluster81ArchiveFileSource source in gen.GetSources()) {

                    currentSource = source;

                    if (filesProcessed >= maxFilesToProcess)
                        break;

                    const string appCountersRequested = "trt";
                    const bool isClusterMode = true;

                    ArchiveParser ap = new ArchiveParserFactory(
                        this.FilerSerialNumber,
                        this.messenger,
                        source,
                        this.nodePath,
                        appCountersRequested,
                        !haveCopiedCcmaMetaToPayload,
                        isClusterMode)  
                        .Build("8.1");

                    ap.processRecords();

                    filesProcessed++;
                    haveCopiedCcmaMetaToPayload = true;

                    this.messenger.sendDebug(string.Format("ClusterNode {0} parse81ArchiveFiles() waiting for progressLock for ProgressIncrement.", this.nodeName));
                    lock (progressLock) {
                        Messenger.updateObject updateObject = Messenger.ProgressIncrement(1, this.nodeName);
                        this.messenger.update(updateObject);
                    }
                    this.messenger.sendDebug(string.Format("ClusterNode {0} parse81ArchiveFiles() done with progressLock for ProgressIncrement.", this.nodeName));

                    if (OdctSettings.Instance.GetDeleteCcmaCopiesAfterProcessing()) {
                        source.deleteLocalFile();
                    }
                }
            } catch (Exception e) {
                this.processingErrorCount++;
                messenger.sendError(string.Format("ClusterNode {0} parse81ArchiveFiles() encountered Exception.", this.nodeName));
                string currentFileName = string.IsNullOrEmpty(currentSource.Filename) ? "currentSource is null or empty" : currentSource.Filename;
                messenger.sendError(string.Format("... on filename: {0}", currentFileName));
                messenger.sendError(e.ToString());
            }

            this.messenger.sendDebug(string.Format("ClusterNode {0} ...parse81ArchiveFiles() processed {1} files.", this.nodeName, filesProcessed));
        }

        private void deleteCcmaFilesFromLocal(string filepath) {
            this.messenger.sendDebug(string.Format("ClusterNode {0} deleteCcmaFilesFromLocal() started.", this.nodeName));
            if ("8.1".Equals(this.cluster.ontapMajorVersion)) {
                string fullfilepath = Path.Combine(getLocalArchiveDirectory(), filepath);
                File.Delete(fullfilepath);
            } else {
                string metafilename = filepath + ".meta";
                string metafilepath = Path.Combine(getLocalArchiveDirectory(), metafilename);
                File.Delete(metafilepath);

                string gzfilename = filepath + ".gz";
                string gzfilepath = Path.Combine(getLocalArchiveDirectory(), gzfilename);
                File.Delete(gzfilepath);
            }
            this.messenger.sendDebug(string.Format("ClusterNode {0} deleteCcmaFilesFromLocal() returned.", this.nodeName));
        }

        private void copyCcmaMetaToPayload(string ccmafilename) {
            this.messenger.sendDebug(string.Format("ClusterNode {0} copyCcmaFilesToLocal() started.", this.nodeName));
            if ("8.1".Equals(this.cluster.ontapMajorVersion)) {
                ClusterMetaFile cmf = new ClusterMetaFile(ccmafilename, "8.1");
                string metadata = cmf.metaXml;
                string[] filebase = ccmafilename.Split('.');
                string destinationFilePath = Path.Combine(getDiagnosticsPath(), filebase[0]);
                File.WriteAllText(destinationFilePath, metadata);
                this.haveCopiedCcmaMetaToPayload = true;
            } else {
                string filename = ccmafilename + ".meta";
                string sourceFilePath = Path.Combine(this.getLocalArchiveDirectory(), filename);
                string destinationFilePath = Path.Combine(getDiagnosticsPath(), filename);
                File.Copy(sourceFilePath, destinationFilePath, true);
                this.haveCopiedCcmaMetaToPayload = true;
            }
            this.messenger.sendDebug(string.Format("ClusterNode {0} copyCcmaFilesToLocal() returned.", this.nodeName));
        }

        public void parseArchiveFile(string ccmafilename, string localpath) {

            this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveFile() started.", this.nodeName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... ccmafilename is {1}", this.nodeName, ccmafilename));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... localpath is {1}", this.nodeName, localpath));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... no frequency specified; all system and volume counters will be collected.", this.nodeName));

            string ccmafilepath = Path.Combine(localpath, ccmafilename);

            ParseParameters parseParameters = this.getAllCountersForOneObject(ccmafilepath, "system", new int[] { 3600000 } );
            parseArchiveAndSaveResult(ccmafilename, localpath, 3600000, "system", "hourly", parseParameters);

            parseParameters = this.getAllCountersForOneObject(ccmafilepath, "volume", new int[] { 3600000 });
            parseArchiveAndSaveResult(ccmafilename, localpath, 3600000, "volume", "hourly", parseParameters);

            parseParameters = this.getAllCountersForOneObject(ccmafilepath, "system", new int[] { 1000, 5000, 60000, 300000 });
            parseArchiveAndSaveResult(ccmafilename, localpath, 300000, "system", "other", parseParameters);

            if (this.cluster.ontapMajorVersion.Equals("8.1") 
                || this.cluster.ontapMajorVersion.Equals("8.2")) {
                parseParameters = this.getAllCountersForOneObject(ccmafilepath, "volume", new int[] { 5000, 60000, 300000 });
                parseArchiveAndSaveResult(ccmafilename, localpath, 60000, "volume", "other", parseParameters);
            } else {
                parseParameters = this.getAllCountersForOneObject(ccmafilepath, "volume", new int[] { 5000, 60000, 300000 });
                parseArchiveAndSaveResult(ccmafilename, localpath, 300000, "volume", "other", parseParameters);
            }
            if ("8.2".Equals(this.cluster.ontapMajorVersion) || "8.3".Equals(this.cluster.ontapMajorVersion)) {
                parseParameters = this.getSpecificCounters(this.cluster.ontapMajorVersion, "cifs", 3600000);
                parseArchiveAndSaveResult(ccmafilename, localpath, 3600000, "cifs", "hourly", parseParameters);
            }

            this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveFile() returned.", this.nodeName));
            return;
        }

        public void parseArchiveAndSaveResult(string ccmafilename, string localpath, Int32 instanceFrequency, string objName, string category, ParseParameters parseParameters ) {
            this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveAndSaveResult() started.", this.nodeName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... ccmafilename is {1}", this.nodeName, ccmafilename));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... localpath is {1}", this.nodeName, localpath));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... instanceFrequency is {1}", this.nodeName, instanceFrequency));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... objName is {1}", this.nodeName, objName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... category is {1}", this.nodeName, category));

            // counterFrequency specifies which counters to gather based on the frequency defined for the counter 
            //   in the .ccma.meta file
            // instanceFrequency specifies which instances to gather based on their frequencies 
            //   in the .ccma.gz binary

            string counterfilename = getCounterFileName(ccmafilename, objName, category);
            string counterfilepath = Path.GetDirectoryName(counterfilename);
            this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveAndSaveResult() waiting for objLock.", this.nodeName));
            lock (objLock) {
                if (!Directory.Exists(counterfilepath)) {
                    Directory.CreateDirectory(counterfilepath);
                }
            }
            this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveAndSaveResult() done with objLock.", this.nodeName));
            parseArchiveFile(ccmafilename, counterfilename, localpath, instanceFrequency, parseParameters);
            this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveAndSaveResult() started.", this.nodeName));
            return;
        }

        public ParseParameters getAllCountersForOneObject(string ccmafilepath, string objName, Int32[] counterFrequency) {
            ParseParameters result = new ParseParameters() {
                objectName = objName,
                numObjects = 1
            };

            string caller = string.Format("{0}", this.nodeName);
            result.countersString = CmaplWrapper.GetCounterNames(ccmafilepath, objName, counterFrequency, this.messenger, caller);
            string[] work = result.countersString.Split(new char[] { ',' });
            result.numCounters = string.Format("{0}", work.Length);

            return result;
        }

        public ParseParameters getSpecificCounters(string ontapMajorVersion, string objName, Int32 frequency) {
            ParseParameters result = new ParseParameters() {
                objectName = objName,
                numObjects = 1
            };

            if ((frequency == 3600000)
                && "cifs".Equals(objName)
                && (("8.2".Equals(ontapMajorVersion) || "8.3".Equals(ontapMajorVersion)))) {
                result.countersString = "cifs_read_ops,cifs_write_ops,cifs_read_latency_hist,cifs_write_latency_hist";
                result.numCounters = "4";
            }
            return result;
        }

        public void parseArchiveFile(string ccmafilename, string targetfilename, string localpath, Int32 instanceFrequency, ParseParameters parseParameters) {
            this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveFile() started.", this.nodeName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... ccmafilename is {1}", this.nodeName, ccmafilename));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... localpath is {1}", this.nodeName, localpath));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... instanceFrequency is {1}", this.nodeName, instanceFrequency));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... parseParameters.objectName is {1}", this.nodeName, parseParameters.objectName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... parseParameters.numObjects is {1}", this.nodeName, parseParameters.numObjects));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... parseParameters.countersString is {1}", this.nodeName, parseParameters.countersString));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... parseParameters.numCounters is {1}", this.nodeName, parseParameters.numCounters));

            string ccmaGzFilePath = Path.Combine(localpath, ccmafilename);

            string objectName = parseParameters.objectName;
            int numObjects = parseParameters.numObjects;
            string countersString = parseParameters.countersString;
            string numCounters = parseParameters.numCounters;

            this.messenger.sendDebug(string.Format("ClusterNode {0}... calling CmaplWrapper.ParseFile()", this.nodeName));
            CmaplWrapper.GetCsv(ccmaGzFilePath, targetfilename, objectName, numObjects, countersString, numCounters, instanceFrequency, this.messenger, string.Format("ClusterNode {0}", this.nodeName));

            this.messenger.sendDebug(string.Format("ClusterNode {0} parseArchiveFile() returned.", this.nodeName));
            return;
        }

        public string getCounterFileName(string filename, string objName, string periodsCollected) {
            this.messenger.sendDebug(string.Format("{0} getCounterFileName() started.", this.nodeName));
            Regex regex = new Regex("\\d{6}_\\d{6}_(?<unixDate>\\d{13})_\\d{6,7}.ccma");
            Match m = regex.Match(filename);
            Int64 unixDateInMsecs = Int64.Parse(m.Groups["unixDate"].Value);
            Int64 unixDateInSecs = unixDateInMsecs / 1000;
            DateTime dt = TimeUtil.ConvertFromUnixTime((Int32)(unixDateInSecs));

            string work = Path.Combine(getCounterPath(), String.Format("{0:yyyy-MM-dd}", dt));
            work = Path.Combine(work, objName);
            work = Path.Combine(work, String.Format("hour_{0:HH_mm}_{1}.csv", dt, periodsCollected));

            this.messenger.sendDebug(string.Format("ClusterNode {0} getCounterFileName returned {1}", this.nodeName, work));
            return work;
        }

        public List<string> getArchiveBaseListWithRetry(
            string httpStringFormat = null, 
            string regexStringFormat = null) 
        {
            List<string> result = new List<string>();
            const int retryCount = 3;
            int currentRetry = 0;

            for ( ; ; ) {
                try {
                    result = getArchiveBaseList(httpStringFormat, regexStringFormat);
                    break;
                } catch (Exception e) {
                    this.messenger.sendError(string.Format(
                        "ClusterNode {0} getArchiveBaseListWithRetry() encountered error {1} on retry {2}", 
                        this.nodeName, e.Message, currentRetry));
                    this.messenger.sendError(string.Format("ClusterNode{0} Exception.ToString(): ",
                        this.nodeName, e.ToString()));
                    currentRetry++;
                    if (currentRetry > retryCount || !isTransientExeption(e)) {
                        throw new OdctClusterCounterCollectionFailedException("getArchiveBaseList Failed.", e);
                    }                    
                }
                Thread.Sleep(50);
            }
            return result;
        }

        public bool isTransientExeption(Exception e) {
            if (e is System.AccessViolationException) return true;
            if (e is OdctSpiProbablyNotConfiguredException) return false;
            if (e is OdctSpiWebException) return true;
            if (e is OdctSpiOtherException) return false;
            return false;
        }

        public List<string> getArchiveBaseList(
            string httpStringFormat = null, 
            string regexStringFormat = null) 
        {
            this.messenger.sendDebug(string.Format("ClusterNode {0} getArchiveBaseList() started.", this.nodeName));

            string httpString= "http://{0}/spi/{1}/etc/log/stats/ccma/kernel/";
            if (!string.IsNullOrEmpty(httpStringFormat)) {
                httpString = httpStringFormat;
            }
            string regexString = "<A HREF=\"(?<filebase>\\d{6}_\\d{6}_\\d{13}_\\d{6,7}).ccma.meta\">";
            if (!string.IsNullOrEmpty(regexStringFormat)) {
                regexString = regexStringFormat;
            }

            string html;
            using (WebClient client = new WebClient()) {
                client.Credentials = new NetworkCredential(this.credentials.Username, this.credentials.Password);
                try {
                    string httpDownloadString = string.Format(httpString, this.nodeMgmtIp, this.nodeName);
                    html = client.DownloadString(httpDownloadString);

                    Regex regex = new Regex(regexString);
                    MatchCollection matches = regex.Matches(html);
                    List<string> filebases = new List<string>();

                    foreach (Match match in matches) {
                        string possibleMatch = match.Groups["filebase"].Value;
                        if (!"00000000".Equals(possibleMatch.Substring(14,8))) {
                            filebases.Add(possibleMatch);
                        }
                    }

                    filebases.Sort();

                    this.messenger.sendInfo(string.Format("ClusterNode {0} getArchiveBaseList() returned {1} filenames", this.nodeName, filebases.Count));
                    return filebases;

                } catch (WebException webException) {
                    if (webException.GetType() == typeof(HttpWebResponse)) {
                        HttpWebResponse httpWebResponse = (HttpWebResponse) webException.Response;
                        int httpResponseCode = (int) httpWebResponse.StatusCode;
                        if ((webException.Status == WebExceptionStatus.ProtocolError) &&
                            (httpResponseCode == 404)) {
                            string msg = string.Format("{0} Encountered '404 Not Found' when attempting spi interface for node {0}.", this.nodeName);
                            this.messenger.sendError(msg);
                            this.messenger.sendError(string.Format("{0} Filer may not be configured for spi web access.", this.nodeName));
                            this.messenger.sendError(string.Format("{0} Spi is required for odct", this.nodeName));
                            this.messenger.sendError(string.Format("{0} 'vserver services web' commands may be used to enable spi, if it is not now enabled.", this.nodeName));
                            this.messenger.sendError(string.Format("see https://library.netapp.com/ecmdocs/ECMP1196798/html/GUID-447FAEB0-CD24-4430-968E-6DC4B33DBE83.html", this.nodeName));
                            this.messenger.sendError(string.Format("ClusterNode {0} WebException.ToString(): {1}",
                                this.nodeName, webException.ToString()));
                            throw new OdctSpiProbablyNotConfiguredException(msg, webException);
                        } 
                    }
                    string msg2 = string.Format("{0} Unexpected WebException attempting spi connect to {0}.", this.nodeName);
                    this.messenger.sendError(msg2);
                    this.messenger.sendError(string.Format("{0} WebException.Status: {1}", this.nodeName, webException.Status));
                    throw new OdctSpiWebException(msg2); 
                } catch (Exception e) {
                    string msg3 = string.Format("{0} Unexpected Other Exception attempting spi connect to {0}.", this.nodeName);
                    this.messenger.sendError(msg3);
                    this.messenger.sendError(string.Format("{0} Exception.Message: {1}", this.nodeName, e.Message));
                    this.messenger.sendError(string.Format("ClusterNode {0} Exception.ToString(): {1}",
                        this.nodeName, e.ToString()));
                    throw new OdctSpiOtherException(msg3, e);
                }
            }

        }

        public string copyCcmaFilesToLocal(string filebase) {
            this.messenger.sendDebug(string.Format("ClusterNode {0} copyCcmaFilesToLocal() started.", this.nodeName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... filebase is {1}", this.nodeName, filebase));
            string localpath = getLocalArchiveDirectory();
            string spiPath = string.Format("http://{0}/spi/{1}/etc/log/stats/ccma/kernel",
                    this.nodeMgmtIp,
                    this.nodeName);
            string filename = string.Format("{0}.ccma", filebase);
            string metaFn = string.Format("{0}.ccma.meta", filebase);
            string gzFn = string.Format("{0}.ccma.gz", filebase);

            copySpiFileToLocal(Path.Combine(spiPath, metaFn), Path.Combine(localpath, metaFn));
            copySpiFileToLocal(Path.Combine(spiPath, gzFn), Path.Combine(localpath, gzFn));

            this.messenger.sendDebug(string.Format("ClusterNode {0} copyCcmaFilesToLocal() returned filename {1}", this.nodeName, filename));
            return filename;
        }

        public void copySpiFileToLocal(string sourcepath, string localpath) {

            this.messenger.sendDebug(string.Format("ClusterNode {0} copySpiFileToLocal() started.", this.nodeName));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... sourcepath is {1}", this.nodeName, sourcepath));
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... localpath is {1}", this.nodeName, localpath));

            using (WebClient client = new WebClient()) {
                client.Credentials = new NetworkCredential(this.credentials.Username, this.credentials.Password);
                DownloadFileWithRetry(client, sourcepath, localpath);
            }

            this.messenger.sendDebug(string.Format("ClusterNode {0} copySpiFileToLocal() returned.", this.nodeName));
        }

        public void DownloadFileWithRetry(WebClient c, string sourcepath, string destinationpath) {

            const int retryCount = 3;
            int currentRetry = 0;

            for (; ; ) {
                try {
                    c.DownloadFile(sourcepath, destinationpath);
                    break;
                } catch (Exception e) {
                    this.messenger.sendError(string.Format("ClusterNode {0} on retry {1} DownloadFilesWithRetry encountered error {2}", this.nodeName, currentRetry, e.ToString()));
                    currentRetry++;
                    if (currentRetry > retryCount || !isTransientDownloadException(e)) {
                        throw new OdctSpiDownloadFailed("Download Failed.", e);
                    }
                }
                Thread.Sleep(50);
            }
        }

        public bool isTransientDownloadException(Exception e) {
            if (e is WebException) return true;
            return false;
        }

        public string getLocalArchiveDirectory() {
            this.messenger.sendDebug(string.Format("ClusterNode {0} getLocalArchiveDirectory() started.", this.nodeName));
            TempFolder tempFolder = new TempFolder();
            string localpath = Path.Combine(tempFolder.PathName, nodeName);

            this.messenger.sendDebug(string.Format("ClusterNode {0} getLocalArchiveDirectory() waiting for objLock.", this.nodeName));
            lock (objLock) {
                if (!Directory.Exists(localpath)) {
                    Directory.CreateDirectory(localpath);
                }
            }
            this.messenger.sendDebug(string.Format("ClusterNode {0} getLocalArchiveDirectory() done with objLock.", this.nodeName));

            this.messenger.sendDebug(string.Format("ClusterNode {0} ... returned localpath {1}", this.nodeName, localpath));
            return localpath;
        }

        public string getCounterPath() {
            this.messenger.sendDebug(string.Format("ClusterNode {0} getCounterPath() started.", this.nodeName));
            if (!Directory.Exists(this.nodeCountersPath)) {
                Directory.CreateDirectory(this.nodeCountersPath);
            }
            this.messenger.sendDebug(string.Format("ClusterNode {0} ... returned counterPath {1}", this.nodeName, this.nodeCountersPath));
            return this.nodeCountersPath;
        }

        public string getDiagnosticsPath() {
            this.messenger.sendDebug(string.Format("ClusterNode {0} getDiagnosticsPath() started.", this.nodeName));
            string diagnosticsPath = Path.Combine(this.nodePath, "diagnostics");
            if (!Directory.Exists(diagnosticsPath)) {
                Directory.CreateDirectory(diagnosticsPath);
            }
            this.messenger.sendDebug(string.Format("CLusterNode {0} ... returned diagnosticsPath {1}", this.nodeName, diagnosticsPath));
            return diagnosticsPath;
        }

        public void CollectAndSaveConfig() {
            this.messenger.sendDebug("");
            this.messenger.sendDebug(string.Format("CLusterNode {0} CollectAndSaveConfig() started.", this.nodeName));
            CollectConfiguration();
            SaveConfig();
        }

        public void CollectConfiguration() {
            this.messenger.sendInfo(string.Format("ClusterNode {0} CollectConfiguration() started.", this.nodeName));
            this.clusterNodeConfig = new ClusterNodeConfiguration(this);
            this.clusterNodeConfig.Build(this.credentials, this.messenger);
        }

        public void SaveConfig() {
            this.messenger.sendInfo(string.Format("ClusterNode {0} SaveConfig() started.", this.nodeName));
            this.nodeConfigPath = Path.Combine(this.nodePath, "configuration");
            Directory.CreateDirectory(this.nodeConfigPath);

            foreach (var s in clusterNodeConfig.Sections) {
                var xo = s.GetNaElement();
                if (xo != null) {
                    var xoPrettyString = xo.ToPrettyString("");
                    int len = 500;
                    this.messenger.sendDebug(string.Format("{0} ****** {1}.ToPrettyString first {2} characters ******", this.nodeName, s.GetSectionTitle(), len));
                    this.messenger.sendDebug(string.Format("{0}\n{1}", this.nodeName, xoPrettyString.Substring(0, Math.Min(len, xoPrettyString.Length - 1))));
                    this.messenger.sendDebug(string.Format("{0} ------ end {1}.ToPrettyString sample ------", this.nodeName, s.GetSectionTitle()));

                    var filePath = Path.Combine(
                        nodeConfigPath,
                        string.Format("{0}.xml", s.GetSectionTitle()));

                    File.WriteAllText(filePath, xoPrettyString);
                }
            }
            this.messenger.sendDebug(string.Format("{0} SaveConfig() returned.", this.nodeName));
        }

        public void CollectPerformanceCounters() {
            this.messenger.sendDebug(string.Format("ClusterNode {0} CollectPerformanceCounters() started.", this.nodeName));
            try {
                if ("8.1".Equals(this.cluster.ontapMajorVersion)) {
                    parse81ArchiveFiles();
                } else {
                    parseArchiveFiles();
                }
            } catch (OdctClusterCounterCollectionFailedException) {
                this.performanceCollectionFailed = true;
            } catch (Exception e) {
                this.messenger.sendError(string.Format("ClusterNode {0} CollectPerformanceCounters() caught Exception with parseArchiveFiles().", this.nodeName));
                this.messenger.sendError(string.Format("e.Message: {0}", e.Message));
                this.messenger.sendError(string.Format("e.StackTrace: {0}", e.StackTrace));
            }
            this.messenger.sendDebug(string.Format("ClusterNode {0} CollectPerformanceCounters() returned.", this.nodeName));
        }
        
    }
}
