﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;

namespace odct.Clusters {
    public class CmaplWrapper {
        private const string cmaplFileName = "cmapl_api.dll";
        private const string iconvFileName = "iconv.dll";
        private const string lib2xmlFileName = "libxml2.dll";
        private const string pocrtFileName = "pocrt.dll";
        private const string zlibFileName = "zlib1.dll";

        private static List<EmbeddedUnmanagedDll> dlls = new List<EmbeddedUnmanagedDll>();

        unsafe public static void GetCsv(string inFilePath, string outFilePath, string objectName, int numObjects, string countersString, string numCounters, int frequency, Messenger messenger, string caller) {

            const int retryCount = 3;
            int currentRetry = 0;

            for (; ; ) {
                try {
                    get_csv(
                        inFilePath,
                        outFilePath,
                        objectName,
                        numObjects,
                        countersString,
                        numCounters,
                        frequency);
                    break;
                } catch (Exception e) {
                    messenger.sendError(string.Format("ClusterNode {0} retry {1} encountered error {2}", caller, currentRetry, e.ToString()));
                    currentRetry++;
                    if (currentRetry > retryCount || !isTransient(e)) {
                        if (e is System.AccessViolationException) {
                            messenger.sendError("System.AccessViolationException encountered during call to GetCsv");
                            messenger.sendError("with the following parameters:");
                            messenger.sendError(string.Format("{0}      inFilePath {1}", inFilePath, caller));
                            messenger.sendError(string.Format("{0}     outFilePath {1}", outFilePath, caller));
                            messenger.sendError(string.Format("{0}      objectName {1}", objectName, caller));
                            messenger.sendError(string.Format("{0}      numObjects {1}", numObjects, caller));
                            messenger.sendError(string.Format("{0}  countersString {1}", countersString, caller));
                            messenger.sendError(string.Format("{0}     numCounters {1}", numCounters, caller));
                            messenger.sendError(string.Format("{0}       frequency {1}", frequency, caller));
                            messenger.sendError(string.Format("{0}   e.Message is: {1}", e.Message, caller));
                            messenger.sendError(e.ToString());
                            throw new OdctCmaplFailureException("Encountered error during CmaplWrapper.GetCsv", e);
                        }

                        messenger.sendError("Other Exception encountered during call to GetCsv");
                        messenger.sendError("with the following parameters:");
                        messenger.sendError(string.Format("{0}      inFilePath {1}", inFilePath, caller));
                        messenger.sendError(string.Format("{0}     outFilePath {1}", outFilePath, caller));
                        messenger.sendError(string.Format("{0}      objectName {1}", objectName, caller));
                        messenger.sendError(string.Format("{0}      numObjects {1}", numObjects, caller));
                        messenger.sendError(string.Format("{0}  countersString {1}", countersString, caller));
                        messenger.sendError(string.Format("{0}     numCounters {1}", numCounters, caller));
                        messenger.sendError(string.Format("{0}       frequency {1}", frequency, caller));
                        messenger.sendError(string.Format("{0}   e.Message is: {1}", e.Message, caller));
                        messenger.sendError(e.ToString());
                        throw new OdctCmaplFailureException("Encountered error during CmaplWrapper.GetCsv", e);

                    }
                }
                Thread.Sleep(50);
            }

        }

        public static bool isTransient(Exception e) {
            if (e is System.AccessViolationException) return true;
            return false;
        }

        unsafe public static string GetCounterNames(string inFilePath, string objectName, int[] frequency, Messenger messenger, string caller) {

            return get_counters(
                inFilePath,
                objectName,
                frequency.Length,
                frequency);

        }

        public static void LoadDllsIfNecessary() {
            CmaplWrapper.LoadDllIfNecessary(Properties.Resources.ZLIB1_DLL, CmaplWrapper.zlibFileName);
            CmaplWrapper.LoadDllIfNecessary(Properties.Resources.ICONV_DLL, CmaplWrapper.iconvFileName);
            CmaplWrapper.LoadDllIfNecessary(Properties.Resources.LIBXML2_DLL, CmaplWrapper.lib2xmlFileName);
            CmaplWrapper.LoadDllIfNecessary(Properties.Resources.POCRT_DLL, CmaplWrapper.pocrtFileName);
            CmaplWrapper.LoadDllIfNecessary(Properties.Resources.CMAPL_DLL, CmaplWrapper.cmaplFileName);
        }

        public static void CleanupDlls() {
            for (int x = 0; x < 20; x++) {
                int index = CmaplWrapper.dlls.Count - 1;

                if (index < 0) {
                    break;
                }

                while (index >= 0) {
                    var dll = CmaplWrapper.dlls[index];
                    if (dll.Cleanup() == true) {
                        CmaplWrapper.dlls.RemoveAt(index);
                    }
                    Thread.Sleep(10);
                    index--;
                }
            }
        }

        private static bool IsDllLoaded(string fileName) {
            foreach (var dll in CmaplWrapper.dlls) {
                if (dll.PathName.EndsWith(fileName)) {
                    return true;
                }
            }

            return false;
        }

        private static bool LoadDllIfNecessary(byte [] dllBytes, string fileName) {
            if (CmaplWrapper.IsDllLoaded(fileName)) {
                return true;
            } else {
                var dll = new EmbeddedUnmanagedDll(dllBytes, fileName);

                if (dll != null) {
                    if (dll.PathName.EndsWith(fileName)) {
                        dlls.Add(dll);
                        return true;
                    }
                }
            }

            return false;
        }
        
        [DllImport(CmaplWrapper.cmaplFileName, CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        [return: MarshalAs(UnmanagedType.LPStr)]
        unsafe private static extern string get_counters(
            [MarshalAs(UnmanagedType.LPStr)]string in_file_path,
            [MarshalAs(UnmanagedType.LPStr)]string objectname, 
            int num_frequencies,
            [MarshalAs(UnmanagedType.LPArray, SizeParamIndex=2)]int[] frequency);

        [DllImport(CmaplWrapper.cmaplFileName, CallingConvention = CallingConvention.Cdecl)]
        unsafe private static extern void get_csv(
            [MarshalAs(UnmanagedType.LPStr)]string in_file_path,
            [MarshalAs(UnmanagedType.LPStr)]string out_file_path,
            [MarshalAs(UnmanagedType.LPStr)]string objectname,
            int num_objects,
            [MarshalAs(UnmanagedType.LPStr)]string counters_str,
            [MarshalAs(UnmanagedType.LPStr)]string num_counters,
            int frequency);
    }
}
