﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Diagnostics;
using System.IO;
using odct;
using odct.FilerConfig;
using odct.Settings;
using odct.Counters;
using odct.Clusters;
using odct.Json;

namespace odct {
    public class Collector {
        public const int ExpectedArchiveFilesPer7ModeFiler = 24 * 31;
        public const int ExpectedArchiveFilesPerCModeFiler = 24 * 30 * 4;
        private ICollectorOptions Options;
        private Messenger messenger;
        private Dictionary<string, CollectedFilerBookmark> serialsCollected;
        private Dictionary<string, CollectionResults> collectionResultsByServer;
        public Dictionary<string, Cluster> clusterByClusterUuid;
        public Dictionary<string, CollectionResults> collectionResultsByClusterUuid;
        private int maxFilesToProcess;
        private List<Thread> threads;
        private static readonly object objLock = new object();

        private struct CollectedFilerBookmark {
            public FilerConfiguration config;
            public string filerTargetPath;
            public Credentials credentials;
        }

        public Collector(Messenger m, ICollectorOptions opts) {

            this.messenger = m;
            this.Options = opts;

            this.serialsCollected = new Dictionary<string, CollectedFilerBookmark>();
            this.collectionResultsByServer = new Dictionary<string, CollectionResults>();
            this.collectionResultsByClusterUuid = new Dictionary<string, CollectionResults>();
            this.clusterByClusterUuid = new Dictionary<string, Cluster>();

            this.maxFilesToProcess = OdctSettings.Instance.GetPerformanceFilesCollectionLimit();

            CmaplWrapper.LoadDllsIfNecessary();
        }

        public void Collect(ICollection<Credentials> creds, string targetArchiveFilePath, string targetTempPath) {
            this.messenger.sendInfo(" ");
            this.messenger.sendInfo("Collector.Collect() started.");

            try {
                if (creds.Count > 0) {
                    NetApp.Manage.NaServer.TRANSPORT_TYPE transportType = NetApp.Manage.NaServer.TRANSPORT_TYPE.HTTPS;

                    foreach (var c in creds) {
                        if (c.verifyResults == null) c.Verify(this.messenger);  
                        if (c.verifyResults.IsClusterMode) {
                            this.messenger.sendInfo(string.Format("Collector.Collect() processing cluster credentials {0}.", c.Server));
                            Cluster cluster = new Cluster(c, this.messenger, targetTempPath);
                            this.messenger.sendInfo(string.Format("Collector.Collect() processing cluster {0}, credentials {1}.", cluster.clusterName, c.Server));
                            if (collectionResultsByClusterUuid.ContainsKey(cluster.clusterUuid)) {
                                this.messenger.sendDebug(string.Format("Collector.Collect() processing duplicate credentials for cluster {0}.", cluster.clusterName));
                                this.collectionResultsByClusterUuid[cluster.clusterUuid] = this.ProcessClusterModeDuplicateCredentials(cluster);
                            } else {
                                this.messenger.sendDebug(string.Format("Collector.Collect() processing non-duplicate credentials for cluster {0}.", cluster.clusterName));
                                this.collectionResultsByClusterUuid[cluster.clusterUuid] = cluster.CollectConfig(this);
                            }
                            continue;
                        }
                        if (collectionResultsByServer.ContainsKey(c.Server)) {
                            this.ProcessDuplicate7ModeCredentials(c);
                            continue;
                        };
                        this.Process7ModeCredentials(creds, c, targetTempPath, ref transportType);
                    }

                    this.calculateExpectedFilesToProcess();
                    Messenger.StartTimer();

                    foreach (Cluster cluster in this.clusterByClusterUuid.Values)
                    {
                        CollectionResults cr = this.collectionResultsByClusterUuid[cluster.clusterUuid];
                        if (cr.configCollected)
                            this.CollectPerformanceCountersForCluster(cluster, ref cr);
                        this.collectionResultsByClusterUuid[cluster.clusterUuid] = cr;
                    }

                    if (!string.IsNullOrEmpty(this.Options.getAppCountersRequested()))
                    {
                        this.CollectPerformanceCountersForAll7ModeFilers(transportType);
                    }

                    this.WriteManifest(targetTempPath);
                    this.Write7ModeProcessingSummary();
                    this.WriteClusterProcessingSummary();
                    this.CloseMessengerAndZipPayload(targetTempPath, targetArchiveFilePath);

                } else {
                    messenger.sendError("No host credentials configured");
                    messenger.closeMessenger();
                }

            } catch (Exception e) {
                this.messenger.sendException(e);
                messenger.closeMessenger();
                ConsoleUI.WriteError(e.ToString());
            }
        }

        public void CollectPerformanceCountersForCluster(Cluster cluster, ref CollectionResults cr)
        {
            try
            {
                this.threads = new List<Thread>();
                var threadsToWaitOn = new List<EventWaitHandle>();

                foreach (ClusterNode cn in cluster.nodeNamesToNodes.Values)
                {
                    if (cn.collectionFailed)
                        continue;
                    var waiter = new EventWaitHandle(false, EventResetMode.ManualReset);
                    threadsToWaitOn.Add(waiter);

                    Thread newt = new Thread(new ThreadStart(delegate
                    {
                        cn.CollectPerformanceCounters();
                        waiter.Set();
                    }));

                    newt.IsBackground = true;

                    this.threads.Add(newt);

                    this.messenger.sendDebug(string.Format("Collector.Collect() starting thread for ClusterNode {0}.",
                        cn.nodeName));
                    newt.Start();

                }
                if (threadsToWaitOn.Count > 0)
                {
                    Mutex.WaitAll(threadsToWaitOn.ToArray());
                }
                this.messenger.sendDebug(string.Format("Collector.Collect() done with Mutex.WaitAll during collection of cluster {0}.", cluster.clusterName));
                cr.collected = true;
            }
            catch (OdctClusterCounterCollectionFailedException e)
            {
                this.messenger.sendDebug("Collector.Collect() caught OdctClusterCounterCollectionFailedException during cluster collection.");
                cr.collected = false;
                cr.description = e.Message;
            }
        }

        public CollectionResults ProcessClusterModeDuplicateCredentials(Cluster cluster)
        {
            this.messenger.sendError(string.Format("Collector.Collect() encountered duplicate credentials for cluster {0}.", cluster.credentials.Server));
            this.messenger.sendError("...Further processing of this duplicate halted.");
            this.messenger.sendError("...Processing of other servers continuing.");

            CollectionResults oldCr = collectionResultsByClusterUuid[cluster.clusterUuid];
            oldCr.servers.Add(cluster.credentials.Server);
            oldCr.count++;
            oldCr.description = "Duplicate cluster credentials encountered.";
            return oldCr;
        }

        public void Process7ModeCredentials(ICollection<Credentials> creds, Credentials c, string targetTempPath, ref NetApp.Manage.NaServer.TRANSPORT_TYPE transportType)
        {
            try
            {
                Messenger.updateObject updateObject = Messenger.ProgressIncrement(0);
                this.messenger.update(updateObject);
                this.messenger.sendInfo(" Collector.Collect() calling CollectFilerConfigurationAndGetPartnerAddress() for self.");
                VerifyResults verifyResults;

                var partner = this.CollectFilerConfigurationAndGetPartnerAddress(c, targetTempPath, out verifyResults);
                if (this.ShouldCollectPartner(creds, partner))
                {
                    this.messenger.sendInfo(" Collector.Collect() calling CollectFilerConfigurationAndGetPartnerAddress() for partner.");
                    this.CollectFilerConfigurationAndGetPartnerAddress(
                        new Credentials(partner, c.Username, c.Password),
                        targetTempPath, out verifyResults);
                }
                if ((!verifyResults.IsHttps) && (!OdctSettings.Instance.DenyHttp()))
                {
                    transportType = NetApp.Manage.NaServer.TRANSPORT_TYPE.HTTP;
                }
            }
            catch (FilerConfigException fce)
            {
                CollectionResults cr = new CollectionResults();
                cr.collected = false;
                cr.description = fce.Message;
                cr.count = 1;
                this.collectionResultsByServer.Add(c.Server, cr);
            }
        }

        public void ProcessDuplicate7ModeCredentials(Credentials c)
        {
            this.messenger.sendError(string.Format("Duplicate Hostname / IP Address {0} encountered.", c.Server));
            this.messenger.sendError("...Further processing of this duplicate halted.");
            this.messenger.sendError("...Processing of other servers continuing.");

            CollectionResults cr = collectionResultsByServer[c.Server];
            cr.servers.Add(c.Server);
            cr.count++;
            collectionResultsByServer[c.Server] = cr;
        }

        public void CollectPerformanceCountersForAll7ModeFilers(NetApp.Manage.NaServer.TRANSPORT_TYPE transportType)
        {
            this.threads = new List<Thread>();
            var threadsToWaitOn = new List<EventWaitHandle>();
            foreach (var s in this.serialsCollected)
            {
                var a = s.Value;

                var waiter = new EventWaitHandle(false, EventResetMode.ManualReset);
                threadsToWaitOn.Add(waiter);

                Thread newt = new Thread(new ThreadStart(delegate
                {
                    this.CollectPerformanceCounters(a.config, a.filerTargetPath, a.credentials, transportType);
                    waiter.Set();
                }));

                newt.IsBackground = true;

                this.threads.Add(newt);

                newt.Start();
            }
            if (threadsToWaitOn.Count > 0)
            {
                Mutex.WaitAll(threadsToWaitOn.ToArray());
            }
        }

        public void CloseMessengerAndZipPayload(string targetTempPath, string targetArchiveFilePath)
        {
            messenger.sendInfo(string.Format("\nCollection done outputting to {0}.", targetTempPath));

            string tempDirectory = Directory.GetParent(targetTempPath).ToString();
            string targetFilename = Path.GetFileName(targetArchiveFilePath);
            string tempPath = Path.Combine(tempDirectory, targetFilename);

            bool targetArchiveFilePathIsReady = true;
            try
            {
                File.WriteAllText(targetArchiveFilePath, "test");
                File.Delete(targetArchiveFilePath);
                this.messenger.sendInfo(string.Format("Payload will be zipped to requested location: {0}", targetArchiveFilePath));
            }
            catch (Exception e)
            {
                targetArchiveFilePathIsReady = false;
                TempFolder.setFolderLockTrue(tempDirectory, "Payload zipped here because requested path was not available.");
                this.messenger.sendError(string.Format("WARNING!: Requested Payload destination \"{0}\" is not available.", targetArchiveFilePath));
                this.messenger.sendError(string.Format("WARNING!: Error encountered was {0}", e.Message));
                this.messenger.sendError(string.Format("WARNING!: Payload can be found at \"{0}\".", tempPath));
            }

            messenger.sendInfo("Closing Messenger");
            messenger.closeMessenger();

            Payload.Zip(targetTempPath, tempPath);
            if (targetArchiveFilePathIsReady)
            {
                File.Move(tempPath, targetArchiveFilePath);
            }
            Directory.Delete(targetTempPath, true);
        }

        public void Write7ModeProcessingSummary()
        {
            messenger.sendInfo("");
            messenger.sendInfo("Summary of Filers Processed:");
            messenger.sendInfo("----------------------------");
            foreach (var kvp in this.collectionResultsByServer)
            {
                CollectionResults cr = kvp.Value;

                messenger.sendInfo(
                    string.Format("{0} : {1}",
                    kvp.Key,
                    cr.description));

                for (int i = 1; i < cr.count; i++)
                {
                    messenger.sendInfo(
                        string.Format("{0} : {1}",
                        kvp.Key,
                        "Duplicate filer collection request identified, duplicate not processed."));
                }
            };
        }

        public void WriteClusterProcessingSummary()
        {
            messenger.sendInfo("");
            messenger.sendInfo("Summary of Clusters Processed:");
            messenger.sendInfo("------------------------------");
            foreach (var kvp in this.collectionResultsByClusterUuid)
            {
                CollectionResults collectionResults = kvp.Value;
                Cluster cluster = this.clusterByClusterUuid[kvp.Key];
                messenger.sendInfo(
                    string.Format("{0} : {1} ({2}: {3})",
                    cluster.credentials.Server,
                    cluster.credentials.verifyResults.resultDescription,
                    cluster.clusterName,
                    cluster.clusterUuid));

                string[] serverArray = collectionResults.servers.ToArray();
                for (int i = 1; i < collectionResults.count; i++)
                {
                    messenger.sendInfo(
                        string.Format("{0} : {1} ({2}: {3})",
                        serverArray[i],
                        "Duplicate cluster collection request, not processed.",
                        cluster.clusterName,
                        cluster.clusterUuid));
                }
            };

        }

        private int calculateExpectedFilesToProcess()
        {
            int expectedFilesToProcessEach7ModeNode = Math.Min(Collector.ExpectedArchiveFilesPer7ModeFiler, this.maxFilesToProcess);
            int expectedFilesToProcessEachCModeNode = Math.Min(Collector.ExpectedArchiveFilesPerCModeFiler, this.maxFilesToProcess);

            int expectedTotal7ModeFilesToProcess = 0;
            int expectedTotalCModeFilesToProcess = 0;

            foreach (KeyValuePair<string,CollectionResults> kvp in this.collectionResultsByClusterUuid)
            {
                string clusterUuid = kvp.Key;
                Cluster cluster = this.clusterByClusterUuid[clusterUuid];
                CollectionResults cr = kvp.Value;
                if (cr.configCollected)
                {
                    foreach (ClusterNode cn in cluster.nodeNamesToNodes.Values)
                    {
                        if (cn.isNodeHealthy && !cn.collectionFailed)
                        {
                            expectedTotalCModeFilesToProcess += expectedFilesToProcessEachCModeNode;
                            this.messenger.sendInfo(string.Format("ClusterNode {0}: Expect to process {1} files", cn.nodeName, expectedFilesToProcessEachCModeNode));
                        }
                    }
                }
            }
            foreach (KeyValuePair<string, CollectionResults> kvp in this.collectionResultsByServer)
            {
                string server = kvp.Key;
                CollectionResults cr = kvp.Value;
                if (cr.configCollected)
                {
                    expectedTotal7ModeFilesToProcess += expectedFilesToProcessEach7ModeNode;
                    this.messenger.sendInfo(string.Format("Server {0}: Expect to process {1} files", server, expectedFilesToProcessEach7ModeNode));
                }
            }
            Messenger.updateObject u = Messenger.ProgressReset(expectedTotal7ModeFilesToProcess + expectedTotalCModeFilesToProcess);
            this.messenger.update(u);
            return expectedTotal7ModeFilesToProcess + expectedTotalCModeFilesToProcess;
        }

        public void Cancel() {
            if (this.threads != null) {
                foreach(var t in this.threads) {
                    if (t.IsAlive) {
                        t.Interrupt();
                    }
                }
            }
        }

        private void WriteManifest(string targetTempPath) {
            JsonObject manifest = new JsonObject();
            foreach (var e in this.serialsCollected) {
                var f = e.Value.config;
                manifest.Add(getFilerSuccessManifestEntry(f));
            }
            foreach (var kvp in this.collectionResultsByServer) {
                foreach (KeyValuePair<JsonString, JsonValue> entry in getFilerFailureManifestEntries(kvp)) {
                    manifest.Add(entry);
                }
            }
            foreach (Cluster cluster in this.clusterByClusterUuid.Values) {
                if (collectionResultsByClusterUuid[cluster.clusterUuid].collected) {
                    foreach (KeyValuePair<JsonString, JsonValue> nodeEntry in getClusterSuccessManifestEntries(cluster)) {
                        manifest.Add(nodeEntry);
                    }
                }
            }
            foreach (var kvp in this.collectionResultsByClusterUuid) {
                foreach (KeyValuePair<JsonString, JsonValue> nodeEntry in getClusterFailureManifestEntries(kvp)) {
                    manifest.Add(nodeEntry);
                }
            }
            var w = new StreamWriter(Path.Combine(targetTempPath, "manifest.json"));
            w.Write(manifest.toString());
            w.Close();
        }

        private List<KeyValuePair<JsonString,JsonValue>> getFilerFailureManifestEntries(KeyValuePair<string, CollectionResults> kvp) { 
            List<KeyValuePair<JsonString,JsonValue>> results = new List<KeyValuePair<JsonString, JsonValue>>();
            string filerIp = kvp.Key;
            CollectionResults cr = (CollectionResults) kvp.Value;
            if (cr.collected) {
                for (int i = 1; i < cr.count; i++) {
                    JsonObject failureValues = new JsonObject();
                    failureValues.Add("collected", new JsonFalse());
                    const string failure = "duplicate filer collection request identified, duplicate not processed.";
                    failureValues.Add("failure-description", new JsonString(failure));
                    results.Add(new KeyValuePair<JsonString,JsonValue>(new JsonString(filerIp), failureValues));
                }
            } else {
                    JsonObject failureValues = new JsonObject();
                    failureValues.Add("collected", new JsonFalse());
                    failureValues.Add("failure-description", new JsonString(cr.description));
                    results.Add(new KeyValuePair<JsonString,JsonValue>(new JsonString(filerIp), failureValues));
            }
            return results;
        }

        private List<KeyValuePair<JsonString, JsonValue>> getClusterFailureManifestEntries(KeyValuePair<string, CollectionResults> kvp) {
            List<KeyValuePair<JsonString, JsonValue>> results = new List<KeyValuePair<JsonString, JsonValue>>();
            string clusterUuid = kvp.Key;
            CollectionResults cr = (CollectionResults) kvp.Value;
            if (cr.collected) {
                for (int i = 1; i < cr.count; i++) {
                    JsonObject failureValues = new JsonObject();
                    failureValues.Add("collected", new JsonFalse());
                    failureValues.Add("collected-config", new JsonFalse());
                    const string failure = "duplicate cluster collection request identified, duplicate not processed.";
                    failureValues.Add("failure-description", new JsonString(failure));
                    results.Add(new KeyValuePair<JsonString, JsonValue>(new JsonString(clusterUuid), failureValues));
                }
            } else {
                JsonObject failureValues = new JsonObject();
                failureValues.Add("collected", new JsonFalse());

                JsonValue configCollected = JsonBool.getJsonBool(cr.configCollected);
                failureValues.Add("collected-config", configCollected);

                failureValues.Add("failure-description", new JsonString(cr.description));
                results.Add(new KeyValuePair<JsonString, JsonValue>(new JsonString(clusterUuid), failureValues));
            }
            return results;
        }

        private KeyValuePair<JsonString, JsonValue> getFilerSuccessManifestEntry(FilerConfiguration f) {
            JsonObject filerValues = new JsonObject();
            filerValues.Add("collected", new JsonTrue());
            filerValues.Add("system-id", new JsonString(f.SystemId));
            filerValues.Add("system-name", new JsonString(f.SystemName));
            filerValues.Add("system-model", new JsonString(f.SystemModel));

            string modeString = f.Is7Mode ? " 7-Mode" : "";
            string sysVersion = string.Format("{0}{1}", f.SystemVersion, modeString);
            filerValues.Add("system-version", new JsonString(sysVersion));
            filerValues.Add("address", new JsonString(f.Address));
            filerValues.Add("partner-name", new JsonString(f.PartnerName));
            filerValues.Add("partner-system-id", new JsonString(f.PartnerSystemId));
            filerValues.Add("partner-address", new JsonString(f.PartnerNetworkAddress));

            return new KeyValuePair<JsonString, JsonValue>(new JsonString(f.SystemSerialNumber), filerValues);
        }

        private List<KeyValuePair<JsonString, JsonValue>> getClusterSuccessManifestEntries(Cluster c) {
            List<KeyValuePair<JsonString, JsonValue>> resultEntries = new List<KeyValuePair<JsonString, JsonValue>>();
            foreach (ClusterNode cn in c.nodeNamesToNodes.Values) {
                resultEntries.Add(getNodeSuccessEntry(cn));
            }
            return resultEntries;
        }

        private JsonObject getClusterObject(Cluster c) {
            JsonObject clusterObject = new JsonObject();
            clusterObject.Add("id", new JsonString(c.clusterUuid));
            clusterObject.Add("cluster-location", new JsonString(c.clusterLocation));
            clusterObject.Add("cluster-name", new JsonString(c.clusterName));
            clusterObject.Add("cluster-serial-number", new JsonString(c.clusterSerialNumber));
            string clusterVersionString = c.ontapFullVersion;
            if (c.ontapGeneration == 8) {
                clusterVersionString = string.Format("{0} clustered Data ONTAP", c.ontapFullVersion);
            }
            if (c.ontapGeneration == 9) {
                clusterVersionString = string.Format("ONTAP {0}", c.ontapFullVersion);
            }
            clusterObject.Add("cluster-version", new JsonString(clusterVersionString));
            return clusterObject;
        }

        private KeyValuePair<JsonString, JsonValue> getNodeSuccessEntry(ClusterNode cn) {
            JsonObject clusterNodeValues = new JsonObject();
            bool collectedValue = true;
            bool collectedConfigValue = true;
            if (cn.performanceCollectionFailed)
                collectedValue = false;
            if (cn.collectionFailed) {
                collectedValue = false;
                collectedConfigValue = false;
            }
            clusterNodeValues.Add("collected", JsonBool.getJsonBool(collectedValue));
            clusterNodeValues.Add("collected-config", JsonBool.getJsonBool(collectedConfigValue));
            clusterNodeValues.Add("system-id", new JsonString(cn.FilerSystemId));
            clusterNodeValues.Add("cluster", getClusterObject(cn.cluster));
            clusterNodeValues.Add("system-name", new JsonString(cn.nodeName));
            clusterNodeValues.Add("system-model", new JsonString(cn.systemModel));
            clusterNodeValues.Add("address", new JsonString(cn.nodeMgmtIp));
            clusterNodeValues.Add("partner-name", new JsonString(cn.PartnerName));
            string partnerSystemId = cn.cluster.nodeNamesToNodes[cn.PartnerName].FilerSystemId;
            clusterNodeValues.Add("partner-system-id", new JsonString(cn.cluster.nodeNamesToNodes[cn.PartnerName].FilerSystemId));
            clusterNodeValues.Add("partner-address", new JsonString(cn.cluster.nodeNamesToNodes[cn.PartnerName].nodeMgmtIp));
            return new KeyValuePair<JsonString,JsonValue>(new JsonString(cn.FilerSerialNumber), clusterNodeValues);
        }

        private bool ShouldCollectPartner(ICollection<Credentials> creds, string partner) 
        {
            if (!this.Options.GetAutoCollectHaPartners() || string.IsNullOrEmpty(partner)) {
                return false;
            }

            foreach (var c in creds) {
                if (partner.Equals(c.Server)) {
                    this.messenger.sendDebug(string.Format(
                        "Ha-Partner {0} already defined explicitly. Prefer explicit credentials.",
                        partner));
                    return false;
                }
            }

            return true;
        }

        public string CollectFilerConfigurationAndGetPartnerAddress(Credentials credentials, string targetTempPath, out VerifyResults verifyResults) {
            CollectionResults cr = new CollectionResults();
            cr.servers.Add(credentials.Server);
            verifyResults = credentials.Verify(this.messenger);

            if (verifyResults.IsVersion8_0) {
                cr.description = "Version 8.0 system identified, this version is not supported by Onsite Data Collection Tool.";
                this.collectionResultsByServer.Add(credentials.Server, cr);
                verifyResults.IsVerified = false;
                return null;
            }
            if (!verifyResults.IsFound) {
                cr.description = "Failed to connect to filer at address provided.";
                this.collectionResultsByServer.Add(credentials.Server, cr);
                verifyResults.IsVerified = false;
                return null;
            }
            if (!verifyResults.IsAuthenticated) {
                cr.description = "Filer refused access with the credentials provided.";
                this.collectionResultsByServer.Add(credentials.Server, cr);
                verifyResults.IsVerified = false;
                return null;
            }
            if (!verifyResults.IsVerified) {
                cr.description = "Failed to verify credentials with credentials provided.";
                this.collectionResultsByServer.Add(credentials.Server, cr);
                verifyResults.IsVerified = false;
                return null;
            }
            var config = this.CollectFilerConfiguration(credentials, targetTempPath);
            cr.configCollected = true;
            cr.description = "Configuration collected successfully with credentials provided.";
            this.collectionResultsByServer.Add(credentials.Server, cr);
            verifyResults.IsVerified = true;
            return config.PartnerNetworkAddress;
        }

        public VerifyResults VerifyConnection(Credentials credentials) {
            this.messenger.sendInfo("Collector.VerifyConnection() started.");
            VerifyResults verifyResults = credentials.Verify(this.messenger);
            verifyResults.logResults(this.messenger, "Collector", credentials.Server);
            return verifyResults;
        }

        public FilerConfiguration CollectFilerConfiguration(Credentials credentials, string targetTempPath) {
            var networkHost = credentials.Server;
            this.messenger.sendInfo("Collector.CollectFilerConfiguration() started.");
            var config = this.GetFilerConfig(credentials, targetTempPath);

            if (this.serialsCollected.ContainsKey(config.SystemSerialNumber)) {
                this.messenger.sendDebug(string.Format(
                    "Already collected serial number {0} skipping recollection...",
                    config.SystemSerialNumber));
            } else {
                var filerTargetPath = Path.Combine(targetTempPath, config.SystemSerialNumber);
                var args = new CollectedFilerBookmark() { 
                    config = config, 
                    credentials = credentials, 
                    filerTargetPath = filerTargetPath };

                this.serialsCollected.Add(config.SystemSerialNumber, args); // TODO: wait until verify connection before adding
                this.SaveFilerConfig(config, filerTargetPath);

            }

            this.messenger.sendInfo("Collector.CollectFilerConfiguration() returned.");
            return config;
        }

        public FilerConfiguration GetFilerConfig(Credentials cred, string targetPath) {
            this.messenger.sendInfo("Collector.GetFilerConfig() started.");
            var f = new FilerConfigurationFactory(cred, this.messenger);
            var config = f.Build();
            this.messenger.sendInfo("Collector.GetFilerConfig() returned.");

            return config;
        }

        public void SaveFilerConfig(FilerConfiguration conf, string filerTargetPath) {
            this.messenger.sendInfo("Collector.SaveFilerConfig() started.");
            var configurationTargetPath = Path.Combine(filerTargetPath, "configuration");
            Directory.CreateDirectory(configurationTargetPath);

            foreach (var s in conf.Sections) {
                var xo = s.GetNaElement();
                if (xo != null) {
                    var xoPrettyString = xo.ToPrettyString("");
                    int len = 500;
                    this.messenger.sendDebug(string.Format("****** {0}.ToPrettyString first {1} characters ******", s.GetSectionTitle(), len));
                    this.messenger.sendDebug(xoPrettyString.Substring(0, Math.Min(len, xoPrettyString.Length - 1)));
                    this.messenger.sendDebug(string.Format("------ end {0}.ToPrettyString sample ------", s.GetSectionTitle()));
                    var filePath = Path.Combine(
                        configurationTargetPath,
                        string.Format("{0}.xml", s.GetSectionTitle()));

                    File.WriteAllText(filePath, xoPrettyString);
                }
            }
            this.messenger.sendInfo("Collector.SaveFilerConfig() returned.");
        }

        public void CollectPerformanceCounters(FilerConfiguration config, string targetPath, Credentials credentials, NetApp.Manage.NaServer.TRANSPORT_TYPE transportType) {
            this.messenger.sendInfo("Collector.CollectPerformanceCounters() started.");
            
            // TODO be able to pass it credentials only, it worries about connect
            var gen = new ZapiSourceGenerator("", new ZapiContext(credentials, this.messenger, transportType));
            try {
                // TODO : pass block to gen.Run
                if (gen.Count() < 1) {
                    this.messenger.sendError(string.Format("<{0}> CollectPerformanceCounters() ArchiveSourceGenerator failed!", config.SystemSerialNumber));
                    this.messenger.sendError(string.Format("<{0}> gen.Count() is: {1}", config.SystemSerialNumber, gen.Count()));
                } 
                       
                var filesProcessed = 0;
                foreach (var src in gen.GetSources()) {

                    bool writeXmlMetaDataToPayload = (filesProcessed == 0);  // write it once only.

                    ArchiveParser ap = new ArchiveParserFactory(
                        config.SystemSerialNumber,
                        messenger, 
                        src, 
                        targetPath,
                        this.Options.getAppCountersRequested(),
                        writeXmlMetaDataToPayload,
                        false).Build(config.SystemVersion);

                    ap.processRecords();
                    filesProcessed++;
                    lock (objLock) {
                        Messenger.updateObject updateObject = Messenger.ProgressIncrement(1);
                        this.messenger.update(updateObject);
                    }
                    if (filesProcessed >= (this.maxFilesToProcess)) break;  // will loop maxFilesToProcess times
                }
                // TODO : end pass block to gen.Run
            } catch (Exception e) {
                this.messenger.sendError(
                    string.Format("<{0}> Collector.CollectPerformanceCounters() encountered unexpected error {1}.", config.SystemSerialNumber, e.Message));
                this.messenger.sendException(e);
            }
            this.messenger.sendInfo(string.Format("<{0}> ", config.SystemSerialNumber));
            this.messenger.sendInfo(string.Format("<{0}>  Collector.CollectPerformanceCounters() returned.", config.SystemSerialNumber));
            this.messenger.sendInfo(string.Format("<{0}> ", config.SystemSerialNumber));
            this.messenger.sendInfo(string.Format("<{0}> ", config.SystemSerialNumber));
        }

        public static string GetTempPath(string targetFilePath) {
            string targetfilename = Path.GetFileNameWithoutExtension(targetFilePath);
            TempFolder tempFolder = new TempFolder();
            string collectionwork = Path.Combine(tempFolder.PathName, 
                string.Format("{0}-{1}", targetfilename, "odct-temp"));
            return collectionwork;
        }
    }
}
