﻿using System;
using System.Diagnostics;
using System.IO;
using System.IO.Compression;
using System.Text;
using System.Xml;
using System.Xml.XPath;
using System.Collections.Generic;
using NetApp.Manage;
using odct.Settings;

namespace odct.Counters {
    public abstract class ArchiveParser {
        protected string SystemSerialNo;
        protected Messenger messenger;
        public ArchiveFileSource source;
        public string outputPath;
        public byte[] allBytes;
        public int xmlEnd = 0;
        protected CounterDefinitions counterDefinitions;
        protected List<int> neededOids;
        protected Dictionary<long, CounterDefinitionLookup> counterDefinitionLookupByOid;
        protected Dictionary<long, List<CounterDefinition>> counterDefinitionListByOid;
        protected Dictionary<long, StreamWriter> writer;
        protected Dictionary<int, int> rowsWrittenByOid;
        protected Dictionary<string, int> requestedFrequencyByDottedId;
        protected Dictionary<string, long> collectedFrequencyByDottedId;
        protected Dictionary<int, long> neededFrequencyByOid;
        protected Dictionary<int, bool> countersAreRequestedByOid;
        private bool writeXmlToPayload;

        public ArchiveParser(string serial, Messenger m, ArchiveFileSource s, string p, Dictionary<string, int> requestedFrequencyByDottedId, bool writeXmlToPayload) {
            this.SystemSerialNo = serial;
            this.messenger = m;
            this.messenger.sendInfo(string.Format("<{0}> ArchiveParser constructor started with output path: {1}.", this.SystemSerialNo, p));
            this.source = s;
            this.outputPath = p;
            this.requestedFrequencyByDottedId = requestedFrequencyByDottedId;
            this.writeXmlToPayload = writeXmlToPayload;
            this.messenger.sendDebug(string.Format("<{0}> ArchiveParser constructor now reading: {1}", this.SystemSerialNo, this.source.Filename));
            this.allBytes = this.source.ReadAsBytes();
            this.xmlEnd = findXmlEnd();

            this.messenger.sendDebug(string.Format("<{0}> ArchiveParser constructor now creating counterDefinitions.", this.SystemSerialNo));
            this.counterDefinitions = new CounterDefinitions(
                getArchiveXmlAsString(),
                requestedFrequencyByDottedId);
            neededOids = this.counterDefinitions.GetOidsToCollect();

            counterDefinitionLookupByOid = new Dictionary<long, CounterDefinitionLookup>();
            counterDefinitionListByOid = new Dictionary<long, List<CounterDefinition>>();
            this.messenger.sendDebug(string.Format("<{0}> ArchiveParser constructor now creating counterDefinition lookups.", this.SystemSerialNo));
            foreach (var objectId in neededOids) {
                counterDefinitionLookupByOid[objectId]
                    = new CounterDefinitionLookup(getArchiveXmlAsString(), counterDefinitions.GetObjectNameByOid(objectId));
                counterDefinitionListByOid[objectId]
                    = counterDefinitionLookupByOid[objectId].GetCounterDefinitionListSortedByCounterId();
            }

            this.messenger.sendDebug(string.Format("<{0}> ArchiveParser constructor now calculating collectedFrequencyByDottedId.", this.SystemSerialNo));
            this.collectedFrequencyByDottedId = calcCollectedFrequencyByDottedItem(requestedFrequencyByDottedId);
            this.messenger.sendDebug(string.Format("<{0}> ArchiveParser constructor now calculating neededFrequencyByDottedId.", this.SystemSerialNo));
            this.neededFrequencyByOid = calcNeededFrequencyByOid(collectedFrequencyByDottedId);
            this.messenger.sendDebug(string.Format("<{0}> ArchiveParser constructor now calculating countersAreRequestedByOid.", this.SystemSerialNo));
            this.countersAreRequestedByOid = calcCountersAreRequestedByOid(collectedFrequencyByDottedId);

            writer = new Dictionary<long, StreamWriter>();
            rowsWrittenByOid = new Dictionary<int, int>();

        }

        protected Dictionary<int, bool> calcCountersAreRequestedByOid(Dictionary<string, long> collectedFrequencyByDottedId) {
            Dictionary<int, bool> calc = new Dictionary<int, bool>();
            foreach (var kvp in collectedFrequencyByDottedId) {
                CounterDefinition cd = this.counterDefinitions.GetCounterDefinitionByDottedName(kvp.Key);
                calc[cd.oid] = true;
            }
            return calc;
        }

        public Dictionary<int, long> calcNeededFrequencyByOid(Dictionary<string, long> collected) {
            Dictionary<int, long> calculated = new Dictionary<int, long>();
            foreach (var kvp in collected) {
                string thisDottedItem = kvp.Key;
                var s = thisDottedItem.Split('.');
                string thisObjectName = s[0];
                int thisOid = (int)counterDefinitions.GetOidByObjectName(thisObjectName);
                if (calculated.ContainsKey(thisOid)) {
                    calculated[thisOid] = calculated[thisOid] | collected[thisDottedItem]; 
                } else {
                    calculated[thisOid] = collected[thisDottedItem];
                };
            }
            return calculated;
        }

        public Dictionary<string, long> calcCollectedFrequencyByDottedItem(Dictionary<string,int> requestedFrequencyByDottedItem) {
            // This method calculates the frequency to be collected based on the frequency requested. Collected 
            // frequency may be different than requested frequency.  If, for example, a user requests that a counter
            // be collected every second (FREQUENCY_1), but it is in the archive only every five minutes (FREQUENCY_4),
            // it can only be collected every five minutes, since more frequent data doesn't exist - it simply isn't in the archive.
            var result = new Dictionary<string, long>();
            foreach (var kvp in requestedFrequencyByDottedItem) {

                string resultKey = kvp.Key;
                CounterDefinition cd = counterDefinitions.GetCounterDefinitionByDottedName(resultKey);
                long cdFrequency = cd.flag & Flag.FREQUENCY_MASK;
                long resultValue;
                switch (kvp.Value) {
                    case 0:
                        resultValue = Flag.CONFIG_FLAG;
                        break;
                    case 1:
                        if ((cdFrequency == Flag.CONFIG_FLAG) || (cdFrequency > Flag.FREQUENCY_1_FLAG)) {
                            resultValue = cdFrequency;
                            break;
                        } else {
                            resultValue = Flag.FREQUENCY_1_FLAG;
                            break;
                        }
                    case 2:
                        if (cdFrequency == Flag.CONFIG_FLAG || (cdFrequency > Flag.FREQUENCY_2_FLAG)) {
                            resultValue = cdFrequency;
                            break;
                        } else {
                            resultValue = Flag.FREQUENCY_2_FLAG;
                            break;
                        }
                    case 3:
                        if (cdFrequency == Flag.CONFIG_FLAG || (cdFrequency > Flag.FREQUENCY_3_FLAG)) {
                            resultValue = cdFrequency;
                            break;
                        } else {
                            resultValue = Flag.FREQUENCY_3_FLAG;
                            break;
                        }
                    case 4:
                        if (cdFrequency == Flag.CONFIG_FLAG) {
                            resultValue = cdFrequency;
                            break;
                        } else {
                            resultValue = Flag.FREQUENCY_4_FLAG;
                            break;
                        }
                    default:
                        resultValue = Flag.CONFIG_FLAG;
                        break;
                }
                result.Add(resultKey, resultValue);
            }

            return result;
        }

        public string getArchiveXmlAsString() {
            byte[] xmlArray = this.allBytes.getSubArray(0, xmlEnd + 1);
            return Encoding.UTF8.GetString(xmlArray);
        }

        public static string getArchiveXmlAsString(string filepath) {

            byte[] rawXmlArray = File.ReadAllBytes(filepath);

            int rawXmlEnd = 0;
            const string searchString = @"</archive>";
            byte[] searchSequence = Encoding.ASCII.GetBytes(searchString);
            rawXmlArray.Contains(ref searchSequence, out rawXmlEnd);

            byte[] cleanXmlArray = rawXmlArray.getSubArray(0, rawXmlEnd + 1);
            return Encoding.UTF8.GetString(cleanXmlArray);
        }

        public int writeXMLAsFile() {
            string xmlFileName = this.source.Filename + ".xml";
            byte[] xmlArray = allBytes.getSubArray(0, xmlEnd + 1);
            File.WriteAllBytes(xmlFileName, xmlArray);
            return 1;
        }

        public int findXmlEnd() {
            int position = 0;
            const string searchString = @"</archive>";
            byte[] searchSequence = Encoding.ASCII.GetBytes(searchString);
            if (allBytes.Contains(ref searchSequence, out position)) {
                xmlEnd = position;
                return position;
            }
            return -1;
        }

        public int processRecords() {

            if (this.writeXmlToPayload)
                writeXmlMetadataToPayload();
            int currentRecordStart = xmlEnd + 2;
            int currentRecordEnd;

            PrepareOutputFiles();

            while ((currentRecordStart < allBytes.Length)) {
                currentRecordEnd = getRecordEnd(currentRecordStart);
                InstanceHeader ih = new InstanceHeader(ref allBytes, currentRecordStart);
                if (recordIsNeeded(ih)) {
                    writeDataRow(currentRecordStart);
                }
                currentRecordStart = currentRecordEnd + 1;
            }
            CloseOutputFiles();
            return 1;
        }

        private void writeXmlMetadataToPayload() {

            string diagnosticspath = Path.Combine(outputPath, "diagnostics");
            string filename = string.Format("{0}.xml", Path.GetFileNameWithoutExtension(this.source.Filename));
            string diagnosticsfile = Path.Combine(diagnosticspath, filename);

            if (!Directory.Exists(diagnosticspath)) {
                Directory.CreateDirectory(diagnosticspath);
            }

            File.WriteAllText(diagnosticsfile, getArchiveXmlAsString());
        }

        protected abstract void writeHeaderRows();
        protected abstract void writeDataRow(int thisRecordStart);

        protected bool recordIsNeeded(InstanceHeader ih) {
            if (!neededOids.Contains((int)ih.oid)) return false;
            if ((neededFrequencyByOid[(int)ih.oid] & ih.instanceFlag.flagValue) == 0) return false;
            return true;
        }

        protected void CloseOutputFiles() {
            foreach (var oid in writer.Keys) { 
                writer[oid].Close(); 
            }
        }

        protected void PrepareOutputFiles() {

            string path = Path.Combine(outputPath, "counters");
            path = Path.Combine(path, (String.Format("{0:yyyy-MM-dd}", getArchiveDateTime())));

            foreach (int oid in neededOids) {
                string objectPath = Path.Combine(path, counterDefinitions.GetObjectNameByOid(oid));
                if (!Directory.Exists(objectPath)) {
                    Directory.CreateDirectory(objectPath);
                }
                if (countersAreRequestedByOid[oid]) {
                    string wpath = Path.Combine(objectPath, (String.Format("hour_{0:HH}.csv", getArchiveDateTime())));
                    this.messenger.sendDebug(string.Format("<{0}> ...writing payload {1}", this.SystemSerialNo, wpath));
                    writer[oid] = new StreamWriter(wpath);
                    rowsWrittenByOid[oid] = 0;
                }

            }
            writeHeaderRows();
        }

        public List<string> getCounterNames(List<CounterDefinition> counters) {
            List<string> result = new List<string>();
            foreach (var c in counters) {
                if (!c.isHistogram()) {
                    result.Add(c.counterName);
                } else {
                    for (int i = 0; i < c.x_name.Length; i++) {
                        result.Add(string.Format("{0}:{1}", c.counterName, c.x_name[i]));
                    };
                }
            }
            return result;
        }

        public List<string> getCounterValues(
                ref byte[] bytes, List<CounterDefinition> counterDefinitionList, Flag instanceFlag, ref int counterAreaStart, int recordStart) {
            int offset = counterAreaStart;
            InstanceHeader ih = new InstanceHeader(ref allBytes, recordStart);
            string objectName = counterDefinitions.GetObjectNameByOid((int) ih.oid);
            List<string> counterValues = new List<string>();
            foreach (var counterDef in counterDefinitionList) {
                string dottedName = string.Format("{0}.{1}", objectName, counterDef.counterName);
                bool counterDefIsRequested = collectedFrequencyByDottedId.ContainsKey(dottedName);
                if (!counterDef.isHistogram()) {
                    if (counterDefIsRequested && (collectedFrequencyByDottedId[dottedName] & ih.instanceFlag.flagValue) > 0) {
                        CounterInstance counterInstance = new CounterInstance(ref bytes, counterDef.ctype, offset, counterDef.size);
                        counterValues.Add(counterInstance.CounterAsString);
                    } else {
                        // not a histogram; not to be collected.
                        counterValues.Add("");
                    }
                } else {
                    // is a histogram
                    counterValues.AddRange(getHistogramValues(ref bytes, counterDef, instanceFlag, offset, counterDefIsRequested));
                }
                if ((counterDef.flag & ih.instanceFlag.flagValue) > 0) {
                    offset += counterDef.size;
                }
            }
            counterAreaStart = offset;
            return counterValues;
        }

        public List<string> getHistogramValues(
                ref byte[] bytes, CounterDefinition counterDef, Flag instanceFlag, int histogramStart, bool valuesAreRequested) {
            var offset = histogramStart;
            CounterInstance counterInstance = null;
            List<string> result = new List<string>();
            int histogramCount = (int) counterDef.size / (int) counterDef.cTypeSize;
            if (valuesAreRequested && (instanceFlag.flagValue & counterDef.flag) > 0) {
                for (int i = 0; i < histogramCount; i++) {
                    counterInstance = new CounterInstance(ref bytes, counterDef.ctype, offset, (int) counterDef.cTypeSize);
                    result.Add(counterInstance.CounterAsString);
                    offset += (int) counterDef.cTypeSize;
                }
            } else {
                for (int i = 0; i < histogramCount; i++) {
                    result.Add("");
                }
            }
            return result;
        }

        public bool hasCounterValues(
                ref byte[] bytes, List<CounterDefinition> counterDefinitionList, Flag instanceFlag, int counterAreaStart, int recordStart) {
            int offset = counterAreaStart;
            InstanceHeader ih = new InstanceHeader(ref allBytes, recordStart);
            string objectName = counterDefinitions.GetObjectNameByOid((int) ih.oid);
            foreach (var counterDef in counterDefinitionList) {

                string dottedName = string.Format("{0}.{1}", objectName, counterDef.counterName);

                if (collectedFrequencyByDottedId.ContainsKey(dottedName)
                    && ((collectedFrequencyByDottedId[dottedName] & ih.instanceFlag.flagValue) > 0)) {
                    return true;
                }
               
                if ((counterDef.flag & ih.instanceFlag.flagValue) > 0) {
                    offset += counterDef.size;
                }
            }
            return false;
        }

        public static string getStringListAsCsv(List<String> inputList) {
            string csv = "";
            bool isFirstPass = true;
            foreach (string s in inputList) {
                if (isFirstPass) {
                    csv = s;
                    isFirstPass = false;
                } else {
                    csv += "," + s;
                }
            }

            return csv;
        }

        public int getRecordEnd(int recordStart) {
            int recordLength = (int) allBytes.getBytesAsUInt64(recordStart + 8, 4);
            return recordStart + recordLength - 1;
        }

        public DateTime getDateTimeFomUnixTime(long unixTime) {
            var unixStartTime = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc);
            return unixStartTime.AddSeconds(unixTime);
        }

        public StringReader getArchiveFileXmlAsStringReader() {
            return new StringReader(ASCIIEncoding.ASCII.GetString(allBytes, 0, (xmlEnd + 1)));
        }

        public string getArchiveAttribute(string attributeName) {
            XPathDocument xpd = new XPathDocument(getArchiveFileXmlAsStringReader());
            XPathNavigator nav = xpd.CreateNavigator();
            XPathNavigator node = nav.SelectSingleNode("/archive/@" + attributeName);
            return node.Value;
        }

        public DateTime getArchiveDateTime() {
            return getDateTimeFomUnixTime(getArchiveUnixTimestamp());
        }

        public long getArchiveUnixTimestamp() {
            return Convert.ToInt64(getArchiveAttribute("unix_timestamp"));
        }

        public long getArchiveMsecsSinceBoot() {
            return Convert.ToInt64(getArchiveAttribute("msecs_since_boot"));
        }

        public long getTimestampFromMsecsSinceBoot(long instance_msecs) {
            long epochTimeInMsecs = getArchiveUnixTimestamp() * 1000;
            long timestampInMsecs = epochTimeInMsecs + instance_msecs - getArchiveMsecsSinceBoot();
            return timestampInMsecs;
        }
    }
}
