﻿using System;
using System.Collections.Generic;
using System.Xml;

namespace odct.Counters {
    public class CounterDefinitionCollector : ICounterDefinitionCollector {
        #region Private Member Variables
        private string archiveXml;

        //Xpath related constants
        private static string CounterNameXpath = ".//counter[normalize-space(name)='{0}']";
        private static string CounterIdXpath = ".//counter[normalize-space(id)='{0}']";
        private static string ObjectNameXpath = "//object[normalize-space(name)='{0}']";
        private static string CounterNodeXpath = ".//counter";

        //Tag name related constants
        private static string IdNodeName = "id";
        private static string FlagNodeName = "flag";
        private static string CtypeNodeName = "ctype";
        private static string SizeNodeName = "size";
        private static string BaseIdNodeName = "base_id";
        private static string ElementNameNodeName = "name";
        #endregion

        #region Constructor
        public CounterDefinitionCollector(string archiveXml) {
            this.archiveXml = archiveXml;
        }
        #endregion Constructor

        #region Public Member Functions
        public Dictionary<int, CounterDefinition> getCounterDefinitionByCounterId(string objectName) {
            var metaDataXmlDoc = getMetaDataXml(this.archiveXml);

            return getCounterDefinitionByCounterId(metaDataXmlDoc, objectName);
        }

        public XmlDocument getMetaDataXml(string xmlContent) {
            var metaDataXml = (XmlDocument) null;
            try {
                metaDataXml = new XmlDocument();
                metaDataXml.LoadXml(xmlContent);
            } catch {
                Console.WriteLine("Invalid XML");
                return null;
            }
            return metaDataXml;
        }

        public List<string> getCountersByObjectName(XmlDocument xmlDoc, string objectName) {
            if (xmlDoc == null) {
                return null;
            }

            XmlNode objectNode = getObjectNode(xmlDoc, objectName);
            if (objectNode == null) {
                Console.WriteLine("Invalid object to look up");
                return null;
            }

            List<string> counters = new List<string>();
            foreach (XmlNode node in objectNode.SelectNodes(CounterNodeXpath)) {
                counters.Add(node.SelectSingleNode(ElementNameNodeName).InnerText.Trim());
            }

            return counters;
        }

        public Dictionary<int, CounterDefinition> getCounterDefinitionByCounterId(XmlDocument xmlDoc, string objectName) {
            if (xmlDoc == null) {
                return null;
            }

            var counterDefinitionByCounterId = new Dictionary<int, CounterDefinition>();
            var objectNode = getObjectNode(xmlDoc, objectName);

            if (objectNode == null) {
                Console.WriteLine("Invalid object to look up");
                return null;
            }

            var oid = int.Parse(objectNode.SelectSingleNode("id").InnerText.Trim());
            var counters = getCountersByObjectName(xmlDoc, objectName); // TODO: ensure these come in from xml sorted

            foreach (string counterName in counters) {
                try {
                    var counterMeta = getCounterInformationFromObjectNode(objectNode, objectName, oid, counterName);
                    if (counterDefinitionByCounterId.ContainsKey(counterMeta.id)) {
                        throw new DataMisalignedException("Unexpectedly re-encountered same counter. Cannot guarantee offset.");
                    }

                    counterDefinitionByCounterId[counterMeta.id] = counterMeta;
                } catch (Exception ex) {
                    Console.WriteLine(ex.Message);
                }
            }

            return counterDefinitionByCounterId;
        }

        public CounterDefinition getCounterInformationFromObjectNode(XmlNode objectNode, string objectName, int oid, string counter) {
            string locationPath = string.Format(CounterNameXpath, counter);
            XmlNode counterNode = objectNode.SelectSingleNode(locationPath);

            if (counterNode == null) {
                throw new Exception("CounterInstance is not available");
            }

            int id = int.Parse(counterNode.SelectSingleNode(IdNodeName).InnerText.Trim());
            long flag = long.Parse(counterNode.Attributes[FlagNodeName].Value.Trim());
            long ctype = long.Parse(counterNode.Attributes[CtypeNodeName].Value.Trim());
            int size = int.Parse(counterNode.SelectSingleNode(SizeNodeName).InnerText.Trim());
            int base_id = int.Parse(counterNode.SelectSingleNode(BaseIdNodeName).InnerText.Trim());
            string baseCounterName = base_id > 0 ? getCounterNameById(objectNode, base_id) : string.Empty;

            string[] xNames = getXNamesIfExists(counterNode);

            return new CounterDefinition() {
                objectName = objectName,
                id = id,
                oid = oid,
                counterName = counter,
                flag = flag,
                ctype = ctype,
                size = size,
                base_id = base_id,
                x_name = xNames,
            };
        }

        public string[] getXNamesIfExists(XmlNode counterNode) {
            List<string> xNames = new List<string>();
            foreach (XmlNode xNode in counterNode.SelectNodes(".//x_name")) {
                xNames.Add(xNode.InnerText.Trim());
            }

            return xNames.Count == 0 ? null : xNames.ToArray();
        }
        #endregion Public Member Functions

        #region Private Member Functions
        private string getCounterNameById(XmlNode objectNode, int id) {
            string locationPath = string.Format(CounterIdXpath, id);
            XmlNode counterNode = objectNode.SelectSingleNode(locationPath);
            return counterNode.SelectSingleNode(ElementNameNodeName).InnerText.Trim();
        }

        private XmlNode getObjectNode(XmlDocument xmlDoc, string objectName) {
            string locationPath = string.Format(ObjectNameXpath, objectName);
            return xmlDoc.DocumentElement.SelectSingleNode(locationPath);
        }
        #endregion Private Member Functions
    }
}
