﻿using System;
using System.Text.RegularExpressions;
using NetApp.Manage;
using odct.Settings;

namespace odct {
    public class Credentials {
        public const char CREDENTIAL_DELIMITER = ',';
        public string Server { get; set; }
        public string Username { get; set; }
        public string Password { get; set; }

        [System.ComponentModel.Browsable(false)]
        public VerifyResults verifyResults { get; private set; }

        [System.ComponentModel.Browsable(false)]
        public NaServer naServer { get; set; }

        [System.ComponentModel.Browsable(false)]
        public Messenger messenger { get; set; }

        [System.ComponentModel.Browsable(false)]
        public int ontapiMajorVersion = 1;

        [System.ComponentModel.Browsable(false)]
        public int ontapiMinorVersion = 1;

        [System.ComponentModel.Browsable(false)]
        public bool isClusterMode = false;

        public Credentials(string delimitedHostCredentials, char delimiter = CREDENTIAL_DELIMITER) {
            var credParts = delimitedHostCredentials.Split(delimiter);
            this.Initialize(credParts[0].Trim(), credParts[1].Trim(), credParts[2].Trim());
        }

        public Credentials(string server, string username, string password) {
            this.Initialize(server, username, password);
        }

        public Credentials(Credentials source) {
            if (source != null) {
                this.Initialize(source.Server, source.Username, source.Password);
            } else {
                this.Initialize(null, null, null);
            }
        }

        public Credentials() {
            this.Initialize(null, null, null);
        }

        private void Initialize(string server, string username, string password) {
            this.Server = (server != null) ? server : string.Empty;
            this.Username = (username != null) ? username : string.Empty;
            this.Password = (password != null) ? password : string.Empty;
        }

        private void initializeNaServer() {

            NaServer newServer = new NaServer(this.Server, ontapiMajorVersion, ontapiMinorVersion);
            newServer.ServerType = NaServer.SERVER_TYPE.FILER;
            newServer.TransportType = NaServer.TRANSPORT_TYPE.HTTPS;
            newServer.Port = 443;
            newServer.Style = NaServer.AUTH_STYLE.LOGIN_PASSWORD;
            newServer.SetAdminUser(this.Username, this.Password);

            this.naServer = newServer;
        }

        private void updateZapiVersion() {

            NaElement api = new NaElement("system-get-ontapi-version");
            NaElement xo = naServer.InvokeElem(api);

            this.ontapiMajorVersion = xo.GetChildIntValue("major-version", 1);
            this.ontapiMinorVersion = xo.GetChildIntValue("minor-version", 12);

            // all I want to do is update ontapi version; unfortunately it looks like I need to replace naServer.
            //  Java server.setApiVersion doesn't seem to be available in c#.

            NaServer newServer = new NaServer(this.Server, ontapiMajorVersion, ontapiMinorVersion);
            newServer.ServerType = NaServer.SERVER_TYPE.FILER;
            newServer.TransportType = naServer.TransportType;
            newServer.Port = naServer.Port;
            newServer.Style = NaServer.AUTH_STYLE.LOGIN_PASSWORD;
            newServer.SetAdminUser(this.Username, this.Password);

            this.naServer = newServer;
        }

        public VerifyResults Verify(Messenger m) {
            this.messenger = m;
            this.initializeNaServer();
            VerifyResults results = this.VerifyConnection();
            if (!results.IsFound) {
                if (!OdctSettings.Instance.DenyHttp()) {
                    this.naServer.Port = 80;
                    this.naServer.TransportType = NaServer.TRANSPORT_TYPE.HTTP;
                    results = this.VerifyConnection();
                }
            }
            results.setResultDescription();
            verifyResults = results;
            return results;
        }

        private VerifyResults VerifyConnection() {

            if (this.messenger == null)
                throw new OdctNullMessengerException("VerifyConnection called with no messenger defined");

            // First test for cluster mode.

            NaElement api = new NaElement("system-get-version");
            NaElement xo;

            VerifyResults verifyResults = new VerifyResults() {
                IsHttps = (naServer.TransportType == NaServer.TRANSPORT_TYPE.HTTPS),
            };

            try {
                xo = naServer.InvokeElem(api);
            } catch (NaAuthException e) {
                this.messenger.sendError("Authorization Failed during Credentials.Verify(): " + e.Message);
                this.messenger.sendDebug("Credentials.Verify() returned false.");
                this.messenger.sendException(e);
                verifyResults.VerifyException = e;
                verifyResults.IsFound = true;
                return verifyResults;
            } catch (NaApiFailedException e) {
                this.messenger.sendError("API FAILED during Credentials.Verify() " + e.Message);
                this.messenger.sendDebug("Credentials.Verify() returned false.");
                this.messenger.sendException(e);
                verifyResults.VerifyException = e;
                return verifyResults;
            } catch (Exception e) {
                this.messenger.sendError("Other Error during Credentials.Verify(): " + e.Message);
                this.messenger.sendDebug("Credentials.Verify() returned false.");
                this.messenger.sendException(e);
                verifyResults.VerifyException = e;
                return verifyResults;
            }

            updateZapiVersion();

            NaElement isClusteredElement = xo.GetChildByName("is-clustered");
            if (isClusteredElement != null && "true".Equals(isClusteredElement.GetContent())) {

                verifyResults.IsClusterMode = true;
                verifyResults.IsVerified = true;
                verifyResults.IsAuthenticated = true;
                verifyResults.IsHttps = (naServer.TransportType == NaServer.TRANSPORT_TYPE.HTTPS);
                verifyResults.IsFound = true;
                verifyResults.setResultDescription();

                return verifyResults;
            }

            // Next collect basic system info.

            NaElement api2 = new NaElement("system-get-info");
            NaElement xo2;

            try {
                xo2 = naServer.InvokeElem(api2);
            } catch (NaAuthException e) {
                this.messenger.sendError("Authorization Failed during Credentials.Verify(): " + e.Message);
                this.messenger.sendDebug("Credentials.Verify() returned false.");
                verifyResults.VerifyException = e;
                this.messenger.sendException(e);
                return verifyResults;
            } catch (NaApiFailedException e) {
                this.messenger.sendError("API FAILED during Credentials.Verify() " + e.Message);
                this.messenger.sendDebug("Credentials.Verify() returned false.");
                this.messenger.sendException(e);
                verifyResults.VerifyException = e;
                return verifyResults;
            } catch (Exception e) {
                this.messenger.sendError("Other Error during Credentials.Verify(): " + e.Message);
                this.messenger.sendDebug("Credentials.Verify() returned false.");
                this.messenger.sendException(e);
                verifyResults.VerifyException = e;
                return verifyResults;
            }

            NaElement systemInfoElement = xo2.GetChildByName("system-info");
            verifyResults = new VerifyResults() {
                IsVerified = true,
                IsAuthenticated = true,
                IsHttps = (naServer.TransportType == NaServer.TRANSPORT_TYPE.HTTPS),
                IsFound = true,
                FilerSystemId = systemInfoElement.GetChildContent("system-id"),
                FilerSerialNumber = systemInfoElement.GetChildContent("system-serial-number"),
                PartnerSystemId = systemInfoElement.GetChildContent("partner-system-id")
            };

            // 8.0 7-mode systems are not supported.  Identify those.

            NaElement versionElement = xo.GetChildByName("version");
            string versionString = versionElement.GetContent();
            if (Regex.Match(versionString, " 8.0.").Success) {
                verifyResults.IsVerified = false;
                verifyResults.IsVersion8_0 = true;
            }

            this.messenger.sendDebug("Credentials.Verify() returned true.");
            this.messenger.sendDebug(xo2.ToPrettyString(""));

            return verifyResults;
        }

        public void resetVfilerTunneling() {
            NaServer newServer = new NaServer(this.Server, this.ontapiMajorVersion, this.ontapiMinorVersion);
            newServer.ServerType = this.naServer.ServerType;
            newServer.TransportType = this.naServer.TransportType;
            newServer.Style = this.naServer.Style;
            newServer.SetAdminUser(this.Username, this.Password);
            newServer.Port = this.naServer.Port;
            this.naServer = newServer;
        }

    }
}
