﻿namespace odct.FilerConfig {
    using System;
    using System.Collections.Generic;
    using System.Text.RegularExpressions;
    using NetApp.Manage;
    using odct.Settings;

    public class FilerConfiguration {
        private static TimeSpan ALLOWED_SLEW = TimeSpan.FromDays(7);
        public IEnumerable<Section> Sections;
        public string SystemSerialNumber { get; private set; }
        public string SystemId { get; private set; }
        public string SystemName { get; private set; }
        public string SystemModel { get; private set; }
        public string SystemVersion { get; private set; }
        public bool Is7Mode { get; private set; }
        public string Address { get; private set; }
        public string PartnerSystemId { get; private set; }
        public string PartnerNetworkAddress { get; private set; }
        public string PartnerName { get; private set; }

        public string[] SimpleZapiCommands;
        public static readonly string[] SimpleZapiCommands_7_3 = { 
                                    "system-get-ontapi-version",
                                    "aggr-space-list-info",
                                    "snapshot-reserve-list-info",
                                    "options-list-info",
                                    "storage-shelf-environment-list-info",
                                    "net-ifconfig-get",
                                    "fcp-adapter-list-info",
                                    "iscsi-adapter-list-info",
                                    "license-list-info",
                                    "disk-list-info" };
        public static readonly string[] SimpleZapiCommands_8_2 = { 
                                    "system-get-ontapi-version",
                                    "aggr-space-list-info",
                                    "snapshot-reserve-list-info",
                                    "options-list-info",
                                    "storage-shelf-environment-list-info",
                                    "net-ifconfig-get",
                                    "fcp-adapter-list-info",
                                    "iscsi-adapter-list-info",
                                    "license-v2-list-info",
                                    "disk-list-info" };
        // "partition-disk-info", --only in cdot

        public void Build(Credentials credentials, Messenger messenger) {
            messenger.sendDebug("FilerConfiguration.Build() started.");
            this.Address = credentials.Server;
            var sections = new List<Section>();
            //            getAsup();

            var systemGetVersion = new SystemGetVersion(credentials, messenger);
            this.SystemVersion = systemGetVersion.SystemVersion;
            this.Is7Mode = systemGetVersion.Is7Mode;
            sections.Add(systemGetVersion);

            sections.Add(new SystemCli("sysconfig -a", credentials, messenger));
            sections.Add(new SystemCli("sysconfig -ac", credentials, messenger));

            var clock = new Clock(credentials, messenger);

            if (!TimeUtil.TimeSlewWithinRange(clock.FilerCurrentDateTime, DateTime.UtcNow, ALLOWED_SLEW)) {
                throw new OperationCanceledException("Local system time is too far out of range.");
            }

            sections.Add(clock);

            var status = new Status(credentials, messenger);

            if (status.isHaPartner) {

                sections.Add(status);

                var partner = new Partner(credentials, messenger);
                sections.Add(partner);

                this.PartnerName = partner.Name;
                var partnerNetResolve = new PartnerNetResolve(credentials, messenger, partner.Name);
                this.PartnerNetworkAddress = partnerNetResolve.Address;
                sections.Add(partnerNetResolve);

            }

            var systemInfo = new SystemInfo(credentials, messenger);
            this.SystemSerialNumber = systemInfo.SystemSerialNumber;
            this.SystemId = systemInfo.SystemId;
            this.SystemName = systemInfo.SystemName;
            this.SystemModel = systemInfo.SystemModel;
            this.PartnerSystemId = systemInfo.PartnerSystemId;

            sections.Add(systemInfo);

            var adapters = new StorageAdapterGetAdapterList(credentials, messenger);
            sections.Add(adapters);
            foreach (var adapter in adapters.AdapterNames) {
                sections.Add(new StorageAdapterGetAdapterInfo(credentials, messenger, adapter));
            }

            var aggregates = new AggregatesGetAggregateList(credentials, messenger);
            sections.Add(aggregates);
            foreach (var aggregate in aggregates.AggregateNames) {
                sections.Add(new AggregatesGetAggregateOptions(credentials, messenger, aggregate));
            }

            var volumes = new VolumeGetList(credentials, messenger);
            sections.Add(volumes);
            foreach (var volume in volumes.VolumeNames) {
                sections.Add(new VolumeGetOptions(credentials, messenger, volume));
            }
            foreach (var volume in volumes.VolumeNames) {
                sections.Add(new VolumeGetContainer(credentials, messenger, volume));
            }

            if (Regex.Match(SystemVersion, "^7.3").Success || Regex.Match(SystemVersion, "^8.1").Success ) {
                SimpleZapiCommands = SimpleZapiCommands_7_3;
            } else {
                SimpleZapiCommands = SimpleZapiCommands_8_2;
            } 

            foreach (var s in SimpleZapiCommands) {
                sections.Add(new SimpleZapiCommand(s, credentials, messenger));
            }

            this.Sections = sections;
        }
    }
}
