﻿using System;
using System.Diagnostics;
using System.Globalization;
using System.Text;

namespace odct {
    public abstract class Messenger {

        public struct updateObject {
            public int? progress { get; set; }
            public string message;
            public string level;
            public int? resetTo;
            public int? adjustMaxBy;
        }

        private static Stopwatch watch;
        private static int filesToProcess;
        private static int filesProcessed;

        public abstract void sendDebug(string message);
        public abstract void sendInfo(string message);
        public abstract void sendWarning(string message);
        public abstract void sendError(string message);

        public void sendException(Exception e) {
            this.sendError("Exception encountered:");
            this.sendError(string.Format("   e.Message: {0}", e.Message));
            this.sendError(string.Format("    e.Source: {0}", e.Source));
            this.sendError("e.StackTrace:");
            this.sendError(e.StackTrace);
        }

        public virtual void update(updateObject o) {
            if (o.message != null) {
                switch (o.level) {
                    case "DEBUG":
                        this.sendDebug(o.message);
                        break;
                    case "INFO":
                        this.sendInfo(o.message);
                        break;
                    case "WARNING":
                        this.sendWarning(o.message);
                        break;
                    case "ERROR":
                        this.sendError(o.message);
                        break;
                    default:
                        this.sendDebug(o.message);
                        break;
                }
            }
        }

        public static updateObject ProgressReset(int newMaxWork) {
            filesToProcess = newMaxWork;
            string msg = string.Format("ProgressReset, expected files to process set to: {0} .", newMaxWork);
            return new updateObject() { 
                message = msg,
                level = "INFO",
                resetTo = newMaxWork };
        }

        public static updateObject ProgressIncrement(int additionalWorkDone, string nodename = null) {
            filesProcessed += additionalWorkDone;
            if (additionalWorkDone == 0)
                return new updateObject() {message = string.Empty};

            long elapsedMs = Messenger.watch.ElapsedMilliseconds;
            double elapsedSec = (double) elapsedMs / 1000;
            double averageSec = elapsedSec / filesProcessed;

            string msg = "";
            string node = "Have";
            if (!string.IsNullOrEmpty(nodename)) {
                node = nodename;
            }
            if (filesProcessed == 1) {
                msg = string.Format("{0} processed file {1} of {2} files in {3:F2} seconds.", 
                    node, filesProcessed, filesToProcess, elapsedSec);
            } else {
                msg = string.Format("{0} processed file {1} of {2} files in {3:F2} seconds.  Averaging {4:F2} seconds per file.",
                    node, filesProcessed, filesToProcess, elapsedSec, averageSec);
            }

            return new updateObject() { 
                message = msg, 
                level = "INFO",
                progress = additionalWorkDone
            };

        }

        public static updateObject ProgressAdjustMaximum(int adjustment) {
            var oldFilesToProcess = filesToProcess;
            filesToProcess += adjustment;
            string msg = string.Format("Adjusting expected files to process from {0} to {1}", oldFilesToProcess, filesToProcess);

            return new updateObject() { 
                message = msg, 
                level = "INFO",
                adjustMaxBy = adjustment
            };

        }

        public virtual void closeMessenger() {}

        public static void StartTimer() {
            watch = Stopwatch.StartNew();
        }

        public static string timeNow() {
            DateTime localDate = DateTime.Now;
            var culture = new CultureInfo("en-US");
            return localDate.ToString(culture);
        }

    }

    public enum messageLevels : byte {
        debug,
        info,
        warning,
        error
    };

}
