﻿namespace odct {
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Threading;
    using System.Windows.Forms;
    using odct.Settings;
    using odct.Clusters;
    using odct.View;
    using System.Runtime.InteropServices;


    public class Program {

        [DllImport("kernel32.dll", SetLastError = true)]
        static extern bool AllocConsole();

        [DllImport("kernel32.dll")]
        static extern bool AttachConsole(int pid);

        [DllImport("kernel32.dll")]
        static extern IntPtr GetConsoleWindow();

        [DllImport("user32.dll")]
        static extern bool ShowWindow(IntPtr hWnd, int nCmdShow);

        public const int EXIT_DECLINED_LICENSE = 1;
        public const int EXIT_CANCELED_UI = 2;

        const int SW_HIDE = 0;
        const int SW_SHOW = 5;

        private static Messenger messenger;
        private static string outputPath = null;

        private delegate void MessageAction(string title, string message);

        [MTAThread]
        public static void Main(string[] args) {

            OdctSettings.InitializeInstance(Program.getSettingsFile(args));

            if (Program.HasHelpFlag(args)) {
                if (!AttachConsole(-1)) {
                    AllocConsole();
                }
                Program.DisplayHelp();
            } else if (Program.UserPassedArgsToRunFromCommandLine(args)) {
                if (!AttachConsole(-1)) {
                    AllocConsole();
                }
                Program.RunFromCommandLine(args);
            } else {
                Thread guiThread = new Thread(new ThreadStart(delegate { Program.RunFromUserInterface(args); }));
                guiThread.SetApartmentState(ApartmentState.STA);
                guiThread.Start();
            }
        }

        public static bool UserPassedArgsToRunFromCommandLine(string[] args) {
            return hasArgFlag("-hosts", args) && hasArgFlag("-out", args) && !hasArgFlag("-gui", args);
        }

        public static void RunFromCommandLine(string[] args) {
            Program.outputPath = Program.getOutputPath(args);
            Program.messenger = new MessageRepeater(new Messenger[] {
                new ConsoleUI(messageLevels.debug),
                new PayloadLogger(GetTargetTempPath())
            });

            Program.ShowLicenseAgreementDialogAndExitIfDeclined(args);

            var validator = new ProgramVersionValidator(Program.messenger);

            if (!validator.CheckEverything()) {
                Program.messenger.closeMessenger();
                return;
            }

            var hostDefinitionsFile = getHostDefinitionsFile(args);

            var creds = hostDefinitionsFile != null
                ? getHostDefinitionsFromFile(hostDefinitionsFile)
                : OdctSettings.Instance.GetHostConnections();

            if (hasArgFlag("-verify", args)) {
                VerifyAll(creds);
            } else {
                var col = new Collector(
                    messenger,
                    OdctSettings.Instance);

                col.Collect(creds, GetTargetArchiveFilePath(), GetTargetTempPath());
            }

            CmaplWrapper.CleanupDlls();
            (new TempFolder()).Cleanup();
        }

        public static void VerifyAll(ICollection<Credentials> creds) {
            Dictionary<string, string> serverToVerifyResult = new Dictionary<string, string>();
            foreach (Credentials c in creds) {

                Program.messenger.sendInfo("Program.VerifyAll() started.");
                VerifyResults verifyResults = c.Verify(Program.messenger);
                verifyResults.logResults(Program.messenger, "Program", c.Server);

                serverToVerifyResult.Add(c.Server, c.verifyResults.resultDescription);
            }
            Program.messenger.sendInfo("Summary of Filers Processed:");
            Program.messenger.sendInfo("-----------------------------");
            foreach (KeyValuePair<string, string> kvp in serverToVerifyResult) {
                Program.messenger.sendInfo(string.Format("Filer {0}: {1}", kvp.Key, kvp.Value));
            }
            Program.messenger.closeMessenger();
            return;
        }

        public static void RunFromUserInterface(string[] args) {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);

            Program.ShowLicenseAgreementDialogAndExitIfDeclined(args);

            var validator = new ProgramVersionValidator();

            if (!validator.CheckEverything()) {
                return;
            }

            Program.ShowDataEntryDialogAndExitIfCanceled(args);
        }

        public static void ShowLicenseAgreementDialogAndExitIfDeclined(string[] args) {
            if (hasArgFlag("-a", args)) {
                return;
            }

            if (!eulaAcceptanceIsFound()) {
                LicenseAgreementWindow licenseWindow = new LicenseAgreementWindow();
                DialogResult dlgResult = licenseWindow.ShowDialog();
                if (dlgResult != DialogResult.OK) {
                    Environment.Exit(Program.EXIT_DECLINED_LICENSE);
                }
                setEulaAcceptanceToTrue();
            }

        }

        public static void setEulaAcceptanceToTrue() {
            File.WriteAllText(getEulaFile(), "OK");
        }

        public static string getEulaPath() {
            var EulaPath = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "NetApp");
            if (!Directory.Exists(EulaPath))
                Directory.CreateDirectory(EulaPath);
            return EulaPath;
        }

        public static string getEulaFile() {
            return Path.Combine(getEulaPath(), "OnsiteDataCollection_EULA.txt");
        }

        public static bool eulaAcceptanceIsFound() {
            bool eulaAcceptanceIsFound = false;
            try {
                string textRead = System.IO.File.ReadAllText(getEulaFile());
                if ("OK".Equals(textRead)) {
                    eulaAcceptanceIsFound = true;
                }
            } catch (Exception) {
                eulaAcceptanceIsFound = false;
            }
            return eulaAcceptanceIsFound;
        }

        public static void ShowDataEntryDialogAndExitIfCanceled(string[] args) {
            var hosts = getHostDefinitionsFile(args);
            var outputPath = getOutputPath(args);
            MainWindow mainView = new MainWindow(hosts, outputPath);
            DialogResult dlgResult = mainView.ShowDialog();

            if (dlgResult != DialogResult.OK) {
                Environment.Exit(Program.EXIT_CANCELED_UI);
            }
        }

        public static bool hasArgFlag(string argFlag, string[] args) {
            for (int i = 0; i < args.Length; i++) {
                if (argFlag.Equals(args[i])) {
                    return true;
                }
            }

            return false;
        }

        public static string getArgValue(string argKey, string[] args) {
            for (int i = 0; i < args.Length; i++) {
                if (argKey.Equals(args[i]) && args.Length > (i + 1)) {
                    return args[i + 1];
                }
            }

            return null;
        }

        public static string getMustPreExistFilePathArgValue(string argKey, string[] args) {
            var path = getArgValue(argKey, args);
            return File.Exists(path) ? path : null;
        }

        public static string getHostDefinitionsFile(string[] args) {
            return getMustPreExistFilePathArgValue("-hosts", args);
        }
        public static string getOutputPath(string[] args) {
            return getArgValue("-out", args);
        }

        public static string getSettingsFile(string[] args) {
            return getMustPreExistFilePathArgValue("-settings", args);
        }

        public static bool HasHelpFlag(string[] args) {
            return hasArgFlag("--help", args) || hasArgFlag("-help", args) || hasArgFlag("-?", args);
        }

        public static void DisplayHelp() {
            Console.WriteLine("NetApp Onsite Data Collector Tool v1.0.0");
            Console.WriteLine("usage: ");
            Console.WriteLine("\t<missing -hosts -out>\tRun GUI");
            Console.WriteLine("\t-gui <other arguments>\tForce GUI to run");
            Console.WriteLine("\t-a\t\t\tAccept and bypass EULA dialog");
            Console.WriteLine("\t-hosts <csv/tsv-file>\tPath to delimited text file of: host-or-ip,username,password");
            Console.WriteLine("\t\t\t\tWhen running from the command line the hosts file MUST be tab separated (tsv)");
            Console.WriteLine("\t-out <output-zip>\tPath to output payload");
            Console.WriteLine("\t-settings <xml>\t\tPath to XML file with advanced options");
            Console.WriteLine("\t-verify\t\t\tVerify connection credentials only and exit");
            Console.WriteLine();
            Console.WriteLine("example: odct -hosts .\\filer-hosts-list.tsv");
            Console.WriteLine("example: odct -settings .\\advanced.xml -hosts .\\filer-hosts-list.tsv");
        }

        private static ICollection<Credentials> getHostDefinitionsFromFile(string hostDefinitionsFilename) {
            var hosts = OdctSettings.Instance.GetValue(
                OdctSettings.FilerSection,
                OdctSettings.Hosts,
                string.Empty)
                .Split(';');

            var creds = new List<Credentials>();

            var file = new StreamReader(hostDefinitionsFilename);
            for (var line = file.ReadLine(); line != null; line = file.ReadLine()) {
                Credentials c = new Credentials(line, ControlChars.Tab);
                c.messenger = messenger;
                creds.Add(c);
            }

            return creds;
        }

        public static string GetTargetTempPath() {
            return Collector.GetTempPath(GetTargetArchiveFilePath());
        }

        public static string GetTargetArchiveFilePath() {
            var p = Program.outputPath ?? OdctSettings.Instance.GetTargetPath();
            return Path.HasExtension(p) ? p : Path.Combine(p, "odct-payload.zip");
        }
    }
}