﻿namespace odct.Settings {
    using System;
    using System.Collections.Generic;
    using System.IO;
    using System.Net;
    using System.Reflection;
    using System.Xml;

    public class OdctVersion {
        private const string RootElementName = "xml";
        private XmlDocument onlineXmlVersion = null;
        private string dataRootDirectory = string.Empty;
        private string workingDirectory = string.Empty;

        public const int LATEST_VERSION = 0;
        public const int NEW_VERSION_AVAILABLE = 1;
        public const int BELOW_MIN_VERSION = 2;

        public OdctVersion() {
            this.Initialize();
        }

        // TODO extract these functions into XmlGetter and extension XmlGetterSetter for OdctVersion and SettingsFile
        public List<string> GetSectionKeys(string sectionKey) {
            List<string> keys = new List<string>();

            foreach (XmlNode node in this.RootNode.ChildNodes) {
                keys.Add(node.Name);
            }

            return keys;
        }

        public string GetValue(string section, string key, string defaultValue = null) {
            string value = defaultValue;
            XmlNode sectionNode = this.GetSectionNode(section);

            if (sectionNode != null) {
                XmlNode valueNode = sectionNode.SelectSingleNode(key);

                if (valueNode != null) {
                    value = valueNode.InnerText;
                }
            }

            return value;
        }

        public int GetIntValue(string section, string key, int defaultValue) {
            string rawValue = this.GetValue(section, key, null);
            int value = defaultValue;

            if (rawValue != null) {
                if (!int.TryParse(rawValue, out value)) {
                    value = defaultValue;
                }
            }

            return value;
        }

        public bool GetBooleanValue(string section, string key, bool defaultValue) {
            string rawValue = this.GetValue(section, key, null);
            bool value = defaultValue;

            if (rawValue != null) {
                if ((StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "true") == 0) ||
                    (StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "yes") == 0) ||
                    (StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "1") == 0)) {
                    value = true;
                } else if ((StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "false") == 0) ||
                    (StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "no") == 0) ||
                    (StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "0") == 0)) {
                    value = false;
                }
            }

            return value;
        }

        private string XmlLocation {
            get {
                return OdctSettings.Instance.GetValue(
                    OdctSettings.DebugSection, 
                    "VersionUriOverride", 
                    "https://setools.netapp.com/noauth/odct/version.xml");
            }
        }

        private XmlNode RootNode {
            get {
                return this.onlineXmlVersion.SelectSingleNode(string.Format("/{0}", OdctVersion.RootElementName));
            }
        }

        private void Initialize() {
            this.DetermineWorkingDirectory();
            this.LoadXml();
        }

        private XmlNode GetSectionNode(string section) {
            return this.RootNode.SelectSingleNode(section);
        }

        private void DetermineWorkingDirectory() {
            this.workingDirectory = Directory.GetCurrentDirectory();
        }

        private void LoadXml() {
            //ServicePointManager.SecurityProtocol = SecurityProtocolType.Ssl3;

            var certValidationCallback = new System.Net.Security.RemoteCertificateValidationCallback((s, ce, ch, ssl) => true);
            var reader = new XmlTextReader(this.XmlLocation);

            // ignore bad certificates
            ServicePointManager.ServerCertificateValidationCallback += certValidationCallback;

            this.onlineXmlVersion = new XmlDocument();
            
            // Load the Xml file...

            try {
                this.onlineXmlVersion.Load(reader);
            } catch {
                this.onlineXmlVersion.RemoveAll();
            }

            ServicePointManager.ServerCertificateValidationCallback -= certValidationCallback;

            var rootNode = this.RootNode;

            if (rootNode == null) {
                string xml = string.Format("<{0} />", OdctVersion.RootElementName);
                this.onlineXmlVersion.LoadXml(xml);
            }
        }

        public Version GetAssemblyVersion(Assembly assembly) {
            AssemblyName assemblyName = assembly.GetName();
            return this.GetAssemblyVersion(assemblyName);
        }

        public Version GetAssemblyVersion(string assemblyName) {
            Version assemblyVersion = null;

            try {
                AssemblyName assName = AssemblyName.GetAssemblyName(assemblyName);
                assemblyVersion = this.GetAssemblyVersion(assName);
            } catch {
                assemblyVersion = new Version(0, 0, 0, 0);
            }

            return assemblyVersion;
        }

        private Version GetAssemblyVersion(AssemblyName assemblyName) {
            Version assemblyVersion = null;

            if (assemblyName != null) {
                assemblyVersion = assemblyName.Version;
            }

            return assemblyVersion;
        }

        public string GetAssemblyCompany(Assembly assembly) {
            string company = string.Empty;

            if (assembly != null) {
                object[] attributes = assembly.GetCustomAttributes(typeof(AssemblyCompanyAttribute), false);

                if (attributes.Length > 0) {
                    company = ((AssemblyCompanyAttribute) attributes[0]).Company;
                }
            }

            return company;
        }

        public Version GetVersion() {
            return Assembly.GetExecutingAssembly().GetName().Version;
        }

        public bool IsAccessible() {
            return this.onlineXmlVersion != null 
                && this.RootNode != null 
                && this.RootNode.ChildNodes.Count > 0;
        }

        public Version Minimum {
            get {
                return new Version(this.GetValue("version", "min", "0.0.0.0"));
            }
        }

        public Version Latest {
            get {
                return new Version(this.GetValue("version", "latest", "0.0.0.0"));
            }
        }

        public bool IsBelowMinimum() {
            return this.GetVersion() < this.Minimum;
        }

        public bool IsBelowLatest() {
            return this.GetVersion() < this.Latest;
        }

        public int GetVersionStatus() {
            int status = OdctVersion.LATEST_VERSION;

            if (this.IsAccessible()) {
                if (this.IsBelowMinimum()) {
                    status = OdctVersion.BELOW_MIN_VERSION;
                } else if (this.IsBelowLatest()) {
                    status = OdctVersion.NEW_VERSION_AVAILABLE;
                }
            }

            return status;
        }
    }
}