﻿namespace odct.Settings {
    using System;
    using System.Collections;
    using System.Collections.Generic;
    using System.Data;
    using System.Diagnostics;
    using System.IO;
    using System.Reflection;
    using System.Xml;

    public class SettingsFile {
        private const string SettingsRootElementName = "Settings";
        private XmlDocument xmlSettings = null;
        private string dataRootDirectory = string.Empty;
        private string workingDirectory = string.Empty;

        public SettingsFile(string fileName, bool autoSave) {
            this.FileName = fileName;
            this.AutoSave = autoSave;

            this.Initialize();
        }

        public string FileName { get; set; }

        public bool AutoSave { get; set; }

        public List<string> GetSectionKeys(string sectionKey) {
            List<string> keys = new List<string>();

            foreach (XmlNode node in this.SettingsRootNode.ChildNodes) {
                keys.Add(node.Name);
            }

            return keys;
        }

        public string GetValue(string section, string key, string defaultValue = null) {
            string value = defaultValue;
            XmlNode sectionNode = this.GetSectionNode(section);

            if (sectionNode != null) {
                XmlNode valueNode = sectionNode.SelectSingleNode(key);

                if (valueNode != null) {
                    value = valueNode.InnerText;
                }
            }

            return value;
        }

        public int GetIntValue(string section, string key, int defaultValue) {
            string rawValue = this.GetValue(section, key, null);
            int value = defaultValue;

            if (rawValue != null) {
                if (!int.TryParse(rawValue, out value)) {
                    value = defaultValue;
                }
            }

            return value;
        }

        public bool GetBooleanValue(string section, string key, bool defaultValue) {
            string rawValue = this.GetValue(section, key, null);
            bool value = defaultValue;

            if (rawValue != null) {
                if ((StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "true") == 0) ||
                    (StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "yes") == 0) ||
                    (StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "1") == 0)) {
                    value = true;
                } else if ((StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "false") == 0) ||
                    (StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "no") == 0) ||
                    (StringComparer.InvariantCultureIgnoreCase.Compare(rawValue, "0") == 0)) {
                    value = false;
                }
            }

            return value;
        }

        public bool SetValue(string section, string key, string value) {
            bool added = false;
            XmlNode sectionNode = this.GetSectionNode(section);

            if (sectionNode == null) {
                sectionNode = this.xmlSettings.CreateElement(section);
                this.SettingsRootNode.AppendChild(sectionNode);
            }

            XmlNode valueNode = sectionNode.SelectSingleNode(key);

            if (valueNode == null) {
                valueNode = this.xmlSettings.CreateElement(key);
                sectionNode.AppendChild(valueNode);
            }

            if (value == null) {
                sectionNode.RemoveChild(valueNode);
            } else {
                valueNode.InnerText = value;
                added = true;
            }

            if (this.AutoSave) {
                this.Save();
            }

            return added;
        }

        public bool SetIntValue(string section, string key, int value) {
            return this.SetValue(section, key, value.ToString());
        }

        public bool SetBooleanValue(string section, string key, bool value) {
            return this.SetValue(section, key, value ? "True" : "False");
        }

        public void Save() {
            XmlWriterSettings xmlWriterSettings = new XmlWriterSettings();
            xmlWriterSettings.Encoding = System.Text.Encoding.UTF8;
            xmlWriterSettings.Indent = true;
            xmlWriterSettings.IndentChars = "    ";
            xmlWriterSettings.NewLineHandling = NewLineHandling.Replace;
            xmlWriterSettings.NewLineChars = ControlChars.NewLine;
            xmlWriterSettings.OmitXmlDeclaration = true;

            XmlWriter xmlWriter = XmlWriter.Create(this.SettingsXmlFileName, xmlWriterSettings);
            this.xmlSettings.Save(xmlWriter);
            xmlWriter.Close();
        }

        private string SettingsXmlFileName {
            get {
                return Path.Combine(this.DataDirectory, this.FileName);
            }
        }

        private string DataRootDirectory {
            get { return this.dataRootDirectory; }
        }

        public string OldDataDirectory {
            get { return this.DataDirectory; }
        }

        public string DataDirectory {
            get { return this.workingDirectory; }
        }

        private XmlNode SettingsRootNode {
            get {
                return this.xmlSettings.SelectSingleNode(string.Format("/{0}", SettingsFile.SettingsRootElementName));
            }
        }

        private void Initialize() {
            this.DetermineWorkingDirectory();
            this.LoadOrCreateSettingsXmlFile();
        }

        private XmlNode GetSectionNode(string section) {
            return this.SettingsRootNode.SelectSingleNode(section);
        }

        private void DetermineWorkingDirectory() {
            this.workingDirectory = Directory.GetCurrentDirectory();
        }

        private void LoadOrCreateSettingsXmlFile() {
            // Load the Xml file...
            this.xmlSettings = new XmlDocument();

            if (File.Exists(this.SettingsXmlFileName)) {
                try {
                    this.xmlSettings.Load(this.SettingsXmlFileName);
                } catch {
                    this.xmlSettings.RemoveAll();
                }
            }

            XmlNode settingsRootNode = this.SettingsRootNode;

            if (settingsRootNode == null) {
                string xml = string.Format("<{0} />", SettingsFile.SettingsRootElementName);
                this.xmlSettings.LoadXml(xml);
            }
        }
    }
}