﻿namespace odct.View {
    using System;
    using System.ComponentModel;
    using System.IO;
    using System.Text.RegularExpressions;
    using System.Windows.Forms;
    using odct.Settings;
    using odct.Clusters;
    using System.Drawing;

    public partial class MainWindow : Form {
        public BindingList<Credentials> Credentials { get; private set; }
        private Collector collector;
        private Messenger messenger;
        private string hostsFilepath = null;
        private string outputFilepath = null;
        private readonly Image trashcanSelectedImage = Properties.Resources.Trashcan_Selected;
        private readonly Image trashcanDeselectedImage = Properties.Resources.Trashcan_Deselected;


        public MainWindow(string hostsFilepath, string outputFilepath) {
            this.InitializeComponent();
            this.AutoFindControllersCheckBox.Checked = OdctSettings.Instance.GetAutoCollectHaPartners();

            this.ControllerDataGrid.AutoGenerateColumns = false; // VERY IMPORTANT!!!

            this.hostsFilepath = hostsFilepath;
            this.outputFilepath = outputFilepath;

            // where should GUI interaction messages get recorded, should we keep them?
            this.messenger = new MessageKeeper();

            this.Credentials = new BindingList<Credentials>();

            var hasPrePopulatedHosts = !string.IsNullOrEmpty(this.hostsFilepath);
            var hasPrePopulatedOutputPath = !string.IsNullOrEmpty(this.outputFilepath);

            if (hasPrePopulatedHosts) {
                var srdr = new StreamReader(this.hostsFilepath);
                this.ImportHostsFromDelimitedTextFile(srdr.BaseStream);
            }

            if (hasPrePopulatedOutputPath) {
                this.SaveFileTextBox.Text = this.outputFilepath;
            }

            this.StartButton.Enabled = hasPrePopulatedHosts && hasPrePopulatedOutputPath && this.AreAnySpecifiedHostsValid() && this.IsPayloadOutputFileSpecified;
        }

        private bool AreAnySpecifiedHostsValid() {
            foreach (DataGridViewRow row in this.ControllerDataGrid.Rows) {
                if (row.Cells["ResultColumn"].Value != null 
                    && !string.IsNullOrEmpty(row.Cells["ResultColumn"].Value.ToString())
                    && row.Cells["ResultColumn"].Value.ToString().Equals("SUCCESS")) {
                        return true;
                }
            }

            return false;
        }

        private void MainWindow_Load(object sender, EventArgs e) {
            this.ControllerDataGrid.DataSource = new BindingSource(this.Credentials, null);
            this.ControllerDataGrid.RowHeadersWidthSizeMode = DataGridViewRowHeadersWidthSizeMode.DisableResizing;

            this.InitializeDeleteImageColumn();
        }

        private void InitializeDeleteImageColumn() {
            DataGridViewCellStyle defaultStyle = this.TrashColumn.DefaultCellStyle;

            this.TrashColumn.Image = this.trashcanDeselectedImage;
            defaultStyle.NullValue = this.trashcanDeselectedImage;
            defaultStyle.SelectionBackColor = this.ControllerDataGrid.DefaultCellStyle.BackColor;
        }

        private void SaveFileButton_Click(object sender, EventArgs e) {
            try {
                SaveFileDialog fdlg = new SaveFileDialog();
                fdlg.Title = "Select payload output zip file";
                //                fdlg.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
                fdlg.Filter = "Zip (*.zip)|*.zip|All files (*.*)|*.*";
                fdlg.FilterIndex = 1;
                fdlg.RestoreDirectory = true;
                if (fdlg.ShowDialog() == DialogResult.OK) {
                    this.SaveFileTextBox.Text = fdlg.FileName;
                    this.StartButton.Enabled = this.AreAnySpecifiedHostsValid() && IsPayloadOutputFileSpecified;
                }
            } catch (Exception exc) {
                MessageBox.Show(exc.Message);
            }
        }

        private void TestButton_Click(object sender, EventArgs e) {
            var validator = new ProgramVersionValidator(false);

            if (!validator.CheckEverything()) {
                return;
            }

            this.TestButton.Enabled = false;
            Cursor.Current = Cursors.WaitCursor;
            this.TestCredentials(this.messenger);
            Cursor.Current = Cursors.Default;
            this.TestButton.Enabled = true;
        }

        private void TestCredentials(Messenger testMessenger) {
            var credentialsSpecified = 0;
            var failures = 0;
            foreach (DataGridViewRow r in this.ControllerDataGrid.Rows) {
                if (r.DataBoundItem == null) {
                    continue;
                }

                Credentials cred = new Credentials((Credentials) r.DataBoundItem);  // make a copy

                if (cred != null) {
                    credentialsSpecified++;

                    if (cred.Username.Trim().Length == 0) {
                        cred.Username = this.LoginIdTextBox.Text;
                    }

                    if (cred.Password.Trim().Length == 0) {
                        cred.Password = this.PasswordTextBox.Text;
                    }

                    r.Cells["ResultColumn"].Value = "Testing...";
                    ControllerDataGrid.Refresh();

                    VerifyResults vr = cred.Verify(this.messenger);
                    if (vr.IsVerified) {
                        r.Cells["ResultColumn"].Value = "SUCCESS";
                        ControllerDataGrid.Refresh();
                    } else if (vr.IsClusterMode) {
                        r.Cells["ResultColumn"].Value = "cDOT not supported";
                        ControllerDataGrid.Refresh();
                        failures++;
                    } else if (vr.IsVersion8_0) {
                        r.Cells["ResultColumn"].Value = "v8.0 not supported";
                        ControllerDataGrid.Refresh();
                        failures++;
                    } else if (!vr.IsFound) {
                        r.Cells["ResultColumn"].Value = "Host not found";
                        ControllerDataGrid.Refresh();
                        failures++;
                    } else if (!vr.IsAuthenticated) {
                        r.Cells["ResultColumn"].Value = "Credentials rejected";
                        ControllerDataGrid.Refresh();
                        failures++;
                    } else {
                        r.Cells["ResultColumn"].Value = "Unknown error";
                        ControllerDataGrid.Refresh();
                        failures++;
                    }
                } else {
                    r.Cells["ResultColumn"].Value = string.Empty;
                }
            }

            var success = failures < 1;
            if (this.SaveFileTextBox.InvokeRequired) {
                this.StartButton.Invoke((MethodInvoker) delegate {
                    this.StartButton.Enabled = IsPayloadOutputFileSpecified;
                });
            } else {
                this.StartButton.Enabled = IsPayloadOutputFileSpecified;
            }

            if (!success) {
                string title = string.Empty;
                string msg = string.Empty;

                if (failures >= credentialsSpecified) {
                    title = Properties.Resources.ALL_HOST_CONNECTIONS_FAILED_TITLE;
                    msg = Properties.Resources.ALL_HOST_CONNECTIONS_FAILED_MSG;

                    MessageBox.Show(msg, title, MessageBoxButtons.OK, MessageBoxIcon.Error);
                    this.StartButton.Enabled = false;
                } else {
                    title = Properties.Resources.SOME_HOST_CONNECTIONS_FAILED_TITLE;
                    msg = string.Format(Properties.Resources.SOME_HOST_CONNECTIONS_FAILED_MSG, failures, credentialsSpecified);

                    MessageBox.Show(msg, title, MessageBoxButtons.OK, MessageBoxIcon.Warning);
                }
            }
        }

        private bool IsPayloadOutputFileSpecified {
            get {
                return !string.IsNullOrEmpty(this.SaveFileTextBox.Text);
            }
        }

        private void StartButton_Click(object sender, EventArgs e) {
            var validator = new ProgramVersionValidator(false);

            if (!validator.CheckEverything()) {
                return;
            }

            var target = this.SaveFileTextBox.Text;
            var temp = Collector.GetTempPath(target);

            var progress = new MessageProgress();
            var logFile = new PayloadLogger(temp);
            var collectionMessenger = new MessageRepeater(new Messenger[] {
                progress,
                logFile
            });

            var creds = this.CopyCredentialsAndFillInDefaultValues(this.Credentials, this.LoginIdTextBox.Text, this.PasswordTextBox.Text);

            // dump this.messenger kept messages into collectionMessenger now that we have a place for them?
            // ((Messenger) collectionMessenger).sendDebug(((MessageKeeper) this.messenger).GetMessages());

            progress.SetWork(delegate { this.startCollection(creds, collectionMessenger, target, temp); });

            int maxFilesToCopy = OdctSettings.Instance.GetPerformanceFilesCollectionLimit();

            int filesPerFiler = Math.Min(maxFilesToCopy, Collector.ExpectedArchiveFilesPer7ModeFiler);
            progress.InitAndShow(this.Credentials.Count * filesPerFiler, delegate {
                this.messenger.sendWarning("Collection Cancelled");
                logFile.closeMessenger();
                this.collector.Cancel();
            });
        }

        private void startCollection(BindingList<Credentials> creds, Messenger collectionMessenger, string target, string temp) {
            var opts = new CollectorOptions {
                AutoCollectHaPartner = this.AutoFindControllersCheckBox.Checked,
                AppCountersRequested = "trt"
            };

            this.collector = new Collector(collectionMessenger, opts);

            this.collector.Collect(creds, target, temp);
        }

        private void ControllerDataGrid_CellClick(object sender, DataGridViewCellEventArgs e) {
            if (e.ColumnIndex >= 0) {
                DataGridViewColumn column = (DataGridViewColumn) ControllerDataGrid.Columns[e.ColumnIndex];

                if (e.RowIndex >= 0) {                                               // this is not the header row
                    if (e.RowIndex != ControllerDataGrid.NewRowIndex) {              // and this is not the New Row 
                        if (column.Index == this.TrashColumn.Index) {  // and this is the trashcan column
                            ControllerDataGrid.Rows.RemoveAt(e.RowIndex);
                        }
                    }
                }
            }
        }

        private void TryDeleteSelectedRow() {
            Int32 selectedRowCount = ControllerDataGrid.Rows.GetRowCount(DataGridViewElementStates.Selected);

            if (selectedRowCount == 0) {
                return;
            }

            for (int i = 0; i < selectedRowCount; i++) {
                ControllerDataGrid.Rows.RemoveAt(i);
            }
        }

        private void ControllerDataGrid_KeyDown(object sender, KeyEventArgs e) {
            if (e.KeyCode == Keys.Delete) {
                TryDeleteSelectedRow();
            }

            try {
                if (e.Control && e.KeyCode == Keys.C) {
                    DataObject clipData = ControllerDataGrid.GetClipboardContent();
                    Clipboard.SetDataObject(clipData);
                    e.Handled = true;
                } else if (e.Control && e.KeyCode == Keys.V) {
                    PasteRecordsInDataGrid(Clipboard.GetText(TextDataFormat.Text));
                }
            } catch (Exception ex) {
                messenger.sendError("Error while trying to copy/paste: " + ex.Message);
                messenger.sendDebug("Copy/paste returned error.");
                messenger.sendException(ex);
            }
        }

        private void PasteRecordsInDataGrid(string credentialsData) {
            try {
                StartButton.Enabled = false;
                var credentails = new BindingList<Credentials>(PeformDeepCopy(this.Credentials));

                string[] rowLines = credentialsData.Split(new string[1] { "\r" + "\n" }, StringSplitOptions.None);

                for (int rowLine = 0; rowLine <= rowLines.Length - 1; rowLine++) {
                    if (rowLine == rowLines.Length - 1 && string.IsNullOrEmpty(rowLines[rowLine])) {
                        break;
                    }

                    var delimiter = DetermineSplitDelimiter(rowLines[rowLine]);
                    if (delimiter == null) {
                        messenger.sendError("Unable to determine pasted contents format delimiter.");
                        return;
                    }

                    string[] columnsData = rowLines[rowLine].Split(new string[1] { delimiter }, StringSplitOptions.None);
                    if (columnsData != null) {
                        Credentials credential = new odct.Credentials();
                        credential.Server = columnsData.Length >= 0 ? columnsData[0] : string.Empty;
                        credential.Username = columnsData.Length >= 1 ? columnsData[1] : string.Empty;
                        credential.Password = columnsData.Length >= 2 ? columnsData[2] : string.Empty;
                        credentails.Add(credential);
                    }
                }

                this.Credentials = credentails;

                var datasource = this.ControllerDataGrid.DataSource as BindingSource;

                if (datasource != null) {
                    datasource.DataSource = this.Credentials;
                    datasource.ResetBindings(false);
                    this.ControllerDataGrid.Refresh();
                }
            } catch (Exception ex) {
                messenger.sendError("Error while trying to paste to Grid: " + ex.Message);
                messenger.sendDebug("Paste to grid returned error.");
                messenger.sendException(ex);
            }
        }

        public static string DetermineSplitDelimiter(string sample) {
            if (Regex.Match(sample, "^.*\t.*\t.*$").Success)
                return "\t";
            else if (Regex.Match(sample, "^.* .* .*$").Success)
                return " ";
            else if (Regex.Match(sample, "^.*,.*,.*$").Success)
                return ",";
            return null;
        }

        private void PasteFromClipboardButton_Click(object sender, EventArgs e) {
            this.PasteRecordsInDataGrid(Clipboard.GetText(TextDataFormat.Text));
        }

        private void ImportFromFileButton_Click(object sender, EventArgs e) {
            StartButton.Enabled = IsPayloadOutputFileSpecified && AreAnySpecifiedHostsValid();

            OpenFileDialog openFileDialog = new OpenFileDialog();

            openFileDialog.Title = "Select host credentials import file.";
            openFileDialog.Filter = "Delimited text (*.csv;*.tsv;*.ssv)|*.csv;*.tsv;*.ssv|All files (*.*)|*.*";

            openFileDialog.CheckFileExists = true;
            openFileDialog.CheckPathExists = true;

            if (openFileDialog.ShowDialog() == DialogResult.OK) {
                // Open the selected file to read.
                Stream fileStream = openFileDialog.OpenFile();

                this.ImportHostsFromDelimitedTextFile(fileStream);
            }
        }

        private void ImportHostsFromDelimitedTextFile(Stream fileStream) {
            using (var reader = new StreamReader(fileStream)) {
                // Read the first line from the file.
                PasteRecordsInDataGrid(reader.ReadToEnd());
            }

            fileStream.Close();
        }
        private BindingList<Credentials> PeformDeepCopy(BindingList<Credentials> oldCopy) {
            var newCopy = new BindingList<Credentials>();
            foreach (Credentials cred in oldCopy) {
                if (!string.IsNullOrEmpty(cred.Server) || !string.IsNullOrEmpty(cred.Username) || !string.IsNullOrEmpty(cred.Password))
                    newCopy.Add(cred);
            }

            return newCopy;
        }

        private void ControllerDataGrid_CellFormatting(object sender, DataGridViewCellFormattingEventArgs e) {
            if ((e.ColumnIndex == this.PasswordColumn.Index) && (e.Value != null)) {
                this.ControllerDataGrid.Rows[e.RowIndex].Tag = e.Value;
                e.Value = chkShowPassword.Checked ? e.Value : new String('*', e.Value.ToString().Length);
            } else if ((e.ColumnIndex == this.TrashColumn.Index) && (e.Value != this.trashcanDeselectedImage) && (e.Value != this.trashcanSelectedImage)) {
                e.Value = this.trashcanDeselectedImage;
            }
        }

        private void ControllerDataGrid_EditingControlShowing(object sender, DataGridViewEditingControlShowingEventArgs e) {
            if (e.Control is TextBox) {
                var textbox = e.Control as TextBox;

                this.StartButton.Enabled = false;

                if ((this.ControllerDataGrid.CurrentRow.Tag != null) && (this.ControllerDataGrid.CurrentCell.ColumnIndex == this.PasswordColumn.Index)) {
                    this.FormatPasswordText(textbox, chkShowPassword.Checked);
                    e.Control.Text = this.ControllerDataGrid.CurrentRow.Tag.ToString();
                } else if (this.ControllerDataGrid.CurrentCell.ColumnIndex != this.PasswordColumn.Index) {
                    this.FormatPasswordText(textbox, true);
                }
            }
        }

        private void FormatPasswordText(TextBox textbox, bool checkboxChecked) {
            textbox.UseSystemPasswordChar = !checkboxChecked;
            textbox.PasswordChar = checkboxChecked ? '\0' : '*';
        }

        private void chkShowPassword_CheckedChanged(object sender, EventArgs e) {
            FormatPasswordText(this.PasswordTextBox, chkShowPassword.Checked);
            this.ShowHideGridPassword();
        }

        private void ShowHideGridPassword() {
            foreach (DataGridViewRow row in this.ControllerDataGrid.Rows) {
                row.Selected = true;
            }
        }

        private void UploadDataButton_Click(object sender, EventArgs e) {

        }

        private void UploadHyperlink_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e) {

        }

        private void ControllerDataGrid_CellEnter(object sender, DataGridViewCellEventArgs e) {
            this.SetTrashcanImage(e.RowIndex, e.ColumnIndex, this.trashcanSelectedImage);
        }

        private void ControllerDataGrid_CellMouseEnter(object sender, DataGridViewCellEventArgs e) {
            this.SetTrashcanImage(e.RowIndex, e.ColumnIndex, this.trashcanSelectedImage);
        }

        private void ControllerDataGrid_CellLeave(object sender, DataGridViewCellEventArgs e) {
            this.SetTrashcanImage(e.RowIndex, e.ColumnIndex, this.trashcanDeselectedImage);
        }

        private void ControllerDataGrid_CellMouseLeave(object sender, DataGridViewCellEventArgs e) {
            this.SetTrashcanImage(e.RowIndex, e.ColumnIndex, this.trashcanDeselectedImage);
        }

        private void SetTrashcanImage(int rowIndex, int columnIndex, Image image) {
            if ((rowIndex >= 0) && (columnIndex == this.TrashColumn.Index)) {
                DataGridViewCell cell = this.ControllerDataGrid.Rows[rowIndex].Cells[columnIndex];

                cell.Value = image;
                this.ControllerDataGrid.InvalidateCell(cell);
            }
        }

        private BindingList<Credentials> CopyCredentialsAndFillInDefaultValues(BindingList<Credentials> sourceList, string defaultUserId, string defaultPassword) {
            var destList = new BindingList<Credentials>();

            foreach (var sourceCred in sourceList) {
                var destCred = new Credentials(sourceCred);

                if (destCred.Username.Trim().Length <= 0) {
                    destCred.Username = defaultUserId;
                }

                if (destCred.Password.Trim().Length <= 0) {
                    destCred.Password = defaultPassword;
                }

                destList.Add(destCred);
            }

            return destList;
        }

        private void MainWindow_FormClosed(object sender, FormClosedEventArgs e) {
            CmaplWrapper.CleanupDlls();
            (new TempFolder()).Cleanup();
        }
    }
}
